#
#  DS_Post_Fac.moo
#

mod_type pop

sscale        0.04   # degr/pix (this is different from stim)
tscale        0.002  # sec/frame
xn           32      # x size
yn           32      # y size
tn         1024      # t size

<area>
  name   v1          # Area name
  x0    10.0         # Origin wrt stim grid
  y0    10.0         #
  xn    12           # width
  yn    12           # height
  umx   64.0 (um)    # microns per Area grid unit (for now, Area = Stim grid)
</area>

<map>
  name orimap          # ID for reference
  type ori             # Type of map
  ncol_x 3             # (2) Number of ori columns in X-direction
  phase_x 90.0  (deg)  # Phase of pinwheel centers along x-axis, 0-360
  phase_y 90.0  (deg)  # Phase of pinwheel centers along y-axis, 0-360
  phase_p 135.0 (deg)  # (135) Phase of color around pinwheel center, 0-360
</map>

model_noise_seed 231965       # Controls trial seed

#--------------------------------------.--------------------------------------#
#                                                                             #
#                                      IN                                     #
#                                                                             #
#-----------------------------------------------------------------------------#
<pop>
  name in
  area v1
  <icon>
    shape circle
    r 0.4
    g 0.4
    b 0.8
  </icon>

  map_ori orimap      # Set orientations using the named map

  <geometry>
    zn    4           # (1) Depth of layer
  </geometry>

  <mech>
    name ex
    type  psg_alpha    # Mechanism type
    tau   0.003        # Tau for excit PSC (sec)
    amp   1.0          # Amplitude for excit PSC (nS)
    syntype  1         # 1-excit, 2-inhib
  <>

  <mech>
    name in
    type  psg_alpha    # Mechanism type
    tau   0.004        # Tau for inhib PSC (sec)
    amp   4.0          # (2.0) Amplitude for inhib PSC (nS)
    syntype  2         # 1-excit, 2-inhib
  <>

  <mech>  ### Was lgn2ctx...
    name lgn_ex
    type  psg_doe    # Mechanism type

    ### EPSG from LGN to cortex, Diff Of Exp

    tau_r  1.0   # (ms) rise time constant
    tau_f  4.0   # (ms) fall time constant
    amp    1     # (nS) amplitude [Troyer et al: 0.89nS for 3 connections]
    sdf    1.0   # (  ) synaptic depression, ampl after spike
    sdtau  99.0  # (ms) synaptic depression, recovery time constant
  <>

  <input>  # LGN input
    type lgn_on_off
    receptor lgn_ex

    pop_origin lgn     # Population of origin of LGN inputs [lgn]

    sf       2.0   # (1.0) (cyc/deg) SF
    sd_orth  0.125 # (deg) Gaussian SD for Gabor RF orthog to ori
    sd_par   0.215 # (deg) Gaussian SD for Gabor RF parallel to ori
    phseed 23092   # for chosing phase (random in 1 of 4 phases)
    ph0      0.0   # (degr) initial phase
    seed  1237     # for chosing DOG sampling

    phtype   1     # (0) 0-const, 1-regular phase steps, 4-regular quads
    phstep  90.0   # (degr) phase step

    nsamp   30     # (50) number of samples to select as inputs
    maxrep   1     # number of times any one connection can be sampled
    eps      0.05  # no connections where profile < this frac of max
    balance  0     # Balance the # of ON and OFF connections
  </input>

  <input>
    type       bg     # Background input spikes
    ex_rate   100.0   # spikes/s
    ex_amp      2.0   # relative to _psc_amp_ex
    in_rate    50.0   # spikes/s
    in_amp      2.0   # relative to _psc_amp_in
  </input>


  <spike_gen>
    type         ifc   #
    v_spike     10.0   # (mV) Spike height
    v_th_x     -52.5   # (mV) Spike threshold
    v_reset_x  -57.8   # (mV) Spike reset voltage [TKPM98 -57.8]
    tau_r_ad     1.00  # (ms) Adaptation rise time
    tau_f_ad    80.0   # (ms) Adaptation fall time
    gbar_ad      0.0   # (nS) Adaptation conductance [TKPM98 0]
    v_ex         0.0   # (mV) Excitatory reversal
    v_in       -70.0   # (mV) Inhibitory reversal
    v_ad       -90.0   # (mV) Adaptation reversal
    v_leak_x   -81.6   # (mV) Leakage reversal [TKPM98 -81.6]
    g_leak_x    18.0   # (nS) Leakage conductance [TKPM98 18]
    c_x        214.0   # (pF) Membrane capacitance [TKPM98 214]
    trefr_x      1.5   # (ms) Refractory period
    gx_scale     2.0   # (  ) Scale excit g, before adding bias 
    gx_bias      0.0   # (nS) Constant added to g, after any scaling
    <gx_noise>
      type   gfg       #  gfg = Gaussian filtered Gaussian noise
      mean   0.0 (nS)  #  mean
      sd     4.0 (nS)  #  standard deviation
      tsd    2.0 (ms)  #  temporal SD
    <>
    grect        1     # half-wave rectify gx, gi after adding noise
  </spike_gen>
</pop>

#--------------------------------------.--------------------------------------#
#                                                                             #
#                                     EXS                                     #
#                                                                             #
#-----------------------------------------------------------------------------#
<pop>
  name exs
  area v1
  <icon>
    shape stellate
    r 0.0
    g 0.8
    b 0.2
    npoint 5
  </icon>

  map_ori orimap       # Set orientations using the named map

  <geometry>
    zn    4
  </geometry>

  <mech>  # Incoming excitatory spikes
    name ex
    type  psg_alpha    # Mechanism type
    tau  0.003         # Tau for excit PSC (sec)
    amp  1.0           # (1.0) Amplitude for excit PSC (nS)
    syntype  1         # 1-excit, 2-inhib
  <>

  <mech>  # Incoming inhibitory spikes
    name in
    type  psg_alpha    # Mechanism type
    tau  0.004         # Tau for inhib PSC (sec)
    amp  4.00          # (4.0) Amplitude for inhib PSC (nS)
    syntype  2         # 1-excit, 2-inhib
  <>

  <mech>  ### Was lgn2ctx...
    name lgn_ex
    type  psg_doe    # Mechanism type

    ### EPSG from LGN to cortex, Diff Of Exp

    tau_r  1.0   # (ms) rise time constant
    tau_f  4.0   # (ms) fall time constant
    amp    1.0   # was 1.0 (nS) amplitude [Troyer: 0.89nS for 3 connections]
    sdf    1.0   # synaptic depression, spike ampl after spike
    sdtau  99.0  # (ms) synaptic depression, recovery time constant
  <>

  <input>
    type       bg     # Background input spikes
    ex_rate     0.0   # spikes/s
    ex_amp      1.0   # relative to _psc_amp_ex
    in_rate   200.0   # spikes/s
    in_amp      6.0   # relative to _psc_amp_in
  </input>

  <input>  # Ex layer input connections (from LGN)
    name lgn
    type lgn_on_off
    receptor lgn_ex

    pop_origin lgn

    sf       2.00  # (2.00) was 1.0 (cyc/deg) SF
    sd_orth  0.125 # (0.125) (deg) Gaussian SD for Gabor RF, orthog to ori
    sd_par   0.215 # (0.215) (deg) Gaussian SD for Gabor RF, parallel to ori
    phseed 485401  # for chosing phase (random in 1 of 4 phases)
    seed  1777     # for chosing DOG sampling
    nsamp   30     # (50) number of samples to select as inputs
    maxrep   1     # number of times any one connection can be sampled
    eps      0.03  # no connections where profile < this frac of max

    phtype   1     # (0) 0-const, 1-regular phase steps, 4-regular quads
    ph0      0.0   # (degr) initial phase
    phstep  90.0   # (degr) phase step
    balance  0     # Balance the # of ON and OFF connections
  </input>

  <input>  ### i2e inputs
    type regular        #
    pop_origin in       # Where it comes from
    receptor   in       # Where it lands

    <distrib>
      type corr_lgn         # Correlation
      corr_sign -1          # 1-correlated, -1-anticorrelated
      algorithm 1           # run new picking algorithm, 0 calls old
      minw      0.5         # minimum weight
      cdist    500.0 (um)   # Cutoff distance (probably need to mult x 2 amp)

      normw     10.0        # (12,30) 39.0       # Based on cell center
    </distrib>
  </input>

  <spike_gen>
    type         ifc   #
    v_spike     10.0   # (mV) Spike height
    v_th_x     -52.5   # (mV) Spike threshold [TKPM98 -52.5]
    v_reset_x  -56.5   # (mV) Spike reset voltage [TKPM98 -56.5]
    tau_r_ad     1.00  # (ms) Adaptation rise time [TKPM98 1.0]
    tau_f_ad    83.3   # (ms) Adaptation fall time [TKPM98 83.3]
    gbar_ad      0.0   # (nS) Adaptation conductance [TKPM98 3.0]
    v_ex         0.0   # (mV) Excitatory reversal [TKPM98 0]
    v_in       -70.0   # (mV) Inhibitory reversal [TKPM98 -70]
    v_ad       -90.0   # (mV) Adaptation reversal [TKPM98 -90]
    v_leak_x   -73.6   # (mV) Leakage reversal [TKPM98 -73.6]
    g_leak_x    25.0   # (nS) Leakage conductance [TKPM98 25]
    c_x        500.0   # (pF) Membrane capacitance [TKPM98 500]
    trefr_x      2.5   # (ms) Refractory period
    gx_scale     5.0   # (3.5) (  ) Scale excit g, before adding bias
    gx_bias      0.0   # (0.0) (nS) Constant added to g, after any scaling
    <gx_noise>
      type   gfg       #  gfg = Gaussian filtered Gaussian noise
      mean   0.0 (nS)  #  mean
      sd     2.0 (nS)  #  standard deviation
      tsd    1.0 (ms)  #  temporal SD
    <>
    grect        1     # half-wave rectify gx, gi after adding noise

  </spike_gen>
</pop>

#--------------------------------------.--------------------------------------#
#                                                                             #
#                                      DS                                     #
#                                                                             #
#-----------------------------------------------------------------------------#
<pop>
  name ds
  area v1
  <icon>
    shape triangle
    r 0.8
    g 0.8
    b 0.2
  </icon>

  map_ori       orimap       # Set orientations using the named map
  map_dir_flag  ori          # for SI02

  <geometry>
    xn   2            # Layer dimensions
    yn   2
    x0   5
    y0   5
  </geometry>

  <mech>  # Incoming excitatory spikes
    name ex
    type  psg_alpha    # Mechanism type
    tau  0.003         # Tau for excit PSC (sec)
    amp  1.0           # Amplitude for excit PSC (nS)
    syntype  1         # 1-excit, 2-inhib
  <>

  <mech>  # Incoming inhibitory spikes
    name in
    type  psg_alpha    # Mechanism type
    tau  0.004         # Tau for inhib PSC (sec)
    amp  4.0           # Amplitude for inhib PSC (nS)
    syntype  2         # 1-excit, 2-inhib
  <>

  ##
  ##  Synaptic Interaction Mechanism 'SI_DS02'
  ##
  ##  This mechanism will convert each synaptic input (i1) to a pair of
  ##  two interacting synaptic inputs (i1 i2).  The input i2 added to form
  ##  the pair must be at the same x,y location as i1 and have an appropriate
  ##  spatial phase shift of its RF with respect to i1.  The phase shift is
  ##  controled by 'ph_shift' and 'ph_dev'.
  ##  
  ##  The interaction will involve setting up a temporal mask following
  ##  spikes from i2 that will be used to modulate the conductance 
  ##  induced by spikes from i1.
  ##  The parameter 'dt' controls the timing of the peak of the positive
  ##  interaction, and 'masktype' determines the shape of the mask function.
  ##

  <mech>  # Direction selective mechanism
    name  ds02
    type  si_ds02       #  Synaptic interaction

    dt          0.020   #  Peak interaction time (sec)
    masktype    2       #  1-sinusoid, 2-maxwell

    mask_amp    1.0     # [1.0 default] mask amplitude
    mask_comp   prod    # ['prod'] 'prod', 'prod1m'= product 1-mask

    ph_shift   90.0     # Target phase shift between cells, -180 to 180 [90.0]
    ph_dev      5.0     # Deviation around phase shift [Broken near 180]
  <>

  ##
  ##  DS Inputs
  ##
  <input>
    type regular        #
    pop_origin exs      # Where it comes from
    receptor   ex       # Where it lands

    syn_int ds02        # Synaptic interaction

    ##  
    ##  This controls the distribution across the population of the
    ##  inputs.  'cdist' is the SD of a Gaussian over distance in the
    ##  cortex, and 'cori' is the SD of a Gaussian in orientation.
    ##  
    <distrib>
      type   ori_dist
      cdist    58.0    # (um) crit connection distance (SD)
      cori     30.0    # (degr) crit orientation difference (max)

      ori_mean_flag 0  # 0-self, 1-offset, 2-fixed absolute [0]
      cori_offset   0  # Only used if 'ori_mean_flag' is 1

      minw      0.05   # () suppress connections with lower weight
      normw    50.0    # () normalize weights to this many unitary
                       #  0.0 - no normalization
      seed    32510    ########
      prob     0.35
    </distrib>
  </input>


  <input>
    type       bg      # Background input spikes
    ex_rate    150.0   # spikes/s
    ex_amp       5.0   # relative to _psc_amp_ex
    in_rate    200.0   # spikes/s
    in_amp       2.0   # relative to _psc_amp_in
  </input>


  <spike_gen ID=ds_ifc>
    type         ifc   #
    v_spike     10.0   # (mV) Spike height
    v_th_x     -52.5   # (mV) Spike threshold [TKPM98 -52.5]
    v_reset_x  -56.5   # (mV) Spike reset voltage [TKPM98 -56.5]
    tau_r_ad     1.00  # (ms) Adaptation rise time [TKPM98 1.0]
    tau_f_ad    83.3   # (ms) Adaptation fall time [TKPM98 83.3]
    gbar_ad      0.0   # (nS) Adaptation conductance [TKPM98 3.0]
    v_ex         0.0   # (mV) Excitatory reversal [TKPM98 0]
    v_in       -70.0   # (mV) Inhibitory reversal [TKPM98 -70]
    v_ad       -90.0   # (mV) Adaptation reversal [TKPM98 -90]
    v_leak_x   -73.6   # (mV) Leakage reversal [TKPM98 -73.6]
    g_leak_x    25.0   # (nS) Leakage conductance [TKPM98 25]
    c_x        500.0   # (pF) Membrane capacitance [TKPM98 500]
    trefr_x      2.5   # (ms) Refractory period
    gx_scale     3.5   # (  ) Scale excit g, before adding bias      (only LGN)
    gx_bias      0.0   # (nS) Constant added to g, after any scaling (only LGN)
    <gx_noise>
      type   gfg       #  gfg = Gaussian filtered Gaussian noise
      mean   0.0 (nS)  #  mean
      sd     2.0 (nS)  #  standard deviation
      tsd    1.0 (ms)  #  temporal SD
    <>
    grect        1     # half-wave rectify gx, gi after adding noise

  </spike_gen>
</pop>

#--------------------------------------.--------------------------------------#
#                                                                             #
#                                     LGN                                     #
#                                                                             #
#  DoG.000.moo                                                                #
#                                                                             #
#-----------------------------------------------------------------------------#
<pop>
  name lgn
  type lgn            # Layer type
  area lgn
  <icon>
    shape cs
    r 1.0
    g 1.0
    b 1.0
    r1 0.5
    g1 0.5
    b1 0.5
  </icon>

  <geometry>
    zn    2           # (1) Depth of layer
  </geometry>

  binocular   0       # 0-monocular, 1-binocular
  mesh_flag   0       # (0) use mesh if 1

  sig1        0.05    # SD Center (deg)
  sig2        0.25    # SD Surr (deg)
  amp1        4.00    # Amp Center
  amp2        0.10    # Amp Surr
  tsig       25.00    # SD Gaussian to multiply temporal filter (ms)
  tab_k     200.00    # ABTemporal filter 'k'
  tab_n       3.00    # ABTemporal filter 'n'
  cs_delay_s  0.008   # Delay of surround w.r.t. center (s)
  tdelay      0.010   # Time delay, to simulate latency (s)

  write_dog_temporal  NULL  #  Set a name to do the dump [NULL]
  ifc_dump_center  0        # IFC dump for middle unit

  <spike_gen>
    type         ifc   # Coductance based I&F
    v_spike     10.0   # (mV) Spike height
    v_th_x     -52.5   # (mV) Spike threshold
    v_reset_x  -56.5   # (mV) Spike reset voltage
    tau_r_ad     1.00  # (ms) Adaptation rise time
    tau_f_ad    80.0   # (ms) Adaptation fall time (fast avoids biphasic onset)
    gbar_ad      0.0   # (nS) Adaptation conductance (zero to avoid transient)
    v_ex         0.0   # (mV) Excitatory reversal
    v_in       -70.0   # (mV) Inhibitory reversal
    v_ad       -90.0   # (mV) Adaptation reversal
    v_leak_x   -73.6   # (mV) Leakage reversal
    g_leak_x    75.0   # (nS) Leakage conductance
    c_x        200.0   # (pF) Membrane capacitance
    trefr_x      0.5   # (ms) Refractory period
    trefr_x_sd   2.0   # (ms) Add rectified Gaussian noise to refr period
    gx_scale     3.5   # (260) Scale exc conductance, before adding bias
    gx_bias     28.5   # (0.0) (nS) Constant added to cond, after any scaling
    <gx_noise>
      type   gfg       #  gfg = Gaussian filtered Gaussian noise
      mean   0.0 (nS)  #  mean
      sd     2.0 (nS)  #  (4.0) standard deviation
      tsd    1.0 (ms)  #  (1.0) temporal SD
    <>
    grect        1     # half-wave rectify gx, gi after adding noise
    dump         0     # 
  </spike_gen>
</pop>
