/*C  (C) Copyright  Wyeth Bair 1993-2010                                     */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <X11/Xlib.h>
#include "my_util.h"
#include "nr_util.h"
#include "iarray_util.h"
#include "farray_util.h"
#include "dot_util.h"
#define IM1 2147483563
#define IM2 2147483399
#define AM (1.0/IM1)
#define IMM1 (IM1-1)
#define IA1 40014
#define IA2 40692
#define IQ1 53668
#define IQ2 52774
#define IR1 12211
#define IR2 3791
#define NTAB 32
#define NDIV (1+IMM1/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)
/*C  (C) Copr. 1986-92 Numerical Recipes Software #.3.                       */
float dot_util_ran2(idum)
     int *idum;
{
  int j;
  long k;
  static long idum2=123456789;
  static long iy=0;
  static long iv[NTAB];
  float temp;
  if (*idum <= 0) {
    if (-(*idum) < 1) *idum=1;
    else *idum = -(*idum);
    idum2=(*idum);
    for (j=NTAB+7;j>=0;j--) {
      k=(*idum)/IQ1;
      *idum=IA1*(*idum-k*IQ1)-k*IR1;
      if (*idum < 0) *idum += IM1;
      if (j < NTAB) iv[j] = *idum;
    }
    iy=iv[0];
  }
  k=(*idum)/IQ1;
  *idum=IA1*(*idum-k*IQ1)-k*IR1;
  if (*idum < 0) *idum += IM1;
  k=idum2/IQ2;
  idum2=IA2*(idum2-k*IQ2)-k*IR2;
  if (idum2 < 0) idum2 += IM2;
  j=iy/NDIV;
  iy=iv[j]-idum2;
  iv[j] = *idum;
  if (iy < 1) iy += IMM1;
  if ((temp=AM*iy) > RNMX) return RNMX;
  else return temp;
}
#undef IM1
#undef IM2
#undef AM
#undef IMM1
#undef IA1
#undef IA2
#undef IQ1
#undef IQ2
#undef IR1
#undef IR2
#undef NTAB
#undef NDIV
#undef EPS
#undef RNMX
void my_delay(n)
     long n;
{
  long i;
  float x,y;
  for(i=0;i<n;i++){
    x = 1777.0;
    y = 21.0 * 2.0/x;
    x /= y;
  }
}
struct dot_struct *dot_util_get_init_dot_struct(nframes,ndots)
     int nframes,ndots;
{
  int i;
  struct dot_struct *d;
  d = (struct dot_struct *)myalloc(sizeof(struct dot_struct));
  d->nframes = nframes;
  d->n = (int *)myalloc(nframes*sizeof(int));
  d->x = (float **)myalloc(nframes*sizeof(float *));
  d->y = (float **)myalloc(nframes*sizeof(float *));
  if (ndots >= 0){
    for(i=0;i<nframes;i++){
      d->n[i] = ndots;
      d->x[i] = (float *)myalloc(ndots*sizeof(float));
      d->y[i] = (float *)myalloc(ndots*sizeof(float));
    }
  }
  return d;
}
struct dot_struct *init_copy_dot_struct(dot)
     struct dot_struct *dot;
{
  int i;
  struct dot_struct *d;
  d = (struct dot_struct *)myalloc(sizeof(struct dot_struct));
  d->nframes = dot->nframes;
  d->n = (int *)myalloc(dot->nframes*sizeof(int));
  d->x = (float **)myalloc(dot->nframes*sizeof(float *));
  d->y = (float **)myalloc(dot->nframes*sizeof(float *));
  for(i=0;i<dot->nframes;i++){
    d->n[i] = dot->n[i];
    d->x[i] = (float *)myalloc(dot->n[i]*sizeof(float));
    d->y[i] = (float *)myalloc(dot->n[i]*sizeof(float));
  }
  return d;
}
void get_min_max_dot(d,xmin,xmax,ymin,ymax)
     struct dot_struct *d;
     float *xmin,*xmax,*ymin,*ymax;
{
  int i,j;
  int done;
  if (d->nframes <= 0)
    exit_error("GET_MIN_MAX_DOT","Frame count <= 0");
  i = 0;
  done = 0;
  while((i<d->nframes)&&(!done)){
    if (d->n[i] > 0){
      *xmin = *xmax = d->x[i][0];
      *ymin = *ymax = d->y[i][0];
      done = 1;
    }else
      i += 1;
  }
  if (i==d->nframes)
    exit_error("GET_MIN_MAX_DOT","No dots found in structure");
  for(i=0;i<d->nframes;i++)
    for(j=0;j<d->n[i];j++){
      if (d->x[i][j] < *xmin)
	*xmin = d->x[i][j];
      if (d->x[i][j] > *xmax)
	*xmax = d->x[i][j];
      if (d->y[i][j] < *ymin)
	*ymin = d->y[i][j];
      if (d->y[i][j] > *ymax)
	*ymax = d->y[i][j];
    }
}
void normalize_dot(d)
     struct dot_struct *d;
{
  int i,j;
  float xmin,ymin,xmax,ymax;
  get_min_max_dot(d,&xmin,&xmax,&ymin,&ymax);
  for(i=0;i<d->nframes;i++)
    for(j=0;j<d->n[i];j++){
      d->x[i][j] -= xmin;
      d->y[i][j] -= ymin;
    }
}
struct dot_struct *read_dots(infile)
     char infile[];
{
  FILE *fopen(),*fin;
  int i,j;
  int tn,tx,ty,nframes,ns;
  struct dot_struct *dot;
  printf("  READ_DOTS\n");
  dot = (struct dot_struct *)myalloc(sizeof(struct dot_struct));
  fin = fopen(infile,"r"); 
  nframes = 0;
  while (fscanf(fin,"%d",&tn) != EOF){
    for (i=0;i<tn;i++)
      ns = fscanf(fin,"%d %d",&tx,&ty);
    nframes += 1;
  }
  fclose(fin);
  printf("    %d frames found.\n",nframes);
  dot->nframes = nframes;
  dot->n = (int *)myalloc(dot->nframes*sizeof(int));
  dot->x = (float **)myalloc(dot->nframes*sizeof(float *));
  dot->y = (float **)myalloc(dot->nframes*sizeof(float *));
  fin = fopen(infile,"r"); 
  for (i=0;i<dot->nframes;i++){
    ns = fscanf(fin,"%d",&(dot->n[i]));
    if (dot->n[i] > 0){
      dot->x[i] = (float *)myalloc(dot->n[i]*sizeof(float));
      dot->y[i] = (float *)myalloc(dot->n[i]*sizeof(float));
      for (j=0;j<dot->n[i];j++)
	ns = fscanf(fin,"%f %f",&(dot->x[i][j]),&(dot->y[i][j]));
    }else{
      dot->x[i] = NULL;
      dot->y[i] = NULL;
    }
  }
  fclose(fin);
  return dot;
}
float ***get_3space_dots(dot,xs,xn,ys,yn,ts,tn)
     struct dot_struct *dot;
     int xs,xn,ys,yn,ts,tn; 
{
  int i,j,k;
  int x,y;
  float ***data;
  data = get_3d_farray(xn,yn,tn);
  for(i=0;i<xn;i++)
    for(j=0;j<yn;j++)
      for(k=0;k<tn;k++)
	data[i][j][k] = 0.0;
  for(i=ts;i<ts+tn;i++)
    for(j=0;j<dot->n[i];j++){
      x = (int)dot->x[i][j] - xs;
      y = (int)dot->y[i][j] - ys;
      if ((x >= 0) && (x < xn) && (y >= 0) && (y < yn)){
	data[x][y][i] = 1.0;
      }
    }
  return data;
}
float ***get_3space_xyt_dots(dotx,doty,dott,dotn,xs,xn,ys,yn,ts,tn)
     int *dotx,*doty,*dott,dotn;
     int xs,xn,ys,yn,ts,tn; 
{
  int i,j,k;
  int x,y,t;
  float ***data;
  data = f3tensor(1,xn,1,yn,1,tn); 
  for(i=1;i<=xn;i++)
    for(j=1;j<=yn;j++)
      for(k=1;k<=tn;k++)
	data[i][j][k] = 0.0;
  for(i=0;i<dotn;i++){
    x = dotx[i] - xs + 1;
    y = doty[i] - ys + 1;
    t = dott[i] - ts + 1;
    if ((x>0) && (x<=xn) && (y>0) && (y<=yn) && (t>0) && (t<=tn))
      data[x][y][t] = 1.0;
  }
  return data;
}
float ***get_3space_xyt_random_dots(xn,yn,tn,ndots,dx,dy,dt,seed)
     int xn,yn,tn; 
     int ndots,dx,dy,dt;
     int *seed;
{
  int i,j,k;
  int x,y,t,x1,y1,t1;
  float ***data;
  data = f3tensor(1,xn,1,yn,1,tn); 
  for(i=1;i<=xn;i++)
    for(j=1;j<=yn;j++)
      for(k=1;k<=tn;k++)
	data[i][j][k] = 0.0;
  if (*seed > 0)
    *seed *= -1;
  dot_util_ran2(seed);
  for(i=0;i<ndots;i++){
    x = 1 + (int)(xn * dot_util_ran2(seed));
    y = 1 + (int)(yn * dot_util_ran2(seed));
    t = 1 + (int)(tn * dot_util_ran2(seed));
    data[x][y][t] = 1.0;
    x1 = x + dx;
    y1 = y + dy;
    t1 = t + dt;
    if ((x1>0) && (x1<=xn) && (y1>0) && (y1<=yn) && (t1>0) && (t1<=tn))
      data[x1][y1][t1] = 1.0;
  }
  return data;
}
void write_dots(outfile,x,y,t,n)
     char outfile[];
     int *x,*y,*t,n;
{
  int i;
  FILE *fopen(),*fout;
  fout = fopen(outfile,"w");
  for(i=0;i<n;i++)
    fprintf(fout,"%4d %4d %4d\n",x[i],y[i],t[i]);
  fclose(fout);
}
void get_random_dots(xn,yn,tn,dens,sscale,tscale,rx,ry,rt,rn,seed)
     int xn,yn,tn;
     float dens; 
     float sscale,tscale;
     int **rx,**ry,**rt,*rn;
     int *seed;
{
  int i;
  int *x,*y,*t,n;
  n = dens * (float)xn*sscale*(float)yn*sscale*(float)tn*tscale;
  printf("  Making %d dots.\n",n);
  x = (int *)myalloc(n*sizeof(int));
  y = (int *)myalloc(n*sizeof(int));
  t = (int *)myalloc(n*sizeof(int));
  if (*seed > 0)
    *seed *= -1;
  dot_util_ran2(seed);
  for(i=0;i<n;i++){
    x[i] = (int)(xn * dot_util_ran2(seed));
    y[i] = (int)(yn * dot_util_ran2(seed));
    t[i] = (int)(tn * dot_util_ran2(seed));
  }
  *rx = x;
  *ry = y;
  *rt = t;
  *rn = n;
}
struct dot_struct *dot_util_get_coherence_frame(sscale,framerate,coher,speed,
					       theta,diam,nframes,dpf,dt,tseed)
     float sscale,framerate,coher,speed,theta,diam;
     int nframes,dpf,dt,tseed;
{
  int i,j,k;
  int n,seed;
  float x,y,dx,dy,size;
  struct dot_struct *ds;
  ds = dot_util_get_init_dot_struct(nframes,dpf);
  seed = tseed;
  if (seed > 0)
    seed *= -1;
  size = diam / sscale; 
  dy = speed * (float)dt/framerate / sscale * sin(theta/180.0*M_PI);
  dx = speed * (float)dt/framerate / sscale * cos(theta/180.0*M_PI);
  for (i=0;i<nframes;i++){
    n = 0; 
    k = i - dt; 
    if (k >= 0){ 
      for (j=0;j<dpf;j++) 
	if (dot_util_ran2(&seed) < coher){ 
	  x = ds->x[k][j] + dx;
	  y = ds->y[k][j] + dy;
	  if (x < 0.0) x += size; else if (x >= size) x -= size;
	  if (y < 0.0) y += size; else if (y >= size) y -= size;
	  ds->x[i][n] = x; 
	  ds->y[i][n] = y;
	  n += 1;
	}
    }
    while (n < dpf){ 
      ds->x[i][n] = dot_util_ran2(&seed)*size;
      ds->y[i][n] = dot_util_ran2(&seed)*size;
      n += 1;
    }
  }
  return ds;
}
void make_newsome_dots(sscale,tscale,coher,speed,theta,diam,period,dens,dt,
		       tseed,rx,ry,rt,rn)
     float sscale,tscale,coher,speed,theta,diam,period,dens,dt;
     int tseed;
     int **rx,**ry,**rt,*rn;
{
  int i,k;
  int *x,*y,*t,ndots,nmax;
  double dx,dy,xx,yy,tt,width,height,r;
  int seed;
  seed = tseed;
  width = height = 20.0; 
  if ((dt<0.0)&&(dens<0.0)){
    ndots = 300; 
    dt = 0.045; 
    dens = (float)ndots/dt/width/height; 
  }else{
    ndots = 0;
    exit_error("MAKE_NEWSOME_DOTS","Other densities, dt's, not available");
  }
  if ((diam>width)||(diam>height))
    exit_error("MAKE_NEWSOME_DOTS","Diameter too large");
  dy = speed * dt * sin(theta/180.0*M_PI);
  dx = speed * dt * cos(theta/180.0*M_PI);
  nmax = ndots * (period/dt + 1.0);
  x = (int *)myalloc(nmax*sizeof(int));
  y = (int *)myalloc(nmax*sizeof(int));
  t = (int *)myalloc(nmax*sizeof(int));
  if (seed > 0)
    seed *= -1;
  dot_util_ran2(&seed);
  k = 0;
  for(i=0;i<ndots;i++){
    tt = dt * (float)i/(float)ndots; 
    xx = width * dot_util_ran2(&seed) - width/2.0; 
    yy = height * dot_util_ran2(&seed) - height/2.0; 
    while (tt<period){
      r = sqrt(xx*xx + yy*yy);
      if (r <= diam/2.0){ 
	x[k] = (int)(xx/sscale + 0.5);
	y[k] = (int)(yy/sscale + 0.5);
	t[k] = (int)(tt/tscale + 0.5);
	k += 1;
      }
      if (dot_util_ran2(&seed) < coher){
	xx += dx; yy += dy;
	if (xx >= width/2.0) xx -= width;
	if (xx < -width/2.0) xx += width;
	if (yy >= height/2.0) yy -= height;
	if (yy < -height/2.0) yy += height;
      }else{
	xx = width * dot_util_ran2(&seed) - width/2.0; 
	yy = height * dot_util_ran2(&seed) - height/2.0; 
      }
      tt += dt;
    }
  }
  *rx = x; *ry = y; *rt = t; *rn = k;
}
void make_newsome_dots_float(sscale,tscale,coher,speed,theta,diam,period,dens,
			     dt,tseed,rx,ry,rt,rn)
     float sscale,tscale,coher,speed,theta,diam,period,dens,dt;
     int tseed;
     float **rx,**ry,**rt;
     int *rn;
{
  int i,k;
  int ndots,nmax;
  float *x,*y,*t;
  double dx,dy,xx,yy,tt,width,height,r;
  int seed;
  seed = tseed;
  width = height = 20.0; 
  if ((dt<0.0)&&(dens<0.0)){
    ndots = 300; 
    dt = 0.045; 
    dens = (float)ndots/dt/width/height; 
  }else{
    ndots = 0;
    exit_error("MAKE_NEWSOME_DOTS","Other densities, dt's, not available");
  }
  if ((diam>width)||(diam>height))
    exit_error("MAKE_NEWSOME_DOTS","Diameter too large");
  dy = speed * dt * sin(theta/180.0*M_PI);
  dx = speed * dt * cos(theta/180.0*M_PI);
  nmax = ndots * (period/dt + 1.0);
  x = (float *)myalloc(nmax*sizeof(float));
  y = (float *)myalloc(nmax*sizeof(float));
  t = (float *)myalloc(nmax*sizeof(float));
  if (seed > 0)
    seed *= -1;
  k = 0;
  for(i=0;i<ndots;i++){
    tt = dt * (float)i/(float)ndots; 
    xx = width * dot_util_ran2(&seed) - width/2.0; 
    yy = height * dot_util_ran2(&seed) - height/2.0; 
    while (tt<period){
      r = sqrt(xx*xx + yy*yy);
      if (r <= diam/2.0){ 
	x[k] = xx/sscale;
	y[k] = yy/sscale;
	t[k] = tt/tscale;
	k += 1;
      }
      if (dot_util_ran2(&seed) < coher){
	xx += dx; yy += dy;
	if (xx >= width/2.0) xx -= width;
	if (xx < -width/2.0) xx += width;
	if (yy >= height/2.0) yy -= height;
	if (yy < -height/2.0) yy += height;
      }else{
	xx = width * dot_util_ran2(&seed) - width/2.0; 
	yy = height * dot_util_ran2(&seed) - height/2.0; 
      }
      tt += dt;
    }
  }
  *rx = x; *ry = y; *rt = t; *rn = k;
}
void get_translating_dots(xn,yn,tn,dens,sscale,tscale,speed,theta,dt,
			  rx,ry,rt,rn,seed)
     int xn,yn,tn;
     float dens; 
     float sscale,tscale,speed,theta,dt;
     int **rx,**ry,**rt,*rn;
     int *seed;
{
  int i,j,k;
  int *x,*y,*t,n,nn,nplots;
  float dx,dy,xx,yy,tt,dtt;
  dy = (speed * dt * sin(theta/180.0*M_PI))/sscale;
  dx = (speed * dt * cos(theta/180.0*M_PI))/sscale;
  dtt = dt/tscale;
  nn = dens * (float)xn*sscale*(float)yn*sscale;
  nplots = (int)((float)tn*tscale/dt);
  n = nn * nplots;
  printf("  Making %d dots.\n",n);
  x = (int *)myalloc(n*sizeof(int));
  y = (int *)myalloc(n*sizeof(int));
  t = (int *)myalloc(n*sizeof(int));
  if (*seed > 0)
    *seed *= -1;
  dot_util_ran2(seed);
  k = 0;
  for(i=0;i<nn;i++){
    xx = xn * dot_util_ran2(seed);
    yy = yn * dot_util_ran2(seed);
    tt = dtt * dot_util_ran2(seed); 
    x[k] = (int)xx;
    y[k] = (int)yy;
    t[k] = (int)tt;
    k += 1;
    for(j=1;j<nplots;j++){
      xx += dx; yy += dy; tt += dtt;
      if (xx >= (float)xn) xx -= (float)xn;
      if (xx < 0.0) xx += (float)xn;
      if (yy >= (float)yn) yy -= (float)yn;
      if (yy < 0.0) yy += (float)yn;
      x[k] = (int)xx;
      y[k] = (int)yy;
      t[k] = (int)tt;
      k += 1;
    }      
  }
  *rx = x;
  *ry = y;
  *rt = t;
  *rn = n;
}
void read_xyt_dots(infile,rx,ry,rt,rn,rnx,rny,rnt,rsscale,rtscale)
     char infile[];
     int **rx,**ry,**rt,*rn,*rnx,*rny,*rnt;
     float *rsscale,*rtscale;
{
  FILE *fopen(),*fin;
  int i;
  int n,*x,*y,*t,ns;
  printf("  READ_XYT_DOTS\n");
  fin = fopen(infile,"r"); 
  ns = fscanf(fin,"%*s %f",rsscale);
  ns = fscanf(fin,"%*s %f",rtscale);
  ns = fscanf(fin,"%*s %d %d %d",rnx,rny,rnt);
  n = 0;
  while (fscanf(fin,"%*d %*d %*d") != EOF)
    n += 1;
  fclose(fin);
  printf("    %d dots found.\n",n);
  x = (int *)myalloc(n*sizeof(int));
  y = (int *)myalloc(n*sizeof(int));
  t = (int *)myalloc(n*sizeof(int));
  fin = fopen(infile,"r"); 
  ns = fscanf(fin,"%*s %*f");
  ns = fscanf(fin,"%*s %*f");
  ns = fscanf(fin,"%*s %*d %*d %*d"); 
  for (i=0;i<n;i++)
    ns = fscanf(fin,"%d %d %d",&x[i],&y[i],&t[i]);
  fclose(fin);
  *rx = x; *ry = y; *rt = t; *rn = n;
}
void read_xyt_dots_float(infile,rx,ry,rt,rn,rnx,rny,rnt,rsscale,rtscale)
     char infile[];
     float **rx,**ry,**rt;
     int *rn,*rnx,*rny,*rnt;
     float *rsscale,*rtscale;
{
  FILE *fopen(),*fin;
  int i;
  int n,ns;
  float *x,*y,*t;
  printf("  READ_XYT_DOTS\n");
  fin = fopen(infile,"r"); 
  ns = fscanf(fin,"%*s %f",rsscale);
  ns = fscanf(fin,"%*s %f",rtscale);
  ns = fscanf(fin,"%*s %d %d %d",rnx,rny,rnt);
  n = 0;
  while (fscanf(fin,"%*f %*f %*f") != EOF)
    n += 1;
  fclose(fin);
  printf("    %d dots found.\n",n);
  x = (float *)myalloc(n*sizeof(float));
  y = (float *)myalloc(n*sizeof(float));
  t = (float *)myalloc(n*sizeof(float));
  fin = fopen(infile,"r"); 
  ns = fscanf(fin,"%*s %*f");
  ns = fscanf(fin,"%*s %*f");
  ns = fscanf(fin,"%*s %*d %*d %*d"); 
  for (i=0;i<n;i++)
    ns = fscanf(fin,"%f %f %f",&x[i],&y[i],&t[i]);
  fclose(fin);
  *rx = x; *ry = y; *rt = t; *rn = n;
}
void write_xyt_dots(outfile,x,y,t,n,nx,ny,nt,sscale,tscale)
     char outfile[];
     int *x,*y,*t,n,nx,ny,nt;
     float sscale,tscale;
{
  FILE *fopen(),*fout;
  int i;
  printf("  WRTE_XYT_DOTS\n");
  fout = fopen(outfile,"w"); 
  fprintf(fout,"sscale %.6f\n",sscale);
  fprintf(fout,"tscale %.6f\n",tscale);
  fprintf(fout,"xyt_limits %d %d %d\n",nx,ny,nt);
  for (i=0;i<n;i++)
    fprintf(fout,"%4d %4d %4d\n",x[i],y[i],t[i]);
  fclose(fout);
}
void write_xyt_dots_float(outfile,x,y,t,n,nx,ny,nt,sscale,tscale)
     char outfile[];
     float *x,*y,*t;
     int n,nx,ny,nt;
     float sscale,tscale;
{
  FILE *fopen(),*fout;
  int i;
  printf("  WRTE_XYT_DOTS\n");
  fout = fopen(outfile,"w"); 
  fprintf(fout,"sscale %.6f\n",sscale);
  fprintf(fout,"tscale %.6f\n",tscale);
  fprintf(fout,"xyt_limits %d %d %d\n",nx,ny,nt);
  for (i=0;i<n;i++)
    fprintf(fout,"%f %f %f\n",x[i],y[i],t[i]);
  fclose(fout);
}
void translate_xyt_dots(dotx,doty,dott,n,tx,ty,tt)
     int *dotx,*doty,*dott,n;
     int tx,ty,tt;
{
  int i;
  for(i=0;i<n;i++){
    dotx[i] += tx;
    doty[i] += ty;
    dott[i] += tt;
  }
}
void translate_xyt_dots_float(dotx,doty,dott,n,tx,ty,tt)
     float *dotx,*doty,*dott;
     int n;
     float tx,ty,tt;
{
  int i;
  for(i=0;i<n;i++){
    dotx[i] += tx;
    doty[i] += ty;
    dott[i] += tt;
  }
}
void rescale_xyt_dots(dotx,doty,dott,n,fsscale,ftscale)
     int *dotx,*doty,*dott,n;
     float fsscale,ftscale;
{
  int i;
  for(i=0;i<n;i++){
    dotx[i] = (int)rint((float)dotx[i]*fsscale);
    doty[i] = (int)rint((float)doty[i]*fsscale);
    dott[i] = (int)rint((float)dott[i]*ftscale);
  }
}
void get_rescale_xyt_dots(dotx,doty,dott,n,fsscale,ftscale,rdotx,rdoty,rdott)
     int *dotx,*doty,*dott,n;
     float fsscale,ftscale;
     int **rdotx,**rdoty,**rdott;
{
  int i;
  int *dx,*dy,*dt;
  dx = (int *)myalloc(n*sizeof(int));
  dy = (int *)myalloc(n*sizeof(int));
  dt = (int *)myalloc(n*sizeof(int));
  for(i=0;i<n;i++){
    dx[i] = (int)rint((float)dotx[i]*fsscale);
    dy[i] = (int)rint((float)doty[i]*fsscale);
    dt[i] = (int)rint((float)dott[i]*ftscale);
  }
  *rdotx = dx; *rdoty = dy; *rdott = dt;
}
void rotate_xyt_dots(dotx,doty,dott,n,x0,y0,theta)
     int *dotx,*doty,*dott,n;
     float x0,y0,theta; 
{
  int i;
  float x,y,a,b,c,d;
  a =  cos(M_PI/180.0*theta); b = sin(M_PI/180.0*theta); 
  c = -sin(M_PI/180.0*theta); d = cos(M_PI/180.0*theta);
  for(i=0;i<n;i++){
    x = (float)dotx[i] - x0;
    y = (float)doty[i] - y0;
    dotx[i] = (int)rint(a*x + c*y + x0);
    doty[i] = (int)rint(b*x + d*y + y0);
  }
}
void get_rotate_xyt_dots(dotx,doty,dott,n,x0,y0,theta,rdotx,rdoty)
     int *dotx,*doty,*dott,n;
     float x0,y0,theta; 
     int **rdotx,**rdoty;
{
  int i;
  int *dx,*dy;
  float x,y,a,b,c,d;
  a =  cos(M_PI/180.0*theta); b = sin(M_PI/180.0*theta); 
  c = -sin(M_PI/180.0*theta); d = cos(M_PI/180.0*theta);
  dx = (int *)myalloc(n*sizeof(int));
  dy = (int *)myalloc(n*sizeof(int));
  for(i=0;i<n;i++){
    x  = (float)dotx[i] - x0;
    y  = (float)doty[i] - y0;
    dx[i] = (int)rint(a*x + c*y + x0);
    dy[i] = (int)rint(b*x + d*y + y0);
  }
  *rdotx = dx;
  *rdoty = dy;
}
void get_scale_rotate_array(sarray,tarray,darray,sn,tn,dn,sscale,tscale,
			    dotx,doty,dott,dotn,rsx,rsy,rst)
     float *sarray,*tarray,*darray; 
     int sn,tn,dn; 
     float sscale,tscale; 
     int *dotx,*doty,*dott,dotn;
     int ****rsx,****rsy,***rst;
{
  int i,j;
  int ***sx,***sy,**st; 
  int *stx,*sty;
  float f;
  st = (int **)myalloc(tn*sizeof(int *));
  for(i=0;i<tn;i++){
    f = tscale/tarray[i];
    st[i] = multiply_iarray(dott,dotn,f);
  }
  sx = get_3d_iarray(sn,dn,-1);
  sy = get_3d_iarray(sn,dn,-1);
  for(i=0;i<sn;i++){
    f = sscale/sarray[i];
    stx = multiply_iarray(dotx,dotn,f);
    sty = multiply_iarray(doty,dotn,f);
    for(j=0;j<dn;j++)
      get_rotate_xyt_dots(stx,sty,dott,dotn,0.0,0.0,darray[j],&sx[i][j],
			  &sy[i][j]);
    myfree(stx); myfree(sty);
  }
  *rsx = sx; *rsy = sy; *rst = st;
}
void convert_xyt_to_frames(x,y,t,n,nx,ny,nt,rfx,rfy,rfcnt,nframes)
     int *x,*y,*t,n,nx,ny,nt;
     int ***rfx,***rfy,**rfcnt,nframes;
{
  int i,k;
  int *fcnt,*tcnt,**fx,**fy; 
  printf("  CONVERT_XYT_TO_FRAMES\n");
  fx = (int **)myalloc(nframes*sizeof(int *));
  fy = (int **)myalloc(nframes*sizeof(int *));
  fcnt = get_zero_iarray(nframes);
  for (i=0;i<n;i++){ 
    k = (int)((float)t[i]/(float)nt * (float)nframes);
    if ((k < 0)||(k >= nframes)){
      printf("k=%d  t[i]=%d  nt=%d  nframes=%d\n",k,t[i],nt,nframes);
      exit_error("CONVERT_XYT_TO_FRAMES","Frame index out of bounds");
    }
    fcnt[k] += 1;
  }
  for (i=0;i<nframes;i++){
    fx[i] = (int *)myalloc(fcnt[i]*sizeof(int));
    fy[i] = (int *)myalloc(fcnt[i]*sizeof(int));
  }
  tcnt = get_zero_iarray(nframes); 
  for (i=0;i<n;i++){ 
    k = (int)((float)t[i]/(float)nt * (float)nframes);
    fx[k][tcnt[k]] = x[i];
    fy[k][tcnt[k]] = y[i];
    tcnt[k] += 1;
  }
  myfree(tcnt);
  *rfx = fx; *rfy = fy; *rfcnt = fcnt;
}
void convert_frames_to_xyt(ds,rdotx,rdoty,rdott,rdotn)
     struct dot_struct *ds;
     int **rdotx,**rdoty,**rdott,*rdotn;
{
  int i,j,k;
  int n;
  printf("  CONVERT_FRAMES_TO_XYT\n");
  n = 0;
  for(i=0;i<ds->nframes;i++)
    n += ds->n[i];
  printf("    Total dots = %d\n",n);
  *rdotx = (int *)myalloc(n*sizeof(int));
  *rdoty = (int *)myalloc(n*sizeof(int));
  *rdott = (int *)myalloc(n*sizeof(int));
  k = 0;
  for(i=0;i<ds->nframes;i++)
    for(j=0;j<ds->n[i];j++){
      (*rdotx)[k] = ds->x[i][j];
      (*rdoty)[k] = ds->y[i][j];
      (*rdott)[k] = i;
      k += 1;
    }
  *rdotn = n;
}
void convert_frames_to_xytf(ds,rdotx,rdoty,rdott,rdotn)
     struct dot_struct *ds;
     float **rdotx,**rdoty,**rdott;
     int *rdotn;
{
  int i,j,k;
  int n;
  printf("  CONVERT_FRAMES_TO_XYTF\n");
  n = 0;
  for(i=0;i<ds->nframes;i++)
    n += ds->n[i];
  *rdotx = (float *)myalloc(n*sizeof(float));
  *rdoty = (float *)myalloc(n*sizeof(float));
  *rdott = (float *)myalloc(n*sizeof(float));
  k = 0;
  for(i=0;i<ds->nframes;i++)
    for(j=0;j<ds->n[i];j++){
      (*rdotx)[k] = ds->x[i][j];
      (*rdoty)[k] = ds->y[i][j];
      (*rdott)[k] = (float)i;
      k += 1;
    }
  *rdotn = n;
}
void draw_xwin_dots(delay,display,win,gc,gc_erase,fx,fy,fcnt,nframes,height,
		    sflag,eflag)
     float delay;    
     Display *display;
     Window win;
     GC gc,gc_erase;
     int **fx,**fy,*fcnt,nframes,height,sflag,eflag;
{
  int i,j;
  i=0;
  for (j=0;j<fcnt[i];j++){ 
    XDrawPoint(display,win,gc,fx[i][j],height-fy[i][j]);
    if (sflag == 1){
      XDrawPoint(display,win,gc,fx[i][j]+1,height-fy[i][j]);
      XDrawPoint(display,win,gc,fx[i][j]  ,height-fy[i][j]+1);
      XDrawPoint(display,win,gc,fx[i][j]+1,height-fy[i][j]+1);
    }
  }
  XFlush(display);
  my_delay((long)(delay*10000.0));
  for (i=1;i<nframes;i++){
    if (eflag == 0){
      for (j=0;j<fcnt[i-1];j++){ 
	XDrawPoint(display,win,gc_erase,fx[i-1][j],height-fy[i-1][j]);
	if (sflag == 1){
	  XDrawPoint(display,win,gc_erase,fx[i-1][j]+1,height-fy[i-1][j]);
	  XDrawPoint(display,win,gc_erase,fx[i-1][j]  ,height-fy[i-1][j]+1);
	  XDrawPoint(display,win,gc_erase,fx[i-1][j]+1,height-fy[i-1][j]+1);
	}
      }
    }
    for (j=0;j<fcnt[i];j++){ 
      XDrawPoint(display,win,gc,fx[i][j],height-fy[i][j]);
      if (sflag == 1){
	XDrawPoint(display,win,gc,fx[i][j]+1,height-fy[i][j]);
	XDrawPoint(display,win,gc,fx[i][j]  ,height-fy[i][j]+1);
	XDrawPoint(display,win,gc,fx[i][j]+1,height-fy[i][j]+1);
      }
    }
    XFlush(display);
    my_delay((long)(delay*10000.0));
  }
  if (eflag == 0){
    for (j=0;j<fcnt[i-1];j++){ 
      XDrawPoint(display,win,gc_erase,fx[i-1][j],height-fy[i-1][j]);
      if (sflag == 1){
	XDrawPoint(display,win,gc_erase,fx[i-1][j]+1,height-fy[i-1][j]);
	XDrawPoint(display,win,gc_erase,fx[i-1][j]  ,height-fy[i-1][j]+1);
	XDrawPoint(display,win,gc_erase,fx[i-1][j]+1,height-fy[i-1][j]+1);
      }
    }
  }
}
