/*C  (C) Copyright  Wyeth Bair 1993-2010                                     */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "my_util.h"
#include "iarray_util.h"
#include "carray_util.h"
#include "farray_util.h"
void read_farray_plot_file(infile,rxdata,rydata,rzdata,rname,m,n)
     char infile[];
     float ***rxdata,***rydata,***rzdata;
     char ***rname;
     int *m,**n; 
{
  int i,j,k;
  int flag,plot_count,*len,tlen,min,max;
  FILE *fopen(),*fin;
  char line[SLEN],t1[SLEN],t2[SLEN],**name;
  float **xdata,**ydata,**zdata;
  printf("  READ_FARRAY_PLOT_FILE\n");
  if ((fin = fopen(infile,"r")) != NULL){
    printf("    Reading %s.\n",infile);
    plot_count = 0;
    flag = 1;
    while(fgets(line,SLEN,fin) != NULL){
      if (count_items_in_string(line) > 0){
	if (line[0] == '/'){
	  flag = 1;
	}else{
	  if (flag == 1){
	    plot_count += 1;
	    flag = 0;
	  }
	}
      }
    }
    fclose(fin);
    printf("    %d plots found.\n",plot_count);
    len = get_zero_iarray(plot_count);
    fin = fopen(infile,"r");
    flag = 1;
    k = 0;
    tlen = 0;
    while(fgets(line,SLEN,fin) != NULL){
      if (count_items_in_string(line) > 0){
	if (line[0] == '/'){
	  flag = 1; 
	}else{
	  if (flag == 1){
	    if (k-1 >= 0){
	      len[k-1] = tlen;  
	    }
	    k += 1;             
	    flag = 0;           
	    tlen = 1;
	  }else{
	    tlen += 1;
	  }
	}
      }
    }
    fclose(fin);
    len[k-1] = tlen;
    printf(" %d LENGTH is %d\n",k-1,tlen);
    min = min_of_iarray(len,plot_count);
    max = max_of_iarray(len,plot_count);
    if (min == max)
      printf("    All plots have same length, %d.\n",min);
    else
      printf("    Plot lengths vary from %d to %d\n",min,max);
    name = get_2d_carray(plot_count,SLEN);
    xdata = (float **)myalloc(plot_count*sizeof(float *));
    ydata = (float **)myalloc(plot_count*sizeof(float *));
    zdata = (float **)myalloc(plot_count*sizeof(float *));
    for(i=0;i<plot_count;i++){
      xdata[i] = (float *)myalloc(len[i]*sizeof(float));
      ydata[i] = (float *)myalloc(len[i]*sizeof(float));
      zdata[i] = (float *)myalloc(len[i]*sizeof(float));
      name[i][0] = '\0';
    }
    fin = fopen(infile,"r");
    flag = 1;
    j = -1;
    k = 0;
    while(fgets(line,SLEN,fin) != NULL){
      if (count_items_in_string(line) > 0){
	if (line[0] == '/'){
	  flag = 1;
	  sscanf(line,"%1s %8s",t1,t2);
	  if (strcmp(t2,"plotname")==0)
	    sscanf(line,"%1s %8s %s",t1,t2,name[j+1]);
	}else{
	  if (flag == 1){
	    j += 1;
	    k = 0;
	    flag = 0;
	  }
	  if (count_items_in_string(line) == 3)
	    sscanf(line,"%f %f %f\n",&xdata[j][k],&ydata[j][k],&zdata[j][k]);
	  else{
	    sscanf(line,"%f %f\n",&xdata[j][k],&ydata[j][k]);
	    zdata[j][k] = 0.0;
	  }
	  k += 1;
	}
      }
    }
    fclose(fin);
    *rxdata = xdata;
    *rydata = ydata;
    *rzdata = zdata;
    *rname = name;
    *m = plot_count;
    *n = len;
  }else{
    printf("    WARNING:  Couldn't open %s.\n",infile);
    *m = -1;
  }
}
void get_displacements_xy_data(xdata,ydata,n,min_gap,sigma,crit,
			       rx,ry,rdx,rdy,rv,rdt,rt,rn)
     float *xdata,*ydata;
     int n,min_gap;
     float sigma,crit,**rx,**ry,**rdx,**rdy,**rv,**rdt;
     int **rt,*rn;
{
  int i,j,k;
  int maxvi,*idata,nd,*t,flag,rlen;
  float *deriv,*tf,maxv,*x0,*y0,*dx,*dy,*dt,*v;
  deriv = get_smooth_abs_deriv_xy_data(xdata,ydata,n,sigma);
  tf = get_binary_threshold_farray(deriv,n,crit);
  idata = get_round_to_int_farray(tf,n);
  myfree(tf);
  replace_min_runs_middle_iarray(idata,n,0,1,min_gap); 
  nd = count_min_runs_middle_iarray(idata,n,1,1); 
  x0 = (float *)myalloc(nd*sizeof(float));  
  y0 = (float *)myalloc(nd*sizeof(float));  
  dx = (float *)myalloc(nd*sizeof(float));  
  dy = (float *)myalloc(nd*sizeof(float));  
  dt = (float *)myalloc(nd*sizeof(float));  
  v = (float *)myalloc(nd*sizeof(float));   
  t = (int *)myalloc(nd*sizeof(int));       
  j = 0; 
  flag = 0;
  while((j<n)&&(!flag)){
    if (idata[j] == 0)
      flag = 1;
    j += 1;
  } 
  maxvi = -1; 
  maxv = -1.0;
  k = 0; 
  rlen = 0; 
  for(i=j;i<n;i++){
    if (idata[i]==0){
      if (rlen>0){ 
	x0[k] = xdata[i-rlen-1]; 
	y0[k] = ydata[i-rlen-1];
	dx[k] = xdata[i] - xdata[i-rlen-1];
	dy[k] = ydata[i] - ydata[i-rlen-1];
	dt[k] = (float)(rlen+1);
	t[k] = i-rlen; 
	v[k] = (deriv[maxvi-1] + deriv[maxvi] + deriv[maxvi+1])/3.0;
	k += 1;
	rlen = 0;
	maxv = -1.0;
      }
    }else{ 
      rlen += 1;
      if (deriv[i] > maxv){ 
	maxv = deriv[i];
	maxvi = i;
      }
    }
  }
  myfree(deriv); myfree(idata);
  *rx=x0; *ry=y0; *rdx=dx; *rdy=dy; *rdt=dt; *rt=t; *rv=v; *rn=nd;
}
