/*C  (C) Copyright  Wyeth Bair 1993-2010                                     */
#include <stdio.h>
#include <math.h>
#include "my_util.h"
double func_alpha(x,t0,alpha)
     double x,t0,alpha;
{
  if (x < t0)
    return 0.0;
  else
    return (alpha*exp(1.0) * (x-t0) * exp(-alpha*(x-t0)));
}
double func_boxcar(x,t0,width)
     double x,t0,width;
{
  if ((x >= t0)&&(x < (t0+width)))
    return 1.0;
  else
    return 0.0;
}
double func_diff_exp(x,t0,tau1,tau2)
     double x,t0,tau1,tau2;
{
  if (x >= t0)
    return exp((t0-x)/tau1) - exp((t0-x)/tau2);
  else
    return 0.0;
}
double func_gamma_n3(x,s)
     double x,s;
{
  if (x < 0.0)
    return 0.0;
  else
    return (x*x*x/(s*s*s) * exp(-x/s));
}
double func_inv_gamma(x,alpha,beta)
     double x,alpha,beta;
{
  exit_error("FUNC_INV_GAMMA","gamma() is deprecated?");
  return 0.0;
}
double func_gaussian(x,mu,sigma)
     double x,mu,sigma;
{
  return (1.0/(sqrt(2.0*M_PI)*sigma)*exp(-0.5*((x-mu)/sigma)*((x-mu)/sigma)));
}
double func_gaussian_one(x,mu,sigma)
     double x,mu,sigma;
{
  return exp(-0.5*((x-mu)/sigma)*((x-mu)/sigma));
}
double func_2d_gaussian(x,y,m1,m2,s1,s2,normflag)
     double x,y,m1,m2,s1,s2;
     int normflag;
{
  if (normflag)
    return (1.0/(2.0*M_PI*s1*s2) *
	    exp(-0.5*(((x-m1)/s1)*((x-m1)/s1) + ((y-m2)/s2)*((y-m2)/s2))));
  else
    return exp(-0.5*(((x-m1)/s1)*((x-m1)/s1) + ((y-m2)/s2)*((y-m2)/s2)));
}
double func_logistic(x,beta)
     double x,beta;
{
  return (1.0/(1.0 + exp(-2.0*beta*x)));
}
double func_maxwell(x,s)
     double x,s;
{
  if (x < 0.0)
    return 0.0;
  else
    return ((sqrt(2.0/M_PI)/s) * x*x/(s*s) * exp(-0.5*x*x/(s*s)));
}
double func_photoreceptor(x,s)
     double x,s;
{
  if (x < 0.0)
    return 0.0;
  else
    return ((sqrt(2.0/M_PI)/s) * x*x*x*x/(s*s*s*s) * exp(-0.5*x/(s)));
}
double func_primate_photocurrent(x,tau_r,tau_d,tau_p,phi)
     double x,tau_r,tau_d,tau_p,phi;
{
  double tr,tr3,td,tp,phirad;
  tr = x/tau_r;
  tr3 = tr*tr*tr;
  td = x/tau_d;
  tp = x/tau_p;
  phirad = phi * M_PI/180.0;
  if (x < 0.0)
    return 0.0;
  else
    return tr3/(1.0+tr3) * exp(-td*td) * cos(2.0*M_PI*tp + phirad);
}
double func_quick(a,b,c,d,x)
     double a,b,c,d,x;
{
  return (d-(d-c)*exp(-pow((x/a),b)));
}
double func_sigmoid(x)
     double x;
{
  if (x >= 0.0)
    return x/(1.0 + x);
  else
    return x/(1.0 - x);
}
double func_sigmoid_log(x,s,m,T,h)
     double x;
     double s;
     double m;
     double T;
     double h;
{
  float result;
  result = s*(1.0/(1.0+exp(-(x+h)/T)) - 0.5) + m;
  if (isnan(result) && x > 0)
    return s-m;
  else if (isnan(result) && x < 0)
    return -m; 
  else
    return result;
}
double func_sigmoid_tanh(x,c,mu)
     double x,c,mu;
{
  float result,ep,en;
  ep = exp(c*(x-mu)/2.0);
  en = exp(-c*(x-mu)/2.0);
  result = (ep - en)/(ep + en);
  if (isnan(result) && x > 0)
    return 1.0;
  else if (isnan(result) && x < 0)
    return -1.0; 
  else
    return result;
}
double func_one_sided_exp(x,t0,tau)
     double x,t0,tau;
{
  if (x < t0)
    return 0.0;
  else
    return (exp(-(x-t0)/tau));
}
double func_lognormal(x,s)
     double x,s;
{
  double xos;
  xos = x/s;
  if (xos <= 0.0)
    return 0.0;
  else
    return 1.0/xos * exp(-0.5 * log(xos)*log(xos));
}
float *alpha_farray(t0,alpha,ampl,n)
     float t0,alpha,ampl;
     int n;
{
  int i;
  float *data;
  data = (float *)myalloc(n*sizeof(float));
  for(i=0;i<n;i++)
    data[i] = ampl*(float)func_alpha((double)i,(double)t0,(double)alpha);
  return data;
}
float *boxcar_farray(t0,width,ampl,n)
     float t0,width,ampl;
     int n;
{
  int i;
  float *data;
  data = (float *)myalloc(n*sizeof(float));
  for(i=0;i<n;i++)
    data[i] = ampl*(float)func_boxcar((double)i,(double)t0,(double)width);
  return data;
}
float *diff_exp_farray(t0,a,tau1,tau2,n)
     float t0,a,tau1,tau2;
     int n;
{
  int i;
  float *data;
  data = (float *)myalloc(n*sizeof(float));
  for(i=0;i<n;i++)
    data[i] = a * (float)func_diff_exp((double)i,(double)t0,(double)tau1,
				       (double)tau2);
  return data;
}
float *gamma_n3_farray(m,s,ampl,n)
     float m,s,ampl;
     int n;
{
  int i;
  float *data;
  data = (float *)myalloc(n*sizeof(float));
  for(i=0;i<n;i++)
    data[i] = ampl*(float)func_gamma_n3((double)i-(double)m,(double)s);
  return data;
}
float *diff_inv_gamma_farray(scale,alpha,beta,posamp,negamp,n)
     float scale,alpha,beta,posamp,negamp;
     int n;
{
  int i, peak;
  float x;
  float *data;
  data = (float *)myalloc(n*sizeof(float));
  peak = -1;
  for(i=0; i<n; i++){
    x = scale*i;
    data[i] = (float)posamp*func_inv_gamma((double)x,(double)alpha,
					   (double)beta);
    if (i > 0 && peak == -1){
      if (data[i] < data[i-1]){
	peak = i-1;
      }
    }
    if(peak != -1 && i > peak){
      x = scale*(i - peak);
      data[i] -= (float)negamp*func_inv_gamma((double)x,(double)alpha,
					      (double)beta); 
    }
  }
  return data;
}
float *cosine_farray(t0,freq,phase,ampl,n)
     float t0,freq,phase,ampl;
     int n;
{
  int i;
  float *data;
  data = (float *)myalloc(n*sizeof(float));
  for(i=0;i<n;i++)
    data[i] = ampl*cos(freq*((double)i-t0) + phase);
  return data;
}
float *gaussian_farray(mu,sigma,ampl,n)
     float mu,sigma,ampl;
     int n;
{
  int i;
  float *data;
  data = (float *)myalloc(n*sizeof(float));
  for(i=0;i<n;i++)
    data[i] = ampl*(float)func_gaussian((double)i,(double)mu,(double)sigma);
  return data;
}
float *gaussian_one_farray(mu,sigma,ampl,n)
     float mu,sigma,ampl;
     int n;
{
  int i;
  float *data;
  data = (float *)myalloc(n*sizeof(float));
  for(i=0;i<n;i++)
    data[i] = ampl*(float)func_gaussian_one((double)i,(double)mu,
					    (double)sigma);
  return data;
}
float *func_logistic_farray(m,beta,ampl,n)
     float m,beta,ampl;
     int n;
{
  int i;
  float *data;
  data = (float *)myalloc(n*sizeof(float));
  for(i=0;i<n;i++)
    data[i] = ampl*(float)func_logistic((double)i-(double)m,(double)beta);
  return data;
}
float *maxwell_farray(m,s,ampl,n)
     float m;         
     float s;         
     float ampl;      
     int n;           
{
  int i;
  float *data;
  data = (float *)myalloc(n*sizeof(float));
  for(i=0;i<n;i++)
    data[i] = ampl*(float)func_maxwell((double)i-(double)m,(double)s);
  return data;
}
float *func_photoreceptor_farray(m,s,ampl,n)
     float m,s,ampl;
     int n;
{
  int i;
  float *data;
  data = (float *)myalloc(n*sizeof(float));
  for(i=0;i<n;i++)
    data[i] = ampl*(float)func_photoreceptor((double)i-(double)m,(double)s);
  return data;
}
float *func_primate_photocurrent_farray(m,ampl,tr,td,tp,phi,n)
     float m,ampl,tr,td,tp,phi;
     int n;
{
  int i;
  float *data;
  data = (float *)myalloc(n*sizeof(float));
  for(i=0;i<n;i++)
    data[i] = ampl*(float)func_primate_photocurrent((double)i-(double)m,
						    (double)tr,(double)td,
						    (double)tp,(double)phi);
  return data;
}
float *func_quick_farray(s,a,b,c,d,n)
     float s,a,b,c,d;
     int n;
{
  int i;
  float *data;
  data = (float *)myalloc(n*sizeof(float));
  for(i=0;i<n;i++)
    data[i] = (float)func_quick(a,b,c,d,(double)((float)i*s));
  return data;
}
float *func_lognormal_farray(m,s,ampl,n)
     float m,s,ampl;
     int n;
{
  int i;
  float *data;
  data = (float *)myalloc(n*sizeof(float));
  for(i=0;i<n;i++)
    data[i] = ampl*(float)func_lognormal((double)i-(double)m,(double)s);
  return data;
}
float *one_sided_exp_farray(t0,tau,ampl,n)
     float t0,tau,ampl;
     int n;
{
  int i;
  float *data;
  data = (float *)myalloc(n*sizeof(float));
  for(i=0;i<n;i++)
    data[i] = ampl*(float)func_one_sided_exp((double)i,(double)t0,(double)tau);
  return data;
}
