/*C  (C) Copyright  Wyeth Bair 1993-2010                                     */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "my_util.h"
#include "nr_util.h"
#include "farray_util.h"
#include "func_util.h"
#include "myrand_util.h"
#include "data_util.h"
#include "paramfile_util.h"
#include "fft_util.h"
#define PFLAG 0
void kernu_norm_2d_filter(f,xn,yn,targ,rsump,rsumn)
     float **f;            
     int xn,yn;            
     float targ;           
     float *rsump,*rsumn;  
{
  int i,j;
  float sump,sumn,tt;
  sump = sumn = 0.0;
  for(i=0;i<xn;i++){
    for(j=0;j<yn;j++){
      tt = f[i][j];
      if (tt > 0.0)
	sump += tt;
      else
	sumn += tt;
    }
  }
  multiply_2d_farray(f,xn,yn,targ/(sump-sumn));
  sump = sumn = 0.0;
  for(i=0;i<xn;i++){
    for(j=0;j<yn;j++){
      tt = f[i][j];
      if (tt > 0.0)
	sump += tt;
      else
	sumn += tt;
    }
  }
  *rsump = sump;
  *rsumn = sumn;
}
void kernel_util_norm_xyt_filter(f,x0,xn,y0,yn,z0,zn,rsump,rsumn)
     float ***f;
     int x0,xn,y0,yn,z0,zn;
     float *rsump,*rsumn;
{
  int i,j,k;
  float sump,sumn,tt;
  sump = sumn = 0.0;
  for(i=x0;i<=xn;i++){
    for(j=y0;j<=yn;j++){
      for(k=z0;k<=zn;k++){
	tt = f[i][j][k];
	if (tt > 0.0)
	  sump += tt;
	else
	  sumn += tt;
      }
    }
  }
  multiply_3d_farray(f,x0,xn,y0,yn,z0,zn,1.0/sump);
  sump = sumn = 0.0;
  for(i=x0;i<=xn;i++){
    for(j=y0;j<=yn;j++){
      for(k=z0;k<=zn;k++){
	tt = f[i][j][k];
	if (tt > 0.0)
	  sump += tt;
	else
	  sumn += tt;
      }
    }
  }
  *rsump = sump;
  *rsumn = sumn;
}
float *rectangle(n)
     int n;
{
  int i;
  float *data;
  data = (float *)myalloc(n*sizeof(float));
  for(i=0;i<n;i++)
    data[i] = 1.0/(float)n;
  return data;
}
float *decay_exp(n,l1)
     int n;
     float l1;
{
  int i;
  float *data,t;
  data = (float *)myalloc(n*sizeof(float));
  for(i=0;i<n;i++){
    t = (float)i/1000.0;
    data[i] = l1*exp(-l1*t)/1000.0;
  }
  return data;
}
float *doe(n,l1,l2)
     int n;
     float l1,l2;
{
  int i;
  float *data,t;
  data = (float *)myalloc(n*sizeof(float));
  for(i=0;i<n;i++){
    t = (float)i/1000.0;
    data[i] = (l1*exp(-l1*t) - l2*exp(-l2*t))/1000.0;
  }
  return data;
}
float *alpha(n,alpha)
     int n;
     float alpha;
{
  int i;
  float *data,t;
  data = (float *)myalloc(n*sizeof(float));
  for(i=0;i<n;i++){
    t = (float)i;
    data[i] = alpha*alpha*t*exp(-alpha*t);
  }
  return data;
}
float *doa(n,a1,a2)
     int n;
     float a1,a2;
{
  int i;
  float *data,t;
  data = (float *)myalloc(n*sizeof(float));
  for(i=0;i<n;i++){
    t = (float)i/1000.0;
    data[i] = (a1*a1*t*exp(-a1*t) - a2*a2*t*exp(-a2*t))/1000.0;
  }
  return data;
}
float *kernu_o_filter_boxcar(mylogf,fo,tscale,rn)
     char mylogf[];
     struct onode *fo;  
     float tscale;      
     int *rn;           
{
  int i;
  int n,ft0,ftn,ftn2;
  float famp,famp2,*fdata;
  if (*rn == -1)
    n     = onode_getpar_int_exit(fo,"n");
  else
    n = *rn;
  ft0    = onode_getpar_int_exit(fo,"t0");
  ftn    = onode_getpar_int_exit(fo,"tn");
  famp   = onode_getpar_flt_exit(fo,"amp");
  ftn2 = 0;
  if (onode_item(fo,"tn2")==1){               
    ftn2  = onode_getpar_int_exit(fo,"tn2");
    famp2 = onode_getpar_flt_exit(fo,"amp2");
  }
  if (ft0 < 0)
    mylogx(mylogf,"KERNU_O_FILTER_BOXCAR","ft0 < 0 for <filter>");
  if (ft0+ftn+ftn2 > n)
    mylogx(mylogf,"KERNU_O_FILTER_BOXCAR",
	   "boxcar too long for 'n' in <filter>");
  fdata = boxcar_farray((float)ft0,(float)ftn,famp,n);
  if (ftn2 > 0){
    for(i=(ft0+ftn);i<(ft0+ftn+ftn2);i++)
      fdata[i] = famp2;
  }
  *rn = n;
  return fdata;
}
float *kernu_o_filter_exp(mylogf,fo,tscale,rn)
     char mylogf[];
     struct onode *fo;  
     float tscale;      
     int *rn;           
{
  int n;
  float t0,t1,t2,amp2,a,tau1,tau2,*fdata,*d1,*d2,xoff,gsd,gsds,*sm;
  char *orig_str;
  if (*rn == -1)
    n = onode_getpar_int_exit(fo,"n");
  else
    n = *rn;
  t1     = onode_getpar_flt_exit(fo,"tau1");      
  t2     = onode_getpar_flt_dflt(fo,"tau2",-1.0); 
  amp2   = onode_getpar_flt_dflt(fo,"amp2",0.0);  
  xoff   = onode_getpar_flt_dflt(fo,"xoff",0.0);  
  gsd    = onode_getpar_flt_dflt(fo,"gauss_sd",0.0);  
  orig_str = onode_getpar_chr_dflt(fo,"origin","zero");
  if (strcmp(orig_str,"center")==0){
    t0 = (float)(int)((n-1)/2.0);
  }else if (strcmp(orig_str,"zero")==0){
    t0 = 0.0;
  }else
    exit_error("KERNU_O_FILTER_EXP","Unkown 'origin' string");
  t0 += xoff / tscale;
  tau1 = t1/tscale;  
  tau2 = t2/tscale;  
  d1 = one_sided_exp_farray(t0,tau1,1.0,n);
  if ((amp2 != 0.0) && (t2 != -1.0)){   
    d2 = one_sided_exp_farray(0.0,tau2,amp2,n);
    fdata = add_farrays(d1,d2,n);
    myfree(d1);
    myfree(d2);
  }else
    fdata = d1;
  if (gsd > 0.0){
    gsds = gsd/tscale;  
    sm = smooth_with_gaussian(fdata,n,gsds,0.01);
    myfree(fdata);
    fdata = sm;
  }
  if (onode_item(fo,"max")==1){
    a = onode_getpar_flt_exit(fo,"max");
    make_max_const_farray(fdata,n,a);
  }else if (onode_item(fo,"area")==1){
    a = onode_getpar_flt_exit(fo,"area");
    norm_area_farray(fdata,n,a);
  }else{
    norm_area_farray(fdata,n,1.0);  
  }
  *rn = n;
  return fdata;
}
float *kernu_o_filter_alpha(mylogf,fo,tscale,rn)
     char mylogf[];
     struct onode *fo;  
     float tscale;      
     int *rn;           
{
  int n;
  float t0,t1,t2,amp2,a,tau1,tau2,*fdata,*d1,*d2,xoff,gsd,gsds,*sm;
  char *orig_str;
  if (*rn == -1)
    n = onode_getpar_int_exit(fo,"n");
  else
    n = *rn;
  t1     = onode_getpar_flt_exit(fo,"tau1");      
  t2     = onode_getpar_flt_dflt(fo,"tau2",-1.0); 
  amp2   = onode_getpar_flt_dflt(fo,"amp2",0.0);  
  xoff   = onode_getpar_flt_dflt(fo,"xoff",0.0);  
  gsd    = onode_getpar_flt_dflt(fo,"gauss_sd",0.0);  
  orig_str = onode_getpar_chr_dflt(fo,"origin","zero");
  if (strcmp(orig_str,"center")==0){
    t0 = (float)(int)((n-1)/2.0);
  }else if (strcmp(orig_str,"zero")==0){
    t0 = 0.0;
  }else
    exit_error("KERNU_O_FILTER_ALPHA","Unkown 'origin' string");
  t0 += xoff / tscale;
  tau1 = t1/tscale;  
  tau2 = t2/tscale;  
  d1 = alpha_farray(t0,1.0/tau1,1.0,n);
  if ((amp2 != 0.0) && (t2 != -1.0)){   
    d2 = alpha_farray(t0,1.0/tau2,amp2,n);
    fdata = add_farrays(d1,d2,n);
    myfree(d1);
    myfree(d2);
  }else
    fdata = d1;
  if (gsd > 0.0){
    gsds = gsd/tscale;  
    sm = smooth_with_gaussian(fdata,n,gsds,0.01);
    myfree(fdata);
    fdata = sm;
  }
  if (onode_item(fo,"max")==1){
    a = onode_getpar_flt_exit(fo,"max");
    make_max_const_farray(fdata,n,a);
  }else if (onode_item(fo,"area")==1){
    a = onode_getpar_flt_exit(fo,"area");
    norm_area_farray(fdata,n,a);
  }else{
    norm_area_farray(fdata,n,1.0);  
  }
  *rn = n;
  return fdata;
}
float *kernu_o_filter_gaussian(mylogf,fo,tscale,rn)
     char mylogf[];
     struct onode *fo;  
     float tscale;      
     int *rn;           
{
  int n;
  float m1,m2,s1,s2,amp2,a,mu1,mu2,sd1,sd2,*fdata,*d1,*d2;
  if (*rn == -1)
    n = onode_getpar_int_exit(fo,"n");
  else
    n = *rn;
  m1     = onode_getpar_flt_exit(fo,"mean");        
  s1     = onode_getpar_flt_exit(fo,"sd");          
  m2     = onode_getpar_flt_dflt(fo,"mean2",-1.0);  
  s2     = onode_getpar_flt_dflt(fo,"sd2",-1.0);    
  amp2   = onode_getpar_flt_dflt(fo,"amp2",0.0);    
  mu1 = m1/tscale;  
  mu2 = m2/tscale;
  sd1 = s1/tscale;
  sd2 = s2/tscale;
  d1 = gaussian_farray(mu1,sd1,1.0,n);
  if ((amp2 != 0.0) && (m2 != -1.0)){   
    d2 = gaussian_farray(mu2,sd2,amp2,n);
    fdata = add_farrays(d1,d2,n);
    myfree(d1);
    myfree(d2);
  }else
    fdata = d1;
  if (onode_item(fo,"max")==1){
    a = onode_getpar_flt_exit(fo,"max");
    make_max_const_farray(fdata,n,a);
  }else if (onode_item(fo,"area")==1){
    a = onode_getpar_flt_exit(fo,"area");
    norm_area_farray(fdata,n,a);
  }else{
    norm_area_farray(fdata,n,1.0);  
  }
  *rn = n;
  return fdata;
}
float *kernu_o_filter_maxwell(mylogf,fo,tscale,rn)
     char mylogf[];
     struct onode *fo;  
     float tscale;      
     int *rn;           
{
  int n,hidnum;
  float m1,m2,amp2,a,s1,s2,*fdata,*d1,*d2,t0,xoff;
  char *orig_str;
  if (*rn == -1)
    n = onode_getpar_int_exit(fo,"n");
  else
    n = *rn;
  hidnum  = onode_getpar_int_dflt(fo,"hidden",0);   
  if (hidnum == 0){
    m1     = onode_getpar_flt_exit(fo,"m1");        
    m2     = onode_getpar_flt_dflt(fo,"m2",-1.0);   
    xoff   = onode_getpar_flt_dflt(fo,"xoff",0.0);  
    amp2   = onode_getpar_flt_dflt(fo,"amp2",0.0);  
  }else if (hidnum == 1){
    m1     =  0.040;   
    m2     =  0.060;   
    xoff   =  0.0;     
    amp2   = -0.80;    
  }else
    mylog_exit(mylogf,"KERNU_O_FILTER_MAXWELL  Bad 'hidden' index\n");
  orig_str = onode_getpar_chr_dflt(fo,"origin","zero");
  if (strcmp(orig_str,"center")==0){
    t0 = (float)(int)((n-1)/2.0);
  }else if (strcmp(orig_str,"zero")==0){
    t0 = 0.0;
  }else
    exit_error("KERNU_O_FILTER_MAXWELL","Unkown 'origin' string");
  t0 += xoff / tscale;
  s1 = m1/(tscale*sqrt(2.0));  
  s2 = m2/(tscale*sqrt(2.0));  
  d1 = maxwell_farray(t0,s1,1.0,n);
  if ((amp2 != 0.0) && (m2 != -1.0)){   
    d2 = maxwell_farray(t0,s2,amp2,n);
    fdata = add_farrays(d1,d2,n);
    myfree(d1);
    myfree(d2);
  }else
    fdata = d1;
  if (onode_item(fo,"max")==1){
    a = onode_getpar_flt_exit(fo,"max");
    make_max_const_farray(fdata,n,a);
  }else if (onode_item(fo,"area")==1){
    a = onode_getpar_flt_exit(fo,"area");
    norm_area_farray(fdata,n,a);
  }else{
    norm_area_farray(fdata,n,1.0);  
  }
  *rn = n;
  return fdata;
}
float **kernu_o_filt2d_gaussian(mylogf,fo,o1,o2,scale1,scale2,rxn,ryn)
     char mylogf[];
     struct onode *fo;     
     float o1,o2;          
     float scale1,scale2;  
     int *rxn;             
     int *ryn;             
{
  int i,j;
  int xn,yn;
  float m1,m2,s1,s2,x,y,a,**fdata;
  if (*rxn == -1){
    xn = onode_getpar_int_exit(fo,"xn");
  }else
    xn = *rxn;
  if (*ryn == -1){
    yn = onode_getpar_int_exit(fo,"yn");
  }else
    yn = *ryn;
  fdata = get_2d_farray(xn,yn);
  m1  = onode_getpar_flt_exit(fo,"mu_1");
  m2  = onode_getpar_flt_exit(fo,"mu_2");
  s1  = onode_getpar_flt_exit(fo,"sd_1");
  s2  = onode_getpar_flt_exit(fo,"sd_2");
  for(i=0;i<xn;i++){
    for(j=0;j<yn;j++){
      x = ((float)i-o1)*scale1;
      y = ((float)j-o2)*scale2;
      fdata[i][j] = func_2d_gaussian(x,y,m1,m2,s1,s2,1);
    }
  }
  if (onode_item(fo,"max")==1){
    a = onode_getpar_flt_exit(fo,"max");
    make_max_const_2d_farray(fdata,xn,yn,a);
  }else if (onode_item(fo,"area")==1){
    a = onode_getpar_flt_exit(fo,"area");
    norm_area_2d_farray(fdata,xn,yn,a);
  }else{
    norm_area_2d_farray(fdata,xn,yn,1.0);  
  }
  *rxn = xn;
  *ryn = yn;
  return fdata;
}
float *kernu_o_filter(mylogf,fo,tscale,rn)
     char mylogf[];
     struct onode *fo;  
     float tscale;      
     int *rn;           
{
  char *ftype,*dumpfile;
  float *fdata;
  mylog(mylogf,"  KERNU_O_FILTER\n");
  if (fo == NULL)
    mylogx(mylogf,"KERNU_O_FILTER","Null filter onode");
  ftype = onode_getpar_chr_exit(fo,"type");
  if (strcmp(ftype,"boxcar")==0){
    fdata = kernu_o_filter_boxcar(mylogf,fo,tscale,rn);
  }else if (strcmp(ftype,"exp")==0){
    fdata = kernu_o_filter_exp(mylogf,fo,tscale,rn);
  }else if (strcmp(ftype,"Gaussian")==0){
    fdata = kernu_o_filter_gaussian(mylogf,fo,tscale,rn);
  }else if (strcmp(ftype,"Maxwell")==0){
    fdata = kernu_o_filter_maxwell(mylogf,fo,tscale,rn);
  }else if (strcmp(ftype,"alpha")==0){
    fdata = kernu_o_filter_alpha(mylogf,fo,tscale,rn);
  }else{
    mylogx(mylogf,"KERNU_O_FILTER","Unknown 1d <filter> type");
  }
  dumpfile = onode_getpar_chr_dflt(fo,"plot_filename",NULL);
  if (dumpfile != NULL){
    if ((strcmp(dumpfile,"null")!=0)&&(strcmp(dumpfile,"NULL")!=0))
      append_farray_plot(dumpfile,"filter(raw_time_units)",fdata,*rn,1);
    myfree(dumpfile);
  }
  myfree(ftype);
  return fdata;
}
float **kernu_o_filt2d(mylogf,fo,o1,o2,scale1,scale2,rxn,ryn)
     char mylogf[];
     struct onode *fo;    
     float o1,o2;         
     float scale1,scale2; 
     int *rxn,*ryn;       
{
  char *ftype,*dumpfile;
  float **fdata;
  mylog(mylogf,"  KERNU_O_FILT2D\n");
  if (fo == NULL)
    mylogx(mylogf,"KERNU_O_FILT2D","Null filter onode");
  ftype = onode_getpar_chr_exit(fo,"type");
  if (strcmp(ftype,"Gaussian2d")==0){
    fdata = kernu_o_filt2d_gaussian(mylogf,fo,o1,o2,scale1,scale2,rxn,ryn);
  }else{
    mylogx(mylogf,"KERNU_O_FILT2D","Unknown 2d <filter> type");
  }
  dumpfile = onode_getpar_chr_dflt(fo,"plot_filename",NULL);
  if (dumpfile != NULL){
    if ((strcmp(dumpfile,"null")!=0)&&(strcmp(dumpfile,"NULL")!=0))
      write_2d_data(dumpfile,fdata,0,0,*rxn,*ryn,4,2,1,0);
    myfree(dumpfile);
  }
  myfree(ftype);
  return fdata;
}
float ***random_filter(xn,yn,zn,sscale,tscale,sr,st,nsd_r,nsd_t,pseed)
     int xn,yn,zn;         
     float sscale,tscale;  
     float sr,st;
     float nsd_r,nsd_t;    
     int *pseed;
{
  int i,j,k;
  int xc,yc,tc;  
  float ***filter,*tp,*sm,tsig;
  float c,cs,ct,csgn,x,y,t,x2,y2,gs,gt;
  if (PFLAG) printf("  RANDOM_FILTER\n");
  c = 1.0/(pow(2.0*M_PI,1.5)*sr*sr*st);
  cs = 2.0*sr*sr;
  ct = 2.0*st*st;
  xc = (xn-1)/2;
  yc = (yn-1)/2;
  tc = (zn-1)/2;
  tsig = nsd_t / tscale;      
  filter = f3tensor(1,xn,1,yn,1,zn);  
  for(i=0;i<xn;i++){
    x = (float)(i-xc)*sscale;
    x2 = x*x/cs;
    for(j=0;j<yn;j++){
      y = (float)(j-yc)*sscale;
      y2 = y*y/cs;
      gs = exp(-x2-y2);
      tp = &(filter[i+1][j+1][1]);
      for(k=0;k<zn;k++){
	t = (float)(k-tc)*tscale;
	gt = exp(-t*t/ct);   
	if (myrand_util_ran2(pseed) > 0.5)
	  csgn = 1.0;
	else
	  csgn = -1.0;
	tp[k] = csgn * c*gs*gt;
      }
      sm = smooth_with_gaussian(tp,zn,tsig,0.05);
      for(k=0;k<zn;k++)
	tp[k] = sm[k];
      myfree(sm);
    }
  }
  printf("*** WYETH - spatial smoothing not implemented\n");
  printf("*** WYETH - spatial smoothing not implemented\n");
  return filter;
}
float **gaussian_2d(xn,yn,xc,yc,sigma,ampl)
     int xn,yn;
     float xc,yc,sigma;
     float ampl;              
{
  int i,j;
  float **filter,cs,x,y,x2,y2;
  if (ampl == 0.0){
    filter = get_zero_2d_farray(xn,yn);
  }else{
    cs = 2.0*sigma*sigma;
    filter = get_2d_farray(xn,yn);
    for(i=0;i<xn;i++){
      x = ((float)i-xc);
      x2 = x*x/cs;
      for(j=0;j<yn;j++){
	y = ((float)j-yc);
	y2 = y*y/cs;
	filter[i][j] = ampl * exp(-x2-y2);
      }
    }
  }
  return filter;
}
float **gabor_2d_space_raw(xn,yn,xc,yc,sdo,sdp,sf,theta,phase)
     int xn,yn;       
     float xc,yc;     
     float sdo,sdp;   
     float sf,theta;  
     float phase;     
{
  int i,j;
  float **filter,cso,csp,twopifr,x,y,gs,ph;
  float a,b,c,d,thetar,xx,yy; 
  cso = 2.0*sdo*sdo;
  csp = 2.0*sdp*sdp;
  twopifr = 2.0*M_PI*sf;
  ph = (phase/180.0 * M_PI);
  filter = get_2d_farray(xn,yn);
  thetar = -theta * M_PI/180.0;
  a =  cos(thetar); b = sin(thetar);
  c = -sin(thetar); d = cos(thetar);
  for(i=0;i<xn;i++){
    xx = (float)i-xc;
    for(j=0;j<yn;j++){
      yy = (float)j-yc;
      x = a*xx + c*yy;
      y = b*xx + d*yy;
      gs = exp(-x*x/cso - y*y/csp);
      filter[i][j] = gs*cos(twopifr * x + ph);
    }
  }
  return filter;
}
float *gabor_2d_space_raw_irreg(xf,yf,n,xc,yc,sdo,sdp,sf,theta,phase)
     float *xf,*yf;     
     int n;           
     float xc,yc;     
     float sdo,sdp;   
     float sf,theta;  
     float phase;     
{
  int i;
  float *filter,cso,csp,twopifr,x,y,gs,ph;
  float a,b,c,d,thetar,xx,yy; 
  cso = 2.0*sdo*sdo;
  csp = 2.0*sdp*sdp;
  twopifr = 2.0*M_PI*sf;
  ph = (phase/180.0 * M_PI);
  filter = (float *)myalloc(n*sizeof(float));
  thetar = -theta * M_PI/180.0;
  a =  cos(thetar); b = sin(thetar);
  c = -sin(thetar); d = cos(thetar);
  for(i=0;i<n;i++){
    xx = xf[i] - xc;
    yy = yf[i] - yc;
    x = a*xx + c*yy;
    y = b*xx + d*yy;
    gs = exp(-x*x/cso - y*y/csp);
    filter[i] = gs*cos(twopifr * x + ph);
  }
  return filter;
}
float **gauss_2d_space_raw(xn,yn,xc,yc,sdo,sdp,theta)
     int xn,yn;       
     float xc,yc;     
     float sdo,sdp;   
     float theta;
{
  int i,j;
  float **filter,cso,csp,x,y;
  float a,b,c,d,thetar,xx,yy;  
  cso = 2.0*sdo*sdo;
  csp = 2.0*sdp*sdp;
  filter = get_2d_farray(xn,yn);
  thetar = -theta * M_PI/180.0;
  a =  cos(thetar); b = sin(thetar);
  c = -sin(thetar); d = cos(thetar);
  for(i=0;i<xn;i++){
    xx = (float)i-xc;
    for(j=0;j<yn;j++){
      yy = (float)j-yc;
      x = a*xx + c*yy;
      y = b*xx + d*yy;
      filter[i][j] = exp(-x*x/cso - y*y/csp);
    }
  }
  return filter;
}
float **dog_2d_space_raw(xn,yn,xc,yc,sdctr,sdsurr,asurr)
     int xn,yn;            
     float xc,yc;          
     float sdctr,sdsurr;   
     float asurr;          
{
  float **dcent,**dsurr,**filter;
  dcent = gauss_2d_space_raw(xn,yn,xc,yc,sdctr,sdctr,0.0);
  dsurr = gauss_2d_space_raw(xn,yn,xc,yc,sdsurr,sdsurr,0.0);
  norm_area_2d_farray(dcent,xn,yn,1.0);
  norm_area_2d_farray(dsurr,xn,yn,asurr);
  filter = subtract_2d_farrays(dcent,dsurr,xn,yn);
  free_2d_farray(dcent,xn);
  free_2d_farray(dsurr,xn);
  return filter;
}
float **gabor_2d_space(xn,yn,sscale,sr,fr,theta,phase)
     int xn,yn;
     float sscale,sr,fr,theta;
     float phase;  
{
  int i,j;
  int xc,yc; 
  float **filter,cs,nx,ny,twopifr,x,y,x2,y2,gs,frndotr,ph;
  cs = 2.0*sr*sr;
  nx = cos(theta/180.0 * M_PI); 
  ny = sin(theta/180.0 * M_PI);
  twopifr = 2.0*M_PI*fr;
  xc = (xn-1)/2;
  yc = (yn-1)/2;
  ph = (phase/180.0 * M_PI);
  filter = get_2d_farray(xn,yn);
  for(i=0;i<xn;i++){
    x = (float)(i-xc)*sscale;
    x2 = x*x/cs;
    for(j=0;j<yn;j++){
      y = (float)(j-yc)*sscale;
      y2 = y*y/cs;
      frndotr = twopifr * (nx*x + ny*y);
      gs = exp(-x2-y2);
      filter[i][j] = gs*cos(frndotr + ph);
    }
  }
  return filter;
}
float ***gabor_space_time(xn,yn,zn,sscale,tscale,sr,st,fr,ft,theta,phase)
     int xn,yn,zn; 
     float sscale,tscale; 
     float sr,st,fr,ft,theta;
     int phase;  
{
  int i,j,k;
  float ***filter;
  float c,cs,ct,nx,ny,twopifr,twopift;
  float x,y,t,x2,y2,gs,gt,frndotr;
  int xc,yc,tc; 
  if (PFLAG) printf("  GABOR_SPACE_TIME\n");
  c = 1.0/(pow(2.0*M_PI,1.5)*sr*sr*st);
  cs = 2.0*sr*sr;
  ct = 2.0*st*st;
  nx = cos(theta/180.0 * M_PI);
  ny = sin(theta/180.0 * M_PI);
  twopifr = 2.0*M_PI*fr;
  twopift = 2.0*M_PI*ft;
  xc = (xn-1)/2;
  yc = (yn-1)/2;
  tc = (zn-1)/2;
  filter = get_3d_farray(xn,yn,zn);
  if (phase==0){  
    for(i=0;i<xn;i++){
      x = (float)(i-xc)*sscale;
      x2 = x*x/cs;
      for(j=0;j<yn;j++){
	y = (float)(j-yc)*sscale;
	y2 = y*y/cs;
	frndotr = twopifr * (nx*x + ny*y);
	gs = exp(-x2-y2);
	for(k=0;k<zn;k++){
	  t = (float)(k-tc)*tscale;
	  gt = exp(-t*t/ct);
	  filter[i][j][k] = c*gs*gt*cos(frndotr - twopift*t);
	}
      }
    }
  }else if (phase==1){ 
    for(i=0;i<xn;i++){
      x = (float)(i-xc)*sscale;
      x2 = x*x/cs;
      for(j=0;j<yn;j++){
	y = (float)(j-yc)*sscale;
	y2 = y*y/cs;
	frndotr = twopifr * (nx*x + ny*y);
	gs = exp(-x2-y2);
	for(k=0;k<zn;k++){
	  t = (float)(k-tc)*tscale;
	  gt = exp(-t*t/ct);
	  filter[i][j][k] = c*gs*gt*sin(frndotr - twopift*t);
	}
      }
    }
  }else{
    printf("*** GABOR_SPACE_TIME:  No such option.\n");
    exit(0);
  }
  return filter;
}
float ***gabor_space_time_tensor(xn,yn,zn,sscale,tscale,srx,sry,st,fr,ft,theta,
				 phase,tfilt)
     int xn,yn,zn;         
     float sscale,tscale;  
     float srx,sry;        
     float st;             
     float fr,ft;          
     float theta;          
     float phase;          
     char *tfilt;          
{
  int i,j,k;
  float ***filter;
  float c,csx,csy,ct,nx,ny,twopifr,twopift,ph0;
  float x,y,t,x2,y2,gs,gt,frndotr;
  int xc,yc,tc; 
  float *zt;    
  int dumpflag;
  if (PFLAG) printf("  GABOR_SPACE_TIME_TENSOR\n");
  dumpflag = 0; 
  if (dumpflag == 1){
    zt = get_zero_farray(zn);
  }
  c = 1.0/(pow(2.0*M_PI,1.5)*srx*sry*st);
  csx = 2.0*srx*srx;
  csy = 2.0*sry*sry;
  ct = 2.0*st*st;
  nx = cos(theta/180.0 * M_PI);
  ny = sin(theta/180.0 * M_PI);
  twopifr = 2.0*M_PI*fr;
  twopift = 2.0*M_PI*ft;
  xc = (xn-1)/2;      
  yc = (yn-1)/2;
  tc = (zn-1)/2;
  ph0 = M_PI/180.0 * phase;  
  filter = f3tensor(1,xn,1,yn,1,zn);  
  for(i=0;i<xn;i++){
    x = (float)(i-xc)*sscale;  
    x2 = x*x/csx;
    for(j=0;j<yn;j++){
      y = (float)(j-yc)*sscale;  
      y2 = y*y/csy;
      frndotr = twopifr * (nx*x + ny*y);
      gs = exp(-x2-y2);
      for(k=0;k<zn;k++){
	t = (float)(k-tc)*tscale;  
	if (strcmp(tfilt,"maxwell")==0){
	  if (t > 0.0)
	    gt = t*t/(st*st) * exp(-t*t/ct);
	  else
	    gt = 0.0;
	}else{
	  gt = exp(-t*t/ct);   
	}
	if (dumpflag == 1){
	  if ((i==0)&&(j==0)){
	    zt[k] = gt;
	  }
	}
	filter[i+1][j+1][k+1] = c*gs*gt*cos(frndotr - twopift*t - ph0);
      }
    }
  }
  if (dumpflag == 1)
    append_farray_plot("zz.MEGauss.pl","gt",zt,zn,2);
  return filter;
}
float ***gabor_space_time_tensor_curve(xn,yn,zn,sscale,tscale,srx,sry,st,
				       fr,theta,phase,p,tfilt)
     int xn,yn,zn;         
     float sscale,tscale;  
     float srx,sry;        
     float st;             
     float fr;             
     float theta;          
     float phase;          
     float p;              
     char *tfilt;          
{
  int i,j,k;
  float ***filter;
  float c,csx,csy,ct,nx,ny,twopifr,ph0;
  float x,y,t,x2,y2,gs,gt,frndotr;
  int xc,yc,tc; 
  float *zt;    
  float pht;
  float tt,crv,trad;
  float max_shift_deg,max_shift_cyc,max_shift_rad;
  if (PFLAG) printf("  GABOR_SPACE_TIME_TENSOR_CURVE\n");
  pht = 0.0;
  c = 1.0/(pow(2.0*M_PI,1.5)*srx*sry*st);
  csx = 2.0*srx*srx;
  csy = 2.0*sry*sry;
  ct = 2.0*st*st;
  nx = cos(theta/180.0 * M_PI);
  ny = sin(theta/180.0 * M_PI);
  twopifr = 2.0*M_PI*fr;
  xc = (xn-1)/2;
  yc = (yn-1)/2;
  tc = (zn-1)/2;
  ph0 = M_PI/180.0 * phase;  
  filter = f3tensor(1,xn,1,yn,1,zn);  
  for(i=0;i<xn;i++){
    x = (float)(i-xc)*sscale;  
    x2 = x*x/csx;
    for(j=0;j<yn;j++){
      y = (float)(j-yc)*sscale;  
      y2 = y*y/csy;
      frndotr = twopifr * (nx*x + ny*y);
      gs = exp(-x2-y2);
      for(k=0;k<zn;k++){
	t = (float)(k-tc)*tscale;  
	if ((t > 0.0) && (t < 3.0*st)){
	  max_shift_deg = 15.0*srx;  
	  max_shift_cyc = max_shift_deg * fr; 
	  max_shift_rad = 2.0*M_PI * max_shift_cyc; 
	  tt = t/(3.0*st);
	  crv =  pow(tt,1.0/p);  
	  pht =  crv * max_shift_rad;   
	  if ((j == yn/2) && (i == xn/2)){
	    printf("k=%d  pht = %f  (tt = %f  crv = %f  p=%f  maxd=%f)\n",
		   k,pht,tt,crv,p,max_shift_deg);
	  }
	}
	if (strcmp(tfilt,"maxwell")==0){
	  if (t > 0.0)
	    gt = t*t/(st*st) * exp(-t*t/ct);
	  else
	    gt = 0.0;
	}else{
	  gt = exp(-t*t/ct);   
	}
	filter[i+1][j+1][k+1] = c*gs*gt*cos(frndotr - pht - ph0);
      }
    }
  }
  return filter;
}
float *kernu_tshift_pow(tso,tn,sscale,tscale,sf)
     struct onode *tso;    
     int tn;               
     float sscale;         
     float tscale;         
     float sf;             
{
  int i;
  int tc; 
  float p,t,t0,t1,tdur,tt,s,*shift,shmax,shmax_cyc,shmax_rad;
  if (PFLAG) printf("  KERNU_TSHIFT_POW\n");
  p     = onode_getpar_flt_exit(tso,"exponent");
  t0    = onode_getpar_flt_exit(tso,"t0");
  t1    = onode_getpar_flt_exit(tso,"t1");
  shmax = onode_getpar_flt_exit(tso,"shift_max");
  if (strcmp("(deg)",onode_getpar_unit(tso,"shift_max"))==0){
    shmax_cyc = shmax * sf;  
  }else if (strcmp("(cyc)",onode_getpar_unit(tso,"shift_max"))==0){
    shmax_cyc = shmax;
  }else
    exit_error("KERNU_TSHIFT_POW","Unknown units on shift_max");
  tc = (tn-1)/2;   
  tdur = t1 - t0;  
  shmax_rad = 2.0*M_PI * shmax_cyc; 
  shift = get_zero_farray(tn);
  for(i=0;i<tn;i++){
    t = (float)(i-tc)*tscale;  
    if ((t > t0) && (t <= t1)){  
      tt = (t-t0)/(float)tdur;   
      s = pow(tt,1.0/p);         
      shift[i] = s * shmax_rad;  
    }
  }
  return shift;
}
float *kernu_tshift_sig(tso,tn,sscale,tscale,sf)
     struct onode *tso;    
     int tn;               
     float sscale;         
     float tscale;         
     float sf;             
{
  int i;
  int tc; 
  float p,t,t0,t1,tdur,tt,s,*shift,shmax,shmax_cyc,shmax_rad;
  if (PFLAG) printf("  KERNU_TSHIFT_SIG\n");
  p     = onode_getpar_flt_exit(tso,"exponent");
  t0    = onode_getpar_flt_exit(tso,"t0");
  t1    = onode_getpar_flt_exit(tso,"t1");
  shmax = onode_getpar_flt_exit(tso,"shift_max");
  if (strcmp("(deg)",onode_getpar_unit(tso,"shift_max"))==0){
    shmax_cyc = shmax * sf;  
  }else if (strcmp("(cyc)",onode_getpar_unit(tso,"shift_max"))==0){
    shmax_cyc = shmax;
  }else
    exit_error("KERNU_TSHIFT_SIG","Unknown units on shift_max");
  tc = (tn-1)/2;   
  tdur = t1 - t0;  
  shmax_rad = 2.0*M_PI * shmax_cyc; 
  shift = get_zero_farray(tn);
 for(i=0;i<tn;i++){
    t = (float)(i-tc)*tscale;  
    if ((t > t0) && (t <= t1)){  
      if (1){
	tt = (t-t0)/(float)tdur;   
	s = -1.0 + 2.0/(1.0 + exp(-(p * tt)));
      }else{
	tt = -1.0 + (t-t0)/(float)tdur;   
	s = 1.0 - 2.0/(1.0 + exp(-(p * tt)));
      }
      shift[i] = s * shmax_rad;  
    }
  }
  return shift;
}
float *kernu_tshift(tso,tn,sscale,tscale,sf)
     struct onode *tso;    
     int tn;               
     float sscale;         
     float tscale;         
     float sf;             
{
  char *stype;
  float *tshift;
  stype = onode_getpar_chr_ptr_exit(tso,"type");
  if (strcmp(stype,"power")==0){
    tshift = kernu_tshift_pow(tso,tn,sscale,tscale,sf);
  }else if (strcmp(stype, "sigmoid")==0){
    tshift = kernu_tshift_sig(tso,tn,sscale,tscale,sf);
  }else{
     exit_error("KERNU_TSHIFT","Unknown shift type");
  }
  return tshift;
}
float ***gabor_space_time_tshift(xn,yn,zn,sscale,tscale,srx,sry,
				 fr,theta,phase,tamp,tshift)
     int xn,yn,zn;         
     float sscale,tscale;  
     float srx,sry;        
     float fr;             
     float theta;          
     float phase;          
     float *tamp;          
     float *tshift;        
{
  int i,j,k;
  int xc,yc;
  float ***filter,c,csx,csy,nx,ny,twopifr,ph0,x,y,x2,y2,gs,frndotr;
  if (PFLAG) printf("  GABOR_SPACE_TIME_TSHIFT\n");
  c = 1.0/(pow(2.0*M_PI,1.5)*srx*sry);
  csx = 2.0*srx*srx;
  csy = 2.0*sry*sry;
  nx = cos(theta/180.0 * M_PI);
  ny = sin(theta/180.0 * M_PI);
  twopifr = 2.0*M_PI*fr;
  xc = (xn-1)/2;  
  yc = (yn-1)/2;  
  ph0 = M_PI/180.0 * phase;  
  filter = f3tensor(1,xn,1,yn,1,zn);  
  for(i=0;i<xn;i++){
    x = (float)(i-xc)*sscale;  
    x2 = x*x/csx;
    for(j=0;j<yn;j++){
      y = (float)(j-yc)*sscale;  
      y2 = y*y/csy;
      frndotr = twopifr * (nx*x + ny*y);
      gs = exp(-x2-y2);
      for(k=0;k<zn;k++){
	filter[i+1][j+1][k+1] = c*gs*tamp[k]*cos(frndotr - tshift[k] - ph0);
      }
    }
  }
  return filter;
}
void rotate_space_time(x,y,t,dir,v,rx,ry,rt)
     float x,y;            
     float t;              
     float dir;            
     float v;              
     float *rx,*ry,*rt;    
{
  float theta,phi;
  float u1,u2,v1,v2,a,b,c,ap,bp,cp,xp,yp,zp;
  float xabc1,xabc2,yabc1,yabc2;
  theta = M_PI * dir/180.0;
  phi = atan(1.0/v);
  u1 = cos(theta);   
  u2 = sin(theta);
  v1 = -sin(theta);  
  v2 = cos(theta);
  a = x*u1 + y* u2;  
  b = x*v1 + y* v2;  
  c = t;             
  ap = a * cos(phi) - c * sin(phi);  
  cp = a * sin(phi) + c * cos(phi);
  bp = b;
  xabc1 = cos(theta);
  xabc2 = -sin(theta);
  yabc1 = sin(theta);
  yabc2 = cos(theta);
  xp = ap * xabc1 + bp * xabc2; 
  yp = ap * yabc1 + bp * yabc2; 
  zp = cp;
  *rx = xp;   
  *ry = yp;
  *rt = zp;
}
float ***gabor_tilt_tensor(xn,yn,zn,sscale,tscale,sr,st,fr,ft,theta,
			   phase,tfilt,shift_frac)
     int xn,yn,zn;         
     float sscale;         
     float tscale;         
     float sr;             
     float st;             
     float fr;             
     float ft;             
     float theta;          
     int phase;            
     char *tfilt;          
     float shift_frac;     
{
  int i,j,k;
  float ***filter;
  float c,cs,ct,nx,ny,twopifr,twopift;
  float x,y,t,x2,y2,gs,gt,frndotr,shx,shy;
  int xc,yc,tc; 
  if (PFLAG) printf("  GABOR_TILT_TENSOR\n");
  c = 1.0/(pow(2.0*M_PI,1.5)*sr*sr*st);
  cs = 2.0*sr*sr;
  ct = 2.0*st*st;
  nx = cos(theta/180.0 * M_PI);
  ny = sin(theta/180.0 * M_PI);
  twopifr = 2.0*M_PI*fr;
  twopift = 2.0*M_PI*ft;
  xc = (xn-1)/2;
  yc = (yn-1)/2;
  tc = (zn-1)/2;
  shx = cos(M_PI*theta/180.0) * ft/fr;  
  shy = sin(M_PI*theta/180.0) * ft/fr;  
  shx *= shift_frac;  
  shy *= shift_frac;  
  filter = f3tensor(1,xn,1,yn,1,zn);  
  for(i=0;i<xn;i++){
    for(j=0;j<yn;j++){
      for(k=0;k<zn;k++){
	t = (float)(k-tc)*tscale;  
	x = (float)(i-xc)*sscale;  
	y = (float)(j-yc)*sscale;  
	frndotr = twopifr * (nx*x + ny*y);
	x -= t*shx;                
	y -= t*shy;                
	x2 = x*x/cs;
	y2 = y*y/cs;
	gs = exp(-x2-y2);
	if (strcmp(tfilt,"maxwell")==0){
	  if (t > 0.0)
	    gt = t*t/(st*st) * exp(-t*t/ct);
	  else
	    gt = 0.0;
	}else{
	  gt = exp(-t*t/ct);   
	}
	if (phase == 0)       
	  filter[i+1][j+1][k+1] = c*gs*gt*cos(frndotr - twopift*t);
	else if (phase == 1)  
	  filter[i+1][j+1][k+1] = c*gs*gt*sin(frndotr - twopift*t);
	else
	  exit_error("GABOR_TILT_TENSOR","No such 'phase'");
      }
    }
  }
  return filter;
}
float ***gabor_tilt2_tensor(xn,yn,zn,sscale,tscale,sr,st,fr,ft,theta,
			   phase,tfilt)
     int xn,yn,zn;         
     float sscale;         
     float tscale;         
     float sr;             
     float st;             
     float fr;             
     float ft;             
     float theta;          
     int phase;            
     char *tfilt;          
{
  int i,j,k;
  float ***filter;
  float c,cs,ct,nx,ny,twopifr,twopift;
  float x,y,t,x2,y2,gs,gt,frndotr,shx,shy;
  int xc,yc,tc; 
  float xraw,yraw,traw,arot,ang,vx,vy,**m,*v,*r,xf; 
  if (PFLAG) printf("  GABOR_TILT2_TENSOR\n");
  if ((theta != 0.0) && (theta != 180.0))
    exit_error("GABOR_TILT2_TENSOR","Only works for 0/180");
  c = 1.0/(pow(2.0*M_PI,1.5)*sr*sr*st);
  cs = 2.0*sr*sr;
  ct = 2.0*st*st;
  nx = cos(theta/180.0 * M_PI);
  ny = sin(theta/180.0 * M_PI);
  twopifr = 2.0*M_PI*fr;
  twopift = 2.0*M_PI*ft;
  xc = (xn-1)/2;
  yc = (yn-1)/2;
  tc = (zn-1)/2;
  arot = 180.0/M_PI * atan2(ft,fr);     
  ang  = theta + 90.0;  
  vx = cos(M_PI/180.0 * ang);
  vy = sin(M_PI/180.0 * ang);
  m = rotation_matrix_3d(-arot,vx,vy,0.0);
  v = get_farray(3);
  xf = 1.0/sin(M_PI/180.0 * (90.0-arot));  
  filter = f3tensor(1,xn,1,yn,1,zn);  
  for(i=0;i<xn;i++){
    for(j=0;j<yn;j++){
      for(k=0;k<zn;k++){
	traw = (float)(k-tc)*tscale;  
	xraw = (float)(i-xc)*sscale;  
	yraw = (float)(j-yc)*sscale;  
	v[0] = xraw;
	v[1] = yraw;
	v[2] = traw;
	r = matrix_times_vector(m,v,3,3);
	x = r[0];
	y = r[1];
	t = r[2];
	myfree(r);
	x *= xf;  
	t /= xf;  
	frndotr = twopifr * (nx*xraw + ny*yraw);  
	x2 = x*x/cs;  
	y2 = y*y/cs;
	gs = exp(-x2-y2);
	if (strcmp(tfilt,"maxwell")==0){
	  if (t > 0.0)
	    gt = t*t/(st*st) * exp(-t*t/ct);
	  else
	    gt = 0.0;
	}else{
	  gt = exp(-t*t/ct);   
	}
	gs = gt = 0.0;
	if (fabs(x) < 0.4)
	  gs = 1.0;
	if (fabs(t) < 0.080)
	  gt = 1.0;
	if (phase == 0)       
	  filter[i+1][j+1][k+1] = c*gs*gt*cos(frndotr - twopift*traw);
	else if (phase == 1)  
	  filter[i+1][j+1][k+1] = c*gs*gt*sin(frndotr - twopift*traw);
	else
	  exit_error("GABOR_TILT_TENSOR","No such 'phase'");
      }
    }
  }
  free_2d_farray(m,3);
  myfree(v);
  return filter;
}
float ***gabor_tilt3_tensor(xn,yn,zn,sscale,tscale,sr,st,fr,ft,theta,
			   phase,tfilt)
     int xn,yn,zn;         
     float sscale;         
     float tscale;         
     float sr;             
     float st;             
     float fr;             
     float ft;             
     float theta;          
     int phase;            
     char *tfilt;          
{
  int i,j,k;
  float ***filter;
  float c,cs,ct,nx,ny,twopifr,twopift;
  float x,y,t,x2,y2,gs,gt,frndotr;
  int xc,yc,tc;                          
  float xraw,yraw,traw;                  
  float vel;
  float rota,rotb,rotc,rotd,xu,tu,rxu,rtu;
  if (PFLAG) printf("  GABOR_TILT3_TENSOR\n");
  if ((theta != 0.0) && (theta != 180.0)) 
    exit_error("GABOR_TILT3_TENSOR","Only works for 0 or 180");
  c = 1.0/(pow(2.0*M_PI,1.5)*sr*sr*st);
  cs = 2.0*sr*sr;
  ct = 2.0*st*st;
  nx = cos(theta/180.0 * M_PI);
  ny = sin(theta/180.0 * M_PI);
  twopifr = 2.0*M_PI*fr;
  twopift = 2.0*M_PI*ft;
  xc = (xn-1)/2;
  yc = (yn-1)/2;
  tc = (zn-1)/2;
  filter = f3tensor(1,xn,1,yn,1,zn);  
  vel = ft/fr;  
  if (theta == 0.0){
    rota =  cos(M_PI/4.0); rotb = sin(M_PI/4.0); 
    rotc = -sin(M_PI/4.0); rotd = cos(M_PI/4.0);
  }else{
    rota =  cos(M_PI/-4.0); rotb = sin(M_PI/-4.0); 
    rotc = -sin(M_PI/-4.0); rotd = cos(M_PI/-4.0);
  }
  for(i=0;i<xn;i++){
    for(j=0;j<yn;j++){
      for(k=0;k<zn;k++){
	traw = (float)(k-tc)*tscale;  
	xraw = (float)(i-xc)*sscale;  
	yraw = (float)(j-yc)*sscale;  
	xu = xraw / vel;  
	tu = traw;        
	rxu = rota*xu + rotc*tu;
	rtu = rotb*xu + rotd*tu;
	x = vel * rxu;   
	t = rtu;         
	y = yraw;        
	frndotr = twopifr * (nx*xraw + ny*yraw);  
	x2 = x*x/cs;
	y2 = y*y/cs;
	gs = exp(-x2-y2);
	if (strcmp(tfilt,"maxwell")==0){
	  if (t > 0.0)
	    gt = t*t/(st*st) * exp(-t*t/ct);
	  else
	    gt = 0.0;
	}else{
	  gt = exp(-t*t/ct);   
	}
	if (phase == 0)       
	  filter[i+1][j+1][k+1] = c*gs*gt*cos(frndotr - twopift*traw);
	else if (phase == 1)  
	  filter[i+1][j+1][k+1] = c*gs*gt*sin(frndotr - twopift*traw);
	else
	  exit_error("GABOR_TILT3_TENSOR","No such 'phase'");
      }
    }
  }
  return filter;
}
float ***gabor_space_time_sep_tensor(ppl,o,xn,yn,tn,sscale,tscale,phase0)
     struct param_pair_list *ppl;
     struct onode *o;
     int xn,yn,tn;         
     float sscale,tscale;  
     float phase0;         
{
  int i,j,k;
  float xc,yc; 
  int dumpflag;
  float ***filter,*tt,sf,ssd,ssdo,ssdp,sfpix,sdo,sdp,**ss,sv,t0,ph,phase;
  float c,cs,nx,ny,twopifr,x,y,x2,y2,frndotr,gs,theta;
  char *tfilt;
  if (PFLAG) printf("  GABOR_SPACE_TIME_SEP_TENSOR\n");
  tt = (float *)myalloc(tn*sizeof(float));
  t0 = (float)((tn-1)/2);  
  if (ppl != NULL){
    sf       = paramfile_get_float_param_or_exit(ppl,"gabor_sf");
    ssdo     = paramfile_get_float_param_or_exit(ppl,"gabor_ssd_orth");
    ssdp     = paramfile_get_float_param_or_exit(ppl,"gabor_ssd_par");
    theta    = paramfile_get_float_param_or_exit(ppl,"gabor_theta");
    phase    = paramfile_get_float_param_or_exit(ppl,"gabor_phase");
    dumpflag = paramfile_get_int_param_default(ppl,"gabor_tfilt_dump",0);
    tfilt    = paramfile_get_char_param_or_exit(ppl,"gabor_tfilt");
  }else{
    sf       = onode_getpar_flt_exit(o,"gabor_sf");
    ssdo     = onode_getpar_flt_exit(o,"gabor_ssd_orth");
    ssdp     = onode_getpar_flt_exit(o,"gabor_ssd_par");
    theta    = onode_getpar_flt_exit(o,"gabor_theta");
    phase    = onode_getpar_flt_exit(o,"gabor_phase");
    dumpflag = onode_getpar_int_dflt(o,"gabor_tfilt_dump",0);
    tfilt    = onode_getpar_chr_exit(o,"gabor_tfilt");
  }
  if (strcmp(tfilt,"gaussian")==0){
    float tsd,sd;
    if (ppl != NULL)
      tsd   = paramfile_get_float_param_or_exit(ppl,"gabor_tsd");
    else
      tsd   = onode_getpar_flt_exit(o,"gabor_tsd");
    sd = tsd / tscale;
    tt = gaussian_one_farray(t0,sd,1.0,tn);
  }else if (strcmp(tfilt,"gabor")==0){
    float tsd,sd,tf,tph0;
    if (ppl != NULL){
      tsd   = paramfile_get_float_param_or_exit(ppl,"gabor_tsd");
      tf    = paramfile_get_float_param_or_exit(ppl,"gabor_tf");
      tph0  = paramfile_get_float_param_or_exit(ppl,"gabor_tphase");
    }else{
      tsd   = onode_getpar_flt_exit(o,"gabor_tsd");
      tf    = onode_getpar_flt_exit(o,"gabor_tf");
      tph0  = onode_getpar_flt_exit(o,"gabor_tphase");
    }
    sd = tsd / tscale;
    tt = gaussian_one_farray(t0,sd,1.0,tn);
    tf *= 2.0*M_PI * tscale;
    tph0 *= M_PI/180.0;
    for(i=0;i<tn;i++){
      x = (float)(i-t0);
      tt[i] *= cos(tf*x - tph0);
    }
  }else if (strcmp(tfilt,"maxwell_diff")==0){
    float tmax,s,s2,tmax2,tamp2,*tt2;
    if (ppl != NULL){
      tmax = paramfile_get_float_param_or_exit(ppl,"gabor_maxwell_tmax");
      tmax2 = paramfile_get_float_param_or_exit(ppl,"gabor_maxwell_tmax_2");
      tamp2 = paramfile_get_float_param_or_exit(ppl,"gabor_maxwell_tamp_2");
    }else{
      tmax = onode_getpar_flt_exit(o,"gabor_maxwell_tmax");
      tmax2 = onode_getpar_flt_exit(o,"gabor_maxwell_tmax_2");
      tamp2 = onode_getpar_flt_exit(o,"gabor_maxwell_tamp_2");
    }
    s  = tmax  / sqrt(2.0) / tscale;
    s2 = tmax2 / sqrt(2.0) / tscale;
    tt  = maxwell_farray(t0,s ,1.0,tn);
    tt2 = maxwell_farray(t0,s2,1.0,tn);
    for(i=0;i<tn;i++)
      tt[i] = tt[i] - tamp2 * tt2[i];
    myfree(tt2);
  }else if (strcmp(tfilt,"maxwell")==0){
    float tmax,s;
    if (ppl != NULL)
      tmax = paramfile_get_float_param_or_exit(ppl,"gabor_maxwell_tmax");
    else
      tmax = onode_getpar_flt_exit(o,"gabor_maxwell_tmax");
    s = tmax / sqrt(2.0) / tscale;
    tt = maxwell_farray(t0,s,1.0,tn);
  }else
    exit_error("GABOR_SPACE_TIME_SEP_TENSOR","Unknown 'gabor_tfilt'");
  if (dumpflag == 1){
    float tsum;
    char tname[SLEN];
    tsum = sum_farray(tt,tn,0,tn);
    printf("GABOR_SPACE_TIME_SEP_TENSOR  tfilter sum =  %f\n",tsum);
    sprintf(tname,"%s_tscale_units",tfilt);
    append_farray_plot("zzz.tfilt.pl",tname,tt,tn,1);
  }
  myfree(tfilt);
  xc = (xn-1)/2;
  yc = (yn-1)/2;
  sdo = ssdo / sscale;  
  sdp = ssdp / sscale;  
  sfpix = sf * sscale;  
  ss = gabor_2d_space_raw(xn,yn,xc,yc,sdo,sdp,sfpix,theta,phase+phase0);
  filter = f3tensor(1,xn,1,yn,1,tn);  
  for(i=0;i<xn;i++){
    for(j=0;j<yn;j++){
      sv = ss[i][j];
      for(k=0;k<tn;k++){
	filter[i+1][j+1][k+1] = sv * tt[k];
      }
    }
  }
  myfree(tt);
  free_2d_farray(ss,xn);
  return filter;
}
float ***gauss_space_exp_time_tensor(mylogf,xn,yn,zn,sscale,tscale,sig,tau)
     char *mylogf;
     int xn,yn,zn; 
     float sscale,tscale; 
     float sig;  
     float tau;  
{
  int i,j,k;
  int xc,yc,tc;
  float ***filter,cs,x,y,t,x2,y2,gs,gt;
  double dsum;
  char tstr[SLEN];
  mylog(mylogf,"  GAUSS_SPACE_EXP_TIME_TENSOR\n");
  sprintf(tstr,"    xn,yn,zn  %d %d %d\n",xn,yn,zn);
  mylog(mylogf,tstr);
  sprintf(tstr,"    sscale %f   tscale %f\n",sscale,tscale);
  mylog(mylogf,tstr);
  sprintf(tstr,"    sig %f   tau %f\n",sig,tau);
  mylog(mylogf,tstr);
  cs = 2.0*sig*sig;
  xc = (xn-1)/2;
  yc = (yn-1)/2;
  tc = 0; 
  filter = f3tensor(1,xn,1,yn,1,zn); 
  dsum = 0.0;
  for(i=0;i<xn;i++){
    x = (float)(i-xc)*sscale;
    x2 = x*x/cs;
    for(j=0;j<yn;j++){
      y = (float)(j-yc)*sscale;
      y2 = y*y/cs;
      gs = exp(-x2-y2);
      for(k=0;k<zn;k++){
	t = (float)(k-tc)*tscale;
	if (t >= 0.0)
	  gt = exp(-t/tau);
	else
	  gt = 0.0;
	filter[i+1][j+1][k+1] = gs*gt;
	dsum += gs*gt;
      }
    }
  }
  sprintf(tstr,"    Sum of filter before norm: %lf\n",dsum);
  mylog(mylogf,tstr);
  for(i=0;i<xn;i++){ 
    for(j=0;j<yn;j++){
      for(k=0;k<zn;k++){
	filter[i+1][j+1][k+1] /= dsum;
      }
    }
  }
  return filter;
}
float ***exp_time_tensor(mylogf,xn,yn,zn,sscale,tscale,tau)
     char *mylogf;
     int xn,yn,zn; 
     float sscale,tscale; 
     float tau;  
{
  int i,j,k;
  int xc,yc,tc;
  float ***filter,cs,x,y,t,gt;
  double dsum;
  char tstr[SLEN];
  mylog(mylogf,"  EXP_TIME_TENSOR\n");
  sprintf(tstr,"    xn,yn,zn  %d %d %d\n",xn,yn,zn);
  mylog(mylogf,tstr);
  sprintf(tstr,"    sscale %f   tscale %f\n",sscale,tscale);
  mylog(mylogf,tstr);
  sprintf(tstr,"    tau %f\n",tau);
  mylog(mylogf,tstr);
  xc = (xn-1)/2;
  yc = (yn-1)/2;
  tc = (zn-1)/2; 
  filter = f3tensor(1,xn,1,yn,1,zn); 
  dsum = 0.0;
  for(i=0;i<xn;i++){
    x = (float)(i-xc)*sscale;       
    for(j=0;j<yn;j++){
      y = (float)(j-yc)*sscale;     
      for(k=0;k<zn;k++){
	t = (float)(k-tc)*tscale;
	if (t >= 0.0)
	  gt = exp(-t/tau);
	else
	  gt = 0.0;
	if ((i == xc) && (j == yc)){
	  filter[i+1][j+1][k+1] = gt;
	  dsum += gt;
	}else{
	  filter[i+1][j+1][k+1] = 0.0;
	}
      }
    }
  }
  sprintf(tstr,"    Sum of filter before norm: %lf\n",dsum);
  mylog(mylogf,tstr);
  for(i=0;i<xn;i++){ 
    for(j=0;j<yn;j++){
      for(k=0;k<zn;k++){
	filter[i+1][j+1][k+1] /= dsum;
      }
    }
  }
  return filter;
}
float ***delta_tensor(mylogf,xn,yn,zn,sscale,tscale)
     char *mylogf;
     int xn,yn,zn;
     float sscale,tscale; 
{
  int i,j,k;
  int xc,yc,tc;
  float ***filter;
  char tstr[SLEN];
  mylog(mylogf,"  EXP_TIME_TENSOR\n");
  sprintf(tstr,"    xn,yn,zn  %d %d %d\n",xn,yn,zn);
  mylog(mylogf,tstr);
  sprintf(tstr,"    sscale %f   tscale %f\n",sscale,tscale);
  mylog(mylogf,tstr);
  xc = xn/2;
  yc = yn/2;
  tc = (zn-1)/2; 
  filter = f3tensor(1,xn,1,yn,1,zn); 
  for(i=0;i<xn;i++){
    for(j=0;j<yn;j++){
      for(k=0;k<zn;k++){
	filter[i+1][j+1][k+1] = 0.0;
      }
    }
  }
  filter[xc+1][yc+1][tc+1] = 1.0;
  return filter;
}
void get_opponent_random(xn,yn,tn,sscale,tscale,sr,st,nsd_r,nsd_t,seed,
			 fpe,fpo,fne,fno)
     int xn,yn,tn;
     float sscale,tscale;
     float sr,st;       
     float nsd_r,nsd_t; 
     int seed;          
     float ****fpe,****fpo,****fne,****fno;
{
  int i,j,k;
  int irev,seed2;
  if (seed > 0)
    seed = -seed;
  seed2 = seed - 113;
  *fpe = random_filter(xn,yn,tn,sscale,tscale,sr,st,nsd_r,nsd_t,&seed);
  *fpo = random_filter(xn,yn,tn,sscale,tscale,sr,st,nsd_r,nsd_t,&seed2);
  *fne = f3tensor(1,xn,1,yn,1,tn);
  *fno = f3tensor(1,xn,1,yn,1,tn);
  for(i=1;i<=xn;i++){
    irev = (xn-1) - i;  
    if (irev < 1)
      irev = 1;
    else if (irev > xn)
      irev = xn;
      for(j=1;j<=yn;j++)
	for(k=1;k<=tn;k++){
	  (*fne)[i][j][k] = (*fpe)[irev][j][k];
	  (*fno)[i][j][k] = (*fpo)[irev][j][k];
	}
  }
  printf("=====> fpe SUM:  %f\n",sum_3d_farray(*fpe,1,xn,1,yn,1,tn));
  printf("=====> fpo SUM:  %f\n",sum_3d_farray(*fpo,1,xn,1,yn,1,tn));
  printf("=====> fne SUM:  %f\n",sum_3d_farray(*fne,1,xn,1,yn,1,tn));
  printf("=====> fno SUM:  %f\n",sum_3d_farray(*fno,1,xn,1,yn,1,tn));
}
void get_opponent_random_x(xn,yn,tn,sscale,tscale,sr,st,nsd_r,nsd_t,seed,
			   fpe,fpo,fpx,fne,fno,fnx)
     int xn,yn,tn;
     float sscale,tscale;
     float sr,st;       
     float nsd_r,nsd_t; 
     int seed;          
     float ****fpe,****fpo,****fpx,****fne,****fno,****fnx;
{
  int i,j,k;
  int irev,seed2,seed3;
  if (seed > 0)
    seed = -seed;
  seed2 = seed - 2113;
  seed3 = seed - 93653;
  printf("seed = %d\n",seed);
  printf("seed2 = %d\n",seed2);
  printf("seed3 = %d\n",seed3);
  *fpe = random_filter(xn,yn,tn,sscale,tscale,sr,st,nsd_r,nsd_t,&seed);
  *fpo = random_filter(xn,yn,tn,sscale,tscale,sr,st,nsd_r,nsd_t,&seed2);
  *fpx = random_filter(xn,yn,tn,sscale,tscale,sr,st,nsd_r,nsd_t,&seed3);
  *fne = f3tensor(1,xn,1,yn,1,tn);
  *fno = f3tensor(1,xn,1,yn,1,tn);
  *fnx = f3tensor(1,xn,1,yn,1,tn);
  for(i=1;i<=xn;i++){
    irev = (xn-1) - i;  
    if (irev < 1)
      irev = 1;
    else if (irev > xn)
      irev = xn;
      for(j=1;j<=yn;j++)
	for(k=1;k<=tn;k++){
	  (*fne)[i][j][k] = (*fpe)[irev][j][k];
	  (*fno)[i][j][k] = (*fpo)[irev][j][k];
	  (*fnx)[i][j][k] = (*fpx)[irev][j][k];
	}
  }
}
void get_quad_opponent_gabor(xn,yn,tn,sscale,tscale,sr,st,fr,ft,theta,tfilt,
			     fpe,fpo,fne,fno)
     int xn,yn,tn;
     float sscale,tscale,sr,st,fr,ft,theta;
     char *tfilt;
     float ****fpe,****fpo,****fne,****fno;
{
  float ntheta;
  ntheta = theta + 180.0;
  *fpe = gabor_space_time_tensor(xn,yn,tn,sscale,tscale,sr,sr,st,fr,ft,theta,
				 0.0,tfilt);
  *fpo = gabor_space_time_tensor(xn,yn,tn,sscale,tscale,sr,sr,st,fr,ft,theta,
				 90.0,tfilt);
  *fne = gabor_space_time_tensor(xn,yn,tn,sscale,tscale,sr,sr,st,fr,ft,ntheta,
				 0.0,tfilt);
  *fno = gabor_space_time_tensor(xn,yn,tn,sscale,tscale,sr,sr,st,fr,ft,ntheta,
				 90.0,tfilt);
}
void get_quad_opponent_gabor_curve(xn,yn,tn,sscale,tscale,sr,st,fr,theta,
				   pwr,tfilt,fpe,fpo,fne,fno)
     int xn,yn,tn;
     float sscale,tscale,sr,st,fr,theta,pwr;
     char *tfilt;
     float ****fpe,****fpo,****fne,****fno;
{
  float ntheta;
  printf("pwr = %f\n",pwr);
  ntheta = theta + 180.0;
  *fpe = gabor_space_time_tensor_curve(xn,yn,tn,sscale,tscale,sr,sr,st,fr,
				       theta,0.0,pwr,tfilt);
  *fpo = gabor_space_time_tensor_curve(xn,yn,tn,sscale,tscale,sr,sr,st,fr,
				       theta,90.0,pwr,tfilt);
  *fne = gabor_space_time_tensor_curve(xn,yn,tn,sscale,tscale,sr,sr,st,fr,
				       ntheta,0.0,pwr,tfilt);
  *fno = gabor_space_time_tensor_curve(xn,yn,tn,sscale,tscale,sr,sr,st,fr,
				       ntheta,90.0,pwr,tfilt);
}
void get_quad_opponent_gabor_tshift(mylogf,fo,xn,yn,tn,sscale,tscale,
				    fpe,fpo,fne,fno)
     char *mylogf;
     struct onode *fo;     
     int xn,yn,tn;         
     float sscale,tscale;  
     float ****fpe,****fpo,****fne,****fno; 
{
  float sdx,sdy,sf,theta,atheta,*tshift,*tamp;
  char *dumpfile;
  struct onode *tso;   
  struct onode *tfo;   
  theta     = onode_getpar_flt_exit(fo,"direction");
  sdx = sdy = onode_getpar_flt_exit(fo,"ssd");
  sf        = onode_getpar_flt_exit(fo,"sf");
  atheta = theta + 180.0;  
  tso = onode_child_get_unique(fo,"shift_v_t");
  tshift = kernu_tshift(tso,tn,sscale,tscale,sf);
  dumpfile = onode_getpar_chr_dflt(tso,"plot_filename",NULL);
  if (dumpfile != NULL){
    if ((strcmp(dumpfile,"null")!=0)&&(strcmp(dumpfile,"NULL")!=0))
      append_farray_plot(dumpfile,"phase_shift_vs_time",tshift,tn,1);
    myfree(dumpfile);
  }
  tfo = onode_child_get_unique(fo,"filter");  
  tamp = kernu_o_filter(mylogf,tfo,tscale,&tn);  
  *fpe = gabor_space_time_tshift(xn,yn,tn,sscale,tscale,sdx,sdy,sf,theta,0.0,
				 tamp,tshift);
  *fpo = gabor_space_time_tshift(xn,yn,tn,sscale,tscale,sdx,sdy,sf,theta,90.0,
				 tamp,tshift);
  *fne = gabor_space_time_tshift(xn,yn,tn,sscale,tscale,sdx,sdy,sf,atheta,0.0,
				 tamp,tshift);
  *fno = gabor_space_time_tshift(xn,yn,tn,sscale,tscale,sdx,sdy,sf,atheta,90.0,
				 tamp,tshift);
  myfree(tshift);  
  myfree(tamp);
}
void get_arb_opponent_gabor(xn,yn,tn,sscale,tscale,sr,st,fr,ft,theta,tfilt,ph,
			    fpe,fpo,fne,fno)
     int xn,yn,tn;
     float sscale,tscale,sr,st,fr,ft,theta;
     char *tfilt;
     float ph;  
     float ****fpe,****fpo,****fne,****fno;
{
  float ntheta;
  ntheta = theta + 180.0;
  *fpe = gabor_space_time_tensor(xn,yn,tn,sscale,tscale,sr,sr,st,fr,ft,theta,
				 0.0,tfilt);
  *fpo = gabor_space_time_tensor(xn,yn,tn,sscale,tscale,sr,sr,st,fr,ft,theta,
				 ph,tfilt);
  *fne = gabor_space_time_tensor(xn,yn,tn,sscale,tscale,sr,sr,st,fr,ft,ntheta,
				 0.0,tfilt);
  *fno = gabor_space_time_tensor(xn,yn,tn,sscale,tscale,sr,sr,st,fr,ft,ntheta,
				 ph,tfilt);
}
void get_quad_opp_tilt_gabor(xn,yn,tn,sscale,tscale,sr,st,fr,ft,a,tfilt,ss,
			     rotflag,fpe,fpo,fne,fno)
     int xn,yn,tn;
     float sscale,tscale,sr,st,fr,ft,a;  
     char *tfilt;
     float ss;      
     int rotflag;
     float ****fpe,****fpo,****fne,****fno;
{
  float na;
  na = a + 180.0;
  if (rotflag == 1){
    *fpe = gabor_tilt3_tensor(xn,yn,tn,sscale,tscale,sr,st,fr,ft,a,0,tfilt);
    *fpo = gabor_tilt3_tensor(xn,yn,tn,sscale,tscale,sr,st,fr,ft,a,1,tfilt);
    *fne = gabor_tilt3_tensor(xn,yn,tn,sscale,tscale,sr,st,fr,ft,na,0,tfilt);
    *fno = gabor_tilt3_tensor(xn,yn,tn,sscale,tscale,sr,st,fr,ft,na,1,tfilt);
  }else{
    *fpe = gabor_tilt_tensor(xn,yn,tn,sscale,tscale,sr,st,fr,ft,a,0,tfilt,ss);
    *fpo = gabor_tilt_tensor(xn,yn,tn,sscale,tscale,sr,st,fr,ft,a,1,tfilt,ss);
    *fne = gabor_tilt_tensor(xn,yn,tn,sscale,tscale,sr,st,fr,ft,na,0,tfilt,ss);
    *fno = gabor_tilt_tensor(xn,yn,tn,sscale,tscale,sr,st,fr,ft,na,1,tfilt,ss);
  }
}
void get_quad_gabor(ppl,o,xn,yn,tn,sscale,tscale,fe,fo)
     struct param_pair_list *ppl;
     struct onode *o;
     int xn,yn,tn;
     float sscale,tscale;
     float ****fe,****fo;
{
  float ph_off;
  ph_off = 0;
  *fe = gabor_space_time_sep_tensor(ppl,o,xn,yn,tn,sscale,tscale,ph_off);
  ph_off = 90.0;
  *fo = gabor_space_time_sep_tensor(ppl,o,xn,yn,tn,sscale,tscale,ph_off);
}
void get_quad_curve_gabor(o,xn,yn,tn,sscale,tscale,fpe,fpo,fae,fao)
     struct onode *o;  
     int xn,yn,tn;
     float sscale,tscale;
     float ****fpe,****fpo,****fae,****fao;
{
  float ph_off;
}
float *wa_temporal(zn,tscale,a,b,n1,n2,tau1,tau2)
     int zn;
     float tscale;
     float a,b;
     int n1,n2;
     float tau1,tau2;
{
  int i;
  float *filter;
  float n1m1fact,n2m1fact;
  float t,f1,f2,ttau,n1m1,n2m1;
  int tc; 
  int phase;
  if (PFLAG) printf("  WA_TEMPORAL\n");
  phase = 0;
  tc = (zn-1)/2;
  n1m1 = (float)(n1-1);
  n2m1 = (float)(n2-1);
  n1m1fact = 1;
  for(i=2;i<n1;i++)
    n1m1fact *= (float)i;
  n2m1fact = 1;
  for(i=2;i<n2;i++)
    n2m1fact *= (float)i;
  filter = (float *)myalloc(zn*sizeof(float));
  if (phase==0){  
    for(i=0;i<zn;i++){
      t = (float)(i-tc)*tscale;
      if (t >= 0.0){
	ttau = t/tau1;
	f1 = pow(ttau,n1m1) / (tau1*n1m1fact) * exp(-ttau);
	ttau = t/tau2;
	f2 = pow(ttau,n2m1) / (tau2*n2m1fact) * exp(-ttau);
      }else{
	f1 = 0.0;
	f2 = 0.0;
      }
      filter[i] = a*(f1 - b*f2);
    }
  }else{
    printf("*** WA_TEMPORAL:  No such option.\n");
    exit(0);
  }
  return filter;
}
float *ab_temporal(zn,tscale,n,k,toff)
     int zn; 
     float tscale;
     int n;
     float k;
     float toff;
{
  int i;
  int tc; 
  float *filter,nfact,np2fact,kt,t;
  if (PFLAG) printf("  AB_TEMPORAL\n");
  tc = (zn-1)/2;
  nfact = 1;
  for(i=2;i<=n;i++)
    nfact *= (float)i;
  np2fact = 1;
  for(i=2;i<=(n+2);i++)
    np2fact *= (float)i;
  filter = (float *)myalloc(zn*sizeof(float));
  for(i=0;i<zn;i++){
    t = (float)(i-tc)*tscale - toff;
    if (t >= 0.0){
      kt = k * t;
      filter[i] = pow(kt,(float)n) * exp(-kt) * (1.0/nfact - kt*kt/np2fact);
    }else{
      filter[i] = 0.0;
    }
  }
  return filter;
}
float *cwq_temporal(int zn,        
		    float tscale,  
		    int a,         
		    float tau,     
		    float fr,      
		    float phi,     
		    float toff){   
  int i;
  int tc; 
  float *f,nfact,ph,tf;
  double t,c;
  if (PFLAG) printf("  CWQ_TEMPORAL\n");
  tc = (zn-1)/2;
  if (a < 1)
    exit_error("CWQ_TEMPORAL","Param 'a' must be >= 1");
  nfact = 1;
  for(i=2;i<a;i++)
    nfact *= (float)i;
  c = 1.0 / ((double)nfact * pow(tau,a));
  ph = phi * M_PI/180;  
  tf = fr * 2.0*M_PI;   
  f = (float *)myalloc(zn*sizeof(float));
  for(i=0;i<zn;i++){
    t = (double)(i-tc)*tscale - toff;  
    if (t >= 0.0){
      f[i] = (float)(c * pow(t,a-1.0) * exp(-t/tau) * cos(tf*t + ph));
    }else{
      f[i] = 0.0;
    }
  }
  return f;
}
float *dd_exp_temporal(zn,tscale,t0,tau1,tau2,amp2,sig)
     int zn;        
     float tscale;  
     float t0;      
     float tau1;    
     float tau2;    
     float amp2;    
     float sig;     
{
  int i;
  int tc;
  float *filter,t,*sm,tsig;
  printf("    DD_EXP_TEMPORAL\n");
  printf("    %f %f %f %f %f %f\n",tau1,tau2,t0,tscale,amp2,sig);
  tc = (zn-1)/2;
  if (PFLAG) printf("  AB_TEMPORAL\n");
  filter = (float *)myalloc(zn*sizeof(float));
  for(i=0;i<zn;i++){
    t = ((float)(i-tc))*tscale - t0;  
    if ((t < 0.00001) && (t > -0.00001))
      t = 0.0; 
    if (t <= 0.0)
      filter[i] = exp(t/tau1);
    else
      filter[i] = -amp2 * exp(-t/tau2);
  }
  tsig = sig/tscale; 
  sm = smooth_with_gaussian(filter,zn,tsig,0.01);
  myfree(filter);
  filter = sm;
  return filter;
}
float *filt_01_temporal(ppl,o,zn,tscale,pflag)
     struct param_pair_list *ppl;
     struct onode *o;
     int zn;        
     float tscale;  
     int pflag;
{
  int i;
  int tc;
  float *filter,t,*sm,tsig,t0,namp,sig,tm;
  if (pflag) printf("    FILT_01_TEMPORAL\n");
  tc = (zn-1)/2;
  if (ppl != NULL){
    t0    = paramfile_get_float_param_default(ppl,"dogt_t0"    ,0.020);
    sig   = paramfile_get_float_param_default(ppl,"dogt_sigsm" ,0.001);
    namp  = paramfile_get_float_param_default(ppl,"dogt_negamp",0.100);
    tm    = paramfile_get_float_param_default(ppl,"dogt_tmult" ,100.0);
  }else{
    t0    = onode_getpar_flt_dflt(o,"dogt_t0"    ,0.020);
    sig   = onode_getpar_flt_dflt(o,"dogt_sigsm" ,0.001);
    namp  = onode_getpar_flt_dflt(o,"dogt_negamp",0.100);
    tm    = onode_getpar_flt_dflt(o,"dogt_tmult" ,100.0);
  }
  filter = (float *)myalloc(zn*sizeof(float));
  for(i=0;i<zn;i++){
    t = ((float)(i-tc))*tscale;  
    if (t < 0.0)
      filter[i] = 0.0;
    else if (t <= t0)
      filter[i] = 1.0 - (t/t0)*(1.0 + namp);
    else if (t > t0)
      filter[i] = -namp * (1.0/(1.0 + (t-t0)*tm));
  }
  if (sig > 0.0){
    tsig = sig/tscale;
    sm = smooth_with_gaussian(filter,zn,tsig,0.01);
    myfree(filter);
    filter = sm;
  }
  return filter;
}
float ***dog_space_time_tensor(xn,yn,zn,sscale,tscale,sig1,sig2,amp1,amp2,
			       coffx,coffy,sdelay,gtsig,gtcent,
			       outfile,ppl,o,pflag,hilb_flag,t0flag,toff,
			       tau_ad)
     int xn,yn,zn; 
     float sscale,tscale; 
     float sig1,sig2,amp1,amp2,coffx,coffy,sdelay;
     float gtsig;
     float gtcent;                
     char outfile[];
     struct param_pair_list *ppl; 
     struct onode *o;             
     int pflag;
     int hilb_flag;  
     int t0flag;     
     float toff;     
     float tau_ad;   
{
  int i,j,k;
  int xc,yc,d; 
  int t2,ti;
  float ***filter;
  float **g1,**g2,*t,tsig,tcent,*gt,sump,sumn,sumg1,sumg2,tt;
  float a,b,tau1,tau2,cxc,cyc;
  int n1,n2,kflag,tst;
  char *tfilt;
  if (pflag){
    printf("  DOG_SPACE_TIME_TENSOR\n");
    printf("    xn,yn,zn   %d,%d,%d\n",xn,yn,zn);
  }
  kflag = 0;
  xc = xn/2;
  yc = yn/2;
  if (sig1 == -1.0){
    if (pflag) printf("  *** sig1 is -1:  USING DELTA FUNCTION at t=0.\n");
    filter = f3tensor(1,xn,1,yn,1,zn); 
    for(i=0;i<xn;i++){
      for(j=0;j<yn;j++){
	for(k=0;k<zn;k++){
	  filter[i+1][j+1][k+1] = 0.0;
	}
      }
    }
    filter[xc+1][yc+1][1] = 11.0; 
    return filter;
  }
  cxc = (float)xc + coffx/sscale; 
  cyc = (float)yc + coffy/sscale; 
  d = my_rint(sdelay);
  tst = my_rint(1000.0*tscale); 
  if (pflag) printf("    amp1= %.2f  amp2= %.2f\n",amp1,amp2);
  tsig = sig1/sscale; 
  if (pflag) printf("    sig1 = %.2f  sscale = %.4f  ==> %.2f pixels\n",
		    sig1,sscale,tsig);
  g1 = gaussian_2d(xn,yn,cxc,cyc,tsig,amp1);
  tsig = sig2/sscale; 
  if (pflag) printf("    sig2 = %.2f  sscale = %.4f  ==> %.2f pixels\n",
		    sig2,sscale,tsig);
  g2 = gaussian_2d(xn,yn,(float)xc,(float)yc,tsig,amp2);
  if (outfile != NULL){
    remove_file(outfile);
    append_farray_plot(outfile,"Gauss_1_raw",g1[my_rint(cxc)],yn,1);
    append_farray_plot(outfile,"Gauss_2_raw",g2[xc],yn,1);
  }
  if (ppl != NULL)
    tfilt = paramfile_get_char_param_default(ppl,"dogt_filter","ab");
  else{
    if (onode_test_ostr(o,"dogt_filter"))
      exit_error("KERNEL_UTIL, DOG_SPACE_TIME_TENSOR",
		 "Use 'tfilter', not 'dogt_filter'");
    tfilt = onode_getpar_chr_dflt(o,"tfilter","ab");
  }
  if (t0flag == 0) 
    ti = zn-1;
  else
    ti = zn-1 - zn/2;
  if (strcmp(tfilt,"wa")==0){ 
    a = 1.0;
    b = 0.8;
    tau1 = 0.002;  
    tau2 = 0.0026; 
    n1 = 9;
    n2 = 10;
    t = wa_temporal(2*zn-1,tscale,a,b,n1,n2,tau1,tau2);
  }else if (strcmp(tfilt,"ab")==0){ 
    float tk;
    int tn;
    if (ppl != NULL){
      tk = paramfile_get_float_param_or_exit(ppl,"tab_k");
      tn = paramfile_get_int_param_or_exit(ppl,"tab_n");
    }else{
      tk = onode_getpar_flt_exit(o,"tab_k");
      tn = onode_getpar_int_exit(o,"tab_n");
    }
    t = ab_temporal(2*zn-1,tscale,tn,tk,toff); 
  }else if (strcmp(tfilt,"dexp")==0){
    float t0,tau1,tau2,tamp2,sigs;
    if (ppl != NULL){
      t0   = paramfile_get_float_param_default(ppl,"dogt_t0"  ,0.020);
      tau1 = paramfile_get_float_param_default(ppl,"dogt_tau1",0.003);
      tau2 = paramfile_get_float_param_default(ppl,"dogt_tau2",0.010);
      tamp2 = paramfile_get_float_param_default(ppl,"dogt_amp2",0.500);
      sigs = paramfile_get_float_param_default(ppl,"dogt_sigs",0.004);
    }else{
      t0   = onode_getpar_flt_dflt(o,"dogt_t0"  ,0.020);
      tau1 = onode_getpar_flt_dflt(o,"dogt_tau1",0.003);
      tau2 = onode_getpar_flt_dflt(o,"dogt_tau2",0.010);
      tamp2 = onode_getpar_flt_dflt(o,"dogt_amp2",0.500);
      sigs = onode_getpar_flt_dflt(o,"dogt_sigs",0.004);
    }
    t = dd_exp_temporal(2*zn-1,tscale,t0,tau1,tau2,tamp2,sigs);
  }else if (strcmp(tfilt,"tfilt_01")==0){
    t = filt_01_temporal(ppl,o,2*zn-1,tscale,pflag);
  }else if (strcmp(tfilt,"maxwell")==0){
    float tmax,s;
    if (ppl != NULL){
      tmax = paramfile_get_float_param_or_exit(ppl,"maxwell_tmax");
    }else{
      tmax = onode_getpar_flt_exit(o,"maxwell_tmax");
    }
    s = tmax / sqrt(2.0) / tscale;
    t = maxwell_farray((float)(zn-1),s,1.0,2*zn-1);
  }else if (strcmp(tfilt,"delta")==0){
    t = get_zero_farray(2*zn-1);
    t[zn-1] = 1.0;
  }else{
    t = NULL;
    exit_error("DOG_SPACE_TIME_TENSOR","Unknown temporal filter");
  }
  myfree(tfilt); 
  if (outfile != NULL){
    append_farray_plot(outfile,"raw_t_filter_(ms)",&(t[ti]),zn,tst);
  }
  if (hilb_flag == 1){  
    float *thil;
    thil = fft_hilbert(t,2*zn-1);
    myfree(t);
    t = thil;
  }
  if (gtsig > 0.0){
    tsig = gtsig / (1000.0*tscale);
    tcent = gtcent / (1000.0*tscale);
    gt = gaussian_one_farray((float)(zn-1)+tcent,tsig,1.0,2*zn-1);
    multiply_farrays_in_place(t,gt,2*zn-1);
    if (outfile != (char *)NULL){
      append_farray_plot(outfile,"Temporal_Gaussian_(ms)",&(gt[ti]),zn,tst);
    }
    myfree(gt);
  }
  if (tau_ad > 0.0){
    int tn;
    float *texp,tau;
    tn = 2*zn-1;
    norm_area_farray(t,tn,1.0);  
    tau = tau_ad / tscale;
    texp = one_sided_exp_farray((float)zn,tau,1.0,tn);
    norm_area_farray(texp,tn,-1.0);  
    for(i=(zn + zn/2);i<tn;i++)
      texp[i] = 0.0;
    add_to_farray(t,texp,tn);
    myfree(texp);
  }
  make_max_const_farray(t,2*zn-1,1.0);
  if (outfile != (char *)NULL){
    append_farray_plot(outfile,"Temporal_filter_(ms)",&(t[ti]),zn,tst);
  }
  filter = f3tensor(1,xn,1,yn,1,zn); 
  sump = 0.0;
  sumn = 0.0;
  sumg1 = 0.0;
  sumg2 = 0.0;
  for(i=0;i<xn;i++){
    for(j=0;j<yn;j++){
      sumg1 += g1[i][j];
      sumg2 += g2[i][j];
      for(k=0;k<zn;k++){
	t2 = ti + k - d;
	if ((t2 < 0) || (t2 >= (2*zn-1))){ 
	  tt = t[ti+k]*g1[i][j];  
	}else{
	  tt = t[ti+k]*g1[i][j] - t[t2]*g2[i][j];
	}
	filter[i+1][j+1][k+1] = tt;
	if (tt > 0.0)
	  sump += tt;
	else
	  sumn += tt;
      }
    }
  }
  if (pflag){
    if (amp2 != 0.0){
      printf("    positive and negative sums: %.4f/%.4f = %.4f\n",
	     sump,sumn,sump/-sumn);
      printf("    ratio of Gaussian area (ctr/sur): %.4f/%.4f = %.4f\n",
	     sumg1,sumg2,sumg1/sumg2);
    }else{
      printf("    positive and negative sums: %.4f and %.4f\n",
	     sump,sumn);
      printf("    Gaussian area, ctr and sur: %.4f and %.4f\n",
	     sumg1,sumg2);
    }
  }
  return filter;
}
float **separable_space(xn,yn,sscale,sigma,sf,phase)
     int xn,yn; 
     float sscale; 
     float sigma,sf;
     int phase;  
{
  int i,j;
  float **filter;
  float twopisf;
  float x,y,gx,gy,s2;
  int xc,yc; 
  if (PFLAG) printf("  SEPARABLE_SPACE\n");
  xc = (xn-1)/2;
  yc = (yn-1)/2;
  twopisf = 2.0*M_PI*sf;
  s2 = 2.0*sigma*sigma;
  filter = get_2d_farray(xn,yn);
  if (phase==0){  
    for(i=0;i<xn;i++){
      x = (float)(i-xc)*sscale;
      gx = exp(-x*x/s2);
      for(j=0;j<yn;j++){
	y = (float)(j-yc)*sscale;
	gy = exp(-y*y/s2) * cos(twopisf*y);
	filter[i][j] = gx * gy;
      }
    }
  }else{
    for(i=0;i<xn;i++){
      x = (float)(i-xc)*sscale;
      gx = exp(-x*x/s2);
      for(j=0;j<yn;j++){
	y = (float)(j-yc)*sscale;
	gy = exp(-y*y/s2) * sin(twopisf*y);
	filter[i][j] = gx * gy;
      }
    }
  }
  return filter;
}
float ***stf3d_01(xn,yn,zn,sscale,tscale,sigma,sf,sphase,n,k,theta,writeflag)
     int xn,yn,zn;
     float sscale,tscale,sigma,sf;
     int sphase,n;
     float k,theta;
     int writeflag;
{
  int i,j,t;
  float ***filter,*tfilter,**sfilter,s,*col;
  char name[SLEN];
  if (PFLAG) printf("  STF3D_01\n");
  tfilter = ab_temporal(zn,tscale,n,k,0.0);
  if (sphase == 0)
    sfilter = gabor_2d_space(xn,yn,sscale,sigma,sf,theta,0.0);
  else
    sfilter = gabor_2d_space(xn,yn,sscale,sigma,sf,theta,90.0);
  if (writeflag){
    sprintf(name,"zzz.k%d.n%d.tfilter.pl",(int)k,n);
    write_farray_plot(name,tfilter,zn);
    sprintf(name,"zzz.sf%.2f.ph%d_tfilter.pl",sf,sphase);
    col = get_column_2d_farray(sfilter,xn,yn,yn/2);
    write_farray_plot(name,col,yn);
    myfree(col);
  }
  filter = get_3d_farray(xn,yn,zn);
  for(i=0;i<xn;i++)
    for(j=0;j<yn;j++){
      s = sfilter[i][j];
      for(t=0;t<zn;t++)
	filter[i][j][t] = s * tfilter[t];
    }
  myfree(tfilter); free_2d_farray(sfilter,xn);
  return filter;
}
float ***stf3d_02(int xn, int yn, int zn,     
		  float sscale, float tscale, 
		  float theta,                
		  float sdorth, float sdpar,  
		  float sf, float sphase,     
		  int a, float tau, float fr, float phi,  
		  int writeflag){             
  int i,j,t;
  float xc,yc,sdo,sdp,sfpix;
  float ***filter,*tfilter,**sfilter,s,*col;
  char name[SLEN];
  if (PFLAG) printf("  STF3D_02\n");
  tfilter = cwq_temporal(zn,tscale,a,tau,fr,phi,0.0);
  xc = (xn-1)/2;
  yc = (yn-1)/2;
  sdo = sdorth / sscale;  
  sdp = sdpar / sscale;  
  sfpix = sf * sscale;  
  sfilter = gabor_2d_space_raw(xn,yn,xc,yc,sdo,sdp,sfpix,theta,sphase);
  if (writeflag){
    sprintf(name,"zzz.tfilter.%d.pl",(int)phi);
    write_farray_plot(name,tfilter,zn);
    sprintf(name,"zzz.sfilter.%d.pl",(int)sphase);
    col = get_column_2d_farray(sfilter,xn,yn,yn/2);
    write_farray_plot(name,col,yn);
    myfree(col);
  }
  filter = get_3d_farray(xn,yn,zn);
  for(i=0;i<xn;i++)
    for(j=0;j<yn;j++){
      s = sfilter[i][j];
      for(t=0;t<zn;t++)
	filter[i][j][t] = s * tfilter[t];
    }
  myfree(tfilter); free_2d_farray(sfilter,xn);
  return filter;
}
void quad_opponent_causal_stf3d_01(xn,yn,tn,sscale,tscale,sigma,sf,
				   n1,n2,k,fuoc,fuec,fdoc,fdec,wrflag)
     int xn,yn,tn;
     float sscale,tscale,sigma,sf;
     int n1,n2;
     float k,****fuoc,****fuec,****fdoc,****fdec;
     int wrflag;
{
  float ***filter1,***filter2,***filter3,***filter4;  
  float ***fuo,***fue,***fdo,***fde; 
  int tn2,sphase;
  if (PFLAG) printf("  QUAD_OPPONENT_CAUSAL_STF3D_01\n");
  tn2 = tn*2;
  sphase = 1;
  filter1 = stf3d_01(xn,yn,tn2,sscale,tscale,sigma,sf,sphase,n1,k,90.0,wrflag);
  filter2 = stf3d_01(xn,yn,tn2,sscale,tscale,sigma,sf,sphase,n2,k,90.0,wrflag);
  sphase = 0;
  filter3 = stf3d_01(xn,yn,tn2,sscale,tscale,sigma,sf,sphase,n2,k,90.0,wrflag);
  filter4 = stf3d_01(xn,yn,tn2,sscale,tscale,sigma,sf,sphase,n1,k,90.0,wrflag);
  fdo = add_3d_farrays(filter1,filter3,0,xn,0,yn,0,tn2);
  fue = add_3d_farrays(filter2,filter4,0,xn,0,yn,0,tn2);
  multiply_3d_farray(filter1,0,xn,0,yn,0,tn2,-1.0);
  multiply_3d_farray(filter2,0,xn,0,yn,0,tn2,-1.0);
  fuo = add_3d_farrays(filter1,filter3,0,xn,0,yn,0,tn2);
  fde = add_3d_farrays(filter2,filter4,0,xn,0,yn,0,tn2);
  free_3d_farray(filter1,xn,yn,tn2); free_3d_farray(filter2,xn,yn,tn2);
  free_3d_farray(filter3,xn,yn,tn2); free_3d_farray(filter4,xn,yn,tn2);
  get_sub_3d_farray(fuo,xn,yn,tn2,fuoc,0,xn,0,yn,tn,tn);
  get_sub_3d_farray(fue,xn,yn,tn2,fuec,0,xn,0,yn,tn,tn);
  get_sub_3d_farray(fdo,xn,yn,tn2,fdoc,0,xn,0,yn,tn,tn);
  get_sub_3d_farray(fde,xn,yn,tn2,fdec,0,xn,0,yn,tn,tn);
  free_3d_farray(fuo,xn,yn,tn2); free_3d_farray(fue,xn,yn,tn2);
  free_3d_farray(fdo,xn,yn,tn2); free_3d_farray(fde,xn,yn,tn2);
  if (PFLAG) printf("  QUAD_OPPONENT_CAUSAL_STF3D_01  Done.\n");
}
void quad_opponent_causal_stf3d_01_tensor(xn,yn,tn,sscale,tscale,sigma,sf,n1,
					  n2,k,theta,fuoc,fuec,fdoc,fdec,
					  wrflag)
     int xn,yn,tn;
     float sscale,tscale,sigma,sf;
     int n1,n2;
     float k,theta,****fuoc,****fuec,****fdoc,****fdec;
     int wrflag;
{
  float ***filter1,***filter2,***filter3,***filter4;  
  float ***fuo,***fue,***fdo,***fde; 
  float sum;
  int sphase;
  sphase = 1;
  filter1 = stf3d_01(xn,yn,tn,sscale,tscale,sigma,sf,sphase,n1,k,theta,wrflag);
  sum = sum_square_3d_farray(filter1,0,xn,0,yn,0,tn);
  multiply_3d_farray(filter1,0,xn,0,yn,0,tn,1.0/sqrt(sum));
  filter2 = stf3d_01(xn,yn,tn,sscale,tscale,sigma,sf,sphase,n2,k,theta,wrflag);
  sum = sum_square_3d_farray(filter2,0,xn,0,yn,0,tn);
  multiply_3d_farray(filter2,0,xn,0,yn,0,tn,1.0/sqrt(sum));
  sphase = 0;
  filter3 = stf3d_01(xn,yn,tn,sscale,tscale,sigma,sf,sphase,n2,k,theta,wrflag);
  sum = sum_square_3d_farray(filter3,0,xn,0,yn,0,tn);
  multiply_3d_farray(filter3,0,xn,0,yn,0,tn,1.0/sqrt(sum));
  filter4 = stf3d_01(xn,yn,tn,sscale,tscale,sigma,sf,sphase,n1,k,theta,wrflag);
  sum = sum_square_3d_farray(filter4,0,xn,0,yn,0,tn);
  multiply_3d_farray(filter4,0,xn,0,yn,0,tn,1.0/sqrt(sum));
  fuo = add_3d_farrays(filter1,filter3,0,xn,0,yn,0,tn);
  fde = add_3d_farrays(filter2,filter4,0,xn,0,yn,0,tn);
  multiply_3d_farray(filter1,0,xn,0,yn,0,tn,-1.0);
  multiply_3d_farray(filter2,0,xn,0,yn,0,tn,-1.0);
  fdo = add_3d_farrays(filter1,filter3,0,xn,0,yn,0,tn);
  fue = add_3d_farrays(filter2,filter4,0,xn,0,yn,0,tn);
  *fuoc = convert_3d_to_tensor(fuo,xn,yn,tn);
  *fuec = convert_3d_to_tensor(fue,xn,yn,tn);
  *fdoc = convert_3d_to_tensor(fdo,xn,yn,tn);
  *fdec = convert_3d_to_tensor(fde,xn,yn,tn);
  free_3d_farray(filter1,xn,yn,tn); free_3d_farray(filter2,xn,yn,tn);
  free_3d_farray(filter3,xn,yn,tn); free_3d_farray(filter4,xn,yn,tn);
  free_3d_farray(fuo,xn,yn,tn); free_3d_farray(fue,xn,yn,tn);
  free_3d_farray(fdo,xn,yn,tn); free_3d_farray(fde,xn,yn,tn);
}
float ***bde_cwq_stf3d_util(int xn, int yn, int tn,      
			    float sscale, float tscale,  
			    float theta,                 
			    float sf,                    
			    float sdo, float sdp,        
			    float ph,                    
			    int a, float tau, float tf,  
			    float tph,                   
			    float eta,                   
			    int write_flag){             
  float sum,***f1,***f2,***f,***ftens;
  f1 = stf3d_02(xn,yn,tn,sscale,tscale,theta,sdo,sdp,sf,ph,
		a,tau,tf,tph,write_flag);
  sum = sum_square_3d_farray(f1,0,xn,0,yn,0,tn);
  multiply_3d_farray(f1,0,xn,0,yn,0,tn,1.0/sqrt(sum));
  ph  -= 90.0;   
  tph -= 90.0;   
  f2 = stf3d_02(xn,yn,tn,sscale,tscale,theta,sdo,sdp,sf,ph,
		a,tau,tf,tph,write_flag);
  sum = sum_square_3d_farray(f2,0,xn,0,yn,0,tn);
  multiply_3d_farray(f2,0,xn,0,yn,0,tn,1.0/sqrt(sum));
  multiply_3d_farray(f2,0,xn,0,yn,0,tn,eta);
  f = add_3d_farrays(f1,f2,0,xn,0,yn,0,tn);
  free_3d_farray(f1,xn,yn,tn);
  free_3d_farray(f2,xn,yn,tn);
  ftens = convert_3d_to_tensor(f,xn,yn,tn);
  free_3d_farray(f,xn,yn,tn);
  return ftens;
}
void bde_cwq_stf3d_tensor(fo,mo,xn,yn,tn,sscale,tscale,fl1,fr1,fl2,fr2)
     struct onode *fo; 
     struct onode *mo; 
     int xn,yn;       
     int tn;          
     float sscale;    
     float tscale;    
     float ****fl1;   
     float ****fr1;
     float ****fl2;
     float ****fr2;
{
  int a,wrflag;
  int qsopp,phopp;  
  float sf,sdo,sdp,theta,tau,tf,phi,eta,phase,ph,phshift,tph;
  wrflag  = onode_getpar_int_dflt(fo,"write_filter",0);
  sf      = onode_getpar_flt_exit(fo,"sf");
  sdo     = onode_getpar_flt_exit(fo,"sd_orth");
  sdp     = onode_getpar_flt_exit(fo,"sd_par");
  phase   = onode_getpar_flt_exit(fo,"phase");
  theta   = onode_getpar_flt_exit(fo,"direction");
  a       = onode_getpar_int_exit(fo,"alpha");
  tau     = onode_getpar_flt_exit(fo,"tau");
  tf      = onode_getpar_flt_exit(fo,"tf");
  phi     = onode_getpar_flt_exit(fo,"phi");
  eta     = onode_getpar_flt_exit(fo,"eta");
  phshift = onode_getpar_flt_dflt(mo,"phase_shift",0.0);
  phopp = onode_getpar_int_dflt(mo,"phase_shift_opp",0);
  qsopp = onode_getpar_int_dflt(mo,"quad_shift_opp",0);
  ph = -phase;
  tph = phi;
  *fl1 = bde_cwq_stf3d_util(xn,yn,tn,sscale,tscale,theta,sf,sdo,sdp,ph,
			   a,tau,tf,tph,eta,wrflag);
  ph = -phase - phshift;  
  tph = phi;
  *fr1 = bde_cwq_stf3d_util(xn,yn,tn,sscale,tscale,theta,sf,sdo,sdp,ph,
			   a,tau,tf,tph,eta,wrflag);
  if (qsopp == 0)
    phase += 90.0;  
  else
    phase -= 90.0;  
  ph = -phase;
  tph = phi;
  *fl2 = bde_cwq_stf3d_util(xn,yn,tn,sscale,tscale,theta,sf,sdo,sdp,ph,
			   a,tau,tf,tph,eta,wrflag);
  if (phopp == 0)
    ph = -phase - phshift;  
  else
    ph = -phase + phshift;  
  tph = phi;
  *fr2 = bde_cwq_stf3d_util(xn,yn,tn,sscale,tscale,theta,sf,sdo,sdp,ph,
			   a,tau,tf,tph,eta,wrflag);
  if (wrflag != 0){
    write_3d_data_part("zzz.fl1.3d",*fl1,1,xn,1,yn,1,tn,4,2,1);
    write_3d_data_part("zzz.fr1.3d",*fr1,1,xn,1,yn,1,tn,4,2,1);
    write_3d_data_part("zzz.fl2.3d",*fl2,1,xn,1,yn,1,tn,4,2,1);
    write_3d_data_part("zzz.fr2.3d",*fr2,1,xn,1,yn,1,tn,4,2,1);
  }
}
float **get_xt_dog(xn,tn,sscale,tscale,amp1,amp2,sig1,sig2,ttype,t1,t2,t3,t4,
		   t5,t6)
     int xn,tn;
     float sscale,tscale,amp1,amp2,sig1,sig2;
     char ttype[];
     float t1,t2,t3,t4,t5,t6;
{
  int i,j;
  float x,xc,*xdata,*tdata,**xtdog;
  xc = (float)xn/2.0 * sscale;
  xdata = (float *)myalloc(xn*sizeof(float));
  for(i=0;i<xn;i++){
    x = (float)i*sscale;
    xdata[i] = (float)(amp1*func_gaussian_one(x,xc,sig1) -
		       amp2*func_gaussian_one(x,xc,sig2));
  }
  append_farray_plot("yyy.xtdog","x",xdata,xn,1);
  if (strcmp(ttype,"ab")==0){
    tdata = ab_temporal(tn,tscale,my_rint(t1),1.0/t2,0.0);
  }else if (strcmp(ttype,"wa")==0){
    tdata = wa_temporal(tn,tscale,t1,t2,my_rint(t3),my_rint(t4),t5,t6);
  }else{
    tdata = NULL;
    exit_error("GET_XT_DOG","Unknown type of temporal filter requested");
  }
  append_farray_plot("yyy.xtdog","t",tdata,tn,1);
  xtdog = get_2d_farray(xn,tn);
  for(i=0;i<xn;i++)
    for(j=0;j<tn;j++)
      xtdog[i][j] = xdata[i]*tdata[j];
  return xtdog;
}
