#include "mpi.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>  
#include <string.h>
#include <X11/Xlib.h>
#include <GL/glx.h>
#include <GL/gl.h>
#include <math.h>
#include <sys/types.h> 
#include <sys/times.h> 
#include <time.h>      
#include "my_util.h"
#include "carray_util.h"
#include "farray_util.h"
#include "glplot_util.h"
#include "mod_util.h"
#include "mod.h"
#include "mm.h"
#define MAX_DATA_N_INT   1000000  
#define MAX_DATA_N_FLOAT 1000000  
struct mm_job_struct ***mm_jtable = NULL; 
int mm_nmod;                              
int mm_nstim;                             
int mm_nrpt;                              
int mm_ndone;                             
int mm_ndoa;                              
unsigned long mm_t0;                      
unsigned long mm_t1;                      
struct mm_proc_struct *mm_ptable = NULL;
int mm_nproc;
char ggstr[LONG_SLEN];
int mm_winw = 500;
int mm_winh = 500;
float mm_zoom = 1.0;
Display *mm_dpy;
Window mm_win;
GLXContext mm_cx;
int mm_dbflag;
float mm_pixel_proc_top;     
float mm_pixel_proc_height;  
int mm_proc_current;         
int mm_gui_hold_flag = 0;    
int mm_monflag = 0;
int mm_moni;            
int mm_mon_mi = 0;      
int mm_sort = 0;        
int mm_winw2 = 1024;
int mm_winh2 = 600;
Display *mm_dpy2;
Window mm_win2;
GLXContext mm_cx2;
int mm_dbflag2;
int   mm_nlines;      
float mm_line_h;      
float mm_spike_h;     
Bool mm_true_test()
{
  return 1;
}
void mm_data_send_iarray(p0,pn,d,n,mylogf)
     int p0;        
     int pn;        
     int *d;        
     int n;         
     char *mylogf;  
{
  int i;
  if (p0<0)
    mylogx(mylogf,"MM_DATA_SEND_IARRAY","Processor ID too low");
  if (pn<1)
    mylogx(mylogf,"MM_DATA_SEND_IARRAY","Processor count too low");
  if (n<0)
    mylogx(mylogf,"MM_DATA_SEND_IARRAY","Invalid length");
  if (n > MAX_DATA_N_INT)
    mylogx(mylogf,"MM_DATA_SEND_IARRAY","Length exceeds arbitrary maximum");
  for(i=p0;i<(p0+pn);i++)
    MPI_Send(d,n,MPI_INT,i,tag_data_int,MPI_COMM_WORLD);
}
void mm_data_recv_iarray(source,mylogf,d,dmax,rn)
     int source;    
     char *mylogf;  
     int *d;        
     int dmax;      
     int *rn;       
{
  MPI_Status stat;
  if (source == -1)
    source = MPI_ANY_SOURCE;
  else if (source < -1)
    mylogx(mylogf,"MM_DATA_RECV_IARRAY","Source ID too low");
  if (dmax<0)
    mylogx(mylogf,"MM_DATA_RECV_IARRAY","Invalid length");
  MPI_Recv(d,MAX_DATA_N_INT,MPI_INT,source,tag_data_int,MPI_COMM_WORLD,&stat);
  MPI_Get_count(&stat,MPI_INT,rn);
  if (*rn > dmax)
    mylogx(mylogf,"MM_DATA_RECV_IARRAY","Too many ints");
}
void mm_data_send_farray(p0,pn,d,n,mylogf)
     int p0;        
     int pn;        
     float *d;      
     int n;         
     char *mylogf;  
{
  int i;
  if (p0<0)
    mylogx(mylogf,"MM_DATA_SEND_FLOAT","Processor ID too low");
  if (pn<1)
    mylogx(mylogf,"MM_DATA_SEND_FLOAT","Processor count too low");
  if (n > MAX_DATA_N_FLOAT)
    mylogx(mylogf,"MM_DATA_SEND_FLOAT","Length exceeds arbitrary maximum");
  for(i=p0;i<(p0+pn);i++)
    MPI_Send(d,n,MPI_FLOAT,i,tag_data_flt,MPI_COMM_WORLD);
}
void mm_data_recv_farray(source,mylogf,d,dmax,rn,rsrc)
     int source;    
     char *mylogf;  
     float *d;      
     int dmax;      
     int *rn;       
     int *rsrc;     
{
  MPI_Status stat;
  if (source == -1)
    source = MPI_ANY_SOURCE;
  else if (source < -1)
    mylogx(mylogf,"MM_DATA_RECV_FARRAY","Source ID too low");
  if (dmax < 0)
    mylogx(mylogf,"MM_DATA_RECV_FARRAY","Invalid length");
  MPI_Recv(d,MAX_DATA_N_FLOAT,MPI_FLOAT,source,tag_data_flt,MPI_COMM_WORLD,
	   &stat);
  if (source == MPI_ANY_SOURCE)
    *rsrc = stat.MPI_SOURCE; 
  MPI_Get_count(&stat,MPI_FLOAT,rn);
  if (*rn > dmax)
    mylogx(mylogf,"MM_DATA_RECV_FARRAY","Too many floats");
}
void mm_mpi_receive_carray(source,tag,nmax,mylogf,rc,rn)
     int source;    
     int tag;       
     int nmax;      
     char mylogf[]; 
     char **rc;     
     int *rn;       
{
  int i;
  int n;
  char *pstr,*c;
  MPI_Status status;
  pstr = (char *)myalloc(nmax*sizeof(char));
  MPI_Recv(pstr,nmax,MPI_CHAR,source,tag,MPI_COMM_WORLD,&status);
  MPI_Get_count(&status,MPI_CHAR,&n);
  c = (char *)myalloc(n*sizeof(char));
  for(i=0;i<n;i++)
    c[i] = pstr[i];
  myfree(pstr);
  *rc = c; *rn = n;
}
void mm_cmd_send(p0,pn,str,mylogf)
     int p0;        
     int pn;        
     char *str;     
     char *mylogf;  
{
  int i;
  if (p0<0)
    mylogx(mylogf,"MM_CMD_SEND","Processor ID too low");
  if (pn<1)
    mylogx(mylogf,"MM_CMD_SEND","Processor count too low");
  if (strlen(str) >= SLEN-1)
    mylogx(mylogf,"MM_CMD_SEND","Command string too long");
  for(i=p0;i<(p0+pn);i++){
    MPI_Send(str,strlen(str)+1,MPI_CHAR,i,tag_cmd_str,MPI_COMM_WORLD);
  }
}
char *mm_cmd_recv(p0,mylogf)
     int p0;        
     char *mylogf;  
{
  int n;
  char *tstr;
  mm_mpi_receive_carray(p0,tag_cmd_str,SLEN,mylogf,&tstr,&n);
  return tstr;
}
int mm_job_get_stimi(jnum)
     int jnum;  
{
  int k;
  int mi;
  mi = (int)(jnum/(mm_nstim * mm_nrpt));  
  k = jnum - mi*(mm_nstim * mm_nrpt);
  return k;
}
void mm_job_get_indices(jnum,rmi,rsi,rri)
     int jnum;  
     int *rmi;  
     int *rsi;  
     int *rri;  
{
  int k;
  int mi,si,ri;
  mi = (int)(jnum/(mm_nstim * mm_nrpt));  
  k = jnum - mi*(mm_nstim * mm_nrpt);
  si = (int)(k/mm_nrpt);  
  ri = k - si*mm_nrpt;    
  *rmi = mi;
  *rsi = si;
  *rri = ri;
}
struct mm_job_struct *mm_job_get_ptr(jnum)
     int jnum;  
{
  int mi,si,ri;
  struct mm_job_struct *jp;
  mm_job_get_indices(jnum,&mi,&si,&ri);
  jp = &(mm_jtable[mi][si][ri]);
  return jp;
}
void mm_mon_help()
{
  printf("\n");
  printf("  MM_MON_HELP\n");
  printf("    h - help (print this list)\n");
  printf("    p - sort trials by processor number\n");
  printf("    s - sort trials by stimulus number\n");
  printf("    > - increase model index for monitor display\n");
  printf("    < - decrease model index for monitor display\n");
  printf("    ? - help (print this list)\n");
  printf("\n");
}
void mm_mon_set_lines_height()
{
  if (mm_sort == 1){ 
    mm_nlines = mm_nstim * mm_nrpt + mm_nproc - 1;
    mm_line_h = 1.8 / (float)mm_nlines;  
  }else{ 
    mm_nlines = mm_nstim * (mm_nrpt + 1) - 1;
    mm_line_h = 1.8 / (float)mm_nlines;  
  }
  mm_spike_h = 0.9 * mm_line_h;  
}
void mm_mon_init(r)
     struct response_struct *r;  
{
  int i;
  int n,li,dur_msec;
  float x,y,w,h,yoff,xtext,colr,colg,colb,tr,tg,tb;
  char tstr[SLEN],chan_name[SLEN];
  struct mm_proc_struct *p;
  strcpy(chan_name,r->nd_name[mm_moni]);
  dur_msec = my_rint(r->tsn*1000.0);
  x = -0.90;
  xtext = -0.95;
  w = 1.8;
  colr = 0.1;
  colg = 0.2;
  colb = 0.2;
  glplot_clear_double(mm_dpy2,mm_win2);
  if (mm_nmod > 1)
    sprintf(tstr,"%s   (model %d)",chan_name,mm_mon_mi);
  else
    sprintf(tstr,"%s",chan_name);
  glplot_draw_text(mm_dpy2,mm_win2,mm_cx2,tstr,0.0,0.95,0.0,1.0,0.0);
  if (mm_sort == 1){ 
    tr = 1.0;
    tg = 1.0;
    tb = 0.0;
    y = 0.90;
    for(i=0;i<mm_nproc;i++){
      p = &(mm_ptable[i]);
      if (p->ndone > 0){
	h = (float)p->ndone * mm_line_h;  
	y -= h;                           
	yoff = (h - mm_line_h) / 2.0;
	glplot_draw_box(colr,colg,colb,x,y,w,h);
	sprintf(tstr,"%3d",i);
	glplot_draw_text(mm_dpy2,mm_win2,mm_cx2,tstr,xtext,y+yoff,tr,tg,tb);
	y -= mm_line_h;            
      }
    }
  }else{  
    tr = 0.4;
    tg = 0.8;
    tb = 0.8;
    n = mm_nstim;
    h = mm_nrpt * mm_line_h;
    yoff = (h - mm_line_h) / 2.0;
    for(i=0;i<n;i++){
      y = -0.90 + (float)i * (h + mm_line_h); 
      glplot_draw_box(colr,colg,colb,x,y,w,h);
      sprintf(tstr,"%3d",n-1-i);
      glplot_draw_text(mm_dpy2,mm_win2,mm_cx2,tstr,xtext,y+yoff,tr,tg,tb);
    }
  }
  sprintf(tstr,"%d ms",dur_msec);
  glplot_draw_text(mm_dpy2,mm_win2,mm_cx2,tstr,0.0,-0.97,tr,tg,tb);
  glClear(GL_DEPTH_BUFFER_BIT);    
}
void mm_mon_update_draw_one(r,pnum,ti,si,ri)
     struct response_struct *r;  
     int pnum;                   
     int ti;                     
     int si;                     
     int ri;                     
{
  int i;
  int n,li;
  float *s,cfx,x,y,x0;
  char tstr[SLEN];
  cfx = 1.8 / (r->tsn * 1000.0);  
  x0 = -0.90;
  li = ri + si * (mm_nrpt + 1);   
  li = mm_nlines - li;            
  y = -0.90 + (float)(li-1) * mm_line_h; 
  s = r->s[0][ti];
  n = r->cnt[0][ti];
  for(i=0;i<n;i++){
    x = x0 + cfx * s[i];
    glBegin(GL_LINE_STRIP);
    glColor3f(1.0,1.0,1.0);   
    glVertex2f(x,y);
    glVertex2f(x,y+mm_spike_h);
    glEnd();
  }
  if (mm_sort == 1)
    sprintf(tstr,"%d",si);
  else
    sprintf(tstr,"%d",pnum);
  glplot_draw_text(mm_dpy2,mm_win2,mm_cx2,tstr,0.93,y,1.0,1.0,0.0);
}
void mm_mon_update_by_proc(r)
     struct response_struct *r;  
{
  int i,j,k;
  int n,ti,si;
  float *s,cfx,x,y,x0;
  char tstr[SLEN];
  struct mm_proc_struct *p;
  struct mm_jel_struct *jel;
  cfx = 1.8 / (r->tsn * 1000.0);  
  x0 = -0.90;
  y = 0.90;
  for(i=0;i<mm_nproc;i++){
    p = &(mm_ptable[i]);
    jel = p->done;
    for(j=0;j<p->ndone;j++){
      ti = jel->j->trial;
      si = jel->j->stim;
      s = r->s[0][ti];
      n = r->cnt[0][ti];
      y -= mm_line_h;                    
      for(k=0;k<n;k++){
	x = x0 + cfx * s[k];
	glBegin(GL_LINE_STRIP);
	glColor3f(1.0,1.0,1.0);  
	glVertex2f(x,y);
	glVertex2f(x,y+mm_spike_h);
	glEnd();
      }
      sprintf(tstr,"%d",si);  
      glplot_draw_text(mm_dpy2,mm_win2,mm_cx2,tstr,0.93,y,0.4,0.8,0.8);
      jel = jel->next;
    }
    if (p->done > 0)
      y -= mm_line_h;                 
  }
}
void mm_mon_update_all(r)
     struct response_struct *r;  
{
  int i,j,k;
  struct mm_job_struct *js;
  if (mm_sort == 1){
    mm_mon_update_by_proc(r);
  }else{
    for(i=0;i<mm_nstim;i++){
      for(j=0;j<mm_nrpt;j++){
	js = &(mm_jtable[mm_mon_mi][i][j]);  
	if (js->tn > -1){  
	  mm_mon_update_draw_one(r,js->proc->id,js->trial,js->stim,js->rpt);
	}
      }
    }
  }
}
int mm_mon_handle_one_event(event)
     XEvent event;
{
  int redrawflag,n;
  XEvent evt;
  XKeyEvent *kevent;
  char kbuf[10];
  redrawflag = 0;
  switch(event.type){
  case ButtonPress:
    switch(event.xbutton.button){
    case 1:
      printf("  left\n");
      break;
    case 2:
      printf("  middle\n");
      break;
    case 3:
      printf("  right\n");
      break;
    }
    break;
  case MotionNotify:
    break;
  case KeyPress:
    kevent = (XKeyEvent *)&event;
    XLookupString(kevent,kbuf,sizeof(kbuf),NULL,NULL);
    if ((kbuf[0]=='h')||(kbuf[0]=='?')){
      mm_mon_help();
    }else if (kbuf[0]=='p'){
      mm_sort = 1; 
      mm_mon_set_lines_height();  
      redrawflag = 1;
    }else if (kbuf[0]=='s'){
      mm_sort = 0; 
      mm_mon_set_lines_height();  
      redrawflag = 1;
    }else if (kbuf[0]=='>'){
      mm_mon_mi += 1;
      if (mm_mon_mi >= mm_nmod)
	mm_mon_mi = 0;
      redrawflag = 1;
    }else if (kbuf[0]=='<'){
      mm_mon_mi -= 1;
      if (mm_mon_mi < 0)
	mm_mon_mi = mm_nmod-1;
      redrawflag = 1;
    }
    break;
  case ConfigureNotify:
    n = XQLength(mm_dpy2);
    while (n > 0){
      XPeekEvent(mm_dpy2,&evt); 
      if (evt.type == ConfigureNotify){
	XNextEvent(mm_dpy2,&event); 
      }else
	XNextEvent(mm_dpy2,&evt);   
      n -= 1;
    }
    if ((event.xconfigure.width != mm_winw2) ||
	(event.xconfigure.height != mm_winh2)){
      glViewport(0,0,event.xconfigure.width,event.xconfigure.height);
      mm_winw2 = event.xconfigure.width;
      mm_winh2 = event.xconfigure.height;
      redrawflag = 1;
    }
  case Expose:  
    break;
  }
  return redrawflag;
}
void mm_mon_event(r)
     struct response_struct *r;  
{
  int redrawflag,tflag;
  XEvent event;
  glplot_util_make_current(mm_dpy2,mm_win2,mm_cx2);
  glDrawBuffer(GL_FRONT_AND_BACK); 
  redrawflag = 0;
  while (XCheckIfEvent(mm_dpy2,&event,mm_true_test,(char *)NULL)){
    tflag = mm_mon_handle_one_event(event);
    if (tflag == 1){  
      glplot_clear_double(mm_dpy2,mm_win2);
      usleep(50000);  
      redrawflag = 1; 
    }
  }
  if (redrawflag){
    mm_mon_init(r);
    mm_mon_update_all(r);
  }
  glXSwapBuffers(mm_dpy2,mm_win2);
  glplot_util_make_current(mm_dpy,mm_win,mm_cx); 
}
char *mm_hms_for_time(t)
     int t; 
{
  int hr,min,sec,s,m;
  char tstr[SLEN],*pstr;
  s = t/1000;     
  m = s/60;       
  hr = m/60;      
  min = m - hr*60;
  sec = s - hr*3600 - min*60;
  sprintf(tstr,"%2d:%2d:%2d",hr,min,sec);
  pstr = strdup(tstr);
  return pstr;
}
void mm_suspend_processor(pnum,targ_state)
     int pnum;   
     int targ_state;  
{
  struct mm_proc_struct *pp;
  if ((pnum <= 0) || (pnum >= mm_nproc))
    exit_error("MM_SUSPEND_PROCESSOR","Invalid processor number");
  if (targ_state == 2){
    pp = &(mm_ptable[pnum]);
    if ((pp->state == 1) || (pp->state == 0)){
      pp->state = 2; 
      printf("    Processor %d suspended by user.\n",pnum);
      if (pp->job != NULL){
	pp->job->proc = NULL;  
	pp->job = NULL;        
	mm_ndoa -= 1;          
      }
    }else{
      printf("    Cannot suspend processor %d, not currently active.\n",pnum);
    }
  }
}
void mm_gui_update()
{
  int i;
  unsigned long tcurr;
  int np,ntot;
  float xname,xndone,xndonet,xntot,xtlast,xcurr,xtime,xtarg,xsr;
  float y,yinc,ytit0,ytit1,ybot0,ybot1,ybot2,ybot3,ytarg;
  float c1r,c1g,c1b,c2r,c2g,c2b;
  float ftop,cf,cfh;
  char tstr[SLEN],*pstr;
  struct mm_proc_struct *pp;
  struct tms buffer;
  glplot_util_make_current(mm_dpy,mm_win,mm_cx);
  ftop = (float)mm_winh;
  cf = 2.0 / ftop; 
  cfh = 2.0 / (float)mm_winw; 
  np = mm_nproc;
  xname   = -0.80;  
  xsr     = -0.55;
  xndonet = -0.20;
  xtarg   = -0.11;
  xndone  =  0.10;
  xntot   =  0.12;
  xtlast  =  0.40;
  xcurr   =  0.60;  
  xtime   =  0.65;
  ytit0 = -1.0 + cf * (ftop - 33.0);  
  ytit1 = -1.0 + cf * (ftop - 50.0);  
  y     = -1.0 + cf * (ftop - 75.0);  
  yinc  = cf * mm_pixel_proc_height;  
  ybot0 = -0.40;
  ybot1 = -0.50;
  ytarg = -0.60;
  ybot0 = -1.0 + cf * 110.0;  
  ybot1 = -1.0 + cf *  90.0;  
  ytarg = -1.0 + cf *  70.0;  
  ybot2 = -1.0 + cf *  50.0;  
  ybot3 = -1.0 + cf *  30.0;  
  c1r = 0.70;  c1g = 0.70;  c1b = 1.00;   
  c2r = 1.00;  c2g = 1.00;  c2b = 0.00;   
  glplot_clear(mm_dpy,mm_win);
  sprintf(tstr,"Processors");
  glplot_draw_text(mm_dpy,mm_win,mm_cx,tstr,xname,ytit1,c1r,c1g,c1b);
  sprintf(tstr,"Jobs");
  glplot_draw_text(mm_dpy,mm_win,mm_cx,tstr,xndone,ytit0,c1r,c1g,c1b);
  sprintf(tstr,"done");
  glplot_draw_text(mm_dpy,mm_win,mm_cx,tstr,xndone,ytit1,c1r,c1g,c1b);
  sprintf(tstr,"Last");
  glplot_draw_text(mm_dpy,mm_win,mm_cx,tstr,xtlast,ytit0,c1r,c1g,c1b);
  sprintf(tstr,"time");
  glplot_draw_text(mm_dpy,mm_win,mm_cx,tstr,xtlast,ytit1,c1r,c1g,c1b);
  sprintf(tstr,"Current");
  glplot_draw_text(mm_dpy,mm_win,mm_cx,tstr,xcurr,ytit0,c1r,c1g,c1b);
  sprintf(tstr,"Mod  Stim  Rpt");
  glplot_draw_text(mm_dpy,mm_win,mm_cx,tstr,xcurr,ytit1,c1r,c1g,c1b);
  ntot = 0;
  for(i=0;i<np;i++){
    pp = &(mm_ptable[i]);
    sprintf(tstr,"%2d.   %s",i,pp->name);
    if (mm_ptable[i].state == 2) 
      glplot_draw_text(mm_dpy,mm_win,mm_cx,tstr,xname,y,1.0,0.0,0.0);
    else if (mm_ptable[i].state == 3) 
      glplot_draw_text(mm_dpy,mm_win,mm_cx,tstr,xname,y,0.5,0.5,0.5);
    else
      glplot_draw_text(mm_dpy,mm_win,mm_cx,tstr,xname,y,c2r,c2g,c2b);
    if (i == mm_proc_current) 
      glplot_draw_disk(0.7,0.7,1.0,xname-cf*10.0,y,cfh*10.0,cf*10.0);
    sprintf(tstr,"%3d",pp->ndone);
    glplot_draw_text(mm_dpy,mm_win,mm_cx,tstr,xndone,y,c2r,c2g,c2b);
    ntot += pp->ndone;
    if (pp->ndone > 0){
      pstr = mm_hms_for_time(pp->donelast->j->tn);
      glplot_draw_text(mm_dpy,mm_win,mm_cx,pstr,xtlast,y,c2r,c2g,c2b);
      myfree(pstr);
    }else
      glplot_draw_text(mm_dpy,mm_win,mm_cx,"  :  :  ",xtlast,y,c2r,c2g,c2b);
    if (pp->job == NULL){
      sprintf(tstr,"    -     -     -");
    }else
      sprintf(tstr,"%4d  %4d  %4d",pp->job->mi,pp->job->stim,pp->job->rpt);
    glplot_draw_text(mm_dpy,mm_win,mm_cx,tstr,xcurr,y,c2r,c2g,c2b);
    y -= yinc;
  }
  sprintf(tstr,"-----");
  glplot_draw_text(mm_dpy,mm_win,mm_cx,tstr,xndone,ybot0,c1r,c1g,c1b);
  sprintf(tstr,"Total done:");
  glplot_draw_text(mm_dpy,mm_win,mm_cx,tstr,xndonet,ybot1,c1r,c1g,c1b);
  sprintf(tstr,"%d",ntot);
  glplot_draw_text(mm_dpy,mm_win,mm_cx,tstr,xntot,ybot1,0.2,1.0,0.2);
  sprintf(tstr,"Target:");
  glplot_draw_text(mm_dpy,mm_win,mm_cx,tstr,xtarg,ytarg,c1r,c1g,c1b);
  sprintf(tstr,"%d",mm_nstim * mm_nrpt);
  glplot_draw_text(mm_dpy,mm_win,mm_cx,tstr,xntot,ytarg,c2r,c2g,c2b);
  tcurr = 10*times(&buffer);  
  pstr = mm_hms_for_time(tcurr - mm_t0);
  glplot_draw_text(mm_dpy,mm_win,mm_cx,pstr,xtime,ybot3,1.0,0.2,0.2);
  myfree(pstr);
  sprintf(tstr,"Stimuli");
  glplot_draw_text(mm_dpy,mm_win,mm_cx,tstr,xname,ybot2,c1r,c1g,c1b);
  sprintf(tstr,"%d",mm_nstim);
  glplot_draw_text(mm_dpy,mm_win,mm_cx,tstr,xsr,ybot2,1.0,1.0,1.0);
  sprintf(tstr,"Repeats");
  glplot_draw_text(mm_dpy,mm_win,mm_cx,tstr,xname,ybot3,c1r,c1g,c1b);
  sprintf(tstr,"%d",mm_nrpt);
  glplot_draw_text(mm_dpy,mm_win,mm_cx,tstr,xsr,ybot3,1.0,1.0,1.0);
  glXSwapBuffers(mm_dpy,mm_win);
}
void mm_gui_init(r)
     struct response_struct *r;  
{
  int i;
  int ns;
  mm_pixel_proc_top = 75.0;     
  mm_pixel_proc_height = 10.0;  
  ns = r->ns;  
  mm_winh = mm_pixel_proc_height * mm_nproc; 
  mm_winh += 60;           
  mm_winh += 120;          
  if (mm_winh < 200)
    mm_winh = 200;
  glplot_init(mm_winw,mm_winh,mm_zoom,"mm",&mm_dpy,&mm_win,&mm_cx,&mm_dbflag,
	      0); 
  glplot_clear_double(mm_dpy,mm_win);
  glplot_set_xfont_generic(mm_dpy,mm_win,mm_cx,"small");
  mm_proc_current = -1;  
  mm_mon_set_lines_height();  
  if (ns > 0){
    mm_monflag = 1; 
    glplot_init(mm_winw2,mm_winh2,mm_zoom,"mm_mon",&mm_dpy2,&mm_win2,&mm_cx2,
		&mm_dbflag2,0); 
    glplot_clear_double(mm_dpy2,mm_win2);
    glplot_set_xfont_generic(mm_dpy2,mm_win2,mm_cx2,"medium");
    glplot_util_make_current(mm_dpy,mm_win,mm_cx);
    mm_moni = -1;
    i = 0;
    while((mm_moni < 0) && (i < r->n)){
      if (strcmp(r->rtype[i],"s")==0){
	mm_moni = i;
      }else{
	i += 1;
      }
    }
    if (mm_moni < 0)
      exit_error("MM_GUI_INIT","Failed to find response index");
    glplot_util_make_current(mm_dpy2,mm_win2,mm_cx2);
    glDrawBuffer(GL_FRONT_AND_BACK); 
    mm_mon_init(r);
    glXSwapBuffers(mm_dpy2,mm_win2);
    glplot_util_make_current(mm_dpy,mm_win,mm_cx); 
  }else{
    exit_error("MM_GUI_INIT","No spike responses requested");
  }
  mm_gui_update();
}
void mm_gui_help()
{
  printf("\n");
  printf("  MM_GUI_HELP\n");
  printf("\n");
  printf("  COMMANDS:\n");
  printf("    h - help (print this list)\n");
  printf("    H - hold display after all responses are done\n");
  printf("    LEFT CLICK - select processor as current\n");
  printf("    s - suspend current processor (reassign any current job)\n");
  printf("    S - reactivate current processor (may accept new jobs)\n");
  printf("    q - exit, writing any response\n");
  printf("    ? - help (print this list)\n");
  printf("\n");
  printf("  COLORS:\n");
  printf("    yellow - processor is active or idle\n");
  printf("    red    - processor has been suspended by user.\n");
  printf("    gray   - user-suspended processor can be reactivated.\n");
  printf("\n");
}
void mm_gui_event()
{
  int pi;
  XEvent event;
  XKeyEvent *kevent;
  XWindowAttributes xwa;
  char kbuf[10];
  glplot_util_make_current(mm_dpy,mm_win,mm_cx);
  if (XCheckIfEvent(mm_dpy,&event,mm_true_test,(char *)NULL)){
    switch(event.type){
    case ButtonPress:
      switch(event.xbutton.button){
      case 1:
	pi = my_rint(((float)event.xbutton.y - mm_pixel_proc_top)/
		     mm_pixel_proc_height);
	if ((pi == mm_proc_current) || (pi < 0) || (pi > mm_nproc))
	  mm_proc_current = -1;
	else
	  mm_proc_current = pi;
	break;
      case 2:
	break;
      case 3:
	break;
      }
      break;
    case MotionNotify:
      break;
    case KeyPress:
      kevent = (XKeyEvent *)&event;
      XLookupString(kevent,kbuf,sizeof(kbuf),NULL,NULL);
      if ((kbuf[0]=='h')||(kbuf[0]=='?')){
	mm_gui_help();
      }else if (kbuf[0]=='H'){
	mm_gui_hold_flag = 1;  
      }else if (kbuf[0]=='q'){
	mm_gui_hold_flag = 0;  
      }else if (kbuf[0]=='s'){
	if ((mm_proc_current > 0) && (mm_proc_current < mm_nproc))
	  mm_suspend_processor(mm_proc_current,2);
      }else if (kbuf[0]=='S'){
	if ((mm_proc_current > 0) && (mm_proc_current < mm_nproc)){
	  if (mm_ptable[mm_proc_current].state == 3){
	    mm_ptable[mm_proc_current].state = 0; 
	    printf("    Processor %d re-activated by user.\n",mm_proc_current);
	  }else if (mm_ptable[mm_proc_current].state == 2){
	    printf("    Cannot reactivate processor %d (no response yet).\n",
		   mm_proc_current);
	  }else{
	    printf("    Processor %d is already active.\n",mm_proc_current);
	  }
	}
      }
      break;
    case ConfigureNotify:
      break;
    case Expose:  
      break;
    }
  }
}
void mm_gui_hold(r,hflag)
     struct response_struct *r;  
     int hflag;
{
  if (hflag != -1)
    mm_gui_hold_flag = hflag;
  while (mm_gui_hold_flag == 1){
    mm_gui_update();
    mm_gui_event();
    mm_mon_event(r); 
    usleep(20000);   
  }
}
int mm_time_elapsed_min()
{
  unsigned long t;
  int s,m;
  struct tms buffer;
  t = 10*times(&buffer);         
  s = ((int)(t - mm_t0))/1000;   
  m = s/60;                      
  return m;
}
int mm_get_proc_state(pi)
     int pi; 
{
  if ((pi < 0) || (pi >= mm_nproc))
    exit_error("MM_GET_PROC_STATE","Bad processor ID");
  return mm_ptable[pi].state;
}
void mm_proc_allow_reactivate(pi)
     int pi; 
{
  int state;
  state = mm_get_proc_state(pi);
  if (state == 2)
    mm_ptable[pi].state = 3;
}
int mm_get_ndone()
{
  return mm_ndone;
}
void mm_print_stat()
{
  int i,j;
  int np;
  struct mm_proc_struct *pp;
  struct mm_jel_struct *jel;
  struct mm_job_struct *jp;
  np = mm_nproc;
  for(i=0;i<np;i++){
    pp = &(mm_ptable[i]);
    printf("  Proc %d did %d jobs (%s)\n",pp->id,pp->ndone,pp->name);
    struct mm_jel_struct *done;     
    jel = pp->done;
    for(j=0;j<pp->ndone;j++){
      jp = jel->j;
      printf("    Mod Stm Rpt %4d %4d %4d   Time %.1f s\n",jp->mi,jp->stim,
	     jp->rpt,(float)jp->tn/1000.0);
      jel = jel->next;
    }
  }
  printf("  Total time = %.1f s\n",(float)(mm_t1-mm_t0)/1000.0);
}
void mm_job_done(r,jnum,pnum,gui_flag,tlogf)
     struct response_struct *r;  
     int jnum;                   
     int pnum;                   
     int gui_flag;
     char *tlogf;     
{
  int mi,si,ri;
  char tstr[SLEN];
  struct mm_job_struct *jp;
  struct mm_proc_struct *pp;
  struct mm_jel_struct *jel;
  struct tms buffer;
  mm_job_get_indices(jnum,&mi,&si,&ri);
  jp = mm_job_get_ptr(jnum);
  pp = &(mm_ptable[pnum]);
  jp->t2 = 10*times(&buffer);        
  mm_t1 = jp->t2;                    
  jp->tn = (int)(jp->t2 - jp->t1);   
  jp->trial = jnum;
  jel = (struct mm_jel_struct *)myalloc(sizeof(struct mm_jel_struct));
  jel->j = jp;
  jel->next = NULL;
  if (pp->ndone == 0){
    pp->done = jel;
  }else{
    pp->donelast->next = jel;
  }
  pp->donelast = jel;
  pp->ndone += 1;    
  pp->job = NULL;    
  pp->state = 0;     
  mm_ndone += 1;
  if (tlogf != NULL){
    sprintf(tstr,"  jnum %d proc %d took %d\n",jnum,pnum,jp->tn);
    append_string_to_file(tlogf,tstr);
  }
  if (gui_flag == 1){
    mm_gui_update();
    if (r->ns > 0){
      glplot_util_make_current(mm_dpy2,mm_win2,mm_cx2);
      glDrawBuffer(GL_FRONT_AND_BACK); 
      if (mm_sort == 1){
	mm_mon_init(r);
	mm_mon_update_by_proc(r);
      }else{
	mm_mon_update_draw_one(r,pnum,jnum,si,ri);
      }
      glXSwapBuffers(mm_dpy2,mm_win2);
      glplot_util_make_current(mm_dpy,mm_win,mm_cx); 
    }
  }
}
unsigned long mm_current_time_ms()
{
  unsigned long ut;
  struct tms buffer;
  ut = 10*times(&buffer);  
  return ut;
}
void mm_assign_job(jnum,dest,mylogf,gui_flag,tlogf)
     int jnum,dest;
     char mylogf[];
     int gui_flag;
     char *tlogf;     
{
  struct mm_job_struct *jp;
  struct mm_proc_struct *pp;
  struct tms buffer;
  char tstr[SLEN];
  if (tlogf != NULL){
    MPI_Send(&jnum,1,MPI_INT,dest,tag_runtrial,MPI_COMM_WORLD);
  }
  if (jnum > -1){ 
    jp = mm_job_get_ptr(jnum);
    pp = &(mm_ptable[dest]);
    jp->proc = pp;
    jp->t1 = 10*times(&buffer);  
    pp->state = 1;  
    pp->job = jp;   
    mm_ndoa += 1;
  }
  if (tlogf != NULL){
    sprintf(tstr,"Assign job %d proc %d\n",jnum,dest);
    append_string_to_file(tlogf,tstr);
  }
  if (gui_flag == 1)
    mm_gui_update();
}
unsigned long mm_get_start_time_for_job(jnum)
     int jnum;
{
  unsigned long tms;
  struct mm_job_struct *jp;
  tms = -1;  
  if (jnum > -1){ 
    jp = mm_job_get_ptr(jnum);
    tms = jp->t1;
  }
  return tms;
}
void mm_ja_log_proc_names(tlogf)
     char *tlogf;
{
  int i;
  char tstr[SLEN];
  if (tlogf != NULL){
    for(i=0;i<mm_nproc;i++){
      sprintf(tstr,"PROC %4d %s\n",i,mm_ptable[i].name);
      append_string_to_file(tlogf,tstr);
    }
    sprintf(tstr,"NSTIM %d\n",mm_nstim);
    append_string_to_file(tlogf,tstr);
    sprintf(tstr,"NRPT %d\n",mm_nrpt);
    append_string_to_file(tlogf,tstr);
  }
}
void mm_init_job_control(nproc,nmod,nstim,nrpt,mylogf,mastername,pnames)
     int nproc;     
     int nmod;      
     int nstim;     
     int nrpt;      
     char mylogf[];      
     char mastername[];  
     char **pnames;      
{
  int i,j,k;
  int nstr;
  char *pstr;
  struct tms buffer;
  mylog(mylogf,"  MM_INIT_JOB_CONTROL\n");
  mm_ndone = mm_ndoa = 0;
  mm_nmod = nmod;
  mm_nstim = nstim;
  mm_nrpt = nrpt;
  mm_jtable = (struct mm_job_struct ***)myalloc(nmod*
				        sizeof(struct mm_job_struct **));
  for(k=0;k<nmod;k++){
    mm_jtable[k] = (struct mm_job_struct **)myalloc(nstim*
					    sizeof(struct mm_job_struct *));
    for(i=0;i<nstim;i++){
      mm_jtable[k][i] = (struct mm_job_struct *)myalloc(nrpt*
						sizeof(struct mm_job_struct));
      for(j=0;j<nrpt;j++){
	mm_jtable[k][i][j].mi = k;
	mm_jtable[k][i][j].stim = i;
	mm_jtable[k][i][j].rpt = j;
	mm_jtable[k][i][j].trial = -1;
	mm_jtable[k][i][j].proc = NULL;
	mm_jtable[k][i][j].t1 = -1;
	mm_jtable[k][i][j].t2 = -1;
	mm_jtable[k][i][j].tn = -1;
      }
    }
  }
  mm_nproc = nproc;
  mm_ptable = (struct mm_proc_struct *)myalloc(nproc*
					       sizeof(struct mm_proc_struct));
  for(i=0;i<nproc;i++){
    mm_ptable[i].name = NULL;
    mm_ptable[i].id = i;
    mm_ptable[i].state = -1;
    mm_ptable[i].job = NULL;
    mm_ptable[i].ndone = 0;
    mm_ptable[i].done = NULL;
    mm_ptable[i].donelast = NULL;
  }
  mm_ptable[0].name = strdup(mastername);
  if (pnames == NULL){
    for(i=1;i<nproc;i++){
      mm_mpi_receive_carray(i,tag_procname,10000,mylogf,&pstr,&nstr);
      mm_ptable[i].name = strdup(pstr);
      mm_ptable[i].state = 0; 
      myfree(pstr);
    }
  }else{
    for(i=1;i<nproc;i++){
      mm_ptable[i].name = strdup(pnames[i]);
      mm_ptable[i].state = 0; 
    }
  }
  mm_t0 = 10*times(&buffer);  
}
int mm_get_fastest_free_proc()
{
  int i;
  int pnum;
  float tmin;
  struct mm_proc_struct *p;
  pnum = -1;
  for(i=1;i<mm_nproc;i++){ 
    p = &(mm_ptable[i]);  
    if (p->state == 0){
      if (pnum < 0){ 
	pnum = i;
	if (p->ndone > 0) 
	  tmin = p->donelast->j->tn; 
	else
	  tmin = -1;      
      }else{
	if (p->ndone > 0){
	  if (tmin > p->donelast->j->tn){ 
	    pnum = i;
	    tmin = p->donelast->j->tn;
	  }
	}else{
	  if (tmin > -1){
	    pnum = i;      
	    tmin = -1;
	  }
	}
      }
    }
  }
  return pnum;
}
int mm_get_next_avail_rpt(mi,stim)
     int mi;
     int stim;
{
  int i;
  int r,done;
  struct mm_job_struct *jp;
  r = -1;
  done = 0;
  i = 0;
  while(!done && (i < mm_nrpt)){
    jp = &(mm_jtable[mi][stim][i]);
    if (jp->proc == NULL){
      done = 1;
      r = i;
    }else
      i += 1;
  }
  return r;
}
struct mm_job_struct *mm_get_next_job_for_proc(pnum)
     int pnum;
{
  int m,s,r;
  int done;
  struct mm_proc_struct *p;
  struct mm_job_struct *jp;
  p = &(mm_ptable[pnum]);   
  jp = NULL;                
  if (p->donelast != NULL){
    m = p->donelast->j->mi;
    s = p->donelast->j->stim;
    r = mm_get_next_avail_rpt(m,s);
    if (r >= 0){
      jp = &(mm_jtable[m][s][r]);
    }
  }
  if (jp == NULL){
    if (p->donelast != NULL){  
      m = p->donelast->j->mi;
      done = 0;
      s = 0;
      while(!done && (s < mm_nstim)){
	if (mm_jtable[m][s][0].proc == NULL){  
	  jp = &(mm_jtable[m][s][0]);
	  done = 1;
	}else{
	  s += 1;
	}
      }
    }else{  
      done = 0;
      m = 0;
      while(!done && (m < mm_nmod)){
	if (mm_jtable[m][0][0].proc == NULL){  
	  jp = &(mm_jtable[m][0][0]);
	  done = 1;
	}else{
	  m += 1;
	}
      }
    }
  }
  if (jp == NULL){
    done = 0;
    m = s = r = 0;
    while(!done){
      if (mm_jtable[m][s][r].proc == NULL){  
	jp = &(mm_jtable[m][s][r]);
	done = 1;
      }else{
	r += 1; 
	if (r >= mm_nrpt){
	  r = 0;
	  s += 1; 
	  if (s >= mm_nstim){
	    s = 0;
	    m += 1;
	    if (m >= mm_nmod){
	      printf("  mm_ndone = %d\n",mm_ndone);
	      printf("  mm_ndoa = %d\n",mm_ndoa);
	      exit_error("MM_GET_NEXT_JOB_FOR_PROC","This should not happen");
	    }
	  }
	}
      }
    }
  }
  return jp;
}
int mm_suggest_next_assignment(rp,rj)
     int *rp;   
     int *rj;   
{
  int i;
  int p,j,flag;
  struct mm_job_struct *jp;
  struct mm_proc_struct *pp;
  time_t rawtime; 
  struct tm *timeinfo;
  for(i=0;i<mm_nproc;i++){
    pp = &(mm_ptable[i]);
    if ((pp->state == 2) && (pp->job != NULL)){
      pp->job->proc = NULL;  
      pp->job = NULL;        
      mm_ndoa -= 1;          
    }
  }
  time(&rawtime);
  timeinfo = localtime(&rawtime);
  if (mm_ndone == (mm_nmod * mm_nstim * mm_nrpt))
    flag = -2;
  else if (mm_ndoa == (mm_nmod * mm_nstim * mm_nrpt))
    flag = -1;
  else{
    p = mm_get_fastest_free_proc();
    if (p <= 0)
      flag = 0;
    else{
      flag = 1;
      jp = mm_get_next_job_for_proc(p);
      if (jp == NULL){
	printf("MM_SUGGEST_NEXT_ASSIGNMENT:  THIS SHOULD NOT HAPPEN\n");
	exit(0);
      }else{
	j = jp->mi * (mm_nstim * mm_nrpt) + jp->stim * mm_nrpt + jp->rpt;
      }
    }
  }
  *rp = p; *rj = j;
  return flag;
}
void mm_fit_assign_job(jnum,dest,mylogf,gui_flag,tlogf)
     int jnum,dest;
     char mylogf[];
     int gui_flag;
     char *tlogf;     
{
  int ccode;
  struct mm_job_struct *jp;
  struct mm_proc_struct *pp;
  struct tms buffer;
  char tstr[SLEN];
}
