/*C  (C) Copyright  Wyeth Bair 1993-2010                                     */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "my_util.h"
#include "nr_util.h"
#include "carray_util.h"
#include "iarray_util.h"
#include "farray_util.h"
#include "plot_util.h"
#include "func_util.h"
#include "myrand_util.h"
#include "data_util.h"
#include "paramfile_util.h"
#include "fft_util.h"
#include "kernel_util.h"
#include "ifc_util.h"
#include "mod_mesh_util.h"
#include "mod_util.h"
#include "pop_util.h"
#include "mod.h" 
#include "paramfile.h"
static int myid = -1;         
static char *mylogf = NULL;   
char ggstr[LONG_SLEN];        
struct mod_dog_filt{ 
  char *name;        
  float sig1;        
  float sig2;        
  float amp1;        
  float amp2;        
  float cs_x;        
  float cs_y;
  float cs_delay;    
  float tsig;
  float tcent;
  float tdelay;
  float tau_ad;
  float power;
  float ***f;        
  float  **s;        
  float ***r;        
  float ***rr;       
  float ***rg;       
  float ***rrg;      
  int oldstimno;     
  float fsum;        
  float psum;        
  float nsum;        
};
int mod_dog_fsn;                   
struct mod_dog_filt **mod_dog_fs;  
int     mod_dog_stim_si = -1;      
float **mod_dog_stim_s  = NULL;    
float ***mod_dog_surr   = NULL;    
float  **mod_dog_surr_s = NULL;    
float ***mod_dog_sup    = NULL;    
float ***mod_dog_sup_r  = NULL;    
float mod_dog_sscale;
float mod_dog_tscale;
int   mod_dog_xn;
int   mod_dog_yn;
int   mod_dog_tn;
float mod_dog_surr_sig;    
float mod_dog_surr_tau;    
int      mod_dog_simp_n;    
int     *mod_dog_simp_nin;  
int    **mod_dog_simp_x;    
int    **mod_dog_simp_y;    
float  **mod_dog_simp_w;    
float ***mod_dog_simp_rf;   
int   ***mod_dog_simp_flag; 
int    mod_dog_cg_flag;     
int    mod_dog_ds_nsub;     
int   *mod_dog_ds_subx1;    
int   *mod_dog_ds_suby1;    
int   *mod_dog_ds_subx2;    
int   *mod_dog_ds_suby2;    
float *mod_dog_ds_subd;     
int mod_dog_sq_flag;        
int mod_dog_sq_dump;        
int mod_dog_squash_flag = 0;  
int    sqtab_n;           
float  sqtab_max;         
float  sqtab_off_pre;     
float  sqtab_dx;          
float *sqtab = NULL;      
int     mod_dog_nl_flag;      
int     mod_dog_nl_num;       
int     mod_dog_nl_length;    
int     mod_dog_nl_sigdump;   
int     mod_dog_nl_fildump;   
float   mod_dog_nl_gstep;     
float   mod_dog_nl_tres;      
float   mod_dog_nl_mag;       
float   mod_dog_nl_off;       
float **mod_dog_nl_filt = NULL;    
float   mod_dog_nl_tau;       
char *mod_dog_logistic_name;     
float mod_dog_logistic_slope;
float mod_dog_logistic_asymp; 
struct mod_dog_filt *mod_dog_fs_by_name(name)
     char *name;              
{
  int i;
  int n;
  struct mod_dog_filt *mdfs;
  mdfs = NULL;
  n = 0;
  for(i=0;i<mod_dog_fsn;i++){
    if (strcmp(name,mod_dog_fs[i]->name)==0){
      mdfs = mod_dog_fs[i];
      n += 1;
    }
  }
  if (n == 0){
    sprintf(ggstr,"  name:  %s\n",name);
    mylog(mylogf,ggstr);
    mylog_exit(mylogf,"MOD_DOG_FS_BY_NAME  population name not found.\n");
  }else if (n > 1){
    sprintf(ggstr,"  found %d times, name:  %s\n",n,name);
    mylog(mylogf,ggstr);
    mylog_exit(mylogf,"MOD_DOG_FS_BY_NAME  population name not unique.\n");
  }
  return mdfs;
}
void model_dog_get_sampling_points(xn,yn,n,seed,ssd,rfs,wsd,eps,rsx,rsy,rn)
     int xn,yn,n,seed;
     float ssd,rfs,wsd,eps;
     int **rsx,**rsy,*rn;
{
  int i,j,k;
  int xx,yy,xi,yi,wxn,wyn,wxc,wyc,wflag,*sx,*sy,nfail,failmax;
  float **rf,min,max,xc,yc;
  float **wm,**wcurr;
  char tstr[SLEN];
  mylog(mylogf,"  MODEL_DOG_GET_SAMPLING_POINTS\n");
  failmax = 400;
  wxn = wyn = 1 + 2.0*my_rint(wsd * 3.0);
  sprintf(ggstr,"    Weight mask is %d x %d\n",wxn,wyn);
  mylog(mylogf,ggstr);
  wxc = (wxn-1)/2;
  wyc = (wyn-1)/2;
  wm = gaussian_2d(wxn,wyn,(float)wxc,(float)wyc,wsd,1.0);
  wcurr = get_zero_2d_farray(xn,yn);
  write_2d_data("zz.rf.wmask.2d",wm,0,0,wxn,wyn,4,2,1,0);
  if (seed > 0)
    seed = -seed;
  xc = (float)(xn - 1)/2.0;
  yc = (float)(yn - 1)/2.0;
  rf = gaussian_2d(xn,yn,xc,yc,ssd,1.0);
  make_max_const_2d_farray(rf,xn,yn,1.0);
  for(i=0;i<xn;i++)
    for(j=0;j<yn;j++)
      if (rf[i][j] < eps)
	rf[i][j] = 0.0;
  max = max_of_2d_farray(rf,xn,yn);
  min = min_of_2d_farray(rf,xn,yn);
  sprintf(ggstr,"    RF min, max =  %f %f\n",min,max);
  mylog(mylogf,ggstr);
  write_2d_data("zz.rf.2d",rf,0,0,xn,yn,4,2,1,0);
  sx = (int *)myalloc(n*sizeof(int));
  sy = (int *)myalloc(n*sizeof(int));
  nfail = 0;
  k = 0;
  while((k<n) && (nfail < failmax)){
    xi = (int)((float)xn * myrand_util_ran2(&seed));
    yi = (int)((float)yn * myrand_util_ran2(&seed));
    if (myrand_util_ran2(&seed) < rf[xi][yi]){
      wflag = 1; 
      for(i=0;i<wxn;i++){
	xx = xi - wxc + i;
	if ((xx >= 0)&&(xx < xn)){
	  for(j=0;j<wyn;j++){
	    yy = yi - wyc + j;
	    if ((yy >= 0)&&(yy < yn)){
	      if (rf[xx][yy] > 0.0){
		if ((wcurr[xx][yy] + wm[i][j]) > (rfs * rf[xx][yy])){
		  wflag = 0;
		}
	      }
	    }
	  }
	}
      } 
      if (wflag == 1){ 
	for(i=0;i<wxn;i++){
	  xx = xi - wxc + i;
	  if ((xx >= 0)&&(xx < xn)){
	    for(j=0;j<wyn;j++){
	      yy = yi - wyc + j;
	      if ((yy >= 0)&&(yy < yn))
		wcurr[xx][yy] += wm[i][j];
	    }
	  }
	}
	sx[k] = xi;
	sy[k] = yi;
	k += 1;
	nfail = 0;
      }else{
	nfail += 1;
      }
    }
  }
  if (nfail < failmax){
    sprintf(ggstr,
	    "    Chose %d sampling points (failure limit not exceeded)\n",k);
    mylog(mylogf,ggstr);
  }else{
    sprintf(ggstr,
	    "    Chose %d sampling points (stopped after %d failures)\n",k,
	    nfail);
    mylog(mylogf,ggstr);
  }
  write_2d_data("zz.rf.weight.2d",wcurr,0,0,xn,yn,4,2,1,0);
  if (myid == -1){
    remove_file("zz.rf.points.pl");
    for(i=0;i<k;i++){
      sprintf(tstr,"%d %d\n",sx[i],sy[i]);
      append_string_to_file("zz.rf.points.pl",tstr);
    }
    remove_file("zz.rf.slice.pl");
    for(i=0;i<yn;i++){
      sprintf(tstr,"%d",i);
      append_farray_plot("zz.rf.slice.pl",tstr,wcurr[i],xn,1);
    }
  }
  free_2d_farray(rf,xn); free_2d_farray(wm,wxn); free_2d_farray(wcurr,xn);
  *rsx = sx; *rsy = sy; *rn = k;
}
void mod_dog_squash_table_prep(m)
     struct model_struct *m; 
{
  int i;
  int dump_flag;
  float x,*sqtabx;
  struct onode *o;
  mylog(mylogf,"  MOD_DOG_SQUASH_TABLE_PREP\n");
  if (m->o != NULL)
    o = onode_get_node_type_item_val(m->o,"pop","name","lgn_on");
  else
    o = NULL;
  if (o == NULL){
    sqtab_max = paramfile_get_float_param_or_exit(m->ppl,"squash_out_max");
    sqtab_off_pre = paramfile_get_float_param_default(m->ppl,
						      "squash_offset_pre",0.0);
  }else{
    sqtab_max = onode_getpar_flt_exit(o,"squash_out_max");
    sqtab_off_pre = onode_getpar_flt_dflt(o,"squash_offset_pre",0.0);
  }
  sqtab_n = 100;         
  sqtab_dx = 4.0* sqtab_max / (float)sqtab_n;      
  if (sqtab != NULL)
    myfree(sqtab);
  sqtab = get_farray(sqtab_n);
  sqtabx = get_farray(sqtab_n);
  x = 0.0;
  for(i=0;i<sqtab_n;i++){
    sqtab[i] = sqtab_max * func_sigmoid_tanh(x*2.0/sqtab_max,1.0,0.0);
    sqtabx[i] = x;
    x += sqtab_dx;
  }
  if (myid == -1){
    if (o == NULL)
      dump_flag = paramfile_get_float_param_default(m->ppl,"squash_dump",0);
    else
      dump_flag = onode_getpar_int_dflt(o,"squash_dump",0);
    if (dump_flag == 1)
      append_farray_xy_plot("zz.sqtab.pl",sqtabx,sqtab,sqtab_n,"Squash_Table");
  }
  myfree(sqtabx);
}
void mod_dog_nonlin1_prep(m)
     struct model_struct *m; 
{
  int i,j; 
  float gain,posamp,negamp,negadj,posadj,minratio; 
  float alpha,beta,a,scale,magadj,mag,negscale,posscale; 
  struct onode *o; 
  sprintf(ggstr,"  MOD_DOG_NONLIN1_PREP\n");
  mylog(mylogf,ggstr);
  if (m->ppl != NULL){
    mod_dog_nl_num = paramfile_get_int_param_default(m->ppl,"nl_num",20);
    mod_dog_nl_length = paramfile_get_int_param_default(m->ppl,"nl_length",40);
    mod_dog_nl_sigdump= paramfile_get_int_param_default(m->ppl,"nl_sigdump",0);
    mod_dog_nl_fildump= paramfile_get_int_param_default(m->ppl,"nl_fildump",0);
    mod_dog_nl_gstep = paramfile_get_float_param_default(m->ppl,"nl_gstep",
							 0.1);
    mod_dog_nl_tres = paramfile_get_float_param_default(m->ppl,"nl_tres",
							 0.1);
    mod_dog_nl_mag = paramfile_get_float_param_default(m->ppl,"nl_mag",
							0.001);
    mod_dog_nl_off = paramfile_get_float_param_default(m->ppl,"nl_offset",
							0.0);
    mod_dog_nl_tau = paramfile_get_float_param_default(m->ppl,"nl_tau",20);
    posamp = paramfile_get_float_param_default(m->ppl,"nl_posamp",1.0);
    negamp = paramfile_get_float_param_default(m->ppl,"nl_negamp",1.0);
    minratio = paramfile_get_float_param_default(m->ppl,"nl_minratio",1.0);
    magadj = paramfile_get_float_param_default(m->ppl,"nl_magadj",1.0);
    alpha = paramfile_get_float_param_default(m->ppl,"nl_alpha",2.0);
    beta = paramfile_get_float_param_default(m->ppl,"nl_beta",0.5);
  }else{
    o = onode_get_node_type_item_val(m->o,"pop","name","lgn_on");
    mod_dog_nl_num = onode_getpar_int_dflt(o,"nl_num",20);
    mod_dog_nl_length = onode_getpar_int_dflt(o,"nl_length",40);
    mod_dog_nl_sigdump = onode_getpar_int_dflt(o,"nl_sigdump",0);
    mod_dog_nl_fildump = onode_getpar_int_dflt(o,"nl_fildump",0);
    mod_dog_nl_gstep = onode_getpar_flt_dflt(o,"nl_gstep",0.1);
    mod_dog_nl_tres = onode_getpar_flt_dflt(o,"nl_tres",0.1);
    mod_dog_nl_mag = onode_getpar_flt_dflt(o,"nl_mag",0.001);
    mod_dog_nl_off = onode_getpar_flt_dflt(o,"nl_offset",0.0);
    mod_dog_nl_tau = onode_getpar_flt_dflt(o,"nl_tau",20);
    posamp = onode_getpar_flt_dflt(o,"nl_posamp",1.0);
    negamp = onode_getpar_flt_dflt(o,"nl_negamp",1.0);
    minratio = onode_getpar_flt_dflt(o,"nl_minratio",1.0);
    magadj = onode_getpar_flt_dflt(o,"nl_magadj",1.0);
    alpha = onode_getpar_flt_dflt(o,"nl_alpha",2.0);
    beta = onode_getpar_flt_dflt(o,"nl_beta",0.5);
  }
  sprintf(ggstr,"   num %i, length %i, gstep %f, gmod %f, tau %f, offset %f\n",
	  mod_dog_nl_num,mod_dog_nl_length,mod_dog_nl_gstep,mod_dog_nl_mag,
	  mod_dog_nl_tau,mod_dog_nl_off);
  mylog(mylogf,ggstr);
  sprintf(ggstr,"    posamp: %f, negamp: %f, alpha: %f, beta: %f\n",posamp,
	  negamp,alpha,beta);
  mylog(mylogf,ggstr);
  if (mod_dog_nl_filt != NULL){ 
    free_2d_farray(mod_dog_nl_filt,mod_dog_nl_num);
  }
  mod_dog_nl_filt = (float **)myalloc(mod_dog_nl_num*sizeof(float *));
  for(i = 0; i < mod_dog_nl_num; i++){
    gain = mod_dog_nl_gstep*i; 
    a = alpha ;
    if (mod_dog_nl_num == 1) {
      negscale = 1;
      posscale = 1;
    } else {
      negscale = (minratio +(1-minratio)*((float)i/(float)(mod_dog_nl_num-1)));
      posscale = (magadj + (1-magadj)*((float)i/(float)(mod_dog_nl_num-1)));
    }
    negadj = negscale*posscale*negamp;
    posadj = posscale*posamp;
    mod_dog_nl_filt[i] = 
      (float *)diff_inv_gamma_farray(mod_dog_tscale,a,beta,posadj,
				     negadj,mod_dog_nl_length);
    if(mod_dog_nl_fildump){
      sprintf(ggstr,"NL_Filter_%i",i);
      append_farray_plot("zz.nl_fildump.pl",ggstr,mod_dog_nl_filt[i],
			 mod_dog_nl_length,1);
    }
  }
}
void mod_dog_nonlin1(m,t,tn)
     struct model_struct *m; 
     float *t;               
     int tn;                 
{
  int i,j, filtnum, inum;
  int *indices, *ival, *icount;
  float sum, meannl, sdevnl, meanl, sdevl, meang, sdevg;
  float *tempt,*gainsums;
  sprintf(ggstr,"    MOD_DOG_NONLIN1\n");
  mylog(mylogf,ggstr);
  tempt = (float *)myalloc(tn*sizeof(float));
  gainsums = (float *)myalloc(tn*sizeof(float));
  indices = (int *)myalloc(tn*sizeof(int));
  ival = (int *)myalloc(mod_dog_nl_num*sizeof(int));
  icount = (int *)myalloc(mod_dog_nl_num*sizeof(int));
  find_nl_bin_index(indices,mod_dog_nl_num,tn,t,mod_dog_nl_tres,
		    mod_dog_nl_tau,gainsums);
  get_unique_count_iarray(indices,tn,&ival,&icount,&inum);
  sprintf(ggstr,"      Index count: ");
  mylog(mylogf,ggstr);
  for(i=0;i<inum;i++) {
    sprintf(ggstr,"%i:%i ",ival[i],icount[i]);
    mylog(mylogf,ggstr);
  }
  sprintf(ggstr,"\n");
  mylog(mylogf,ggstr);
  if(mod_dog_nl_sigdump){
    append_farray_plot("zz.nl_sigdump.pl","Linear_Drive",t,tn,1);
    append_farray_plot("zz.nl_sigdump.pl","Gain_sums",gainsums,tn,1);
  }
  mean_sdev_farray(t,tn,&meanl,&sdevl);
  sprintf(ggstr,"      Linear signal: m: %f, sd: %f\n",meanl,sdevl);
  mylog(mylogf,ggstr);
  mean_sdev_farray(gainsums,tn,&meang,&sdevg);
  sprintf(ggstr,"      Gain signal: m: %f, sd: %f\n",meang,sdevg);
  mylog(mylogf,ggstr);
  for(i=0; i<tn; i++){
    tempt[i] = t[i];
    sum = 0.0;
    for(j=0; j < mod_dog_nl_length && j<=i; j++){
      sum += tempt[i-j]*mod_dog_nl_mag*mod_dog_nl_filt[indices[i]][j];
    }
    t[i] = sum + mod_dog_nl_off;
  }
  mean_sdev_farray(t,tn,&meannl,&sdevnl);
  sprintf(ggstr,"      Non-Linear signal: m: %f, sd: %f\n",meannl,sdevnl);
  mylog(mylogf,ggstr);
  if(mod_dog_nl_sigdump){
    append_farray_plot("zz.nl_sigdump.pl","NonLinear_Output",t,tn,1);
  }
  myfree(tempt);
  myfree(gainsums);
  myfree(indices);
  myfree(ival);
  myfree(icount);
}
void mod_dog_squash_blake(m,t,tn,pflag)
     struct model_struct *m; 
     float *t;               
     int tn;                 
     int pflag; 
{
  int i;
  int squashflag; 
  float severity,midpoint,scale,shift; 
  float inm, insdev, sqm, sqsdev; 
  struct onode *o; 
  if (m->ppl != NULL){
    mod_dog_sq_dump = paramfile_get_int_param_default(m->ppl,"sq_dump",0);
    squashflag = paramfile_get_int_param_default(m->ppl,"sq_type",0);
    midpoint = paramfile_get_float_param_default(m->ppl,"sq_midpoint",0.0);
    severity = paramfile_get_float_param_default(m->ppl,"sq_severity",1.0);
    scale = paramfile_get_float_param_default(m->ppl,"sq_scale",1.0);
    shift = paramfile_get_float_param_default(m->ppl,"sq_shift",0.0);
  }else{
    o = onode_get_node_type_item_val(m->o,"pop","name","lgn_on");
    mod_dog_sq_dump = onode_getpar_int_dflt(o,"sq_dump",0);
    squashflag = onode_getpar_int_dflt(o,"sq_type",0);
    midpoint = onode_getpar_flt_dflt(o,"sq_midpoint",0.0);
    severity = onode_getpar_flt_dflt(o,"sq_severity",1.0);
    scale = onode_getpar_flt_dflt(o,"sq_scale",1.0);
    shift = onode_getpar_flt_dflt(o,"sq_shift",0.0);
  }
  if (pflag){
    sprintf(ggstr,"    MOD_DOG_SQUASH\n");
    mylog(mylogf,ggstr);
    sprintf(ggstr,"      Midpoint: %f, severity: %f, scale: %f, shift: %f\n",
	    midpoint,severity,scale,shift);
    mylog(mylogf,ggstr);
  }
  if(mod_dog_sq_dump)
    append_farray_plot("zz.sqdump.pl","Input_to_Squash",t,tn,1);
  mean_sdev_farray(t,tn,&inm,&insdev);
  if (pflag){
    sprintf(ggstr,"      Input: m: %f, sd: %f\n",inm,insdev);
    mylog(mylogf,ggstr);
  }
  for(i=0; i<tn; i++){
    if(squashflag == 0) 
      t[i] = scale*(midpoint - (float)func_sigmoid_tanh((t[i]+shift),
							severity,0.0));
    else if (squashflag == 1) 
      t[i] = (float)func_sigmoid_log(t[i],scale,midpoint,severity,shift); 
  }
  if(mod_dog_sq_dump)
    append_farray_plot("zz.sqdump.pl","Output_from_Squash",t,tn,1);
  mean_sdev_farray(t,tn,&sqm,&sqsdev);
  if (pflag){
    sprintf(ggstr,"      Squash: m: %f, sd: %f\n",sqm,sqsdev);
    mylog(mylogf,ggstr);
  }
}
void mod_dog_squash(d,n)
     float *d;  
     int n;     
{
  int i,k;
  float x,dx;
  for(i=0;i<n;i++){
    x = d[i] + sqtab_off_pre;
    if (x > 0.0){  
      k = x / sqtab_dx;
      if (k >= (sqtab_n - 1))
	x = sqtab_max;  
      else{ 
	dx = (x - (float)k*sqtab_dx) / sqtab_dx;
	x = sqtab[k] + (sqtab[k+1] - sqtab[k]) * dx;
      }
      d[i] = x - sqtab_off_pre;
    }
  }
}
void model_dog_01_prep(m,r)
     struct model_struct *m;    
     struct response_struct *r; 
{
  int xn,yn,tn,ix,iy,iz,pflag;
  float ***dog,tscale,sscale,surr_sig,surr_tau;
  char *outfile,*prefix;
  struct param_pair_list *mppl;
  struct mod_dog_filt *f;
  myid = mylog_set_id_log(m->process_id,&mylogf);  
  mylog(mylogf,"  MODEL_DOG_01_PREP\n");
  pflag = 0;
  if (myid == -1)
    pflag = 1;
  prefix = paramfile_get_char_param_default(r->ppl,"outfile","zz");
  mppl = m->ppl;
  sscale = paramfile_get_float_param_or_exit(mppl,"sscale");
  tscale = paramfile_get_float_param_or_exit(mppl,"tscale");
  xn = paramfile_get_int_param_or_exit(mppl,"xn");
  yn = paramfile_get_int_param_or_exit(mppl,"yn");
  tn = paramfile_get_int_param_or_exit(mppl,"tn");
  mod_dog_sscale = sscale;
  mod_dog_tscale = tscale;
  mod_dog_xn = xn;
  mod_dog_yn = yn;
  mod_dog_tn = tn;
  f = (struct mod_dog_filt *)myalloc(sizeof(struct mod_dog_filt));
  f->sig1 = paramfile_get_float_param_or_exit(mppl,"sig1");
  f->sig2 = paramfile_get_float_param_or_exit(mppl,"sig2");
  f->amp1 = paramfile_get_float_param_or_exit(mppl,"amp1");
  f->amp2 = paramfile_get_float_param_or_exit(mppl,"amp2");
  f->cs_x = paramfile_get_float_param_default(mppl,"cs_x",0.0);
  f->cs_y = paramfile_get_float_param_default(mppl,"cs_y",0.0);
  f->cs_delay = paramfile_get_float_param_or_exit(mppl,"cs_delay");
  f->tsig = paramfile_get_float_param_or_exit(mppl,"tsig");
  f->tcent = paramfile_get_float_param_default(mppl,"tcent",0.0);
  f->tdelay = paramfile_get_float_param_default(mppl,"tdelay",0.0);
  f->tau_ad = paramfile_get_float_param_default(mppl,"tau_ad",0.0);
  f->power = paramfile_get_float_param_default(mppl,"power",1.0);
  mod_dog_nl_flag = paramfile_get_int_param_default(mppl,"nl_flag",0);
  mod_dog_sq_flag = paramfile_get_int_param_default(mppl,"sq_flag",0);
  mod_dog_cg_flag = paramfile_get_int_param_default(mppl,"contrast_gain",0);
  mylog(mylogf,"    Computing 3D DOG filter.\n");
  outfile = paramfile_make_fname(mppl,"write_dog_temporal",prefix,".dogt.pl");
  f->f = dog_space_time_tensor(xn,yn,tn,sscale,tscale,f->sig1,f->sig2,f->amp1,
			       f->amp2,f->cs_x,f->cs_y,f->cs_delay,f->tsig,
			       f->tcent,outfile,mppl,m->o,pflag,0,0,0.0,
			       f->tau_ad);
  kernel_util_norm_xyt_filter(f->f,1,xn,1,yn,1,tn,&(f->psum),&(f->nsum));
  printf("  Filter 1, sum pos,neg  %f %f\n",f->psum,f->nsum);
  dog = f->f;
  mod_dog_fsn = 1;
  mod_dog_fs = (struct mod_dog_filt **)myalloc(mod_dog_fsn*
					       sizeof(struct mod_dog_filt *));
  mod_dog_fs[0] = f;
  if (outfile != (char *)NULL)
    myfree(outfile);
  f->fsum = sum_3d_farray(dog,1,xn,1,yn,1,tn);
  sprintf(ggstr,"    sum over 3d filter = %f",f->fsum);
  mylog(mylogf,ggstr);
  get_max_coord_3d_farray(dog,1,xn,1,yn,1,tn,&ix,&iy,&iz);
  sprintf(ggstr,"    Max at (%d,%d,%d) in [1,%d][1,%d][1,%d]\n",
	  ix,iy,iz,xn,yn,tn);
  mylog(mylogf,ggstr);
  outfile = paramfile_make_fname(mppl,"write_trace_filter",prefix,".trf.pl");
  if (outfile != (char *)NULL){
    int wx,wy,wz,tdn;
    float *td;
    wx = paramfile_get_nth_int_param_or_exit(mppl,"write_trace_filter",1);
    wy = paramfile_get_nth_int_param_or_exit(mppl,"write_trace_filter",2);
    wz = paramfile_get_nth_int_param_or_exit(mppl,"write_trace_filter",3);
    if (wx == -1){
      td = get_1d_from_3d_farray(dog,1,xn,wy,1,wz,1);
      tdn = xn;
    }else if (wy == -1){
      td = get_1d_from_3d_farray(dog,wx,1,1,yn,wz,1);
      tdn = yn;
    }else if (wz == -1){
      td = get_1d_from_3d_farray(dog,wx,1,wy,1,1,tn);
      tdn = tn;
    }else{
      td = NULL;
      tdn = 0;
      mylog_exit(mylogf,"MODEL_DOG_01_PREP  Must set one index to -1\n");
    }
    write_farray_plot(outfile,td,tdn);
    myfree(td);
    myfree(outfile);
  }
  surr_sig = paramfile_get_float_param_default(mppl,"surr_sd",0.0);
  if (surr_sig > 0.0){
    surr_tau = paramfile_get_float_param_or_exit(mppl,"surr_tau");
    mod_dog_surr = gauss_space_exp_time_tensor(mylogf,xn,yn,tn,sscale,tscale,
					       surr_sig,surr_tau);
    mod_dog_surr_sig = surr_sig;
    mod_dog_surr_tau = surr_tau;
    mylog(mylogf,"    Computing FFT of Surround filter.\n");
    mod_dog_surr_s = matrix(1,xn,1,2*yn);
    contort_real_3d_farray(mod_dog_surr,1,1,1,xn,yn,tn);
    rlft3(mod_dog_surr,mod_dog_surr_s,xn,yn,tn,1);  
  }
  outfile = paramfile_make_fname(mppl,"write_3d_filter",prefix,".dog.3d");
  if (outfile != (char *)NULL){
    write_3d_data_part(outfile,dog,1,xn,1,yn,1,tn,4,2,1);
    myfree(outfile);
  }
  mod_dog_squash_flag = paramfile_get_int_param_default(mppl,"squash_flag",0);
  if (mod_dog_squash_flag)
    mod_dog_squash_table_prep(m);
  myfree(prefix);
}
void mod_dog_filt_struct_free_contents(f)
     struct mod_dog_filt *f;
{
  int xn,yn,tn;
  if (f->name != NULL)  myfree(f->name);
  xn = mod_dog_xn;
  yn = mod_dog_yn;
  tn = mod_dog_tn;
  if (f->f   != NULL)  free_f3tensor(f->f,1,xn,1,yn,1,tn);
  if (f->s   != NULL)  free_matrix(f->s,1,xn,1,2*yn);
  if (f->r   != NULL)  free_f3tensor(f->r,1,xn,1,yn,1,tn);
  if (f->rr  != NULL)  free_f3tensor(f->rr,1,xn,1,yn,1,tn);
  if (f->rg  != NULL)  free_3d_farray(f->rg,xn,yn,tn);
  if (f->rrg != NULL)  free_3d_farray(f->rrg,xn,yn,tn);
}
struct mod_dog_filt *mod_dog_get_filter_struct(po,pflag)
     struct onode *po;        
     int pflag;               
{
  int xn,yn,tn,ix,iy,iz;
  float sscale,tscale;
  char *dumpfile;
  struct mod_dog_filt *f;
  f = (struct mod_dog_filt *)myalloc(sizeof(struct mod_dog_filt));
  f->name     = onode_getpar_chr_exit(po,"name");  
  f->sig1     = onode_getpar_flt_exit(po,"sig1");
  f->sig2     = onode_getpar_flt_exit(po,"sig2");
  f->amp1     = onode_getpar_flt_exit(po,"amp1");
  f->amp2     = onode_getpar_flt_exit(po,"amp2");
  f->cs_x     = onode_getpar_flt_dflt(po,"cs_x",0.0);
  f->cs_y     = onode_getpar_flt_dflt(po,"cs_y",0.0);
  if (onode_test_ostr(po,"cs_delay") == 1)
    f->cs_delay = onode_getpar_flt_exit(po,"cs_delay");
  else{
    f->cs_delay = onode_getpar_flt_exit(po,"cs_delay_s");
    f->cs_delay /= mod_dog_tscale;
  }
  f->tsig     = onode_getpar_flt_exit(po,"tsig");
  f->tcent    = onode_getpar_flt_dflt(po,"tcent",0.0);
  f->tdelay   = onode_getpar_flt_dflt(po,"tdelay",0.0);
  f->power    = onode_getpar_flt_dflt(po,"power",1.0);
  f->tau_ad   = onode_getpar_flt_dflt(po,"tau_ad",0.0);
  if (myid == -1){
    dumpfile = onode_getpar_chr_dflt(po,"write_dog_temporal",NULL);
  if (dumpfile != NULL){
    if (strcmp(dumpfile,"NULL")==0) 
      dumpfile = NULL;
  }
  }else
    dumpfile = NULL;
  xn = mod_dog_xn;
  yn = mod_dog_yn;
  tn = mod_dog_tn;
  sscale = mod_dog_sscale;
  tscale = mod_dog_tscale;
  f->f = dog_space_time_tensor(xn,yn,tn,sscale,tscale,f->sig1,f->sig2,
			       f->amp1,f->amp2,f->cs_x,f->cs_y,f->cs_delay,
			       f->tsig,f->tcent,
			       dumpfile,NULL,po,pflag,0,0,0.0,f->tau_ad);
  kernel_util_norm_xyt_filter(f->f,1,xn,1,yn,1,tn,&(f->psum),&(f->nsum));
  f->fsum = sum_3d_farray(f->f,1,xn,1,yn,1,tn);
  sprintf(ggstr,"    sum over 3d filter = %f",f->fsum);
  mylog(mylogf,ggstr);
  get_max_coord_3d_farray(f->f,1,xn,1,yn,1,tn,&ix,&iy,&iz);
  sprintf(ggstr,"    Max at (%d,%d,%d) in [1,%d][1,%d][1,%d]\n",
	  ix,iy,iz,xn,yn,tn);
  mylog(mylogf,ggstr);
  f->s  = NULL;
  f->r  = NULL;
  f->rr = NULL;
  f->rg  = NULL; 
  f->rrg = NULL;
  f->oldstimno = -1;
  if (dumpfile != NULL)
    myfree(dumpfile);
  return f;
}
void model_dog_01_prep_o(m,r)
     struct model_struct *m;    
     struct response_struct *r; 
{
  int k;
  int xn,yn,tn,ix,iy,iz,pflag;
  float tscale,sscale,surr_sig,surr_tau;
  char *ltype;
  struct onode *o,*t,*nlo;
  if (mylogf == NULL)  
    myid = mylog_set_id_log(m->process_id,&mylogf);  
  mylog(mylogf,"  MODEL_DOG_01_PREP_O\n");
  mod_dog_fsn = 0;  
  pflag = 0;
  if (myid == -1)
    pflag = 1;
  mod_dog_sscale = sscale = onode_getpar_flt_exit(m->o,"sscale");
  mod_dog_tscale = tscale = onode_getpar_flt_exit(m->o,"tscale");
  mod_dog_xn     = xn     = onode_getpar_int_exit(m->o,"xn");
  mod_dog_yn     = yn     = onode_getpar_int_exit(m->o,"yn");
  mod_dog_tn     = tn     = onode_getpar_int_exit(m->o,"tn");
  o = onode_get_node_type_item_val(m->o,"pop","name","lgn");
  if (o == NULL){
    o = onode_get_node_type_item_val(m->o,"pop","name","lgn_on");
    if (o == NULL){
      mylog(mylogf,"  *** Found neither pop 'lgn' nor 'lgn_on'\n");
      return;  
    }else{
      exit_error("MODEL_DOG_01_PREP_O","Old lgn pop format, please update");
    }
  }
  mod_dog_fsn = 0;
  t = onode_get_next_type(m->o->o,"pop");
  while(t != NULL){
    ltype = onode_getpar_chr_ptr(t,"type");
    if (ltype != NULL){
      if (pop_util_check_lclass(ltype,"lgn0"))
	mod_dog_fsn += 1;
    }
    t = onode_get_next_type(t->next,"pop");
  }
  mod_dog_fs = (struct mod_dog_filt **)myalloc(mod_dog_fsn*
					       sizeof(struct mod_dog_filt *));
  sprintf(ggstr,"    %d LGN populations\n",mod_dog_fsn);
  mylog(mylogf,ggstr);
  if (mod_dog_fsn == 0)
    mylog_exit("MODEL_DOG_01_PREP_O  No LGN populations found.\n");
  k = 0;
  t = onode_get_next_type(m->o->o,"pop");
  while(t != NULL){
    ltype = onode_getpar_chr_ptr(t,"type");
    if (ltype != NULL){
      if (pop_util_check_lclass(ltype,"lgn0")){
	mod_dog_fs[k] = mod_dog_get_filter_struct(t,pflag);
	k += 1;
      }
    }
    t = onode_get_next_type(t->next,"pop");
  }
  mod_dog_nl_flag = onode_getpar_int_dflt(o,"nl_flag",0);
  mod_dog_sq_flag = onode_getpar_int_dflt(o,"sq_flag",0);
  mod_dog_cg_flag = onode_getpar_int_dflt(o,"contrast_gain",0);
  surr_sig = onode_getpar_flt_dflt(o,"surr_sd",0.0);
  if (surr_sig > 0.0){
    surr_tau = onode_getpar_flt_exit(o,"surr_tau");
    mod_dog_surr = gauss_space_exp_time_tensor(mylogf,xn,yn,tn,sscale,tscale,
					       surr_sig,surr_tau);
    mod_dog_surr_sig = surr_sig;
    mod_dog_surr_tau = surr_tau;
    mylog(mylogf,"    Computing FFT of Surround filter.\n");
    mod_dog_surr_s = matrix(1,xn,1,2*yn);
    contort_real_3d_farray(mod_dog_surr,1,1,1,xn,yn,tn);
    rlft3(mod_dog_surr,mod_dog_surr_s,xn,yn,tn,1);  
  }
  mod_dog_squash_flag = onode_getpar_int_dflt(o,"squash_flag",0);
  if (mod_dog_squash_flag){
    mod_dog_squash_table_prep(m);
  }
  nlo = onode_child_get_unique(o,"nonlin");
  if (nlo != NULL){
    mod_dog_logistic_name  = onode_getpar_chr_exit(nlo,"name");
    mod_dog_logistic_slope = onode_getpar_flt_exit(nlo,"slope");
    mod_dog_logistic_asymp = onode_getpar_flt_exit(nlo,"asymptote");
    mylog(mylogf,"  LOGISTIC\n");
    sprintf(ggstr,"    name  %s\n",mod_dog_logistic_name);
    mylog(mylogf,ggstr);
    sprintf(ggstr,"    slope  %f\n",mod_dog_logistic_slope);
    mylog(mylogf,ggstr);
    sprintf(ggstr,"    asymp  %f\n",mod_dog_logistic_asymp);
    mylog(mylogf,ggstr);
  }else{
    mod_dog_logistic_name = NULL;
    mod_dog_logistic_slope = -1.0;
    mod_dog_logistic_asymp = 0.0;
  }
}
void mod_dog_cg01_response()
{
  int i,j,k;
  int xn,yn,tn;
  int nm;
  float ***r,*t,*lowpass,*mask,*v,mean,var,*lowpv,*rnorm,*rnew;
  mylog(mylogf,"******* MOD_DOG_CG01_RESPONSE\n");
  mylog(mylogf,"******* WYETH - this attempt should be done w/ quad filts?\n");
  xn = mod_dog_xn;
  yn = mod_dog_yn;
  tn = mod_dog_tn;
  r = mod_dog_fs[0]->r;
  mean = mod_dog_fs[0]->fsum/2.0;
  mylog_exit(mylogf,"WYETH ---- THIS psum was calc'd *BEFORE* norm in past\n");
  var  = mod_dog_fs[0]->psum - mean;
  nm = tn/2;
  if (nm > 256)
    nm = 256;
  mask = alpha_farray(0.0,0.05,1.0,nm);
  norm_area_farray(mask,nm,1.0);
  if (myid == -1)
    append_farray_plot("zz.cg.pl","mask",mask,nm,1);
  i = j = xn/2;
  t = r[i][j];
  rnorm = (float *)myalloc(tn*sizeof(float));
  v = (float *)myalloc(tn*sizeof(float));
  for(k=0;k<tn;k++){
    rnorm[k] = (t[k+1] - mean);
    if (rnorm[k] < 0.0)                 
      v[k] = -rnorm[k]/var;
    else
      v[k] = rnorm[k]/var;
  }
  lowpass = convolve_with_mask_causal(rnorm,tn,mask,nm);
  lowpv = convolve_with_mask_causal(v,tn,mask,nm);
  rnew = (float *)myalloc(tn*sizeof(float));
  for(k=0;k<tn;k++){
    rnew[k] = rnorm[k] - 2.0*lowpv[k] * lowpass[k];
    rnew[k] += mean; 
  }
  if (myid == -1){
    append_farray_plot("zz.cg.pl","v",v,tn,1);
    append_farray_plot("zz.cg.pl","lowpv",lowpv,tn,1);
    append_farray_plot("zz.cg.pl","rnew",rnew,tn,1);
    append_farray_plot("zz.cg.pl","lowpass",lowpass,tn,1);
    append_farray_plot("zz.cg.pl","rnorm",rnorm,tn,1);
    append_farray_plot("zz.cg.pl","r",t+1,tn,1);
  }
  exit(0);
}
void mod_dog_surr_response(eye_flag)
     int eye_flag;  
{
  int i,j,k;
  int xn,yn,tn,xi,yi;
  float ***r,***rsq,*t,***surr,***f,**f_s,mu,y;
  mylog(mylogf,"  MOD_DOG_SURR_RESPONSE\n");
  xn = mod_dog_xn;
  yn = mod_dog_yn;
  tn = mod_dog_tn;
  if (eye_flag == 0)
    r = mod_dog_fs[0]->r;
  else
    r = mod_dog_fs[0]->rr;
  f = mod_dog_surr;
  f_s = mod_dog_surr_s;
  xi = 1 + xn/2;
  yi = 1 + yn/2;
  rsq = f3tensor(1,xn,1,yn,1,tn);
  for(i=1;i<=xn;i++){
    for(j=1;j<=yn;j++){
      t = r[i][j];
      mu = t[1];  
      for(k=1;k<=tn;k++){
	y = t[k] - mu;      
	rsq[i][j][k] = y*y; 
	mu = 0.98*mu + 0.02*t[k];  
      }
    }
  }
  compute_response_single_from_ft(rsq,f,f_s,xn,yn,tn,0,&surr);
  if (eye_flag == 0){
    if (mod_dog_sup != NULL)
      free_f3tensor(mod_dog_sup,1,xn,1,yn,1,tn);
    mod_dog_sup = surr;
  }else{
    if (mod_dog_sup_r != NULL)
      free_f3tensor(mod_dog_sup_r,1,xn,1,yn,1,tn);
    mod_dog_sup_r = surr;
  }
  free_f3tensor(rsq,1,xn,1,yn,1,tn);
}
float ***mod_dog_gain_ptr(name,eyeflag)
     char *name;
     int eyeflag; 
{
  struct mod_dog_filt *mdfs;
  if (name != NULL)
    mdfs = mod_dog_fs_by_name(name);
  else{
    mylog(mylogf,"MOD_DOG_GAIN_PTR  'name' is NULL, using _fs[0].\n");
    mdfs = mod_dog_fs[0];
  }
  if (eyeflag == 0)
    return mdfs->rg;
  else
    return mdfs->rrg;
}
float ***mod_dog_gain(data,xn,yn,tn,mdfs)
     float ***data;
     int xn,yn,tn;
     struct mod_dog_filt *mdfs;
{
  int i,j,k;
  float ***g,*t,*tg,fhalf,fmax;
  fhalf = 0.5 * mdfs->fsum;  
  fmax = (mdfs->psum - mdfs->nsum)/2.0;  
  g = get_3d_farray(xn,yn,tn);
  for(i=0;i<xn;i++){
    for(j=0;j<yn;j++){
      t = &(data[i+1][j+1][1]);
      tg = g[i][j];
      for(k=0;k<tn;k++){
	tg[k] = (t[k]-fhalf)/fmax;
	if (tg[k] < 0.0)
	  tg[k] *= -1.0;
      }
    }
  }
  return g;
}
float ***mod_dog_get_response(s,k,name,gain_flag)
     struct stim_struct *s;   
     int k;                   
     char *name;              
     int gain_flag;           
{
  int xn,yn,tn,pflag;
  float ***dog,**dog_s,***rdog;
  struct mod_dog_filt *mdfs;
  pflag = 0;  
  if (myid == -1)
    pflag = 1;
  if (name != NULL)
    mdfs = mod_dog_fs_by_name(name);
  else{
    mylog(mylogf,"MOD_DOG_GET_RESPONSE  'name' is NULL, using _fs[0].\n");
    mdfs = mod_dog_fs[0];
  }
  xn = mod_dog_xn;
  yn = mod_dog_yn;
  tn = mod_dog_tn;
  dog   = mdfs->f;
  dog_s = mdfs->s;
  if (mdfs->oldstimno != s->stimno){ 
    if (mdfs->r != NULL)
      free_f3tensor(mdfs->r,1,xn,1,yn,1,tn);
    if (mod_dog_stim_si != s->stimno){
      if (mod_dog_stim_s != NULL)
	free_matrix(mod_dog_stim_s,1,xn,1,2*yn);
      mod_dog_stim_s = compute_response_single_from_ft_ret_s(s->d,dog,dog_s,xn,
							     yn,tn,pflag,
							     &rdog);
      if (gain_flag == 1){
	if (mdfs->rg != NULL){
	  free_3d_farray(mdfs->rg,xn,yn,tn);
	}
	mdfs->rg = mod_dog_gain(rdog,xn,yn,tn,mdfs); 
      }
      mod_dog_stim_si = s->stimno;
    }else{
      compute_response_single_from_ft_2(s->d,mod_dog_stim_s,dog,dog_s,
					xn,yn,tn,pflag,&rdog);
    }
    mdfs->r = rdog;
    if (mod_dog_surr_sig > 0.0){
      mod_dog_surr_response(0); 
    }
    if (mod_dog_cg_flag){ 
      mod_dog_cg01_response();
    }
    mdfs->oldstimno = s->stimno;
  }else{
    rdog = mdfs->r; 
  }
  return rdog;
}
float ***mod_dog_get_response_binoc(s,k,name,rright)
     struct stim_struct *s;   
     int k;                   
     char *name;              
     float ****rright;
{
  int xn,yn,tn,pflag;
  float ***dog,**dog_s,***rdog,***rdogr;
  struct mod_dog_filt *mdfs;
  pflag = 0;  
  if (myid == -1)
    pflag = 1;
  if (name != NULL)
    mdfs = mod_dog_fs_by_name(name);
  else{
    mylog(mylogf,"MOD_DOG_GET_RESPONSE  'name' is NULL, using _fs[0].\n");
    mdfs = mod_dog_fs[0];
  }
  xn = mod_dog_xn;
  yn = mod_dog_yn;
  tn = mod_dog_tn;
  dog   = mdfs->f;
  dog_s = mdfs->s;
  if (mdfs->oldstimno != s->stimno){ 
    if (mdfs->r != NULL){
      free_f3tensor(mdfs->r,1,xn,1,yn,1,tn);
      free_f3tensor(mdfs->rr,1,xn,1,yn,1,tn);
    }
    compute_response_single_from_ft(s->d,dog,dog_s,xn,yn,tn,pflag,&rdog);
    mdfs->r = rdog;
    compute_response_single_from_ft(s->d_r,dog,dog_s,xn,yn,tn,pflag,&rdogr);
    mdfs->rr = rdogr;
    if (mod_dog_surr_sig > 0.0){
      mod_dog_surr_response(0); 
      mod_dog_surr_response(1);
    }
    if (mod_dog_cg_flag)
      exit_error("MOD_DOG_GET_RESPONSE_BINOC","CG not imp'd for binoc");
    mdfs->oldstimno = s->stimno;
  }else{
    rdog  = mdfs->r;  
    rdogr = mdfs->rr;
  }
  *rright = rdogr;
  return rdog;
}
void model_dog_01_prep_ft()
{
  int i;
  int xn,yn,tn;
  struct mod_dog_filt *mdfs;
  mylog(mylogf,"  MODEL_DOG_01_PREP_FT\n");
  xn = mod_dog_xn;
  yn = mod_dog_yn;
  tn = mod_dog_tn;
  for(i=0;i<mod_dog_fsn;i++){
    mdfs = mod_dog_fs[i];
    sprintf(ggstr,"    Computing FFT of DOG filter, %s\n",mdfs->name);
    mylog(mylogf,ggstr);
    mdfs->s = matrix(1,xn,1,2*yn);
    contort_real_3d_farray(mdfs->f,1,1,1,xn,yn,tn);
    rlft3(mdfs->f,mdfs->s,xn,yn,tn,1);  
  }
}
void model_dog_01_done(m)
     struct model_struct *m; 
{
  int i;
  int xn,yn,tn;
  struct mod_dog_filt *mdfs;
  mylog(mylogf,"  MODEL_DOG_01_DONE\n");
  xn = mod_dog_xn;
  yn = mod_dog_yn;
  tn = mod_dog_tn;
  if (1){ 
    for(i=0;i<mod_dog_fsn;i++){
      mod_dog_filt_struct_free_contents(mod_dog_fs[i]);
    }
    myfree(mod_dog_fs);
    mod_dog_fs = NULL;
    mod_dog_fsn = 0;
  }else{ 
    for(i=0;i<mod_dog_fsn;i++){
      mdfs = mod_dog_fs[i];
      free_f3tensor(mdfs->f,1,xn,1,yn,1,tn);
      free_matrix(mdfs->s,1,xn,1,2*yn);
      mdfs->f = (float ***)NULL;
      mdfs->s = (float **)NULL;
      if (mdfs->r != NULL)
	free_f3tensor(mdfs->r,1,xn,1,yn,1,tn);
      mdfs->r = (float ***)NULL;
    }
  }
  if (mod_dog_surr != NULL){
    free_f3tensor(mod_dog_surr,1,xn,1,yn,1,tn);
    free_matrix(mod_dog_surr_s,1,xn,1,2*yn);
  }
  if (mod_dog_sup != NULL){
    free_f3tensor(mod_dog_sup,1,xn,1,yn,1,tn);
  }
  mod_dog_surr = (float ***)NULL;
  mod_dog_surr_s = (float **)NULL;
}
void mod_dog_nl_logistic(t,n,fsum)
     float *t;
     int n;
     float fsum;  
{
  int i;
  float a,a2,m,fs2;
  a = mod_dog_logistic_asymp;
  a2 = 2.0 * a;
  fs2 = fsum / 2.0;
  m = 2.0 * mod_dog_logistic_slope / mod_dog_logistic_asymp;
  for(i=0;i<n;i++){
    t[i] = a2/(1.0 + exp(-m * (t[i]-fs2))) - a;
    t[i] += fs2;
  }
}
float *mod_dog_get_resp_trace(r,xi,yi,tn,onflag,layname)
     float ***r;     
     int xi,yi;      
     int tn;         
     int onflag;     
     char *layname;  
{
  int i;
  float *t,*tp,c;
  struct mod_dog_filt *mdfs;
  if (layname != NULL)
    mdfs = mod_dog_fs_by_name(layname);
  else{
    mylog(mylogf,"MOD_DOG_GET_RESP_TRACE  'name' is NULL, using _fs[0].\n");
    mdfs = mod_dog_fs[0];
  }
  tp = r[xi+1][yi+1];  
  t = (float *)myalloc(tn*sizeof(float));
  if (onflag == 1){ 
    for(i=0;i<tn;i++)
      t[i] = tp[i+1];
  }else{
    for(i=0;i<tn;i++) 
      t[i] = mdfs->fsum - tp[i+1];
  }
  if (mod_dog_logistic_name != NULL){
    mod_dog_nl_logistic(t,tn,mdfs->fsum);
  }
  if (mdfs->power != 1.0){
    printf(" POWER HERE %f\n",mdfs->power);
    pow_farray(t,tn,mdfs->power);
  }
  if (mod_dog_surr_sig > 0.0){
    tp = mod_dog_sup[xi+1][yi+1];
    for(i=0;i<tn;i++){
      c =  8.0 * tp[i+1]; 
      if (c < 0.0)
	c = 0.0;
      t[i] *= 1.0 - c;
    }
  }
  return t;
}
void model_dog_01_get_response(m,s,r,k)
     struct model_struct *m;        
     struct stim_struct *s;         
     struct response_struct *r;     
     int k;                         
{
  int i,xi,yi,ri;
  int xn,yn,tn,ns,vflag,*seedlist,rn; 
  float ***rdog,*t,*fs,sampling;
  char *outfile,tname[LONG_SLEN],*prefix,tstr[LONG_SLEN],*sgen,*lname;
  char *dump_name;
  static int oldstimno = -1;
  float *gti,*gtx,samp,*rv;
  struct onode *po,*spko;
  prefix = paramfile_get_char_param_default(r->ppl,"outfile","zz");
  if (s->nvar > 0)
    strcpy(tname,s->name);  
  else
    strcpy(tname,"raw");    
  xn = mod_dog_xn;  
  yn = mod_dog_yn;
  tn = mod_dog_tn;
  if (m->ppl == NULL){
    lname = strdup("lgn"); 
    po = onode_get_node_type_item_val(m->o,"pop","name",lname);
    spko = onode_child_get_unique(po,"spike_gen");
    if (onode_test_int(po,"ifc_dump_center",1)){
      dump_name = tname;
    }else
      dump_name = NULL;
  }else{
    lname = NULL;
    po = NULL;
    spko = NULL;
  }
  rdog = mod_dog_get_response(s,k,lname,0);
  if (m->ppl == NULL)
    outfile = onode_make_fname(po,"write_3d_response",prefix,".resp.3d");
  else
    outfile = paramfile_make_fname(m->ppl,"write_3d_response",prefix,
				   ".resp.3d");
  if (outfile != (char *)NULL){
    write_3d_data_part(outfile,rdog,1,xn,1,yn,1,tn,4,2,1);
    myfree(outfile);
  }
  samp = 1.0/mod_dog_tscale;
  seedlist = get_seeds(m->mseed[r->tsi],1000000,r->n);
  for(ri=0;ri<r->n;ri++){
    if ((strcmp(r->plname[ri],"lgn")==0)||(strcmp(r->plname[ri],"lgn_on")==0)){
      xi = r->xi[ri];
      yi = r->yi[ri];
      if ((xi < 0) || (xi >= xn) || (yi < 0) || (yi >= yn))
	mylog_exit(mylogf,
		   "MODEL_DOG_01_GET_RESPONSE  response x,y out of bounds\n");
      else{
	sprintf(ggstr,"   LGN %d %d\n",xi,yi);
	mylog(mylogf,ggstr);
      }
      t = mod_dog_get_resp_trace(rdog,xi,yi,tn,1,lname);  
      if (mod_dog_nl_flag == 1){
	if (mod_dog_nl_filt == NULL) 
	  mod_dog_nonlin1_prep(m); 
	mod_dog_nonlin1(m,t,tn); 
      }
      if (mod_dog_squash_flag)
	mod_dog_squash(t,tn);
      else if (mod_dog_sq_flag)  
	mod_dog_squash_blake(m,t,tn,1);
      if (myid == -1){
	if (m->ppl == NULL)
	  outfile = onode_make_fname(po,"write_raw",prefix,".raw.pl");
	else
	  outfile = paramfile_make_fname(m->ppl,"write_raw",prefix,".raw.pl");
	if (outfile != (char *)NULL){
	  sprintf(tstr,"%s_%d_%d",tname,xi,yi);
	  append_farray_plot(outfile,tstr,t,tn,1);
	  myfree(outfile);
	}
      }
      vflag = 0;
      if (r->nf > 0)
	vflag = 1;
      if (m->ppl == NULL)
	sgen = onode_getpar_chr_exit(spko,"type");
      else
	sgen = paramfile_get_char_param_default(m->ppl,"spike_gen","poisson");
      if (strcmp(sgen,"ifc")==0){
	gtx = t;
	gti = get_zero_farray(tn);
	if (m->ppl != NULL){
	  ifc_test(myid,m,"ifc",gtx,gti,samp,tn,seedlist[ri],1,tname,
		   &fs,&ns,vflag,&rv,&rn);
	}else{
	  ifc_test(myid,m,"lgn",gtx,gti,samp,tn,seedlist[ri],1,dump_name,
		   &fs,&ns,vflag,&rv,&rn);
	}
	myfree(gti);
      }else if (strcmp(sgen,"poisson")==0){
	ifc_util_poisson(mylogf,m,spko,t,tn,samp,seedlist[ri],0,vflag,&fs,&ns,
			 &rv,&rn);
      }else
	mylog_exit(mylogf,"MODEL_DOG_01_GET_RESPONSE  Unknown spike gen\n");
      myfree(sgen);
      sampling = r->samp[ri]; 
      multiply_farray(fs,ns,sampling/1000.0);
      add_const_farray(fs,ns,mod_dog_fs[0]->tdelay*sampling);
      r->s[ri][r->gtsi] = fs;
      r->cnt[ri][r->gtsi] = ns;
      if (r->nf > 0)
	mylog_exit(mylogf,"MODEL_DOG_01_GET_RESPONSE  - not re-impl'd yet\n");
      if (r->nf > 0){
	mylog(mylogf,"    Storing Vm\n");
	r->f[0][r->gtsi] = rv;
	r->fcnt[0][r->gtsi] = rn;
      }
      myfree(t);
    }else if (strcmp(r->plname[ri],"lgn_off")==0){
      mylog_exit(mylogf,
		 "MODEL_DOG_01_GET_RESPONSE  'lgn_off' not yet implemented\n");
    }
  }
  myfree(seedlist);
  myfree(prefix); 
}
void mod_dog_get_spikes_2d_flag(m,r,rdog,lname,onflag,xn,yn,tn,flag,samp,seed,
				rs2,rcnt2,sav_flag,rx,ri,rv,rvn)
     struct model_struct *m;        
     struct response_struct *r;     
     float ***rdog;                 
     char *lname;                   
     int onflag;                    
     int xn,yn,tn,**flag;           
     float samp;                    
     int seed;                      
     float ****rs2;                 
     int ***rcnt2;                  
     int sav_flag;                  
     float ****rx,****ri,****rv;    
     int *rvn;                      
{
  int i,j,k;
  int **cnt2,flag_count,spike_count,vn,savit,sav_vn;
  int tseed,dumpx,dumpy,*seedlist;
  float ***s2,*t,*gti,*gtx,*v;
  float ***sav_vm,***sav_gx,***sav_gi;
  char pname[SLEN],prefix[SLEN],tstr[SLEN],*dumpname,*dumpname_ptr,*sgen;
  struct onode *po,*spko;
  struct mod_dog_filt *mdfs;
  if (lname != NULL)
    mdfs = mod_dog_fs_by_name(lname);
  else{
    mdfs = mod_dog_fs[0];
  }
  if (m->ppl != NULL){
    strcpy(prefix,"ifc");
    sgen = paramfile_get_char_param_default(m->ppl,"spike_gen","poisson");
  }else{
    strcpy(prefix,lname);
    po = onode_get_node_type_item_val(m->o,"pop","name",lname);
    if (po == NULL)
      mylog_exit(mylogf,"MOD_DOG_GET_SPIKES_2D_FLAG  No lgn layer.");
    spko = onode_child_get_unique(po,"spike_gen");
    sgen = onode_getpar_chr_exit(spko,"type");
  }
  dumpname = (char *)NULL;
  dumpx = dumpy = -1;
  if (myid == -1){
    strcpy(pname,"dog_ifc_dump_coord");
    if (m->ppl != NULL){
      if (paramfile_test_param(m->ppl,pname)){
	dumpx = paramfile_get_nth_int_param_or_exit(m->ppl,pname,0);
	dumpy = paramfile_get_nth_int_param_or_exit(m->ppl,pname,1);
	sprintf(ggstr,"    Dumping LGN results for unit %d %d\n",dumpx,dumpy);
	mylog(mylogf,ggstr);
	if (onflag)
	  sprintf(tstr,"%s_ON_%d_%d",lname,dumpx,dumpy);
	else
	  sprintf(tstr,"%s_OFF_%d_%d",lname,dumpx,dumpy);
	dumpname = strdup(tstr);
      }
    }else{
      if ((onode_test_ostr(po,pname))||
	  onode_test_int(po,"ifc_dump_center",1)){
	dumpx = xn/2;
	dumpy = yn/2;
	sprintf(ggstr,"    Dumping LGN results for unit %d %d\n",dumpx,dumpy);
	mylog(mylogf,ggstr);
	if (onflag)
	  sprintf(tstr,"%s_ON_%d_%d",lname,dumpx,dumpy);
	else
	  sprintf(tstr,"%s_OFF_%d_%d",lname,dumpx,dumpy);
	dumpname = strdup(tstr);
      }
    }
  }
  cnt2 = get_zero_2d_iarray(xn,yn);
  s2 = (float ***)myalloc(xn*sizeof(float **));
  for(i=0;i<xn;i++){
    s2[i] = (float **)myalloc(yn*sizeof(float *));
  }
  gti = get_zero_farray(tn);  
  if (onflag)
    seedlist = get_seeds(seed,1000000,xn*yn);
  else
    seedlist = get_seeds(17*seed,1000000,xn*yn);
  if (mod_dog_nl_flag == 1){
    if (mod_dog_nl_filt == NULL) 
      mod_dog_nonlin1_prep(m);   
  }
  if (sav_flag > 0){
    sav_vm = get_3d_farray(xn,yn,0);
    sav_gx = get_3d_farray(xn,yn,0);
    sav_gi = get_3d_farray(xn,yn,0);
  }else{
    sav_vm = sav_gx = sav_gi = NULL;
    savit = 0;
  }
  flag_count = 0;
  spike_count = 0;
  t = (float *)myalloc(tn*sizeof(float));
  for(i=0;i<xn;i++){
    for(j=0;j<yn;j++){
      if (flag[i][j]){
	if (sav_flag > 0){
	  if (flag[i][j] > 1)
	    savit = 1;
	  else
	    savit = 0;
	}
	t = mod_dog_get_resp_trace(rdog,i,j,tn,onflag,lname);
	if (mod_dog_nl_flag == 1){
	  mod_dog_nonlin1(m,t,tn); 
	}
	if (mod_dog_squash_flag){
	  if ((i==dumpx) && (j==dumpy))   
	    append_farray_plot("zz.squash.dump.pl","before",t,tn,1);
	  mod_dog_squash(t,tn);
	  if ((i==dumpx) && (j==dumpy))   
	    append_farray_plot("zz.squash.dump.pl","after",t,tn,1);
	}else if (mod_dog_sq_flag){
	  mod_dog_squash_blake(m,t,tn,0);
	}
	gtx = t;
	tseed = seedlist[i*yn + j];
	if (strcmp(sgen,"poisson")==0){
	  ifc_util_poisson(mylogf,m,spko,gtx,tn,samp,tseed,0,0,&(s2[i][j]),
			   &(cnt2[i][j]),&v,&vn);
	}else if (strcmp(sgen,"ifc")==0){
	  if ((i==dumpx) && (j==dumpy)){   
	    dumpname_ptr = dumpname;
	  }else
	    dumpname_ptr = (char *)NULL;
	  ifc_test(myid,m,prefix,gtx,gti,samp,tn,tseed,0,dumpname_ptr,
		   &(s2[i][j]),&(cnt2[i][j]),savit,&v,&vn);
	  if (savit){
	    sav_gx[i][j] = copy_farray(gtx,tn);
	    sav_gi[i][j] = copy_farray(gti,tn);
	    sav_vm[i][j] = v;
	    sav_vn = vn;
	  }
	}else{
	  mylog_exit(mylogf,"MOD_DOG_GET_SPIKES_2D_FLAG  Unknown spike gen\n");
	}
	add_const_farray(s2[i][j],cnt2[i][j],mdfs->tdelay*1000.0);
	flag_count += 1;
	spike_count += cnt2[i][j];
	myfree(t);  
      }else{ 
	s2[i][j] = NULL;  
	cnt2[i][j] = -1;  
	if ((i==dumpx) && (j==dumpy)){
	  mylog(mylogf,"  *** LGN cell to be dumped not selected as input\n");
	}
      }
    }
  }
  sprintf(ggstr,
	  "      Spikes generated for %d of %d positions, %.4f spk/cell\n",
	  flag_count,xn*yn,(float)spike_count/(float)flag_count);
  mylog(mylogf,ggstr);
  if (dumpname != (char *)NULL)
    myfree(dumpname);
  myfree(seedlist);
  myfree(gti);
  myfree(sgen);
  if (sav_flag > 0){
    *rx = sav_gx;
    *ri = sav_gi;
    *rv = sav_vm;
    *rvn = sav_vn;
  }else{
    *rv = *rx = *ri = NULL;
    *rvn = -1;
  }
  *rs2 = s2; *rcnt2 = cnt2;
}
void mod_dog_get_gt_prod_pair_sarray(s1,n1,s2,n2,dur,delay,mask,nm,rgt)
     float *s1;
     int n1;
     float *s2;
     int n2;
     int dur,delay;
     float *mask;
     int nm;
     float **rgt;               
{
  int i;
  int t;
  float *sx1,*sx2,*g1,*g2,*gt;
  sx1 = get_zero_farray(dur);
  sx2 = get_zero_farray(dur);
  for(i=0;i<n1;i++){
    t = my_rint(s1[i]);
    if ((t >= 0)&&(t < dur)){
      sx1[t] += 1.0;
    }
  }
  for(i=0;i<n2;i++){
    t = delay + my_rint(s2[i]);  
    if ((t >= 0)&&(t < dur)){
      sx2[t] += 1.0;
    }
  }
  g1 = convolve_with_mask_causal(sx1,dur,mask,nm);
  g2 = convolve_with_mask_causal(sx2,dur,mask,nm);
  gt = get_farray(dur);
  for(i=0;i<dur;i++)
    gt[i] = g1[i] * g2[i];
  myfree(g1); myfree(g2);
  myfree(sx1); myfree(sx2);
  *rgt = gt;
}
void mod_dog_get_gt_sarray(m,s,cnt,n,t1,t2,amp,rgt,rnms)
     struct model_struct *m;    
     float **s;
     int *cnt,n;
     float t1,t2,amp;
     float **rgt;               
     int *rnms;
{
  int i,j;
  int nx,t,tn,nms;
  float *xshape,*gt,*g;
  nx = (int)(t1 * 7.0);        
  xshape = diff_exp_farray(0.0,amp,t1,t2,nx);
  tn = mod_dog_tn;
  nms = (int)((float)tn * 1000.0*mod_dog_tscale); 
  gt = get_zero_farray(nms);
  for(i=0;i<n;i++){
    for(j=0;j<cnt[i];j++){
      t = my_rint(s[i][j]);
      if ((t >= 0)&&(t < nms)){
	gt[t] += 1.0;
      }
    }
  }
  if (myid == -1)
    append_farray_plot("zz.simp_out.pl","spikes_out",gt,nms,1);
  g = convolve_with_mask_causal(gt,nms,xshape,nx);
  myfree(gt); myfree(xshape);
  gt = g;
  if (myid == -1)
    append_farray_plot("zz.simp_out.pl","gtx_to_complex",g,nms,1);
  *rgt = gt; *rnms = nms;
}
void mod_dog_get_gtx_gti(m,s0,s1,cnt0,cnt1,nin,inx,iny,inw,rgtx,rgti,rnms,cn)
     struct model_struct *m;    
     float ***s0,***s1;         
     int **cnt0,**cnt1;
     int nin,*inx,*iny;
     float *inw;
     float **rgtx,**rgti;
     int *rnms;
     int cn;
{
  int i,j;
  int nx,ni,xi,yi,t,cnt,tn,nms;
  float *xshape,*ishape,*s,t1x,t2x,ampx,t1i,t2i,ampi,*gtx,*gti,*g;
  char tstr[SLEN];
  t1x   = paramfile_get_float_param_or_exit(m->ppl,"simp_ex_tau_f");
  t2x   = paramfile_get_float_param_or_exit(m->ppl,"simp_ex_tau_r");
  ampx  = paramfile_get_float_param_or_exit(m->ppl,"simp_ex_amp");
  t1i   = paramfile_get_float_param_or_exit(m->ppl,"simp_in_tau_f");
  t2i   = paramfile_get_float_param_or_exit(m->ppl,"simp_in_tau_r");
  ampi  = paramfile_get_float_param_or_exit(m->ppl,"simp_in_amp");
  nx = (int)(t1x * 7.0);        
  xshape = diff_exp_farray(0.0,ampx,t1x,t2x,nx);
  ni = (int)(t1i * 7.0);        
  ishape = diff_exp_farray(0.0,ampi,t1i,t2i,ni);
  tn = mod_dog_tn;
  nms = (int)((float)tn * 1000.0*mod_dog_tscale); 
  gtx = get_zero_farray(nms);
  gti = get_zero_farray(nms);
  for(i=0;i<nin;i++){
    xi = inx[i];
    yi = iny[i];
    if (inw[i] < 0.0){  
      s = s0[xi][yi];
      cnt = cnt0[xi][yi];
    }else{
      s = s1[xi][yi];
      cnt = cnt1[xi][yi];
    }
    for(j=0;j<cnt;j++){
      t = my_rint(s[j]);
      if ((t >= 0)&&(t < nms)){
	gtx[t] += 1.0;
      }
    }
  }
  if (myid == -1){
    append_farray_plot("zz.simp.pl","gtx_spikes",gtx,nms,1);
    if (cn == 0)
      append_farray_plot("zz.simp.pl","xshape",xshape,nx,1);
  }
  g = convolve_with_mask_causal(gtx,nms,xshape,nx);
  myfree(gtx);
  gtx = g;
  if (myid == -1){
    sprintf(tstr,"gtx_%d",cn);
    append_farray_plot("zz.simp.pl",tstr,gtx,nms,1);
  }
  myfree(xshape); myfree(ishape);
  *rgtx = gtx; *rgti = gti; *rnms = nms;
}
void model_dog_simp_01_prep(mppl,r)
     struct param_pair_list *mppl; 
     struct response_struct *r;    
{
  int i,j,k;
  int xn,yn,nsamp,seed,xi,yi;
  int **flag;
  float **rf,sscale,theta,ssd,sf,ph,min,max,fsamp;
  mylog(mylogf,"  MODEL_DOG_SIMP_01_PREP\n");
  sscale = mod_dog_sscale;
  xn = mod_dog_xn;
  yn = mod_dog_yn;
  flag = get_zero_2d_iarray(xn,yn);   
  ssd   = paramfile_get_float_param_or_exit(mppl,"simp_ssd");
  sf    = paramfile_get_float_param_or_exit(mppl,"simp_sf");
  ph    = paramfile_get_float_param_or_exit(mppl,"simp_phase");
  theta = paramfile_get_float_param_or_exit(mppl,"simp_ori");
  fsamp = paramfile_get_float_param_or_exit(mppl,"simp_rf_fsamp");
  seed  = paramfile_get_int_param_or_exit(mppl,"simp_rf_seed");
  nsamp = paramfile_get_int_param_or_exit(mppl,"simp_rf_nsamp");
  if (seed > 0)
    seed = -seed;
  mod_dog_simp_n = 1;
  mod_dog_simp_nin = get_zero_iarray(1);
  mod_dog_simp_nin[0] = nsamp;
  mod_dog_simp_x = get_2d_iarray(1,nsamp);
  mod_dog_simp_y = get_2d_iarray(1,nsamp);
  mod_dog_simp_w = get_2d_farray(1,nsamp);
  mod_dog_simp_rf = (float ***)myalloc(sizeof(float **));
  mod_dog_simp_flag = (int ***)myalloc(sizeof(int **));
  mod_dog_simp_flag[0] = flag;
  rf = gabor_2d_space(xn,yn,sscale,ssd,sf,theta,ph); 
  mod_dog_simp_rf[0] = rf;
  max = max_of_2d_farray(rf,xn,yn);
  min = min_of_2d_farray(rf,xn,yn);
  sprintf(ggstr,"  min, max =  %f %f\n",min,max);
  mylog(mylogf,ggstr);
  k = 0;
  while(k < nsamp){
    xi = (int)((float)xn * myrand_util_ran2(&seed));
    yi = (int)((float)yn * myrand_util_ran2(&seed));
    if (flag[xi][yi] == 0){
      if ((rf[xi][yi] > fsamp*max)|| (rf[xi][yi] < fsamp*min)){
	flag[xi][yi] = 1;
	mod_dog_simp_x[0][k] = xi;
	mod_dog_simp_y[0][k] = yi;
	mod_dog_simp_w[0][k] = rf[xi][yi];
	k += 1;
      }
    }
  }
  for(i=0;i<mod_dog_simp_nin[0];i++){
    sprintf(ggstr,"%d (%d,%d) %.2f\n",i,mod_dog_simp_x[0][i],
	    mod_dog_simp_y[0][i],mod_dog_simp_w[0][i]);
    mylog(mylogf,ggstr);
  }
  for(i=0;i<yn;i++){
    for(j=0;j<xn;j++)
      if (flag[i][j])
	if (rf[i][j] > 0.0)
	  mylog(mylogf," O");
	else
	  mylog(mylogf," X");
      else
	mylog(mylogf," .");
    mylog(mylogf,"\n");
  }
}
void model_dog_simp_01_get_response(m,s,r,k)
     struct model_struct *m;        
     struct stim_struct *s;         
     struct response_struct *r;     
     int k; 
{
  int xn,yn,tn,ns,vn;
  int **cnt0,**cnt1,*inx,*iny,nin,**flag,gnms,seed,tvn;
  float ***rdog,*fs,sampling,***s0,***s1,*inw,*gtx,*gti,sampms,*v,samp;
  float ***tgx,***tgi,***tgv;
  char *outfile;
  xn = mod_dog_xn;
  yn = mod_dog_yn;
  tn = mod_dog_tn;
  rdog = mod_dog_get_response(s,k,NULL,0); 
  outfile = paramfile_get_char_param_default(m->ppl,"write_3d_response","");
  if (strcmp(outfile,"")!=0)
    write_3d_data_part(outfile,rdog,1,xn,1,yn,1,tn,4,2,1);
  myfree(outfile);
  samp = 1.0/mod_dog_tscale;
  flag = mod_dog_simp_flag[0];
  seed = m->mseed[r->tsi];
  mod_dog_get_spikes_2d_flag(m,r,rdog,NULL,1,xn,yn,tn,flag,samp,seed,
			     &s1,&cnt1,0,&tgx,&tgi,&tgv,&tvn); 
  mod_dog_get_spikes_2d_flag(m,r,rdog,NULL,0,xn,yn,tn,flag,samp,seed,
			     &s0,&cnt0,0,&tgx,&tgi,&tgv,&tvn); 
  nin = mod_dog_simp_nin[0];
  inx = mod_dog_simp_x[0];
  iny = mod_dog_simp_y[0];
  inw = mod_dog_simp_w[0];
  mod_dog_get_gtx_gti(m,s0,s1,cnt0,cnt1,nin,inx,iny,inw,&gtx,&gti,&gnms,0);
  sampms = 1000.0;
  ifc_test(myid,m,"ifc_simp",gtx,gti,sampms,gnms,m->mseed[r->tsi],1,
	   (char *)NULL,&fs,&ns,0,&v,&vn);
  myfree(gti); myfree(gtx);
  sampling = mod_util_resp_get_samp(r,"spikes");
  multiply_farray(fs,ns,sampling/1000.0);
  (void)mod_util_resp_check_store_s(r,"spikes",mylogf,fs,ns,0);
}
void model_dog_comp_01_prep(mppl,r)
     struct param_pair_list *mppl; 
     struct response_struct *r;    
{
  int i,j,k;
  int xn,yn,nsamp,seed,xi,yi,n;
  int **flag;
  float **rf,sscale,theta,ssd,sf,ph,min,max,fsamp,dtheta;
  mylog(mylogf,"  MODEL_DOG_COMP_01_PREP\n");
  sscale = mod_dog_sscale;
  xn = mod_dog_xn;
  yn = mod_dog_yn;
  ssd   = paramfile_get_float_param_or_exit(mppl,"simp_ssd");
  sf    = paramfile_get_float_param_or_exit(mppl,"simp_sf");
  ph    = paramfile_get_float_param_or_exit(mppl,"simp_phase");
  theta = paramfile_get_float_param_or_exit(mppl,"simp_ori");
  fsamp = paramfile_get_float_param_or_exit(mppl,"simp_rf_fsamp");
  seed  = paramfile_get_int_param_or_exit(mppl,"simp_rf_seed");
  nsamp = paramfile_get_int_param_or_exit(mppl,"simp_rf_nsamp");
  if (seed > 0)
    seed = -seed;
  n = paramfile_get_int_param_or_exit(mppl,"comp_nsimp");
  mod_dog_simp_n = n;
  mod_dog_simp_nin = get_const_iarray(n,nsamp); 
  mod_dog_simp_x = get_2d_iarray(n,nsamp);
  mod_dog_simp_y = get_2d_iarray(n,nsamp);
  mod_dog_simp_w = get_2d_farray(n,nsamp);
  mod_dog_simp_rf = (float ***)myalloc(n*sizeof(float **));
  mod_dog_simp_flag = get_zero_3d_iarray(n,xn,yn);
  dtheta = 360.0/(float)n;
  for(i=0;i<n;i++){
    rf = gabor_2d_space(xn,yn,sscale,ssd,sf,theta,ph+(float)i*dtheta);
    mod_dog_simp_rf[i] = rf;
    max = max_of_2d_farray(rf,xn,yn);
    min = min_of_2d_farray(rf,xn,yn);
    sprintf(ggstr,"  min, max =  %f %f\n",min,max);
    mylog(mylogf,ggstr);
    flag = mod_dog_simp_flag[i];
    k = 0;
    while(k < nsamp){
      xi = (int)((float)xn * myrand_util_ran2(&seed));
      yi = (int)((float)yn * myrand_util_ran2(&seed));
      if (flag[xi][yi] == 0){
	if ((rf[xi][yi] > fsamp*max)|| (rf[xi][yi] < fsamp*min)){
	  flag[xi][yi] = 1;
	  mod_dog_simp_x[i][k] = xi;
	  mod_dog_simp_y[i][k] = yi;
	  mod_dog_simp_w[i][k] = rf[xi][yi];
	  k += 1;
	}
      }
    }
    for(j=0;j<mod_dog_simp_nin[i];j++){
      sprintf(ggstr,"%d (%d,%d) %.2f\n",j,mod_dog_simp_x[i][j],
	      mod_dog_simp_y[i][j],mod_dog_simp_w[i][j]);
      mylog(mylogf,ggstr);
    }
    for(j=0;j<yn;j++){
      for(k=0;k<xn;k++)
	if (flag[j][k])
	  if (rf[j][k] > 0.0)
	    mylog(mylogf," O");
	  else
	    mylog(mylogf," X");
	else
	  mylog(mylogf," .");
      mylog(mylogf,"\n");
    }
  }
}
void model_dog_comp_01_get_response(m,s,r,k)
     struct model_struct *m;        
     struct stim_struct *s;         
     struct response_struct *r;     
     int k; 
{
  int i,xi,yi;
  int xn,yn,tn,ns,vn,seed,tvn;
  int *cnts,**cnt0,**cnt1,*inx,*iny,nin,**flag,gnms;
  float ***rdog,*fs,sampling,amp,t1,t2,*v;
  float **ss,***s0,***s1,*inw,*gtx,*gti,sampms,samp;
  float ***tgx,***tgi,***tgv;
  char *outfile;
  sampms = 1000.0;
  xn = mod_dog_xn;
  yn = mod_dog_yn;
  tn = mod_dog_tn;
  rdog = mod_dog_get_response(s,k,NULL,0); 
  outfile = paramfile_get_char_param_default(m->ppl,"write_3d_response","");
  if (strcmp(outfile,"")!=0)
    write_3d_data_part(outfile,rdog,1,xn,1,yn,1,tn,4,2,1);
  myfree(outfile);
  flag = get_zero_2d_iarray(xn,yn);
  for(i=0;i<mod_dog_simp_n;i++){
    for(xi=0;xi<xn;xi++)
      for(yi=0;yi<yn;yi++)
	if (mod_dog_simp_flag[i][xi][yi])
	  flag[xi][yi] = 1;
  }
  samp = 1.0/mod_dog_tscale;
  seed = m->mseed[r->tsi];
  mod_dog_get_spikes_2d_flag(m,r,rdog,NULL,1,xn,yn,tn,flag,samp,seed,
			     &s1,&cnt1,0,&tgx,&tgi,&tgv,&tvn); 
  mod_dog_get_spikes_2d_flag(m,r,rdog,NULL,0,xn,yn,tn,flag,samp,seed,
			     &s0,&cnt0,0,&tgx,&tgi,&tgv,&tvn); 
  cnts = (int *)myalloc(mod_dog_simp_n*sizeof(int));
  ss = (float **)myalloc(mod_dog_simp_n*sizeof(float *));
  for(i=0;i<mod_dog_simp_n;i++){
    nin = mod_dog_simp_nin[i];
    inx = mod_dog_simp_x[i];
    iny = mod_dog_simp_y[i];
    inw = mod_dog_simp_w[i];
    mod_dog_get_gtx_gti(m,s0,s1,cnt0,cnt1,nin,inx,iny,inw,&gtx,&gti,&gnms,i);
    ifc_test(myid,m,"ifc_simp",gtx,gti,sampms,gnms,m->mseed[r->tsi],1,
	     (char *)NULL,&ss[i],&cnts[i],0,&v,&vn);
    myfree(gti); myfree(gtx);
  }
  mylog(mylogf,"    Summing conductance from 'mod_dog_simp_n' simple cells.\n");
  t1   = paramfile_get_float_param_or_exit(m->ppl,"comp_ex_tau_f");
  t2   = paramfile_get_float_param_or_exit(m->ppl,"comp_ex_tau_r");
  amp  = paramfile_get_float_param_or_exit(m->ppl,"comp_ex_amp");
  mod_dog_get_gt_sarray(m,ss,cnts,mod_dog_simp_n,t1,t2,amp,&gtx,&gnms);
  mylog(mylogf,"    Computing complex cell spike train.\n");
  gti = get_zero_farray(gnms);
  ifc_test(myid,m,"ifc_comp",gtx,gti,sampms,gnms,m->mseed[r->tsi],1,
	   (char *)NULL,&fs,&ns,0,&v,&vn);
  myfree(gtx); myfree(gti);
  sampling = mod_util_resp_get_samp(r,"spikes");
  multiply_farray(fs,ns,sampling/1000.0);
  (void)mod_util_resp_check_store_s(r,"spikes",mylogf,fs,ns,0);
}
void mod_dog_ds_get_pool_g(s1,s0,cnt1,cnt0,tf,ds_tmu,ds_tsd,t1,t2,amp,rg,rn)
     float ***s1,***s0;
     int **cnt1,**cnt0;
     float tf,ds_tmu,ds_tsd,t1,t2,amp;
     float **rg;
     int *rn;
{
  int i,j;
  int xi,yi,n,nx,t,tn,nms;
  static int trn = 0;
  float *ts,*xshape,*st0,*st1,*g0,*g1,*gt,*gtot,ampl,theta,period;
  char tstr[SLEN];
  nx = (int)(t1 * 7.0);        
  xshape = diff_exp_farray(0.0,amp,t1,t2,nx);
  tn = mod_dog_tn;
  nms = (int)((float)tn * 1000.0*mod_dog_tscale); 
  st0 = get_zero_farray(nms);
  st1 = get_zero_farray(nms);
  gt  = get_zero_farray(nms);
  gtot = get_zero_farray(nms);
  if (myid == -1){
    if (tf > 0.0){
      append_string_to_file("zz.ampthet.pl","/newplot\n");
      sprintf(tstr,"/plotname p_%d\n",trn);
      append_string_to_file("zz.ampthet.pl",tstr);
    }
  }
  n = mod_dog_ds_nsub;
  for(i=0;i<n;i++){ 
    xi = mod_dog_ds_subx1[i];
    yi = mod_dog_ds_suby1[i];
    ts = s0[xi][yi];
    for(j=0;j<cnt0[xi][yi];j++){ 
      t = my_rint(ts[j]);
      if ((t >= 0)&&(t < nms)){
	st0[t] += 1.0;
      }
    }
    xi = mod_dog_ds_subx2[i];
    yi = mod_dog_ds_suby2[i];
    ts = s0[xi][yi];
    for(j=0;j<cnt0[xi][yi];j++){ 
      t = my_rint(ts[j] + ds_tmu);
      if ((t >= 0)&&(t < nms)){
	st1[t] += 1.0;
      }
    }
    g0 = convolve_with_mask_causal(st0,nms,xshape,nx);
    g1 = convolve_with_mask_causal(st1,nms,xshape,nx);
    if (myid == -1){
      if (i==0){
	append_farray_plot("zz.subunit.pl","s0_subunit_0",st0,nms,1);
	append_farray_plot("zz.subunit.pl","s1_subunit_0",st1,nms,1);
	append_farray_plot("zz.subunit.pl","g0_subunit_0",g0,nms,1);
	append_farray_plot("zz.subunit.pl","g1_subunit_0",g1,nms,1);
      }
    }
    for(j=0;j<nms;j++){
      gt[j] = g0[j]*g1[j];
      gtot[j] += gt[j];
      st0[j] = st1[j] = 0.0;  
    }
    myfree(g0); myfree(g1);
    if (myid == -1){
      if (i==1)
	append_farray_plot("zz.subunit.pl","Prod_subunit_0",gt,nms,1);
    }
    if (myid == -1){
      if (tf > 0.0){
	period = 1000.0/tf;
	get_fourier_harmonic_farray(gt,nms,1,period,&ampl,&theta);
	sprintf(tstr,"%.4f %.4f\n",ampl,theta);
	append_string_to_file("zz.ampthet.pl",tstr);
      }
    }
    ts = s1[xi][yi];
    for(j=0;j<cnt1[xi][yi];j++){ 
      t = my_rint(ts[j]);
      if ((t >= 0)&&(t < nms)){
	st0[t] += 1.0;
      }
    }
    xi = mod_dog_ds_subx2[i];
    yi = mod_dog_ds_suby2[i];
    ts = s1[xi][yi];
    for(j=0;j<cnt1[xi][yi];j++){ 
      t = my_rint(ts[j] + ds_tmu);
      if ((t >= 0)&&(t < nms)){
	st1[t] += 1.0;
      }
    }
    g0 = convolve_with_mask_causal(st0,nms,xshape,nx);
    g1 = convolve_with_mask_causal(st1,nms,xshape,nx);
    if (myid == -1){
      if (i==0){
	append_farray_plot("zz.subunit.pl","s0_subunit_0",st0,nms,1);
	append_farray_plot("zz.subunit.pl","s1_subunit_0",st1,nms,1);
	append_farray_plot("zz.subunit.pl","g0_subunit_0",g0,nms,1);
	append_farray_plot("zz.subunit.pl","g1_subunit_0",g1,nms,1);
      }
    }
    for(j=0;j<nms;j++){
      gt[j] = g0[j]*g1[j];
      gtot[j] += gt[j];
      st0[j] = st1[j] = 0.0;  
    }
    myfree(g0); myfree(g1);
    if (myid == -1){
      if (i==1)
	append_farray_plot("zz.subunit.pl","Prod_subunit_0",gt,nms,1);
    }
    if (myid == -1){
      if (tf > 0.0){
	period = 1000.0/tf;
	get_fourier_harmonic_farray(gt,nms,1,period,&ampl,&theta);
	sprintf(tstr,"%.4f %.4f\n",ampl,theta);
	append_string_to_file("zz.ampthet.pl",tstr);
      }
    }
  }
  myfree(st0); myfree(st1); myfree(gt); myfree(xshape);
  gt = smooth_with_gaussian(gtot,nms,1.0,0.01);
  myfree(gtot);
  gtot = gt;
  if (myid == -1)
    append_farray_plot("zz.subunit.pl","Tot_of_Prods__smoothsig=1",gtot,nms,1);
  trn += 1;
  *rg = gtot; *rn = nms;
}
void model_dog_ds_01_prep(mppl,r)
     struct param_pair_list *mppl; 
     struct response_struct *r;    
{
  int i;
  int xn,yn,seed,gseed,n,*sx,*sy,sn,done;
  float ssd,xmu,xsd,ymu,ysd,xmuo2,ymuo2,xir,yir,rfs,eps,wsd;
  char tstr[SLEN];
  mylog(mylogf,"  MODEL_DOG_DS_01_PREP\n");
  xn = mod_dog_xn;
  yn = mod_dog_yn;
  ssd   = paramfile_get_float_param_or_exit(mppl,"ds_ssd");
  xmu = paramfile_get_float_param_or_exit(mppl,"ds_xmu");
  xsd = paramfile_get_float_param_or_exit(mppl,"ds_xsd");
  ymu = paramfile_get_float_param_or_exit(mppl,"ds_ymu");
  ysd = paramfile_get_float_param_or_exit(mppl,"ds_ysd");
  seed  = paramfile_get_int_param_or_exit(mppl,"ds_seed");
  if (seed > 0)
    seed = -seed;
  gseed = seed;
  n = paramfile_get_int_param_or_exit(mppl,"ds_nsub");
  rfs = paramfile_get_float_param_or_exit(mppl,"ds_rfs");
  wsd = paramfile_get_float_param_or_exit(mppl,"ds_wsd");
  eps = paramfile_get_float_param_or_exit(mppl,"ds_eps");
  model_dog_get_sampling_points(xn,yn,n,seed,ssd,rfs,wsd,eps,&sx,&sy,&sn);
  n = sn;
  mod_dog_ds_nsub = n;
  mod_dog_ds_subx1 = (int *)myalloc(n*sizeof(int));
  mod_dog_ds_suby1 = (int *)myalloc(n*sizeof(int));
  mod_dog_ds_subx2 = (int *)myalloc(n*sizeof(int));
  mod_dog_ds_suby2 = (int *)myalloc(n*sizeof(int));
  xmuo2 = xmu/2.0;
  ymuo2 = ymu/2.0;
  for(i=0;i<n;i++){
    xir = (float)sx[i] + 0.5*nr_util_gasdev(&gseed);
    yir = (float)sy[i] + 0.5*nr_util_gasdev(&gseed);
    done = 0;
    while(!done){
      mod_dog_ds_subx1[i] = my_rint(xir - xmuo2 + xsd*nr_util_gasdev(&gseed));
      mod_dog_ds_suby1[i] = my_rint(yir - ymuo2 + ysd*nr_util_gasdev(&gseed));
      mod_dog_ds_subx2[i] = my_rint(xir + xmuo2 + xsd*nr_util_gasdev(&gseed));
      mod_dog_ds_suby2[i] = my_rint(yir + ymuo2 + ysd*nr_util_gasdev(&gseed));
      if ((mod_dog_ds_subx1[i] >= 0)&&(mod_dog_ds_subx1[i] < xn)&&
	  (mod_dog_ds_subx2[i] >= 0)&&(mod_dog_ds_subx2[i] < xn)&&
	  (mod_dog_ds_suby1[i] >= 0)&&(mod_dog_ds_suby1[i] < yn)&&
	  (mod_dog_ds_suby2[i] >= 0)&&(mod_dog_ds_suby2[i] < yn)){
	done = 1;
      }
    }
  }
  if (myid == -1){
    mylog(mylogf,"    Writing x,y coords for subunits to:  zzz.DS.RF.pl\n");
    remove_file("zzz.DS.RF.pl");
    append_string_to_file("zzz.DS.RF.pl","/newplot\n");
    sprintf(tstr,"0 0\n0 %d\n%d %d\n%d 0\n0 0\n",yn-1,xn-1,yn-1,xn-1);
    append_string_to_file("zzz.DS.RF.pl",tstr);
    for(i=0;i<n;i++){
      append_string_to_file("zzz.DS.RF.pl","/newplot\n");
      sprintf(tstr,"%4d %4d\n%4d %4d\n",mod_dog_ds_subx1[i],
	      mod_dog_ds_suby1[i],mod_dog_ds_subx2[i],mod_dog_ds_suby2[i]);
      append_string_to_file("zzz.DS.RF.pl",tstr);
    }
  }
  myfree(sx); myfree(sy);
}
void model_dog_ds_01_get_response(m,s,r,k)
     struct model_struct *m;        
     struct stim_struct *s;         
     struct response_struct *r;     
     int k;                         
{
  int i,j;
  int xn,yn,tn,ns,**cnt0,**cnt1,**flag,gnms,vn,seed,tvn;
  float ***rdog,*fs,sampling,amp,t1,t2,ds_tmu,ds_tsd,*v;
  float ***s0,***s1,*gtx,*gti,sampms,tf,samp;
  float ***tgx,***tgi,***tgv;
  char *outfile;
  xn = mod_dog_xn;
  yn = mod_dog_yn;
  tn = mod_dog_tn;
  if (paramfile_test_param(s->ppl,"tf"))
    tf = paramfile_get_float_param_or_exit(s->ppl,"tf","");
  else
    tf = -1.0;
  rdog = mod_dog_get_response(s,k,NULL,0); 
  outfile = paramfile_get_char_param_default(m->ppl,"write_3d_response","");
  if (strcmp(outfile,"")!=0)
    write_3d_data_part(outfile,rdog,1,xn,1,yn,1,tn,4,2,1);
  myfree(outfile);
  flag = get_zero_2d_iarray(xn,yn);
  for(i=0;i<mod_dog_ds_nsub;i++){ 
    flag[mod_dog_ds_subx1[i]][mod_dog_ds_suby1[i]] = 1;
    flag[mod_dog_ds_subx2[i]][mod_dog_ds_suby2[i]] = 1;
  }
  for(i=0;i<yn;i++){
    for(j=0;j<xn;j++){
      if (flag[j][i])
	mylog(mylogf," *");
      else
	mylog(mylogf," .");
    }
    mylog(mylogf,"\n");
  }
  samp = 1.0/mod_dog_tscale;
  seed = m->mseed[r->tsi];
  mod_dog_get_spikes_2d_flag(m,r,rdog,NULL,1,xn,yn,tn,flag,samp,seed,
			     &s1,&cnt1,0,&tgx,&tgi,&tgv,&tvn);
  mod_dog_get_spikes_2d_flag(m,r,rdog,NULL,0,xn,yn,tn,flag,samp,seed,
			     &s0,&cnt0,0,&tgx,&tgi,&tgv,&tvn);
  t1  = paramfile_get_float_param_or_exit(m->ppl,"ds_ex_tau_f");
  t2  = paramfile_get_float_param_or_exit(m->ppl,"ds_ex_tau_r");
  amp = paramfile_get_float_param_or_exit(m->ppl,"ds_ex_amp");
  ds_tmu  = paramfile_get_float_param_or_exit(m->ppl,"ds_tmu");
  ds_tsd  = paramfile_get_float_param_or_exit(m->ppl,"ds_tsd");
  mod_dog_ds_get_pool_g(s1,s0,cnt1,cnt0,tf,ds_tmu,ds_tsd,t1,t2,amp,&gtx,&gnms);
  mylog(mylogf,"    Computing complex cell spike train.\n");
  gti = get_zero_farray(gnms);
  sampms = 1000.0;
  ifc_test(myid,m,"ifc_comp",gtx,gti,sampms,gnms,m->mseed[r->tsi],1,
	   (char *)NULL,&fs,&ns,0,&v,&vn);
  myfree(gtx); myfree(gti);
  sampling = mod_util_resp_get_samp(r,"spikes");
  multiply_farray(fs,ns,sampling/1000.0);
  (void)mod_util_resp_check_store_s(r,"spikes",mylogf,fs,ns,0);
}
void model_dog_ds_02_prep(mppl,r)
     struct param_pair_list *mppl; 
     struct response_struct *r;    
{
  int i,j,k;
  int xn,yn,seed,n;
  int **flag,nsimp,nsamp,xi,yi;
  float **rf,sscale,theta,ssd,sf,ph,min,max,fsamp;
  float tph;
  mylog(mylogf,"  MODEL_DOG_DS_02_PREP\n");
  sscale = mod_dog_sscale;
  xn = mod_dog_xn;
  yn = mod_dog_yn;
  n = paramfile_get_int_param_or_exit(mppl,"ds_nsub");
  nsimp = 2*n;
  ssd   = paramfile_get_float_param_or_exit(mppl,"simp_ssd");
  sf    = paramfile_get_float_param_or_exit(mppl,"simp_sf");
  ph    = paramfile_get_float_param_or_exit(mppl,"simp_phase");
  theta = paramfile_get_float_param_or_exit(mppl,"simp_ori");
  fsamp = paramfile_get_float_param_or_exit(mppl,"simp_rf_fsamp");
  seed  = paramfile_get_int_param_or_exit(mppl,"simp_rf_seed");
  nsamp = paramfile_get_int_param_or_exit(mppl,"simp_rf_nsamp");
  if (seed > 0)
    seed = -seed;
  mod_dog_simp_n = nsimp;
  mod_dog_simp_nin = get_const_iarray(nsimp,nsamp);
  mod_dog_simp_x = get_2d_iarray(nsimp,nsamp);
  mod_dog_simp_y = get_2d_iarray(nsimp,nsamp);
  mod_dog_simp_w = get_2d_farray(nsimp,nsamp);
  mod_dog_simp_rf = (float ***)myalloc(nsimp*sizeof(float **));
  mod_dog_simp_flag = get_zero_3d_iarray(nsimp,xn,yn);
  tph = ph;
  for(i=0;i<nsimp;i++){
    if (i%2 == 0){
      rf = gabor_2d_space(xn,yn,sscale,ssd,sf,theta,tph);
    }else{
      rf = gabor_2d_space(xn,yn,sscale,ssd,sf,theta,tph+90.0);
      tph += 90.0; 
    }
    mod_dog_simp_rf[i] = rf;
    max = max_of_2d_farray(rf,xn,yn);
    min = min_of_2d_farray(rf,xn,yn);
    sprintf(ggstr,"  min, max =  %f %f\n",min,max);
    mylog(mylogf,ggstr);
    flag = mod_dog_simp_flag[i];
    k = 0;
    while(k < nsamp){
      xi = (int)((float)xn * myrand_util_ran2(&seed));
      yi = (int)((float)yn * myrand_util_ran2(&seed));
      if (flag[xi][yi] == 0){
	if ((rf[xi][yi] > fsamp*max)|| (rf[xi][yi] < fsamp*min)){
	  flag[xi][yi] = 1;
	  mod_dog_simp_x[i][k] = xi;
	  mod_dog_simp_y[i][k] = yi;
	  mod_dog_simp_w[i][k] = rf[xi][yi];
	  k += 1;
	}
      }
    }
    for(j=0;j<mod_dog_simp_nin[i];j++){
      sprintf(ggstr,"%d (%d,%d) %.2f\n",j,mod_dog_simp_x[i][j],
	      mod_dog_simp_y[i][j],mod_dog_simp_w[i][j]);
      mylog(mylogf,ggstr);
    }
    for(j=0;j<yn;j++){
      for(k=0;k<xn;k++)
	if (flag[j][k])
	  if (rf[j][k] > 0.0)
	    mylog(mylogf," O");
	  else
	    mylog(mylogf," X");
	else
	  mylog(mylogf," .");
      mylog(mylogf,"\n");
    }
  }
}
void mod_dog_get_gt_ds_02(m,s,cnt,n,t1,t2,amp,delay,rgt,rnms)
     struct model_struct *m;   
     float **s;
     int *cnt,n;
     float t1,t2,amp;
     int *delay;               
     float **rgt;              
     int *rnms;                
{
  int i,j;
  int nx,tn,nms;
  float *xshape,*gt,*g,xarea;
  tn = mod_dog_tn;
  nms = (int)((float)tn * 1000.0*mod_dog_tscale); 
  gt = get_zero_farray(nms);
  nx = (int)(t1 * 7.0);        
  xshape = diff_exp_farray(0.0,amp,t1,t2,nx);
  xarea = (float)nx * mean_farray(xshape,nx);
  if (myid == -1)
    append_farray_plot("zz.xshape.pl","xshape_0",xshape,nx,1);
  for(i=0;i<n;i+=2){
    mod_dog_get_gt_prod_pair_sarray(s[i],cnt[i],s[i+1],cnt[i+1],nms,
				    delay[i/2],xshape,nx,&g);
    for(j=0;j<nms;j++)
      gt[j] += g[j];
    myfree(g);
  }
  myfree(xshape);
  t1 *= 1.414;
  t2 *= 1.414;
  nx = (int)(t1 * 7.0);        
  xshape = diff_exp_farray(0.0,amp,t1,t2,nx);
  norm_area_farray(xshape,nx,xarea*1.0);  
  if (myid == -1)
    append_farray_plot("zz.xshape.pl","xshape_0",xshape,nx,1);
  for(i=0;i<n;i+=2){
    mod_dog_get_gt_prod_pair_sarray(s[i],cnt[i],s[i+1],cnt[i+1],nms,
				    delay[i/2]*2,xshape,nx,&g);
    for(j=0;j<nms;j++)
      gt[j] += g[j];
    myfree(g);
  }
  myfree(xshape);
  t1 *= 1.414;
  t2 *= 1.414;
  nx = (int)(t1 * 7.0);        
  xshape = diff_exp_farray(0.0,amp,t1,t2,nx);
  norm_area_farray(xshape,nx,xarea*1.0); 
  if (myid == -1)
    append_farray_plot("zz.xshape.pl","xshape_0",xshape,nx,1);
  for(i=0;i<n;i+=2){
    mod_dog_get_gt_prod_pair_sarray(s[i],cnt[i],s[i+1],cnt[i+1],nms,
				    delay[i/2]*4,xshape,nx,&g);
    for(j=0;j<nms;j++)
      gt[j] += g[j];
    myfree(g);
  }
  myfree(xshape);
  t1 *= 2.0;
  t2 *= 2.0;
  nx = (int)(t1 * 7.0);        
  xshape = diff_exp_farray(0.0,amp,t1,t2,nx);
  norm_area_farray(xshape,nx,xarea*0.2); 
  if (myid == -1)
    append_farray_plot("zz.xshape.pl","xshape_0",xshape,nx,1);
  for(i=0;i<n;i+=2){
    mod_dog_get_gt_prod_pair_sarray(s[i],cnt[i],s[i+1],cnt[i+1],nms,
				    delay[i/2]*8,xshape,nx,&g);
    for(j=0;j<nms;j++)
      gt[j] += g[j];
    myfree(g);
  }
  myfree(xshape);
  if (myid == -1)
    append_farray_plot("zz.ds02.gtot.out.pl","gtx_ds_tot",gt,nms,1);
  *rgt = gt; *rnms = nms;
}
void model_dog_ds_02_get_response(m,s,r,k)
     struct model_struct *m;        
     struct stim_struct *s;         
     struct response_struct *r;     
     int k; 
{
  int i,xi,yi;
  int xn,yn,tn,ns,vn,seed,tvn;
  int *cnts,**cnt0,**cnt1,*inx,*iny,nin,**flag,gnms,*delay,ds_delay;
  int meshflag,sc_simp;
  float ***rdog,*fs,sampling,amp,t1,t2,*v,fscale;
  float **ss,***s0,***s1,*inw,*gtx,*gti,samp,sampms;
  float ***tgx,***tgi,***tgv;
  char *outfile;
  sampms = 1000.0;
  xn = mod_dog_xn;
  yn = mod_dog_yn;
  tn = mod_dog_tn;
  meshflag = paramfile_get_int_param_default(m->ppl,"mesh_flag",0);
  if (meshflag){
    fscale = 22.0;
    sprintf(ggstr,"      Scaling 'rdog' by %f, to match non-mesh ds02.\n",
	    fscale);
    mylog(mylogf,ggstr);
    printf("  ***  See .../libc/archive/mod_mesh_util.* for old code\n");
    exit_error("MODEL_DOG_DS_02_GET_RESPONSE","Old way no longer supported");
  }else{
    rdog = mod_dog_get_response(s,k,NULL,0); 
  }
  { 
    float min,max;
    get_min_max_3d_farray(rdog,1,xn,1,yn,1,tn,&min,&max);
    sprintf(ggstr,"  3D retina response min, max =  %f %f\n",min,max);
    mylog(mylogf,ggstr);
    if (myid == -1)
      append_farray_plot("zz.rdog_mid_v_t.pl","rdog_mid_v_t",
			 rdog[xn/2][yn/2]+1,tn,1);
  }
  outfile = paramfile_get_char_param_default(m->ppl,"write_3d_response","");
  if (strcmp(outfile,"")!=0)
    write_3d_data_part(outfile,rdog,1,xn,1,yn,1,tn,4,2,1);
  myfree(outfile);
  flag = get_zero_2d_iarray(xn,yn);
  for(i=0;i<mod_dog_simp_n;i++){
    for(xi=0;xi<xn;xi++)
      for(yi=0;yi<yn;yi++)
	if (mod_dog_simp_flag[i][xi][yi])
	  flag[xi][yi] = 1;
  }
  samp = 1.0/mod_dog_tscale;
  seed = m->mseed[r->tsi];
  mod_dog_get_spikes_2d_flag(m,r,rdog,NULL,1,xn,yn,tn,flag,samp,seed,
			     &s1,&cnt1,0,&tgx,&tgi,&tgv,&tvn);
  mod_dog_get_spikes_2d_flag(m,r,rdog,NULL,0,xn,yn,tn,flag,samp,seed,
			     &s0,&cnt0,0,&tgx,&tgi,&tgv,&tvn);
  cnts = (int *)myalloc(mod_dog_simp_n*sizeof(int));
  ss = (float **)myalloc(mod_dog_simp_n*sizeof(float *));
  sc_simp = 0; 
  for(i=0;i<mod_dog_simp_n;i++){
    nin = mod_dog_simp_nin[i];
    inx = mod_dog_simp_x[i];
    iny = mod_dog_simp_y[i];
    inw = mod_dog_simp_w[i];
    mod_dog_get_gtx_gti(m,s0,s1,cnt0,cnt1,nin,inx,iny,inw,&gtx,&gti,&gnms,i);
    ifc_test(myid,m,"ifc_simp",gtx,gti,sampms,gnms,m->mseed[r->tsi],0,
	     (char *)NULL,&ss[i],&cnts[i],0,&v,&vn);
    sc_simp += cnts[i];
    myfree(gti); myfree(gtx);
  }
  mylog(mylogf,"  Simple cells\n");
  sprintf(ggstr,"    %d cells, %.4f spikes/cell\n",mod_dog_simp_n,
	  (float)sc_simp/(float)mod_dog_simp_n);
  mylog(mylogf,ggstr);
  t1   = paramfile_get_float_param_or_exit(m->ppl,"comp_ex_tau_f");
  t2   = paramfile_get_float_param_or_exit(m->ppl,"comp_ex_tau_r");
  amp  = paramfile_get_float_param_or_exit(m->ppl,"comp_ex_amp");
  ds_delay  = paramfile_get_int_param_or_exit(m->ppl,"ds_delay");
  delay = get_zero_iarray(mod_dog_simp_n/2);
  for(i=0;i<mod_dog_simp_n/2;i++)
    delay[i] = ds_delay;
  mod_dog_get_gt_ds_02(m,ss,cnts,mod_dog_simp_n,t1,t2,amp,delay,&gtx,&gnms);
  myfree(delay);
  sprintf(ggstr,"  Complex DS cell (%d DS subunits)\n",mod_dog_simp_n/2);
  mylog(mylogf,ggstr);
  gti = get_zero_farray(gnms);
  if (myid == -1)
    append_farray_plot("zz.ifc_comp.pl","gtx",gtx,gnms,1);
  {
    float *gsm;
    gsm = smooth_with_gaussian(gtx,gnms,1.0,0.01);
    myfree(gtx);
    gtx = gsm;
    if (myid == -1)
      append_farray_plot("zz.ifc_comp.pl","gtx_smooth",gtx,gnms,1);
  }
  ifc_test(myid,m,"ifc_comp",gtx,gti,sampms,gnms,m->mseed[r->tsi],1,
	   (char *)NULL,&fs,&ns,0,&v,&vn);
  myfree(gtx); myfree(gti);
  sampling = mod_util_resp_get_samp(r,"spikes");
  multiply_farray(fs,ns,sampling/1000.0);
  (void)mod_util_resp_check_store_s(r,"spikes",mylogf,fs,ns,0);
}
void mod_dog_run_dog_01(m,s,r,k,action)
     struct model_struct *m;    
     struct stim_struct *s;     
     struct response_struct *r; 
     int k;                     
     int action;                
{
  myid = mylog_set_id_log(m->process_id,&mylogf);  
  if (action == 0){
    if (m->ppl == NULL){
      model_dog_01_prep_o(m,r);
    }else{
      model_dog_01_prep(m,r);
    }
    model_dog_01_prep_ft();
  }else if (action == 1){
    model_dog_01_get_response(m,s,r,k);
  }else if (action == -1){
    model_dog_01_done(m);
  }
}
void mod_dog_run_simp_01(m,s,r,k,action)
     struct model_struct *m;    
     struct stim_struct *s;     
     struct response_struct *r; 
     int k;                     
     int action;                
{
  myid = mylog_set_id_log(m->process_id,&mylogf);  
  if (action == 0){
    model_dog_01_prep(m,r);
    model_dog_01_prep_ft();
    model_dog_simp_01_prep(m->ppl,r);
  }else if (action == 1){
    model_dog_simp_01_get_response(m,s,r,k);
  }else if (action == -1){
    model_dog_01_done(m);
  }
}
void mod_dog_run_comp_01(m,s,r,k,action)
     struct model_struct *m;    
     struct stim_struct *s;     
     struct response_struct *r; 
     int k;                     
     int action;                
{
  myid = mylog_set_id_log(m->process_id,&mylogf);  
  if (action == 0){
    model_dog_01_prep(m,r);
    model_dog_01_prep_ft();
    model_dog_comp_01_prep(m->ppl,r);
  }else if (action == 1){
    model_dog_comp_01_get_response(m,s,r,k);
  }else if (action == -1){
    model_dog_01_done(m);
  }
}
void mod_dog_run_ds_01(m,s,r,k,action)
     struct model_struct *m;    
     struct stim_struct *s;     
     struct response_struct *r; 
     int k;                     
     int action;                
{
  myid = mylog_set_id_log(m->process_id,&mylogf);  
  if (action == 0){
    model_dog_01_prep(m,r);
    model_dog_01_prep_ft();
    model_dog_ds_01_prep(m->ppl,r);
  }else if (action == 1){
    model_dog_ds_01_get_response(m,s,r,k);
  }else if (action == -1){
    model_dog_01_done(m);
  }
}
void mod_dog_run_ds_02(m,s,r,k,action)
     struct model_struct *m;    
     struct stim_struct *s;     
     struct response_struct *r; 
     int k;                     
     int action;                
{
  int meshflag;
  myid = mylog_set_id_log(m->process_id,&mylogf);  
  meshflag = paramfile_get_int_param_default(m->ppl,"mesh_flag",0);
  if (action == 0){
    if (meshflag)
      model_mesh_rgc_01_prep(m,s,r);
    model_dog_01_prep(m,r);
    model_dog_01_prep_ft();
    model_dog_ds_02_prep(m->ppl,r);
  }else if (action == 1){
    model_dog_ds_02_get_response(m,s,r,k);
  }else if (action == -1){
    model_dog_01_done(m);
    if (meshflag)
      model_mesh_rgc_01_done(m->ppl);
  }
}
