/*C  (C) Copyright  Wyeth Bair 1993-2010                                     */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <X11/Xlib.h>
#include <GL/glx.h>
#include <GL/gl.h>
#include "my_util.h"
#include "carray_util.h"
#include "iarray_util.h"
#include "farray_util.h"
#include "mod_gr_util.h"   
#include "ifc.h"   
#include "mod.h"   
void mod_gr_draw_poly(gw,gh,gx0,gy0,gdx,gdy,px,py,psz,pn,r,g,b)
     int gw,gh;
     float gx0,gy0,gdx,gdy,px,py,psz;
     int pn;
     float r,g,b;
{
  int i;
  float x,y,tx,ty;
  glColor3f(r,g,b);
  glBegin(GL_POLYGON);
  for(i=0;i<pn;i++){
    x = px + psz * cos((float)i/(float)pn * 2.0 * M_PI);
    y = py + psz * sin((float)i/(float)pn * 2.0 * M_PI);
    tx = gx0 + x * gdx;
    ty = gy0 + y * gdy;
    glVertex2f(tx,ty);
  }
  glEnd();
}
void mod_gr_draw_polygon(x,y,dx,dy,r,g,b,ns,openflag)
     float x,y;    
     float dx,dy;  
     float r,g,b;
     int ns;       
     int openflag;
{
  int i;
  float tx,ty;
  if (openflag == 0)
    glBegin(GL_POLYGON);
  else
    glBegin(GL_LINE_LOOP);
  glColor3f(r,g,b);
  for(i=0;i<ns;i++){
    tx = x+dx/2.0 + dx/2.0 * cos((float)i/(float)ns * 2.0 * M_PI);
    ty = y+dy/2.0 + dy/2.0 * sin((float)i/(float)ns * 2.0 * M_PI);
    glVertex2f(tx,ty);
  }
  glEnd();
}
void mod_gr_draw_star(x,y,dx,dy,r,g,b,npt,openflag)
     float x,y;     
     float dx,dy;   
     float r,g,b;
     int npt;       
     int openflag;
{
  int i;
  float tx,ty,rx1,ry1,rx2,ry2,dth;
  float *xc,*yc;  
  rx1 = dx/2.0;
  ry1 = dy/2.0;
  rx2 = dx/4.0;
  ry2 = dy/4.0;
  dth = M_PI/(float)npt;
  xc = (float *)myalloc(npt*2*sizeof(float));
  yc = (float *)myalloc(npt*2*sizeof(float));
  for(i=0;i<npt;i++){
    tx = x+rx1 + rx1 * cos((float)i/(float)npt * 2.0 * M_PI + M_PI/2.0);
    ty = y+ry1 + ry1 * sin((float)i/(float)npt * 2.0 * M_PI + M_PI/2.0);
    xc[2*i] = tx;
    yc[2*i] = ty;
    tx = x+rx1 + rx2 * cos((float)i/(float)npt * 2.0 * M_PI + M_PI/2.0 + dth);
    ty = y+ry1 + ry2 * sin((float)i/(float)npt * 2.0 * M_PI + M_PI/2.0 + dth);
    xc[2*i+1] = tx;
    yc[2*i+1] = ty;
  }
  glColor3f(r,g,b);
  if (openflag)
    glBegin(GL_LINE_LOOP);
  else
    glBegin(GL_POLYGON);
  glVertex2f(xc[npt*2-1],yc[npt*2-1]);
  glVertex2f(xc[0],yc[0]);
  glVertex2f(xc[1],yc[1]);
  glEnd();
  for(i=0;i<(npt-1);i++){
    if (openflag)
      glBegin(GL_LINE_LOOP);
    else
      glBegin(GL_POLYGON);
    glVertex2f(xc[2*i+1],yc[2*i+1]);
    glVertex2f(xc[2*i+2],yc[2*i+2]);
    glVertex2f(xc[2*i+3],yc[2*i+3]);
    glEnd();
  }
  if (openflag)
    glBegin(GL_LINE_LOOP);
  else
    glBegin(GL_POLYGON);
  for(i=0;i<npt;i++){
    glVertex2f(xc[2*i+1],yc[2*i+1]);
  }
  glEnd();
  myfree(xc); myfree(yc);
}
void mod_gr_draw_unit_icon(x,y,dx,dy,icon,w,openflag)
     float x,y;    
     float dx,dy;  
     struct pop_icon *icon;
     float w;      
     int openflag;  
{
  int i;
  int ns;
  float ex,ey,tx,ty,r,g,b;
  if (w == -1.0){  
    if (strcmp(icon->shape,"cs")==0){
      if ((icon->r + icon->g + icon->b) > 2.6){
	r = g = b = 0.0;
      }else{
	r = g = b = 1.0;
      }
    }else{
      r = icon->r + 0.5;  if (r > 1.0) r = 1.0;
      g = icon->g + 0.5;  if (g > 1.0) g = 1.0;
      b = icon->b + 0.5;  if (b > 1.0) b = 1.0;
      r = g = b = 1.0;  
    }
  }else{
    r = w * icon->r;
    g = w * icon->g;
    b = w * icon->b;
  }
  if (strcmp(icon->shape,"circle")==0){
    mod_gr_draw_polygon(x,y,dx,dy,r,g,b,12,openflag);
  }else if (strcmp(icon->shape,"polygon")==0){
    mod_gr_draw_polygon(x,y,dx,dy,r,g,b,icon->nside,openflag);
  }else if (strcmp(icon->shape,"stellate")==0){
    mod_gr_draw_star(x,y,dx,dy,r,g,b,icon->nside,openflag);
  }else if (strcmp(icon->shape,"triangle")==0){
    ex = 0.20 * dx;
    ey = 0.20 * dy;
    if (openflag)
      glBegin(GL_LINE_LOOP);
    else
      glBegin(GL_POLYGON);
    glColor3f(r,g,b);
    glVertex2f(x+ex,y);
    glVertex2f(x   ,y+ey);
    glVertex2f(x+(dx-ex)/2.0,y+dy);
    glVertex2f(x+(dx+ex)/2.0,y+dy);
    glVertex2f(x+dx,y+ey);
    glVertex2f(x+dx-ex,y);
    glEnd();
  }else if (strcmp(icon->shape,"square")==0){
    if (openflag)
      glBegin(GL_LINE_LOOP);
    else
      glBegin(GL_POLYGON);
    glColor3f(r,g,b);
    glVertex2f(x,y);
    glVertex2f(x+dx,y);
    glVertex2f(x+dx,y+dy);
    glVertex2f(x,y+dy);
    glEnd();
  }else if (strcmp(icon->shape,"cs")==0){
    mod_gr_draw_polygon(x+dx/4.0,y+dy/4.0,dx/2.0,dy/2.0,r,g,b,12,openflag);
    mod_gr_draw_polygon(x,y,dx,dy,icon->r1,icon->g1,icon->b1,12,openflag);
  }else{
    exit_error("MOD_GR_DRAW_UNIT_ICON","Unknown shape");
  }
}
void mod_gr_draw_grid_single(gi,gj,gc,icon,w,scale,openflag)
     int gi,gj;
     struct gui_gpar *gc;
     struct pop_icon *icon;
     float w;       
     float scale;   
     int openflag;
{
  float x,y,bw,bh;
  bw = gc->bw;
  bh = gc->bh;
  if (gc->geom_flag == 0){
    x  = gc->x0 + (float)gi * gc->dx;
    y  = gc->y0 + (float)gj * gc->dy;
    if (gj%2 != 0)
      x += gc->oddxoff;
  }else{
    x = gc->x0 + gc->irr_x[gi]*gc->dx;
    y = gc->y0 + gc->irr_y[gi]*gc->dy;
  }
  if (scale != 1.0){
    x += (1.0 - scale) * 0.5 * bw;
    y += (1.0 - scale) * 0.5 * bh;
    bw *= scale;
    bh *= scale;
  }
  glClear(GL_DEPTH_BUFFER_BIT);
  mod_gr_draw_unit_icon(x,y,bw,bh,icon,w,openflag);
}
void mod_gr_get_grid_params(xn,yn,x0,y0,w,h,winfrac,rgx0,rgy0,rgw,rgh)
     int xn,yn;          
     float x0,y0,w,h;    
     float winfrac;      
     float *rgx0,*rgy0;  
     float *rgw,*rgh;    
{
  *rgx0 = x0 + (1.0-winfrac)*w/2.0;
  *rgy0 = y0 + (1.0-winfrac)*h/2.0;
  *rgw = winfrac * w / (float)xn;
  *rgh = winfrac * h / (float)yn;
}
float mod_gr_get_dx_lay(lay,winfrac,cfw)
     struct pop_layer *lay;
     float winfrac;
     float cfw;
{
  float dx;
  int xn,yn;
  xn = my_rint((float)(lay->xn) * lay->xf);
  yn = my_rint((float)(lay->yn) * lay->yf);
  dx = winfrac * cfw / (float)xn; 
  return dx;
}
float mod_gr_get_dy_lay(lay,winfrac,cfh)
     struct pop_layer *lay;
     float winfrac;
     float cfh;
{
  float dy;
  int xn,yn;
  xn = my_rint((float)(lay->xn) * lay->xf);
  yn = my_rint((float)(lay->yn) * lay->yf);
  dy = winfrac * cfh / (float)yn; 
  return dy;
}
struct gui_gpar *mod_gr_get_grid_params_lay(pl,winfrac,cfx0,cfy0,cfw,cfh,
					    gridfrac)
     struct pop_layer *pl;
     float winfrac;
     float cfx0,cfy0;
     float cfw,cfh;
     float gridfrac;
{
  float gdx,gdy,x0,y0;
  struct gui_gpar *gc;
  gc = (struct gui_gpar *)myalloc(sizeof(struct gui_gpar));
  gc->xn = pl->xn;
  gc->yn = pl->yn;
  gdx = mod_gr_get_dx_lay(pl,winfrac,cfw);
  gdy = mod_gr_get_dy_lay(pl,winfrac,cfh);
  gc->gdx = gdx;
  gc->gdy = gdy;
  x0 = cfx0 + (1.0 - winfrac) * cfw/2.0; 
  y0 = cfy0 + (1.0 - winfrac) * cfh/2.0;
  gc->x0 = x0 + pl->x0 * gdx;  
  gc->y0 = y0 + pl->y0 * gdy;
  gc->bw = gridfrac * gdx;
  gc->bh = gridfrac * gdy;
  gc->dx = pl->xf * gdx;
  gc->dy = pl->yf * gdy;
  gc->oddxoff = pl->oddxoff * gdx;  
  return gc;
}
