/*C  (C) Copyright  Wyeth Bair 1993-2013                                     */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "my_util.h"
#include "farray_util.h"
#include "data_util.h"
#include "spike_util.h"
#include "paramfile_util.h"
#include "fft_util.h"
#include "kernel_util.h"
#include "mod_util.h"
#include "ifc_util.h"
#include "mod.h" 
#include "paramfile.h"
static int myid = -1;         
static char *mylogf = NULL;   
char ggstr[LONG_SLEN];        
int     mod_srf_xn;
int     mod_srf_yn;
int     mod_srf_tn;
float   mod_srf_tscale;
float   mod_srf_sscale;     
int     mod_srf_mean0;        
int     mod_srf_framei;       
int     mod_srf_ttn;          
float   mod_srf_norm_pow;     
float **mod_srf_map = NULL;   
int     mod_srf_fxn;          
int     mod_srf_fyn;          
float   mod_srf_fscale;       
void mod_srf_01_prep(m,r)
     struct model_struct *m;    
     struct response_struct *r; 
{
  int xn,yn,tn,fxn,fyn;
  float sscale,tscale,o1,o2,scale1,scale2,nyqf;
  struct onode *fo;
  myid = mylog_set_id_log(m->process_id,&mylogf);  
  mylog(mylogf,"  MOD_SRF_01_PREP\n");
  xn     = onode_getpar_int_exit(m->o,"xn");
  yn     = onode_getpar_int_exit(m->o,"yn");
  tn     = onode_getpar_int_exit(m->o,"tn");
  tscale = onode_getpar_flt_exit(m->o,"tscale");
  sscale = onode_getpar_flt_exit(m->o,"sscale");
  mod_srf_mean0    = onode_getpar_int_dflt(m->o,"zero_mean",0);
  mod_srf_framei   = onode_getpar_int_dflt(m->o,"frame_index",-1);
  mod_srf_norm_pow = onode_getpar_flt_dflt(m->o,"norm_power",0.0);
  mod_srf_ttn      = onode_getpar_int_dflt(m->o,"tn_response",tn);
  if ((mod_srf_framei < -1) || (mod_srf_framei >= tn))
    exit_error("MOD_SRF_01_PREP","Invalid value for 'mod_srf_framei'");
  mod_srf_xn = xn;
  mod_srf_yn = yn;
  mod_srf_tn = tn;
  mod_srf_tscale = tscale;
  mod_srf_sscale = sscale;
  nyqf = 1.0/(2.0*sscale); 
  mod_srf_fscale = nyqf / ((float)xn/2.0);
  mod_srf_fxn = xn;
  mod_srf_fyn = yn/2;
  fxn = mod_srf_fxn;
  fyn = mod_srf_fyn;
  fo = onode_child_get_unique(m->o,"filter");
  if (fo == NULL)
    exit_error("MOD_SRF_01_PREP","Cannot find <filter>");
  o1 = (float)xn/2.0;
  o2 = 0.0;
  scale1 = scale2 = mod_srf_fscale;  
  mod_srf_map = kernu_o_filt2d(mylogf,fo,o1,o2,scale1,scale2,&fxn,&fyn);
  write_2d_data("zz.srf.2d",mod_srf_map,0,0,fxn,fyn,4,2,1,1);
}
void mod_srf_01_done(m)
     struct model_struct *m;        
{
  mylog(mylogf,"  MOD_SRF_01_DONE\n");
  if (mod_srf_map != NULL)
    free_2d_farray(mod_srf_map,mod_srf_fxn);
}
void mod_srf_01_get_response(m,s,r)
     struct model_struct *m;        
     struct stim_struct *s;         
     struct response_struct *r;     
{
  int i,j;
  int xn,yn,tn,fxn,fyn,n;
  float **d2d,**pow2d,*d,tscale,oldmean,rtot;
  xn     = mod_srf_xn;  
  yn     = mod_srf_yn;
  tn     = mod_srf_tn;
  tscale = mod_srf_tscale;
  fxn    = mod_srf_fxn;
  fyn    = mod_srf_fyn;
  if (mod_srf_framei >= 0){
    d2d = get_2d_farray(xn,yn);
    for(i=0;i<xn;i++)
      for(j=0;j<yn;j++){
	d2d[i][j] = s->d[i+1][j+1][mod_srf_framei+1];
      }
    sprintf(ggstr,"    Using only frame %d of (0..%d) frames\n",mod_srf_framei,
	    tn-1);
    mylog(mylogf,ggstr);
    if (mod_srf_mean0 == 1){
      subtract_mean_2d_farray(d2d,xn,yn,&oldmean);
      sprintf(ggstr,"    Subtracted mean (%f) from frame\n",oldmean);
      mylog(mylogf,ggstr);
    }else
      oldmean = 0.0;
    pow2d = power_2d(d2d,xn,yn,0);  
    free_2d_farray(d2d,xn);
  }else{
    exit_error("MOD_SRF_01_GET_RESPONSE","Full 3D not implemented yet");
  }
  if (mod_srf_norm_pow > 0.0){
    sprintf(ggstr,"    Normalizing total power to %f\n",mod_srf_norm_pow);
    mylog(mylogf,ggstr);
    norm_area_2d_farray(pow2d,xn,yn/2,mod_srf_norm_pow);
  }
  rtot = 0.0;
  for(i=0;i<mod_srf_fxn;i++){
    for(j=0;j<mod_srf_fyn;j++){
      rtot += mod_srf_map[i][j] * pow2d[i][j];
    }
  }
  printf("    rtot = %f\n",rtot);
  n = mod_srf_ttn;  
  d = get_const_farray(n,rtot);
  {
    int ns,vn,seed,cpflag;
    float *fs,*v,samp;
    struct onode *sgo;
    seed = m->mseed[r->tsi];
    samp = 1.0/mod_srf_tscale;
    sgo = onode_child_get_unique(m->o,"spike_gen");
    if (sgo != NULL){
      ifc_util_poisson(mylogf,m,sgo,d,n,samp,seed,0,1,&fs,&ns,&v,&vn);
    }else{
      exit_error("MOD_SRF_01_GET_RESPONSE","Cannot find <spike_gen>");
    }
    cpflag = 0;       
    mod_util_resp_store_s(r,0,fs,ns,cpflag,mylogf);
  }
  myfree(d);
}
void mod_srf_run_01(m,s,r,action)
     struct model_struct *m;    
     struct stim_struct *s;     
     struct response_struct *r; 
     int action;                
{
  myid = mylog_set_id_log(m->process_id,&mylogf);  
  if (action == 0){
    mod_srf_01_prep(m,r);
  }else if (action == 1){
    mod_srf_01_get_response(m,s,r);
  }else if (action == -1){
    mod_srf_01_done(m);
  }
}
