/*C  (C) Copyright  Wyeth Bair 1993-2010                                     */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "my_util.h"
#include "misc_util.h"
#include "carray_util.h"
#include "iarray_util.h"
#include "farray_util.h"
#include "func_util.h"
#include "myrand_util.h"
#define RXCENTER  2000.0 
#define RYCENTER  2000.0
struct cell_struct{ 
  float x,y;
  struct cell_struct *prev,*next;
};
struct h_cell_struct{ 
  float x,y;
  int n;                           
  int ui;                          
  int cid;
  int edge_flag;                   
  struct h_cell_struct **neighbor; 
};
float retutil_get_cone_diam_um(ecc,spec)
     float ecc;
     char spec[];
{
  float radum;  
  if (strcmp(spec,"Tyler85_OD")==0)
    radum = 1.4 * pow((ecc + 0.2),0.2);
  else
    radum = 1.4 * pow((ecc + 0.2),0.2);  
  return radum;
}
float retutil_get_cone_density(c,r)
     float c;  
     float r;  
{
  float a,b,dens,d0,d10;
  d0 = 210000.0; 
  d10 = 3500.0;  
  b = d10*pow(10.0,c)/(d0-d10);
  a = b*d0;
  dens = a/(b+pow(r,c));
  return dens;
}
void minimize_mask_length(mask,nm,tolerance)
     float **mask;
     int *nm;
     float tolerance;
{
  int n;
  float eps,min,max,*t;
  n = *nm;
  get_min_max_farray(*mask,n,&min,&max);
  if (min < 0.0)
    eps = tolerance * mymaxf(-min,max);
  else
    eps = tolerance * max;
  while ((fabs((*mask)[n-1]) < eps) && (n > 1))
    n--;
  if (n == 1)
    exit_error("MINIMIZE_MASK_LENGTH","Mask shrunk to length 1");
  if (n == *nm)
    printf("  *** WARNING: last value in mask exceeds tolerance.\n");
  if (n < *nm){ 
    t = copy_farray(*mask,n);
    myfree(*mask);
    *mask = t;
    *nm = n;
  }
}
float *retutil_get_cone_filter(tau_r,tau_d,tau_p,phi,rn)
     float tau_r,tau_d,tau_p,phi;
     int *rn;
{
  int n;
  float *mask;
  n = 500;
  mask = func_primate_photocurrent_farray(0.0,1.0,tau_r,tau_d,tau_p,phi,n);
  append_farray_plot("zzz.photo.mask.pl","MASK",mask,n,1);
  minimize_mask_length(&mask,&n,0.005); 
  norm_area_farray(mask,n,1.0);
  *rn = n;
  return mask;
}
void retutil_init_retina(flag,n,dist,rcx,rcy,rcn,rb,rbn)
     int flag,n;
     float dist,**rcx,**rcy;
     int *rcn;
     struct cell_struct **rb;
     int *rbn;
{
  int cn,bn;
  float *cx,*cy;
  struct cell_struct *b,*p;
  cx = (float *)myalloc(n*sizeof(float));
  cy = (float *)myalloc(n*sizeof(float));
  if (flag == 1){ 
    cx[0] = RXCENTER;
    cy[0] = RYCENTER;
    cx[1] = cx[0] + dist;
    cy[1] = cy[0];
    cx[2] = cx[0] + dist/2.0;
    cy[2] = cy[0] + dist*sqrt(3.0)/2.0;
    cn = 3;
    b = (struct cell_struct *)myalloc(sizeof(struct cell_struct));
    b->x = cx[0]; b->y = cy[0];
    b->next = (struct cell_struct *)myalloc(sizeof(struct cell_struct));
    p = b->next;
    p->x = cx[1]; p->y = cy[1]; p->prev = b;
    p->next = (struct cell_struct *)myalloc(sizeof(struct cell_struct));
    p->next->prev = p;
    p = p->next;
    p->x = cx[2]; p->y = cy[2];
    p->next = b;
    b->prev = p;
    bn = 3;
  }else if (flag == 2){ 
    cx[0] = RXCENTER; cy[0] = RYCENTER;
    cx[1] = cx[0] + dist; cy[1] = cy[1];
    cx[2] = cx[1]; cy[2] = cy[1] + dist;
    cx[3] = cx[0];  cy[3] = cy[2];
    cn = 4;
    b = (struct cell_struct *)myalloc(sizeof(struct cell_struct));
    b->x = cx[0]; b->y = cy[0];
    b->next = (struct cell_struct *)myalloc(sizeof(struct cell_struct));
    b->next->prev = b;
    p = b->next;
    p->x = cx[1]; p->y = cy[1];
    p->next = (struct cell_struct *)myalloc(sizeof(struct cell_struct));
    p->next->prev = p;
    p = p->next;
    p->x = cx[2]; p->y = cy[2];
    p->next = (struct cell_struct *)myalloc(sizeof(struct cell_struct));
    p->next->prev = p;
    p = p->next;
    p->x = cx[3]; p->y = cy[3];
    p->next = b;
    b->prev = p;
    bn = 4;
  }else{
    cn = bn = 0;
    b = NULL;
    exit_error("INIT_RETINA","Unknown flag value");
  }
  *rcx = cx; *rcy = cy; *rcn = cn; *rb = b; *rbn = bn;
}
float retutil_get_angle(a1,a2,b1,b2,c1,c2)
     float a1,a2,b1,b2,c1,c2;
{
  float x,y,u1,u2,v1,v2,theta;
  u1 = b1-a1; u2 = b2-a2;
  v1 = c1-b1; v2 = c2-b2;
  x =  u1*v1 + u2*v2;
  y = -u2*v1 + u1*v2;
  theta = 180.0/M_PI * (atan2(y,x)+ M_PI);
  return theta;
}
void process_next_edge(rb,bn,cx,cy,cn,dist,dnoise,pseed,k)
     struct cell_struct **rb;
     int *bn;
     float *cx,*cy;
     int *cn;
     float dist,dnoise;
     int *pseed;
     int k;
{
  struct cell_struct *p,*t;
  int flag;
  double a1,a2,b1,b2,c1,c2;
  double u1,u2,v1,v2,vo1,vo2,w1,w2,wo1,wo2,r1,r2,ro1,ro2;
  double m,x,y,mr2,mu2,mv2,mvo2,nx,ny,xnew,ynew,tdist2,distsq;
  double dab2,dnew;
  p = *rb;
  distsq = dist*dist;
  a1 = p->x; a2 = p->y;
  b1 = p->next->x; b2 = p->next->y;
  c1 = p->next->next->x; c2 = p->next->next->y;
  v1 = b1-a1; v2 = b2-a2;
  u1 = c1-b1; u2 = c2-b2;
  w1 = c1-a1; w2 = c2-a2;
  vo1 = v2; vo2 = -v1;
  wo1 = w2; wo2 = -w1;
  mv2 = (v1*v1 + v2*v2); 
  m = (u1*v1 + u2*v2)/mv2;
  r1 = m*v1;
  r2 = m*v2;
  mvo2 = (vo1*vo1 + vo2*vo2); 
  m = (u1*vo1 + u2*vo2)/mvo2;
  ro1 = m*vo1;
  ro2 = m*vo2;
  mr2 = r1*r1 + r2*r2; 
  mu2 = u1*u1 + u2*u2; 
  dab2 = mv2/4.0;
  if (distsq <= dab2)
    dnew = 0.0;
  else
    dnew = sqrt(distsq - dab2);
  m = sqrt(mvo2);
  vo1 /= m;
  vo2 /= m;
  xnew =  dnew*vo1 + (a1 + b1)/2.0;
  ynew =  dnew*vo2 + (a2 + b2)/2.0;
  tdist2 = (xnew-c1)*(xnew-c1) + (ynew-c2)*(ynew-c2);
  if ((ro1*vo1 < 0.0)||(ro2*vo2 < 0.0)||(tdist2 >= distsq)) 
    flag = 2;
  else if (mr2 < 0.25*mu2) 
    flag = 1;
  else if ((r1*v1 > 0.0)||(r2*v2 > 0.0)) 
    flag = 2;
  else 
    flag = 0;
  if (tdist2 < distsq)
    if (100.0 > retutil_get_angle(a1,a2,c1,c2,p->next->next->next->x,
				  p->next->next->next->y)){
      flag = 0;
    }
  if (flag==1){
    if (dnoise > 0.0){
      nx = (myrand_util_ran2(pseed)-0.5)*dnoise;
      ny = (myrand_util_ran2(pseed)-0.5)*dnoise;
    }else
      nx = ny = 0.0;
    dab2 = w1*w1 + w2*w2;
    if (distsq <= dab2/4.0)
      dnew = 0.0;
    else
      dnew = sqrt(distsq - dab2/4.0);
    m = sqrt(dab2);
    wo1 /= m;
    wo2 /= m;
    xnew =  dnew*wo1 + (a1 + c1)/2.0;
    ynew =  dnew*wo2 + (a2 + c2)/2.0;
    x = cx[*cn] = xnew + nx;
    y = cy[*cn] = ynew + ny;
    *cn += 1;
    t = p->next;
    t->x = x; t->y = y;
    *rb = t; 
  }else if (flag==2){
    if (dnoise > 0.0){
      nx = (myrand_util_ran2(pseed)-0.5)*dnoise;
      ny = (myrand_util_ran2(pseed)-0.5)*dnoise;
    }else
      nx = ny = 0.0;
    x = cx[*cn] = nx + xnew;
    y = cy[*cn] = ny + ynew;
    *cn += 1;
    t = (struct cell_struct *)myalloc(sizeof(struct cell_struct));
    t->x = x; t->y = y;
    t->next = p->next;
    t->prev = p;
    p->next->prev = t;
    p->next = t;
    *bn += 1;
    *rb = t; 
  }else{
    t = p->next;
    p->next = t->next;
    p->next->prev = p;
    myfree(t);
    *bn -= 1;
  }
}
void retutil_set_neighbors(clist,n,xi,yi,nw,nh,ci,wideflag)
     struct h_cell_struct *clist;
     int n;
     int xi;  
     int yi;  
     int nw;  
     int nh;  
     int ci;  
     int wideflag;  
{
  int k;
  struct h_cell_struct *hc;
  int f0,f1,f2,f3,f4,f5;
  f0 = f1 = f2 = f3 = f4 = f5 = 1;
  if (wideflag == 1){
    if (xi == 0)
      f0 = f2 = f5 = 0;
    else if (xi == (nw-1))
      f1 = f3 = f4 = 0;
  }else{
    if (xi == 0)
      f0 = 0;
    else if (xi == (nw-2))
      f1 = 0;
  }
  if (yi == 0)
    f4 = f5 = 0;
  else if (yi == (nh-1))
    f2 = f3 = 0;
  hc = &(clist[ci]);
  hc->n = f0 + f1 + f2 + f3 + f4 + f5;
  hc->neighbor = (struct h_cell_struct **)
    myalloc(hc->n*sizeof(struct h_cell_struct *));
  k = 0;
  if (f0 == 1){
    hc->neighbor[k] = &(clist[ci-1]);
    k += 1;
  }
  if (f1 == 1){
    hc->neighbor[k] = &(clist[ci+1]);
    k += 1;
  }
  if (f2 == 1){
    hc->neighbor[k] = &(clist[ci+(nw-1)]);
    k += 1;
  }
  if (f3 == 1){
    hc->neighbor[k] = &(clist[ci+nw]);
    k += 1;
  }
  if (f4 == 1){
    hc->neighbor[k] = &(clist[ci-(nw-1)]);
    k += 1;
  }
  if (f5 == 1){
    hc->neighbor[k] = &(clist[ci-nw]);
    k += 1;
  }
}
void retutil_get_cone_coords_regjit(n,dens,uperpix,dnoise,seed,rn,rclist,
				    rnw,rnh)
     int n;              
     float dens;         
     float uperpix;      
     float dnoise;
     int seed;
     int *rn;            
     struct h_cell_struct **rclist; 
     int *rnw;           
     int *rnh;           
{
  int i,j,k;
  int nw,nh,ncone,nhalf,ymid,wideflag;
  float *cx,*cy,cdist,edist,side_len_mm,dy,yval,rx,ry,d,noise_amp;
  struct h_cell_struct *clist;
  side_len_mm = sqrt(n/dens);     
  edist = sqrt(1.0/(sqrt(3.0)/2.0*dens));  
  cdist = edist * 1000.0/uperpix; 
  dy = sqrt(3.0)/2.0 * cdist;     
  nhalf = (int)(side_len_mm / edist / 2);  
  nw    = nhalf * 2 + 1;                   
  ncone = nw;   
  nh = 1;       
  wideflag = 1; 
  while(ncone < n){
    if (wideflag == 1){  
      ncone += 2 * (nw-1);  
    }else{
      ncone += 2 * nw;      
    }
    wideflag = 1 - wideflag;
    nh += 2;
  }
  clist = (struct h_cell_struct *)myalloc(ncone*sizeof(struct h_cell_struct));
  ymid = (nh-1)/2;
  k = 0;
  for(i=0;i<nh;i++){  
    yval = (i - ymid)*dy;
    if (wideflag == 1){
      for(j=0;j<nw;j++){
	clist[k].x = (j - nhalf)*cdist;
	clist[k].y = yval;
	retutil_set_neighbors(clist,ncone,i,j,nw,nh,k,wideflag);
	k += 1;
      }
    }else{
      for(j=0;j<(nw-1);j++){
	clist[k].x = ((float)(j - nhalf) + 0.5)*cdist;
	clist[k].y = yval;
	retutil_set_neighbors(clist,ncone,i,j,nw,nh,k,wideflag);
	k += 1;
      }
    }
    wideflag = 1 - wideflag;
  }
  if ((dnoise > 0.0) && (dnoise <= 1.0)){
    noise_amp = dnoise * cdist/2.0;
    if (seed > 0)
      seed = -seed;
    for(i=0;i<ncone;i++){
      rx = 2.0*myrand_util_ran2(&seed) - 1.0;  
      ry = 2.0*myrand_util_ran2(&seed) - 1.0;  
      d = rx*rx + ry*ry;
      while(d >= 1.0){
	rx = 2.0*myrand_util_ran2(&seed) - 1.0;  
	ry = 2.0*myrand_util_ran2(&seed) - 1.0;  
	d = rx*rx + ry*ry;
      }
      clist[i].x += rx * noise_amp;
      clist[i].y += ry * noise_amp;
    }
  }else if (dnoise != 0.0){
    exit_error("RETUTIL_GET_CONE_COORDS_REGJIT","Noise value out of range");
  }
  *rn = ncone;
  *rclist = clist;
  *rnw = nw;
  *rnh = nh;
}
void retutil_get_cone_coords(n,dens,uperpix,dnoise,seed,rcx,rcy)
     int n;
     float dens;         
     float uperpix;      
     float dnoise;
     int seed;
     float **rcx,**rcy;
{
  int i;
  int flag,cn,bn,diam;
  float *cx,*cy,cdist,edist;
  struct cell_struct *b;
  if (seed > 0)
    seed = -seed;
  edist = sqrt(1.0/(sqrt(3.0)/2.0*dens));
  cdist = edist * 1000.0/uperpix; 
  flag = 1; 
  retutil_init_retina(flag,n,cdist,&cx,&cy,&cn,&b,&bn);
  diam = (int)(cdist-1.0);
  i = 0;
  while(cn < n){
    process_next_edge(&b,&bn,cx,cy,&cn,cdist,dnoise,&seed,i);
    i += 1;
  }
  for(i=0;i<n;i++){
    cx[i] -= RXCENTER;
    cy[i] -= RYCENTER;
  }
  *rcx = cx; *rcy = cy;
}
void retutil_mosaic_reg_s(betw,clist,n,nw,nh)
     int betw;         
     struct h_cell_struct *clist; 
     int n;            
     int nw;           
     int nh;           
{
  int i,j,k;
  int wide_row,rowi,tw,shift,shv,shift_val;
  if ((nh - 1)%4 == 0){
    wide_row = 1;  
    tw = nw;
  }else{
    wide_row = 0;  
    tw = nw - 1;
  }
  shift_val = 1 + betw/2;
  k = 0;  
  rowi = 0;
  shift = 0;  
  for(i=0;i<n;i++){
    if (rowi % (betw+1) == 0){
      if ((k+shv) % (betw+1) == 0){
	if (clist[i].cid != -1){  
	  clist[i].cid = 0;       
	}
      }
    }
    k += 1;
    if (k == tw){  
      k = 0;
      wide_row = 1 - wide_row;  
      if (wide_row == 1)  
	tw = nw;
      else
	tw = nw - 1;
      if (rowi % (betw+1) == 0){  
	shift = 1 - shift;
	if (shift == 1)
	  shv = shift_val;
	else
	  shv = 0;
      }
      rowi += 1;
    }
  }
}
void retutil_get_mosaic_color(xn,yn,n,dens,mupix,dnoise,seedx,seedc,prs,lmr,
			      mylogf,
			      rcx,rcy,rcid,rcn)
     int xn,yn;          
     int n;              
     float dens;
     float mupix;
     float dnoise;
     int seedx,seedc;    
     float prs,lmr;      
     char *mylogf;       
     float **rcx,**rcy;
     int **rcid,*rcn;
{     
  int i,k;
  int *cid,*tid,cn,nn;
  float *cx,*cy,*tx,*ty,prm,rv;
  char tstr[SLEN];
  retutil_get_cone_coords(n,dens,mupix,dnoise,seedx,&cx,&cy);
  if (seedc > 0)
    seedc = -seedc;
  cid = get_zero_iarray(n);
  cn = 0;
  prm = (1.0 - prs) / (lmr + 1.0); 
  for(i=0;i<n;i++){
    cx[i] += (float)(xn-1)/2.0;
    cy[i] += (float)(yn-1)/2.0;
    if ((cx[i] >= 0.0) && (cx[i] < (double)(xn-1)) &&
	(cy[i] >= 0.0) && (cy[i] < (double)(yn-1))){
      rv = myrand_util_ran2(&seedc);
      if (rv < prs)
	cid[i] = 0;
      else if (rv < (prs + prm))
	cid[i] = 1;
      else
	cid[i] = 2;
      cn += 1;  
    }else
      cid[i] = -1;
  }
  sprintf(tstr,"    %d of %d cones within grid\n",cn,n);
  mylog(mylogf,tstr);
  tx = (float *)myalloc(cn*sizeof(float));
  ty = (float *)myalloc(cn*sizeof(float));
  tid = (int *)myalloc(cn*sizeof(int));
  k = 0;
  for(i=0;i<n;i++){
    if (cid[i] >= 0){
      tx[k] = cx[i];
      ty[k] = cy[i];
      tid[k] = cid[i];
      k += 1;
    }
  }
  myfree(cx); myfree(cy); myfree(cid);
  *rcn = cn;
  *rcx = tx;
  *rcy = ty;
  *rcid = tid;
}
void retutil_get_mosaic_regjit(xn,yn,n,dens,mupix,dnoise,seedx,seedc,prs,lmr,
			       regs,mylogf,
			       rcx,rcy,rcid,rcn,
			       redge0,redge1,redged,redgen,rmow,rmoh)
     int xn,yn;          
     int n;              
     float dens;
     float mupix;
     float dnoise;
     int seedx,seedc;    
     float prs,lmr;      
     int regs;           
     char *mylogf;       
     float **rcx,**rcy;
     int **rcid,*rcn;
     int **redge0,**redge1;  
     double **redged;         
     int *redgen;            
     int *rmow;          
     int *rmoh;          
{     
  int i,j,k;
  int *tid,cn,nn,ne,*edge0,*edge1,edgen,mow,moh;
  float x,y,*tx,*ty,prm,rv,dx,dy;
  double *edgedist;
  char tstr[SLEN];
  struct h_cell_struct *clist,**nlist;
  retutil_get_cone_coords_regjit(n,dens,mupix,dnoise,seedx,&nn,&clist,
				 rmow,rmoh);  
  if (seedc > 0)
    seedc = -seedc;
  if (regs > 0)  
    prs = 0.0;
  cn = 0;  
  prm = (1.0 - prs) / (lmr + 1.0); 
  for(i=0;i<nn;i++){
    x = clist[i].x + (float)(xn-1)/2.0;
    y = clist[i].y + (float)(yn-1)/2.0;
    clist[i].x = x;
    clist[i].y = y;
    if ((x >= 0.0) && (x < (double)(xn-1)) &&
	(y >= 0.0) && (y < (double)(yn-1))){
      rv = myrand_util_ran2(&seedc);
      if (rv < prs)
	clist[i].cid = 0;
      else if (rv < (prs + prm))
	clist[i].cid = 1;
      else
	clist[i].cid = 2;
      clist[i].edge_flag = 1;
      clist[i].ui = cn;
      cn += 1;  
    }else{
      clist[i].cid = -1;  
      clist[i].edge_flag = 0;
      clist[i].ui = -1;   
    }
  }
  sprintf(tstr,"    %d of %d cones within grid\n",cn,nn);
  mylog(mylogf,tstr);
  if (regs > 0)
    retutil_mosaic_reg_s(regs,clist,nn,*rmow,*rmoh);
  edge0 = (int *)myalloc(cn*6*sizeof(int));
  edge1 = (int *)myalloc(cn*6*sizeof(int));
  edgedist = (double *)myalloc(cn*6*sizeof(double));
  k = 0;  
  for(i=0;i<nn;i++){  
    if (clist[i].cid != -1){   
      x = clist[i].x;
      y = clist[i].y;
      ne = clist[i].n;             
      nlist = clist[i].neighbor;   
      for(j=0;j<ne;j++){
	if (nlist[j]->edge_flag == 1){
	  dx = x - nlist[j]->x;
	  dy = y - nlist[j]->y;
	  edgedist[k] = sqrt(dx*dx + dy*dy);
	  edge0[k] = clist[i].ui;
	  edge1[k] = nlist[j]->ui;
	  k += 1;
	}
      }
      clist[i].edge_flag = 0;  
    }
  }
  edgen = k;
  sprintf(tstr,"    %d unique edges\n",edgen);
  mylog(mylogf,tstr);
  tx = (float *)myalloc(cn*sizeof(float));
  ty = (float *)myalloc(cn*sizeof(float));
  tid = (int *)myalloc(cn*sizeof(int));
  k = 0;
  for(i=0;i<nn;i++){
    if (clist[i].cid >= 0){
      tx[k]  = clist[i].x;
      ty[k]  = clist[i].y;
      tid[k] = clist[i].cid;
      if (k != clist[i].ui)
	exit_error("(retina_util)","THIS SHOULD NOT HAPPEN - ui bad");
      k += 1;
    }
  }
  *rcn = cn;
  *rcx = tx;
  *rcy = ty;
  *rcid = tid;
  *redge0 = edge0;   
  *redge1 = edge1;
  *redged = edgedist;
  *redgen = edgen;
}
void retutil_irr2sq_mask(x,y,xn,yn,sig,nsd,rn,rx,ry,rw)
     float  x,y;  
     int  xn,yn;  
     float  sig;  
     float  nsd;  
     int    *rn;  
     int   **rx;  
     int   **ry;  
     float **rw;  
{
  int i,j,k;
  int n,xi,yi,x0,x1,y0,y1,rad,*cx,*cy;
  float *w,fc,dsq;
  rad = my_rint(nsd * sig);
  if (rad < 1)
    rad = 1;
  xi = my_rint(x);
  yi = my_rint(y);
  x0 = xi - rad;
  x1 = xi + rad;
  y0 = yi - rad;
  y1 = yi + rad;
  if (x0 < 0)
    x0 = 0;
  if (x1 > (xn-1))
    x1 = xn-1;
  if (y0 < 0)
    y0 = 0;
  if (y1 > (yn-1))
    y1 = yn-1;
  n = (x1 - x0 + 1) * (y1 - y0 + 1);
  if (n <= 0)
    exit_error("RETUTIL_IRR2SQ_MASK","Invalid n");
  cx = (int *)myalloc(n*sizeof(int));
  cy = (int *)myalloc(n*sizeof(int));
  w  = (float *)myalloc(n*sizeof(float));
  fc = -0.5/(sig*sig);
  k = 0;
  for(i=x0;i<=x1;i++){
    for(j=y0;j<=y1;j++){
      dsq = ((float)i-x)*((float)i-x) + ((float)j-y)*((float)j-y);
      cx[k] = i;
      cy[k] = j;
      w[k] = exp(fc * dsq);  
      k += 1;
    }
  }
  if (k != n)
    exit_error("RETUTIL_IRR2SQ_MASK","Should not happen");
  *rx = cx;
  *ry = cy;
  *rw = w;
  *rn = n;
}
float ***retutil_hex_to_3d(hv,cn,tn,cx,cy,rxn,ryn)
     float **hv;  
     int cn;      
     int tn;      
     float *cx;   
     float *cy;   
     int *rxn;    
     int *ryn;    
{
  int i;
  int ti,nw,nh,nwc,nhc,xi,yi;
  float ***d,xmin,xmax,ymin,ymax,val;
  double dx,dy;
  printf("  RETUTIL_HEX_TO_3D");
  dx = fabs(cx[1] - cx[0]);  
  dy = fabs(cy[2] - cy[0]);  
  get_min_max_farray(cx,cn,&xmin,&xmax);
  get_min_max_farray(cy,cn,&ymin,&ymax);
  nw = my_rint(1.0 + (xmax - xmin) / dx);
  nh = my_rint(1.0 + (ymax - ymin) / dy);
  nwc = 2*nw;
  nhc = 2*nh;
  d = get_zero_3d_farray(nwc,nhc,tn);
  for(ti=0;ti<tn;ti++){
    for(i=0;i<cn;i++){
      val = hv[i][ti];
      xi = my_rint(2.0 * (cx[i] - xmin) / dx);
      yi = my_rint(2.0 * (cy[i] - ymin) / dy);
      d[xi  ][yi  ][ti] = val;
      d[xi  ][yi+1][ti] = val;
      d[xi+1][yi  ][ti] = val;
      d[xi+1][yi+1][ti] = val;
    }
  }
  *rxn = nwc;
  *ryn = nhc;
  return d;
}
float ***retutil_hex_to_3d_new(hv,cn,tn,cx,cy,xn,yn)
     float **hv;  
     int cn;      
     int tn;      
     float *cx;   
     float *cy;   
     int xn;      
     int yn;      
{
  int i,j,k;
  int ti,xi,yi,*wn,**wx,**wy;
  float **ww,***d,***dw,xmin,xmax,ymin,ymax,*td,*tdw,w,*tval;
  printf("  RETUTIL_HEX_TO_3D_NEW\n");
  get_min_max_farray(cx,cn,&xmin,&xmax);
  get_min_max_farray(cy,cn,&ymin,&ymax);
  printf("    Cone  x min,max  %f  %f\n",xmin,xmax);
  printf("    Cone  y min,max  %f  %f\n",ymin,ymax);
  printf("    xn,yn =   %d , %d\n",xn,yn);
  wn = (int *)myalloc(cn*sizeof(int));
  wx = (int **)myalloc(cn*sizeof(int *));
  wy = (int **)myalloc(cn*sizeof(int *));
  ww = (float **)myalloc(cn*sizeof(float *));
  for(i=0;i<cn;i++){
    retutil_irr2sq_mask(cx[i],cy[i],xn,yn,3.0,3.0,&(wn[i]),&(wx[i]),&(wy[i]),
			&(ww[i]));
  }
  printf("DONE getting weights  xn yn tn %d %d %d\n",xn,yn,tn);
  d  = get_zero_3d_farray(xn,yn,tn);
  dw = get_zero_3d_farray(xn,yn,tn);
  for(i=0;i<cn;i++){
    tval = hv[i];
    for(j=0;j<wn[i];j++){  
      xi = wx[i][j];
      yi = wy[i][j];
      w  = ww[i][j];
      td  =  d[xi][yi];  
      tdw = dw[xi][yi];
      for(ti=0;ti<tn;ti++){
	td[ti] += w * tval[ti];   
	tdw[ti] += w;
      }
    }
  }
  for(i=0;i<xn;i++){
    for(j=0;j<yn;j++){
      for(k=0;k<tn;k++){
	if (dw[i][j][k] > 0.0)
	  d[i][j][k] /= dw[i][j][k];
      }
    }
  }
  return d;
}
void retutil_val_dist_plot(hv,cn,cx,cy,cid,t0,outfile)
     float **hv;  
     int cn;      
     float *cx;   
     float *cy;   
     int cid;     
     int t0;      
     char *outfile;  
{
  int i;
  float *val,*dist,x,y,x1,y1;
  char pname[SLEN];
  printf("  RETUTIL_VAL_DIST_PLOT\n");
  val = (float *)myalloc(cn*sizeof(float));
  dist = (float *)myalloc(cn*sizeof(float));
  x = cx[cid];
  y = cy[cid];
  for(i=0;i<cn;i++){
    x1 = cx[i];
    y1 = cy[i];
    dist[i] = sqrt((x1-x)*(x1-x) + (y1-y)*(y1-y));
    val[i] = hv[i][t0];
  }
  sprintf(pname,"cone_%d_time_%d",cid,t0);
  append_farray_xy_plot(outfile,dist,val,cn,pname);
}
void retutil_read_mosaic(infile,rcx,rcy,rcid,rcn)
     char infile[];
     float **rcx,**rcy;
     int **rcid,*rcn;
{
  FILE *fin,*fopen();
  int i;
  int *cid,cn,ns;
  float *cx,*cy,x,y;
  char c;
  if ((fin = fopen(infile,"r"))==NULL){
    printf("  filename = %s\n",infile);
    exit_error("RETUTIL_READ_MOSAIC","Cannot open file");
  }
  cn = 0;
  while(fscanf(fin,"%f %f %c",&x,&y,&c)!=EOF)
    cn += 1;
  fclose(fin);
  cx = (float *)myalloc(cn*sizeof(float));
  cy = (float *)myalloc(cn*sizeof(float));
  cid = (int *)myalloc(cn*sizeof(int));
  fin = fopen(infile,"r");
  for(i=0;i<cn;i++){
    ns = fscanf(fin,"%f %f %c",&x,&y,&c);
    cx[i] = x;
    cy[i] = y;
    if (c=='S')
      cid[i] = 0;
    else if (c=='M')
      cid[i] = 1;
    else if (c=='L')
      cid[i] = 2;
    else
      exit_error("RETUTIL_READ_MOSAIC","Bad cone type");
  }
  fclose(fin);
  *rcx = cx; *rcy = cy; *rcid = cid; *rcn = cn;
}
int retutil_custom_mosaic_cid_file(infile,cid,cn)
     char infile[];
     int *cid;        
     int cn;
{
  FILE *fin,*fopen();
  int i;
  int cnt;
  char c;
  if ((fin = fopen(infile,"r"))==NULL){
    printf("  filename = %s\n",infile);
    exit_error("RETUTIL_CUSTOM_MOSAIC_CID_FILE","Cannot open file");
  }
  cnt = 0;
  while(fscanf(fin,"%d %c",&i,&c)!=EOF){
    if ((i < 0) || (i >= cn)){
      printf("  Cone index value:  %d\n",i);
      exit_error("RETUTIL_CUSTOM_MOSAIC_CID_FILE","Cannot open file");
    }
    if (c=='S')
      cid[i] = 0;
    else if (c=='M')
      cid[i] = 1;
    else if (c=='L')
      cid[i] = 2;
    else{
      printf("  ConeID = %d, expecting 0,1,2\n",cid[i]);
      exit_error("RETUTIL_CUSTOM_MOSAIC_CID_FILE","Bad cone type");
    }
    cnt += 1;
  }
  fclose(fin);
  return cnt;  
}
