/*C  (C) Copyright  Wyeth Bair 1993-2010                                     */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "my_util.h"
#include "nr_util.h"
#include "misc_util.h"
#include "carray_util.h"
#include "iarray_util.h"
#include "farray_util.h"
#include "plot_util.h"
#include "myrand_util.h"
#include "spike_util.h"
#include "spike_data.h"    
#define IM1 2147483563
#define IM2 2147483399
#define AM (1.0/IM1)
#define IMM1 (IM1-1)
#define IA1 40014
#define IA2 40692
#define IQ1 53668
#define IQ2 52774
#define IR1 12211
#define IR2 3791
#define NTAB 32
#define NDIV (1+IMM1/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)
/*C  (C) Copr. 1986-92 Numerical Recipes Software #.3.                       */
float spike_util_ran2(idum)
     int *idum;
{
  int j;
  int k;
  static int idum2=123456789;
  static int iy=0;
  static int iv[NTAB];
  float temp;
  if (*idum <= 0) {
    if (-(*idum) < 1) *idum=1;
    else *idum = -(*idum);
    idum2=(*idum);
    for (j=NTAB+7;j>=0;j--) {
      k=(*idum)/IQ1;
      *idum=IA1*(*idum-k*IQ1)-k*IR1;
      if (*idum < 0) *idum += IM1;
      if (j < NTAB) iv[j] = *idum;
    }
    iy=iv[0];
  }
  k=(*idum)/IQ1;
  *idum=IA1*(*idum-k*IQ1)-k*IR1;
  if (*idum < 0) *idum += IM1;
  k=idum2/IQ2;
  idum2=IA2*(idum2-k*IQ2)-k*IR2;
  if (idum2 < 0) idum2 += IM2;
  j=iy/NDIV;
  iy=iv[j]-idum2;
  iv[j] = *idum;
  if (iy < 1) iy += IMM1;
  if ((temp=AM*iy) > RNMX) return RNMX;
  else return temp;
}
#undef IM1
#undef IM2
#undef AM
#undef IMM1
#undef IA1
#undef IA2
#undef IQ1
#undef IQ2
#undef IR1
#undef IR2
#undef NTAB
#undef NDIV
#undef EPS
#undef RNMX
/*C  (C) Copr. 1986-92 Numerical Recipes Software #.3.                       */
float su_gasdev(idum)
     int *idum;
{
  float ran1();
  static int iset=0;
  static float gset;
  float fac,rsq,v1,v2;
  if ((iset == 0) || (*idum <= 0)){  
    do {
      v1=2.0*spike_util_ran2(idum)-1.0;
      v2=2.0*spike_util_ran2(idum)-1.0;
      rsq=v1*v1+v2*v2;
    } while (rsq >= 1.0 || rsq == 0.0);
    fac=sqrt(-2.0*log(rsq)/rsq);
    gset=v1*fac;
    iset=1;
    return v2*fac;
  }else{
    iset=0;
    return gset;
  }
}
int gaussian_int(mu,sigma,seed)
     float mu,sigma;
     int *seed;
{
  return(my_rint(su_gasdev(seed)*sigma + mu));
}
float su_expdev(lambda,idum)
     float lambda;
     int *idum;
{
  if (lambda == 0.0)
    return 0.0;
  else
    return -log(spike_util_ran2(idum))/lambda;
}
float su_gaussian_float(mu,sigma,idum)
     float mu,sigma;
     int *idum;
{
  return(su_gasdev(idum)*sigma + mu);
}
void free_sarray(s,cnt,n)
     int **s,*cnt,n;
{
  int i;
  for(i=0;i<n;i++)
    myfree(s[i]);
  myfree(s); myfree(cnt);
}
void write_spike_plot(outfile,s,n,start,period,sampling)
     char outfile[];
     int *s,n,start,period;
     float sampling;
{
  FILE *fopen(),*fout;
  int i;
  fout = fopen(outfile,"w");
  fprintf(fout,"%d 0\n",(int)((float)start*1000.0/sampling));
  for (i=0;i<n;i++)
    if ((s[i] >= start) && (s[i] < start+period)){
      fprintf(fout,"%d 0\n",(int)((float)s[i]*1000.0/sampling));
      fprintf(fout,"%d 1\n",(int)((float)s[i]*1000.0/sampling));
      fprintf(fout,"%d 0\n",(int)((float)s[i]*1000.0/sampling));
    }
  fprintf(fout,"%d 0\n",(int)((float)(start+period-1)*1000.0/sampling));
  fclose(fout);
}
void write_float_spike_plot(outfile,s,n,start,period,sampling)
     char outfile[];
     float *s;
     int n,start,period;
     float sampling;
{
  FILE *fopen(),*fout;
  int i;
  fout = fopen(outfile,"w");
  fprintf(fout,"%d 0\n",(int)((float)start*1000.0/sampling));
  for (i=0;i<n;i++)
    if ((s[i] >= (float)start) && (s[i] < (float)(start+period))){
      fprintf(fout,"%.3e 0\n",s[i]*1000.0/sampling);
      fprintf(fout,"%.3e 1\n",s[i]*1000.0/sampling);
      fprintf(fout,"%.3e 0\n",s[i]*1000.0/sampling);
    }
  fprintf(fout,"%d 0\n",(int)((float)(start+period-1)*1000.0/sampling));
  fclose(fout);
}
int count_spikes(data,n,start,period)
     int *data,n;
     int start,period;
{
  int i;
  int scount;
  scount = 0;
  for (i=0;i<n;i++)
    if ((data[i] >= start) && (data[i] < start+period))
      scount += 1;
  return scount;
}
int count_spikes_float(data,n,start,period)
     float *data;
     int n;
     float start,period;
{
  int i;
  int scount;
  scount = 0;
  for (i=0;i<n;i++)
    if ((data[i] >= start) && (data[i] < start+period))
      scount += 1;
  return scount;
}
int count_spikes_dbl(data,n,start,period)
     double *data;
     int n;
     double start,period;
{
  int i;
  int scount;
  scount = 0;
  for (i=0;i<n;i++)
    if ((data[i] >= start) && (data[i] < start+period))
      scount += 1;
  return scount;
}
float mean_spike_rate(data,n,start,period,sampling)
     int *data,n;
     int start,period;
     float sampling;
{
  float sr;
  sr = (float)count_spikes(data,n,start,period);
  sr /= (float)period/sampling;
  return sr;
}
float mean_spike_rate_float(data,n,start,period,sampling)
     float *data;
     int n;
     float start,period;
     float sampling;
{
  float sr;
  sr = (float)count_spikes_float(data,n,start,period);
  sr /= period/sampling;
  return sr;
}
int total_spikes_sarray(s,cnt,n,start,period)
     int **s,*cnt,n,start,period;
{
  int i;
  int total;
  total = 0;
  for (i=0;i<n;i++)
    total += count_spikes(s[i],cnt[i],start,period);
  return total;
}
int get_max_time_sarray(s,cnt,n)
     int **s,*cnt,n;
{
  int i;
  int max,tmax;
  max = max_of_iarray(s[0],cnt[0]);
  for(i=1;i<n;i++){
    tmax = max_of_iarray(s[i],cnt[i]);
    if (tmax > max)
      max = tmax;
  }
  return max;
}
int *convert_sampling_units_trial(s,n,f)
     int *s,n;
     float f;
{
  int i;
  int *r;
  float sf;
  r = (int *)myalloc(n*sizeof(int));
  for(i=0;i<n;i++){
    sf = (float)s[i]*f;
    if (sf >= 0.0)
      r[i] = (int)(0.5 + sf);
    else
      r[i] = (int)(-0.5 + sf);
  }
  return r;
}
int **get_sarray_convert_sampling(s,cnt,n,f)
     int **s,*cnt,n;
     float f;
{
  int i;
  int **r;
  r = (int **)myalloc(n*sizeof(int *));
  for(i=0;i<n;i++)
    r[i] = convert_sampling_units_trial(s[i],cnt[i],f);
  return r;
}
void sarray_convert_sampling(s,cnt,n,f)
     int **s,*cnt,n;
     float f;
{
  int i,j;
  float t;
  for(i=0;i<n;i++)
    for(j=0;j<cnt[i];j++){
      t = (float)s[i][j];
      if (t >= 0.0)
	s[i][j] = (int)(0.5 + t*f);
      else
	s[i][j] = (int)(-0.5 + t*f);
    }
}
int get_first_spike(data,n,t)
     int *data,n,t;
{
  int i;
  int first;
  first = t-1;
  i = 0;
  while ((first < t)&&(i<n)){
    if (data[i] >= t)
      first = data[i];
    else
      i += 1;
  }
  return first;
}
int get_first_spike_in_window(data,n,t,tn)
     int *data,n,t,tn;
{
  int i,k;
  int first;
  first = t-1;
  i = 0;
  while ((first < t)&&(i<n)){
    k = data[i] - t;
    if ((k >= 0)&&(k < tn))
      first = data[i];
    else
      i += 1;
  }
  return first;
}
int get_last_spike(data,n,t)
     int *data,n,t;
{
  int i;
  int last;
  last = t+1;
  i = n-1;
  while ((last > t)&&(i>=0)){
    if (data[i] <= t)
      last = data[i];
    else
      i -= 1;
  }
  return last;
}
int *extract_spikes(spike_data,n,start,period,count,zero_flag)
     int *spike_data,n,start,period;
     int *count;
     int zero_flag;  
{
  int i,k,t;
  int *data;
  *count = count_spikes(spike_data,n,start,period);
  data = (int *)myalloc(*count*sizeof(int));
  k = 0;
  for (i=0;i<n;i++){
    t = spike_data[i] - start;
    if ((t >= 0) && (t < period)){
      if (zero_flag)
	data[k] = t;
      else
	data[k] = spike_data[i];
      k += 1;
    }
  }
  return data;
}
int *extract_int_spikes_from_float(s,n,start,period,count,zero_flag,tscale)
     float *s;
     int n;
     float start,period;
     int *count;
     int zero_flag;  
     float tscale;   
{
  int i,k;
  int *data;
  *count = count_spikes_float(s,n,start,period);
  data = (int *)myalloc(*count*sizeof(int));
  k = 0;
  for (i=0;i<n;i++){
    if ((s[i] >= start) && (s[i] < start+period)){
      if (zero_flag)
	data[k] = my_rint(tscale*(s[i] - start));
      else
	data[k] = my_rint(tscale*s[i]);
      k += 1;
    }
  }
  return data;
}
int *extract_int_spikes_from_dbl(s,n,start,period,count,zero_flag,tscale)
     double *s;
     int n;
     double start,period;
     int *count;
     int zero_flag;  
     double tscale;   
{
  int i,k;
  int *data;
  *count = count_spikes_dbl(s,n,start,period);
  data = (int *)myalloc(*count*sizeof(int));
  k = 0;
  for (i=0;i<n;i++){
    if ((s[i] >= start) && (s[i] < start+period)){
      if (zero_flag)
	data[k] = my_rint(tscale*(s[i] - start));
      else
	data[k] = my_rint(tscale*s[i]);
      k += 1;
    }
  }
  return data;
}
void add_const_sarray(s,cnt,n,c)
     int **s,*cnt,n,c;
{
  int i,j;
  for(i=0;i<n;i++)
    for(j=0;j<cnt[i];j++)
      s[i][j] += c;
}
int count_unique_spikes(data,n,start,duration)
     int *data,n;
     int  start,duration;
{
  int i;
  int scount;
  int t,t_old;
  scount = 0;
  t_old = -1;
  for (i=0;i<n;i++){
    t = data[i] - start;
    if ((t >= 0)&&(t < duration))
      if (t != t_old)
	scount += 1;
    t_old = t;
  }
  return scount;
}
int *extract_unique_spikes(data,n,start,duration,n_unique,zero_flag)
     int *data,n;
     int start,duration;
     int *n_unique;
     int zero_flag;  
{
  int i;
  int t,t_old,pos,*udata;
  *n_unique = count_unique_spikes(data,n,start,duration);
  udata = (int *)myalloc(*n_unique*sizeof(int));
  pos = 0;
  t_old = -1;
  for(i=0;i<n;i++){
    t = data[i] - start;
    if ((t >= 0)&&(t < duration))
      if (t != t_old){
	if (zero_flag)
	  udata[pos] = t;
	else
	  udata[pos] = data[i];
	pos += 1;
      }
    t_old = t;
  }
  return udata;
}	
void get_unique_sarray(s,cnt,n,period,rs,rcnt)
     int **s,*cnt,n,period,***rs,**rcnt;
{
  int i;
  int **ts,*tcnt;
  ts = (int **)myalloc(n*sizeof(int *));
  tcnt = (int *)myalloc(n*sizeof(int));
  for(i=0;i<n;i++)
    ts[i] = extract_unique_spikes(s[i],cnt[i],0,period,&tcnt[i],0);
  *rs = ts; *rcnt = tcnt;
}	
void spikeu_expand_spikes_wsd_accum(s,n,xacc,tn,w,sdf,sdtau)
     float *s;      
     int n;         
     float *xacc;   
     int tn;        
     float w;       
     float sdf;     
     float sdtau;   
{
  int i;
  int t,tbad,tmax;
  float a,anew,dt;
  tbad = tmax = 0;
  a = 1.0;      
  dt = 10000.0; 
  for(i=0;i<n;i++){
    t = my_rint(s[i]);
    if ((t >= 0)&&(t < tn)){
      if (i > 0)
	dt = s[i] - s[i-1];
      anew = 1.0 - (1.0 - a*sdf) * exp(-dt/sdtau);
      a = anew;
      xacc[t] += anew * w;        
    }else{
      if (t == tn)
	tmax += 1;
      else{
	tbad += 1;
      }
    }
  }
}
float *expand_spike_array(spike_data,n,start,period)
     int *spike_data,n;
     int start,period;
{
  int i,t;
  float *data;
  data = get_zero_farray(period);
  for (i=0;i<n;i++){
    t = spike_data[i] - start;
    if ((t >= 0) && (t < period))
      data[t] += 1.0;
  }
  return data;
}
int *expand_spike_int_array(spike_data,n,start,period)
     int *spike_data,n;
     int start,period;
{
  int i,t;
  int *data;
  data = get_zero_iarray(period);
  for (i=0;i<n;i++){
    t = spike_data[i] - start;
    if ((t >= 0) && (t < period))
      data[t] += 1;
  }
  return data;
}
float *expand_spike_farray(spike_data,n,start,period)
     float *spike_data;
     int n,start,period;
{
  int     i,t;
  float  *data;
  data = get_zero_farray(period);
  for (i=0;i<n;i++){
    t = my_rint(spike_data[i]) - start;
    if ((t >= 0) && (t < period))
      data[t] += 1.0;
  }
  return data;
}
void get_times_from_expanded_int_spikes(data,n,start,rs,rcnt)
     int *data,n,start,**rs,*rcnt;
{
  int i,j,k;
  int *t;
  if (min_of_iarray(data,n) < 0)
    exit_error("GET_TIMES_FROM_EXPANDED_INT_SPIKES","NEGATIVE VALUE FOUND");
  *rcnt = sum_iarray(data,n,0,n);
  t = (int *)myalloc(*rcnt*sizeof(int));
  k = 0;
  for (i=0;i<n;i++)
    for(j=0;j<data[i];j++){
      t[k] = i + start;
      k += 1;
    }
  *rs = t;
}
void get_times_from_expanded_float_spikes(data,n,start,s,cnt)
     float *data;
     int n,start,**s,*cnt;
{
  int i,j,k;
  int *t;
  if (min_of_farray(data,n) < 0)
    exit_error("GET_TIMES_FROM_EXPANDED_INT_SPIKES","NEGATIVE VALUE FOUND");
  *cnt = sum_farray(data,n,0,n);
  t = (int *)myalloc(*cnt*sizeof(int));
  k = 0;
  for (i=0;i<n;i++)
    for(j=0;j<(int)(data[i]);j++){
      t[k] = i + start;
      k += 1;
    }
  *s = t;
}
int *histogram_expand_sarray(s,n,start,period,binsize,binflag,shiftflag,
			     rn,rtrunc,rshift)
     int *s,n;
     int start,period;
     int binsize;       
     int binflag;       
     int shiftflag;     
     int *rn,*rtrunc,*rshift;
{
  int i,k;
  int hn,t,ntrunc,nshift,*x;
  hn = period / binsize; 
  x = get_zero_iarray(hn);
  for (i=0;i<n;i++){
    t = s[i] - start;
    if ((t >= 0) && (t < period)){
      k = (int)(t/binsize);
      if ((k >= 0)&&(k < hn)){
	x[k] += 1;
      }
    }
  }
  ntrunc = 0;
  nshift = 0;
  if (binflag == 1){
    if (shiftflag == 0){       
      for(i=0;i<hn;i++)
	if (x[i] > 1){
	  x[i] = 1;
	  ntrunc += 1;
	}
    }else{
      for(i=0;i<hn;i++)
	if (x[i] > 1){
	  if (i > 0)           
	    if (x[i-1] == 0){
	      x[i-1] += 1;
	      x[i] -= 1;
	      nshift += 1;
	    }
	  if (x[i] > 1)        
	    if (i < (hn-1))         
	      if (x[i+1] == 0){
		x[i+1] += 1;
		x[i] -= 1;
		nshift += 1;
	      }
	  if (x[i] > 1){        
	    x[i] = 1;
	    ntrunc += 1;
	  }
	}
    }
    for(i=0;i<hn;i++)
      if (x[i] > 1){
	exit_error("HISTOGRAM_EXPAND_SARRAY","Truncation failed");
      }
  }
  *rn = hn; *rtrunc = ntrunc; *rshift = nshift;
  return x;
}
float *merge_spike_arrays_float(s1,n1,s2,n2)
     float *s1;
     int n1;
     float *s2;
     int n2;
{
  int i;
  int n;
  float *s;
  n = n1 + n2;
  s = (float *)myalloc(n*sizeof(float));
  for(i=0;i<n1;i++)
    s[i] = s1[i];
  for(i=0;i<n2;i++)
    s[n1+i] = s2[i];
  if (n > 1)
    sort_farray(s,n);
  return s;
}
int *merge_spike_arrays(s1,n1,s2,n2,start,period,rn)
     int *s1,n1,*s2,n2;
     int start,period;
     int *rn;
{
  int i,j;
  int t,pos,n;
  int *merged,*expanded;
  expanded = get_zero_iarray(period);
  n = 0;
  for (i=0;i<n1;i++){ 
    t = s1[i]-start;
    if ((t>=0) && (t<period)){
      expanded[t] += 1;
      n += 1;
    }
  }
  for (i=0;i<n2;i++){
    t = s2[i]-start;
    if ((t>=0) && (t<period)){
      expanded[t] += 1;
      n += 1;
    }
  }
  merged = get_zero_iarray(n); 
  pos = 0;
  for (i=0;i<period;i++)
    for (j=0;j<expanded[i];j++){ 
      merged[pos] = i+start;
      pos += 1;
    }
  if (pos != n)
    exit_error("MERGE_SPIKE_ARRAY","This never happens");
  myfree(expanded);
  *rn = n;
  return merged;
}
int *merge_sarray(s,cnt,n,start,period,rn)
     int **s,*cnt,n;
     int start,period;
     int *rn;
{
  int i,j;
  int t,pos,count;
  int *merged,*expanded;
  expanded = get_zero_iarray(period);
  count = 0;
  for(i=0;i<n;i++)
    for (j=0;j<cnt[i];j++){ 
      t = s[i][j]-start;
      if ((t>=0) && (t<period)){
	expanded[t] += 1;
	count += 1;
      }
    }
  merged = get_zero_iarray(count); 
  pos = 0;
  for (i=0;i<period;i++)
    for (j=0;j<expanded[i];j++){ 
      merged[pos] = i+start;
      pos += 1;
    }
  if (pos != count)
    exit_error("MERGE_SPIKE_ARRAY","This never happens");
  myfree(expanded);
  *rn = count;
  return merged;
}
int *float_to_int_spike(fdata,fn,n,start,duration,c)
     float *fdata;
     int fn,*n;
     int start,duration; 
     float c; 
{
  int i;
  int *s,t;
  s = (int *)myalloc(fn*sizeof(int));
  *n = 0;
  for (i=0;i<fn;i++){
    t = (int)(0.5 + c*fdata[i]);
    if ((t>=start)&&(t<(start+duration))){
      s[*n] = t;
      *n += 1;
    }
  }
  return s;
}
void print_sarray(s,cnt,n,t,tn)
     int **s,*cnt,n,t,tn;
{
  int i,j;
  int *ss,sn;
  printf("\n%d\n",n);
  for(i=0;i<n;i++){
    ss = extract_spikes(s[i],cnt[i],t,tn,&sn,0);
    printf("%d\n",sn);
    for(j=0;j<sn;j++)
      printf(" %d",ss[j]);
    printf("\n");
    myfree(ss);
  }
}
char *compare_trains(s1,n1,s2,n2)
     int *s1,n1;   
     int *s2,n2;   
{
  int i,j;
  char *t;
  t = NULL;
  return t;
}
void print_sarray_append(outfile,s,cnt,n,t,tn)
     char outfile[];
     int **s,*cnt,n,t,tn;
{
  int i,j;
  int *ss,sn;
  FILE *fout,*fopen();
  if ((fout = fopen(outfile,"a"))==NULL){
    exit_error("PRINT_SARRAY_APPEND","Cannot open file");
  }
  fprintf(fout,"\n%d\n",n);
  for(i=0;i<n;i++){
    ss = extract_spikes(s[i],cnt[i],t,tn,&sn,0);
    fprintf(fout,"%d\n",sn);
    for(j=0;j<sn;j++)
      fprintf(fout," %d",ss[j]);
    fprintf(fout,"\n");
    myfree(ss);
  }
  fclose(fout);
}
void align_to_first_spike_in_window_sarray(s,cnt,n,t,tn,tnew)
     int **s,*cnt,n,t,tn,tnew;
{
  int i,k;
  for(i=0;i<n;i++){
    k = get_first_spike_in_window(s[i],cnt[i],t,tn);
    if (k > t)
      add_const_iarray(s[i],cnt[i],tnew-k);
  }
}
void spike_count_sarrays(s1,cnt1,s2,cnt2,n,start,period,rsc1,rsc2)
     int **s1,*cnt1,**s2,*cnt2,n;
     int start,period;
     float **rsc1,**rsc2;
{
  int i;
  float *c1,*c2,z,mean,sd1,sd2;
  c1 = (float *)myalloc(n*sizeof(float));
  c2 = (float *)myalloc(n*sizeof(float));
  for (i=0;i<n;i++){
    c1[i] = (float)count_spikes(s1[i],cnt1[i],start,period);
    c2[i] = (float)count_spikes(s2[i],cnt2[i],start,period);
  }
  *rsc1 = c1;
  *rsc2 = c2;
}
void spike_count_r_sarrays(s1,cnt1,s2,cnt2,n,rr,prob,start,period,sflag)
     int **s1,*cnt1,**s2,*cnt2,n;
     float *rr,*prob;
     int start,period,sflag;
{
  int i;
  float *c1,*c2,z,mean,sd1,sd2;
  c1 = (float *)myalloc(n*sizeof(float));
  c2 = (float *)myalloc(n*sizeof(float));
  for (i=0;i<n;i++){
    c1[i] = (float)count_spikes(s1[i],cnt1[i],start,period);
    c2[i] = (float)count_spikes(s2[i],cnt2[i],start,period);
  }
  mean_sdev_farray(c1,n,&mean,&sd1);
  mean_sdev_farray(c2,n,&mean,&sd2);
  if ((sd1 == 0.0)||(sd2 == 0.0)||(n < 3))
    *rr = *prob = -2.0;
  else
    if (sflag){
      simple_pearsn(c1-1,c2-1,(long)n,rr);
      *prob = 1.0;
    }else
      pearsn(c1-1,c2-1,(long)n,rr,prob,&z);
  myfree(c1); myfree(c2);
}
void spike_trig_stim_sarray(s,cnt,n,data,doffset,dperiod,sstart,speriod,
			    t0,tn,rtdata,rnspikes)
     int **s,*cnt,n;
     float **data;
     int doffset,dperiod,sstart,speriod,t0,tn;
     float ***rtdata;
     int *rnspikes;
{
  int i,j;
  int t,tspikes,nspikes;
  float **tdata,**ttdata;
  tspikes = 0;
  for(i=0;i<n;i++)
    tspikes += cnt[i];
  tdata = (float **)myalloc(tspikes*sizeof(float *));
  nspikes = 0;
  for(i=0;i<n;i++)
    for(j=0;j<cnt[i];j++){
      t = s[i][j];
      if ((t >= sstart)&&(t < (sstart+speriod))){
	t += t0 - doffset; 
	if ((t>=0) && ((t+tn)<=dperiod)){
	  tdata[nspikes] = copy_farray(data[i]+t,tn);
	  nspikes += 1;
	}
      }
    }
  if (nspikes > 0){
    ttdata = (float **)myalloc(nspikes*sizeof(float *));
    for(i=0;i<nspikes;i++)
      ttdata[i] = tdata[i];
  }else{
    ttdata = NULL;
  }
  myfree(tdata); 
  *rtdata = ttdata;
  *rnspikes = nspikes;
}
void spike_trig_avg_sarray(s,cnt,n,data,doffset,dperiod,sstart,speriod,
			   avg_start,avg_period,rmean,rsdev,rnspikes)
     int **s,*cnt,n;
     float **data;
     int doffset,dperiod,sstart,speriod,avg_start,avg_period;
     float **rmean,**rsdev;
     int *rnspikes;
{
  int i,j;
  int t,tspikes,nspikes;
  float *mean,*sdev,**tdata;
  tspikes = 0;
  for(i=0;i<n;i++)
    tspikes += cnt[i];
  tdata = (float **)myalloc(tspikes*sizeof(float *));
  nspikes = 0;
  for(i=0;i<n;i++)
    for(j=0;j<cnt[i];j++){
      t = s[i][j];
      if ((t >= sstart)&&(t < (sstart+speriod))){
	t += avg_start - doffset; 
	if ((t>=0) && ((t+avg_period)<=dperiod)){
	  tdata[nspikes] = data[i] + t;
	  nspikes += 1;
	}
      }
    }
  if (nspikes > 0){
    mean_sdev_2d_farray(tdata,nspikes,avg_period,&mean,&sdev);
  }else{
    mean = get_zero_farray(avg_period);
    sdev = get_zero_farray(avg_period);
  }
  myfree(tdata); 
  *rmean = mean; *rsdev = sdev; *rnspikes = nspikes;
}
int varcond_check_condition(cond,t0,data,n,outfile)
     struct spike_util_cond_struct *cond;    
     int t0;          
     float *data;
     int n;
     char *outfile;
{
  int k;
  int flag,minmaxflag,tn;
  float x,mu,sd,vcrit,min,max;
  char ct1[SLEN],*ctype;
  tn = cond->winn;
  if ((t0 < 0) || (t0+tn > n)){
    printf(" t0 = %d   cond->winn = %d\n",t0,tn);
    exit_error("VARCOND_CHECK_CONDITION","Condition window error");
  }
  ctype = cond->ctype;
  k = strlen(ctype);
  if ((ctype[k-3]=='m')&&(ctype[k-2]=='i')&&(ctype[k-1]=='n'))
    minmaxflag = -1;
  else if ((ctype[k-3]=='m')&&(ctype[k-2]=='a')&&(ctype[k-1]=='x'))
    minmaxflag = 1;
  else
    minmaxflag = 0;
  flag = 0;
  if (minmaxflag != 0){
    strcpy(ct1,ctype);
    ct1[k-4] = '\0';
    if (strcmp(ct1,"abs_val_avg")==0){
      x = (float)sum_abs_farray(data+t0,tn)/(float)tn;
    }else if (strcmp(ct1,"sqr_avg")==0){
      x = (float)sum_square_farray(data+t0,tn)/(float)tn;
    }else if (strcmp(ct1,"var")==0){
      mean_sdev_farray(data+t0,tn,&mu,&sd);
      x = sd*sd;
    }else if (strcmp(ct1,"sd")==0){
      mean_sdev_farray(data+t0,tn,&mu,&sd);
      x = sd;
    }else if (strcmp(ct1,"avg")==0){
      mean_sdev_farray(data+t0,tn,&mu,&sd);
      x = mu;
    }else if (strcmp(ct1,"val")==0){
      if (minmaxflag == -1)
	x = min_of_farray(data+t0,tn);
      else
	x = max_of_farray(data+t0,tn);
    }else{
      printf("ctype = %s\n",ctype);
      exit_error("VARCOND_CHECK_CONDITION","Unknown condition type");
      x = 0.0;
    }
    if (((minmaxflag == -1) && (x >= cond->vcrit)) ||
	((minmaxflag ==  1) && (x <= cond->vcrit))){
      flag = 1;
    }
  }else if ((strcmp(ctype,"range")==0)||(strcmp(ctype,"range_x")==0)){
    min = min_of_farray(data+t0,tn);
    max = max_of_farray(data+t0,tn);
    if (strcmp(ctype,"range")==0){
      if ((min >= cond->vcrit) && (max <= cond->vcrit1)){
	flag = 1;
	if (outfile != NULL){
	  char ttstr[SLEN],tname[SLEN];
	  float tmean,tsd;
	  mean_sdev_farray(data+t0,tn,&tmean,&tsd);
	  sprintf(ttstr,"%f %f\n",tmean,tsd);
	  sprintf(tname,"%s.dump",outfile);
	  append_string_to_file(tname,ttstr);
	}
      }
    }else{ 
      if ((min < cond->vcrit) && (max > cond->vcrit1)){
	flag = 1;
      }
    }
  }else{
    printf("ctype = %s\n",ctype);
    exit_error("VARCOND_CHECK_CONDITION","Unknown condition type");
  }
  return flag;
}
void spike_trig_avg_sarray_varcond(s,cnt,n,data,doffset,dperiod,sstart,speriod,
				   avgt0,avgtn,ctype,c0,cn,vcrit,
				   rmean,rsdev,rnspikes)
     int **s,*cnt,n;
     float **data;
     int doffset,dperiod,sstart,speriod,avgt0,avgtn;
     char ctype[];
     int c0,cn;
     float vcrit,**rmean,**rsdev;
     int *rnspikes;
{
  int i,j;
  int t,tspikes,nspikes,nreject,x0;
  float *mean,*sdev,**tdata,r;
  struct spike_util_cond_struct cond;    
  cond.ctype = strdup(ctype); 
  cond.win0 = c0;
  cond.winn = cn;
  cond.vcrit = vcrit;
  printf("  SPIKE_TRIG_AVG_SARRAY_VARCOND\n");
  printf("    %s c0=%d cn=%d vcrit=%f\n",ctype,c0,cn,vcrit);
  printf("  ************* WYETH CHECK THIS CODE since struct added \n");
  printf("  ************* WYETH CHECK THIS CODE since struct added \n");
  printf("  ************* WYETH CHECK THIS CODE since struct added \n");
  tspikes = 0;
  for(i=0;i<n;i++)
    tspikes += cnt[i];
  tdata = (float **)myalloc(tspikes*sizeof(float *));
  x0 = c0 - avgt0;    
  nspikes = nreject = 0;
  for(i=0;i<n;i++) 
    for(j=0;j<cnt[i];j++){ 
      t = s[i][j];
      if ((t >= sstart)&&(t < (sstart+speriod))){
	t += avgt0 - doffset; 
	if ((t>=0) && ((t+avgtn)<=dperiod)){
	  tdata[nspikes] = data[i] + t;
	  if (varcond_check_condition(&cond,x0,tdata[nspikes],avgtn,NULL))
	    nspikes += 1;
	  else
	    nreject += 1;
	}
      }
    }
  if (nspikes > 0)
    mean_sdev_2d_farray(tdata,nspikes,avgtn,&mean,&sdev);
  else{
    mean = get_zero_farray(avgtn);
    sdev = get_zero_farray(avgtn);
  }
  r = 100.0 * (float)nspikes/(float)(nspikes+nreject);
  printf("    %.2f %% spikes accepted (%d of %d)\n",r,nspikes,nreject+nspikes);
  myfree(cond.ctype);
  myfree(tdata); 
  *rmean = mean; *rsdev = sdev; *rnspikes = nspikes;
}
void spike_trig_avg_condition_sarray(s,cnt,n,data,doffset,dperiod,sstart,
				     speriod,avgt0,avgtn,min,max,
				     rmean,rsdev,rnspikes)
     int **s,*cnt,n;
     float **data;
     int doffset,dperiod,sstart,speriod,avgt0,avgtn;
     float min,max,**rmean,**rsdev;
     int *rnspikes;
{
  int i,j;
  int t,tspikes,nspikes;
  float *mean,*sdev,**tdata,*tf,tmin,tmax;
  tspikes = 0;
  for(i=0;i<n;i++)
    tspikes += cnt[i];
  tdata = (float **)myalloc(tspikes*sizeof(float *));
  nspikes = 0;
  for(i=0;i<n;i++)
    for(j=0;j<cnt[i];j++){
      t = s[i][j];
      if ((t >= sstart)&&(t < (sstart+speriod))){
	t += avgt0 - doffset; 
	if ((t>=0) && ((t+avgtn)<=dperiod)){
	  tf = data[i] + t;
	  get_min_max_farray(tf,avgtn,&tmin,&tmax);
	  if ((tmin >= min)&&(tmax <= max)){
	    tdata[nspikes] = tf;
	    nspikes += 1;
	  }
	}
      }
    }
  mean_sdev_2d_farray(tdata,nspikes,avgtn,&mean,&sdev);
  myfree(tdata); 
  *rmean = mean; *rsdev = sdev; *rnspikes = nspikes;
}
void spike_trig_distrib_sarray(s,cnt,n,data,doffset,dperiod,sstart,speriod,
			       avg_start,avg_period,vmin,vmax,g80_flag,
			       rdistrib)
     int **s,*cnt,n;
     int **data;
     int doffset,dperiod,sstart,speriod,avg_start,avg_period,vmin,vmax;
     int g80_flag;
     float ***rdistrib;
{
  int i,j,k;
  int t,dk,nspikes,nv;
  float **distrib;
  printf("  SPIKE_TRIG_DISTRIB_SARRAY\n");
  nv = vmax - vmin + 1;
  distrib = get_zero_2d_farray(avg_period,nv);
  nspikes = 0;
  if (g80_flag == 0)
    for(i=0;i<n;i++)
      for(j=0;j<cnt[i];j++){
	t = s[i][j];
	if ((t >= sstart)&&(t < (sstart+speriod))){
	  t = s[i][j] - doffset + avg_start; 
	  if ((t>=0) && ((t+avg_period)<=dperiod)){
	    nspikes += 1;
	    for(k=t;k<(t+avg_period);k++){
	      dk = data[i][k] - vmin;
	      if ((dk >= 0)&&(dk < nv)){
		distrib[k-t][dk] += 1.0;
	      }else
		exit_error("SPIKE_TRIG_DISTRIB_SARRAY",
			   "Stimulus value out of bounds");
	    }
	  }
	}
      }
  else
    for(i=0;i<n;i++)
      for(j=0;j<cnt[i];j++){
	t = s[i][j];
	if ((t >= sstart)&&(t < (sstart+speriod))){
	  t = s[i][j] - doffset + avg_start; 
	  if ((t>=0) && ((t+avg_period)<=dperiod) &&
	      (((t>=1248)&&((t+avg_period)<13600))||(t>=14847))){
	    if ((t>=1248)&&((t+avg_period)<13600))
	      t -= 12;  
	    else
	      t -= 25;  
	    nspikes += 1;
	    for(k=t;k<(t+avg_period);k++){
	      dk = data[i][k] - vmin;
	      if ((dk >= 0)&&(dk < nv)){
		distrib[k-t][dk] += 1.0;
	      }else
		exit_error("SPIKE_TRIG_DISTRIB_SARRAY",
			   "Stimulus value out of bounds");
	    }
	  }
	}
      }
  printf("    %d spikes used as triggers\n",nspikes);
  multiply_2d_farray(distrib,avg_period,nv,1.0/(float)nspikes);
  for(i=0;i<nv;i++) 
    distrib[0][i] = 0.0;
  nspikes = 0;
  for(i=0;i<n;i++)
    for(j=0;j<dperiod;j++){
      dk = data[i][j] - vmin;
      if ((dk >= 0)&&(dk < nv)){
	distrib[0][dk] += 1.0;
	nspikes += 1;
      }else
	exit_error("SPIKE_TRIG_DISTRIB_SARRAY",
		   "Stimulus value out of bounds");
    }
  multiply_farray(distrib[0],nv,1.0/(float)nspikes);
  *rdistrib = distrib;
}
void spike_trig_2d_avg_trial(st,n,data,doffset,dperiod,dn,sstart,speriod,
			     avg_start,avg_period,rmean,rnspikes)
     int *st,n;
     float **data;
     int doffset,dperiod,dn,sstart,speriod,avg_start,avg_period;
     float ***rmean;
     int *rnspikes;
{
  int j;
  int t,nspikes;
  float **mean,***tdata;
  tdata = (float ***)myalloc(n*sizeof(float **));
  nspikes = 0;
  for(j=0;j<n;j++){
    t = st[j];
    if ((t >= sstart)&&(t < (sstart+speriod))){
      t += avg_start - doffset; 
      if ((t>=0) && ((t+avg_period)<=dperiod)){
        tdata[nspikes] = data + t; 
	nspikes += 1;
      }
    }
  }
  if (nspikes > 0)
    mean_3d_farray(tdata,nspikes,avg_period,dn,&mean);
  else
    mean = get_zero_2d_farray(avg_period,dn);
  myfree(tdata); 
  *rmean = mean; *rnspikes = nspikes;
}
void second_order_wiener_kernel_sarray(s,cnt,n,data,doffset,dperiod,sstart,
				       speriod,k0,kn,rw2)
     int **s,*cnt,n;
     float **data;
     int doffset,dperiod,sstart,speriod,k0,kn;
     float ***rw2;
{
  int i,j,k,l;
  int t1,t2,count;
  float **w2;
  w2 = get_zero_2d_farray(kn,kn);
  for(i=0;i<kn;i++){ 
    printf("  tau_1 = %d\n",i);
    for(j=i;j<kn;j++){ 
      count = 0;
      for(k=0;k<n;k++) 
	for(l=0;l<cnt[k];l++){ 
	  t1 = s[k][l] + k0 + i - doffset;
	  t2 = s[k][l] + k0 + j - doffset;
	  if ((t1>0)&&(t1<dperiod)&&(t2>0)&&(t2<dperiod)){
	    w2[i][j] += data[k][t1]*data[k][t2];
	    count += 1;
	  }
	}
      w2[i][j] /= (float)count;
    }
  }
  *rw2 = w2;
}
void get_pattern_trig_sarray_from_sarray(s,cnt,n,stim,ns,start,expf,pat,npat,
					 win0,winn,rs,rcnt,rn)
     int **s,*cnt,n;
     float **stim;
     int *ns;
     int start; 
     float expf; 
     int *pat,npat; 
     int win0,winn; 
     int ***rs,**rcnt,*rn;
{
  int i,j,k,l;
  int **ss,*scnt,sn,period,match,t;
  sn = 0; 
  for(l=0;l<n;l++){ 
    period = (int)((float)ns[l]*expf);
    for(i=0;i<(ns[l]-npat+1);i++){
      k = 0;
      match = 1;
      while((k<npat)&&(match))
	if ((int)stim[l][i+k] != pat[k])
	  match = 0;
	else
	  k += 1;
      t = win0 + start + (int)((float)(i+npat-1)*expf);
      if ((k==npat)&&(t+winn < period)&&(t>=0)){ 
	sn += 1;
      }
    }
  }
  ss = (int **)myalloc(sn*sizeof(int *));
  scnt = (int *)myalloc(sn*sizeof(int));
  j = 0;
  for(l=0;l<n;l++){ 
    period = (int)((float)ns[l]*expf);
    for(i=0;i<(ns[l]-npat+1);i++){
      k = 0;
      match = 1;
      while((k<npat)&&(match))
	if ((int)stim[l][i+k] != pat[k])
	  match = 0;
	else
	  k += 1;
      t = win0 + start + (int)((float)(i+npat-1)*expf);
      if ((k==npat)&&(t+winn < period)&&(t>=0)){ 
	ss[j] = extract_spikes(s[l],cnt[l],t,winn,&scnt[j],1);
	j += 1;
      }
    }
  }
  *rs = ss; *rcnt = scnt; *rn = sn;
}
int pattern_trig_avg_sarray(s,cnt,n,stim,ns,start,expf,pat,npat,win0,winn,rav)
     int **s,*cnt,n; 
     float **stim;   
     int *ns;        
     int start;      
     float expf;     
     int *pat;       
     int npat;       
     int win0,winn;  
     float **rav;    
{
  int i,j,k,l;
  int period,pcount,match,t;
  float *sx,*avg;
  avg = get_zero_farray(winn);  
  pcount = 0; 
  for(l=0;l<n;l++){ 
    period = (int)((float)ns[l]*expf);
    sx = expand_spike_array(s[l],cnt[l],0,period);
    for(i=0;i<(ns[l]-npat+1);i++){  
      k = 0;  
      match = 1; 
      while((k<npat)&&(match)){  
	if ((int)stim[l][i+k] != pat[k])
	  match = 0; 
	else
	  k += 1;    
      }
      t = win0 + start + (int)((float)(i+npat-1)*expf);
      if ((k==npat)&&(t+winn < period)&&(t>=0)){ 
	pcount += 1;
	for(j=0;j<winn;j++)  
	  avg[j] += sx[t+j];
      }
    }
    myfree(sx);
  }
  if (pcount > 0)
    multiply_farray(avg,winn,1.0/(float)pcount); 
  *rav = avg;     
  return pcount;  
}
int pattern_trig_avg_self(n,stim,ns,expf,pat,npat,win0,winn,rav)
     int n;
     float **stim;
     int *ns;
     float expf; 
     int *pat,npat; 
     int win0,winn; 
     float **rav;
{
  int i,j,k,l;
  int period,pcount,match,t,ixf;
  float *sx,*avg;
  ixf = my_rint(expf);
  avg = get_zero_farray(winn);
  pcount = 0;
  for(l=0;l<n;l++){ 
    period = ns[l]*ixf;
    sx = get_zero_farray(period); 
    k = 0;
    for(i=0;i<ns[l];i++){
      for(j=0;j<ixf;j++){
	sx[k] = stim[l][i];
	k += 1;
      }
    }
    for(i=0;i<(ns[l]-npat+1);i++){
      k = 0;
      match = 1;
      while((k<npat)&&(match))
	if ((int)stim[l][i+k] != pat[k])
	  match = 0;
	else
	  k += 1;
      t = win0 + (i+npat-1)*ixf;
      if ((k==npat)&&(t+winn < period)&&(t>=0)){ 
	pcount += 1;
	for(j=0;j<winn;j++)
	  avg[j] += sx[t+j];
      }
    }
    myfree(sx);
  }
  if (pcount > 0)
    multiply_farray(avg,winn,1.0/(float)pcount);
  *rav = avg;
  return pcount;
}
void multi_pattern_trig_avg_sarray(s,cnt,n,stim,ns,start,expf,pat,pcnt,np,
				   win0,winn,evenflag,rav,rnav)
     int **s,*cnt,n;
     float **stim;
     int *ns;
     int start;          
     float expf;         
     int **pat,*pcnt,np; 
     int win0,winn;      
     int evenflag;
     float ***rav;
     int **rnav;
{
  int i,j,k,l,p;
  int i0,iinc,period,*pcount,match,t;
  float *sx,**avg;
  i0 = 0;
  iinc = 1;
  if (evenflag >= 0){
    iinc = 2;
    if (evenflag == 0)
      i0 = 1;
    else if (evenflag == 1)
      i0 = 0;
  }
  avg = get_zero_2d_farray(np,winn);
  pcount = get_zero_iarray(np);
  for(l=0;l<n;l++){ 
    period = (int)((float)ns[l]*expf);
    sx = expand_spike_array(s[l],cnt[l],0,period);
    for(p=0;p<np;p++){ 
      for(i=i0;i<(ns[l]-pcnt[p]+1);i+=iinc){
	k = 0;
	match = 1;
	while((k < pcnt[p]) && match)
	  if ((int)stim[l][i+k] != pat[p][k]) 
	    match = 0;
	  else
	    k += 1;
	t = win0 + start + (int)((float)(i+pcnt[p]-1)*expf);
	if ((k==pcnt[p])&&(t+winn < period)&&(t>=0)){ 
	  pcount[p] += 1;
	  for(j=0;j<winn;j++)
	    avg[p][j] += sx[t+j];
	}
      }
    }
    myfree(sx);
  }
  for(i=0;i<np;i++){
    if (pcount[i] > 0)
      multiply_farray(avg[i],winn,1.0/(float)pcount[i]);
  }
  *rav = avg;
  *rnav = pcount;
}
void multi_pattern_trig_avg_farray(fdata,n,fdn,stim,ns,start,expf,pat,pcnt,np,
				   win0,winn,evenflag,rav,rsd,rnav)
     float **fdata;
     int n,fdn;
     float **stim;
     int *ns;
     int start; 
     float expf; 
     int **pat,*pcnt,np; 
     int win0,winn; 
     int evenflag;
     float ***rav,***rsd;
     int **rnav;
{
  int i,k,l,p;
  int i0,iinc,period,*pcount,match,t;
  float **avg,**sd,***fp;
  i0 = 0;
  iinc = 1;
  if (evenflag >= 0){
    iinc = 2;
    if (evenflag == 0)
      i0 = 1;
    else if (evenflag == 1)
      i0 = 0;
  }
  fp = (float ***)myalloc(np*sizeof(float **));
  for(i=0;i<np;i++)
    fp[i] = (float **)myalloc(fdn*sizeof(float *)); 
  avg = (float **)myalloc(np*sizeof(float *));
  sd = (float **)myalloc(np*sizeof(float *));
  pcount = get_zero_iarray(np);
  for(l=0;l<n;l++){ 
    period = (int)((float)ns[l]*expf);
    for(p=0;p<np;p++){ 
      for(i=i0;i<(ns[l]-pcnt[p]+1);i+=iinc){
	k = 0;
	match = 1;
	while((k<pcnt[p])&&(match))
	  if ((int)stim[l][i+k] != pat[p][k])
	    match = 0;
	  else
	    k += 1;
	t = win0 + start + (int)((float)(i+pcnt[p]-1)*expf);
	if ((k==pcnt[p])&&(t+winn < period)&&(t>=0)){ 
	  fp[p][pcount[p]] = &(fdata[l][t]); 
	  pcount[p] += 1;
	}
      }
    }
  }
  for(i=0;i<np;i++){
    if (pcount[i] > 0)
      mean_sdev_2d_farray(fp[i],pcount[i],winn,&(avg[i]),&(sd[i]));
  }
  for(i=0;i<np;i++)
    myfree(fp[i]);
  myfree(fp);
  *rav = avg; *rsd = sd; *rnav = pcount;
}
void avg_spikes_at_trig_sarray(s,cnt,n,ts,tcnt,toffset,start,period,avg_start,
			   avg_period,sigma,ravg,rnavg)
     int **s,*cnt,n,**ts,*tcnt;
     int toffset,start,period,avg_start,avg_period;
     float sigma,**ravg;
     int *rnavg;
{
  int i,j,k;
  int ia,ib,ss,t,navg;
  float *avg,*smooth;
  avg = get_zero_farray(avg_period);
  navg = 0;
  for(i=0;i<n;i++)
    for(j=0;j<tcnt[i];j++){ 
      t = ts[i][j] + toffset;
      ia = t + avg_start;
      ib = ia + avg_period - 1;
      if ((ia >= start)&&(ib < start+period)){
	navg += 1;
	for(k=0;k<cnt[i];k++){ 
	  ss = s[i][k] - ia;
	  if ((ss >= 0)&&(ss < avg_period))
	    avg[ss] += 1.0;
	}
      }
    }
  if (navg > 0)
    multiply_farray(avg,avg_period,1.0/(float)navg);
  if (sigma > 0.0){
    smooth = smooth_with_gaussian(avg,avg_period,sigma,0.01);
    myfree(avg);
    avg = smooth;
  }
  *ravg = avg; *rnavg = navg;
}
void get_coincidence_spikes(d1,n1,d2,n2,rd1,rn1,rd2,rn2,start,period,winsize)
     int *d1,n1,*d2,n2,**rd1,*rn1,**rd2,*rn2;
     int start,period,winsize;
{
  int i,j;
  int *t1,*t2,tn1,tn2,done,dt,match;
  t1 = (int *)myalloc(period*sizeof(int));
  t2 = (int *)myalloc(period*sizeof(int));
  i = j = 0; 
  done = 0;
  while(!done && (i<n1))
    if (d1[i] < start)
      i += 1;
    else
      done = 1;
  done = 0;
  while(!done && (j<n2))
    if (d2[j] < start)
      j += 1;
    else
      done = 1;
  match = 0; 
  tn1 = tn2 = 0;
  if ((i<n1)&&(j<n2)){
    match = done = 0;
    dt = d1[i] - d2[j];
    if ((dt >= -winsize)&&(dt <= winsize)){
      match = 1;
      t1[tn1] = d1[i];
      tn1 += 1;
      t2[tn2] = d2[j];
      tn2 += 1;
    }
  }else
    done = 1;
  while(!done){
    while(match && !done){
      if ((i < n1-1)&&(j < n2-1)){
	if ((d2[j+1] - d1[i]) <= (d1[i+1] - d2[j])){
	  j += 1;
	  dt = d1[i] - d2[j];
	  if ((dt >= -winsize)&&(dt <= winsize)){
	    t2[tn2] = d2[j];
	    tn2 += 1;
	  }else
	    match = 0;
	}else{
	  i += 1;
	  dt = d1[i] - d2[j];
	  if ((dt >= -winsize)&&(dt <= winsize)){
	    t1[tn1] = d1[i];
	    tn1 += 1;
	  }else
	    match = 0;
	}
      }else if (i < n1-1){
	i += 1;
	dt = d1[i] - d2[j];
	if ((dt >= -winsize)&&(dt <= winsize)){
	  t1[tn1] = d1[i];
	  tn1 += 1;
	}else
	  match = 0;
      }else if (j < n2-1){
	j += 1;
	dt = d1[i] - d2[j];
	if ((dt >= -winsize)&&(dt <= winsize)){
	  t2[tn2] = d2[j];
	  tn2 += 1;
	}else
	  match = 0;
      }else
	done = 1;
    }
    while(!match && !done){
      if (d1[i] <= d2[j])
	i += 1;
      else
	j += 1;
      if ((i >= n1)||(j >= n2))
	done = 1;
      else if ((d1[i] >= start+period)||(d2[j] >= start+period))
	done = 1;
      else{
	dt = d1[i] - d2[j];
	if ((dt >= -winsize)&&(dt <= winsize)){
	  match = 1;
	  t1[tn1] = d1[i];
	  tn1 += 1;
	  t2[tn2] = d2[j];
	  tn2 += 1;
	}
      }
    }
  }
  *rd1 = copy_iarray(t1,tn1);
  *rd2 = copy_iarray(t2,tn2);
  myfree(t1); myfree(t2);
  *rn1 = tn1; *rn2 = tn2;
}
void get_coincidence_sarrays(s1,cnt1,s2,cnt2,n,cs1,ccnt1,cs2,ccnt2,
			     start,period,winsize)
     int **s1,*cnt1,**s2,*cnt2,n,***cs1,**ccnt1,***cs2,**ccnt2;
     int start,period,winsize;
{
  int i;
  int **c1,**c2,*cn1,*cn2;
  c1 = (int **)myalloc(n*sizeof(int *));
  c2 = (int **)myalloc(n*sizeof(int *));
  cn1 = (int *)myalloc(n*sizeof(int));
  cn2 = (int *)myalloc(n*sizeof(int));
  for(i=0;i<n;i++)
    get_coincidence_spikes(s1[i],cnt1[i],s2[i],cnt2[i],&c1[i],&cn1[i],
			   &c2[i],&cn2[i],start,period,winsize);
  *cs1 = c1; *cs2 = c2; *ccnt1 = cn1; *ccnt2 = cn2;
}
void get_suppress_spikes_sarrays(s1,cnt1,s2,cnt2,n,win0,winn,rs,rcnt)
     int **s1,*cnt1,**s2,*cnt2,n,win0,winn,***rs,**rcnt;
{
  int i,j,k;
  int **s,*cnt,flag,t,t1;
  s = (int **)myalloc(n*sizeof(int *));
  cnt = (int *)myalloc(n*sizeof(int));
  for(i=0;i<n;i++){
    s[i] = (int *)myalloc(cnt2[i]*sizeof(int));
    cnt[i] = 0;
    for(j=0;j<cnt2[i];j++){
      flag = 1;
      t = s2[i][j];
      for(k=0;k<cnt1[i];k++){
	t1 = s1[i][k] + win0;
	if ((t >= t1)&&(t < t1+winn))
	  flag = 0;
      }
      if (flag){
	s[i][cnt[i]] = t;
	cnt[i] += 1;
      }
    }
  }
  *rs = s; *rcnt = cnt;
}
void transmission_stat_sarrays(s1,cnt1,s2,cnt2,n,start,period,win0,winn,
			       sflag,rntrig,rntrans)
     int **s1,*cnt1,**s2,*cnt2,n,start,period,win0,winn,sflag;
     int *rntrig,*rntrans;
{
  int i,j,k;
  int flag,t,t2,a,b,spm1,ntrig,ntrans;
  spm1 = start + period - 1;
  if (winn < 0)
    exit_error("TRANSMISSION_STAT_SARRAYS","winn must be greater than 0");
  ntrig = ntrans = 0;
  for(i=0;i<n;i++){
    for(j=0;j<cnt1[i];j++){
      t = s1[i][j];
      a = t + win0;
      b = a + winn - 1;
      if ((t >= start)&&(a >= start)&&(t<=spm1)&&(b<=spm1)){
	ntrig += 1;
	flag = 0;
	for(k=0;k<cnt2[i];k++){
	  t2 = s2[i][k];
	  if ((t2 >= a)&&(t2 <= b)){
	    flag += 1;
	  }
	}
	if (flag > 0)
	  ntrans += 1;
      }
    }
  }
  *rntrig = ntrig; *rntrans = ntrans;
}
void shift_sarray(s,cnt,n,k)
     int **s,*cnt,n,k;
{
  int i,j;
  int **ts,*tcnt;
  if (k >= n)
    exit_error("SHIFT_SARRAY","Shift k > n");
  ts = (int **)myalloc(n*sizeof(int *));
  tcnt = (int *)myalloc(n*sizeof(int));
  for(i=0;i<n;i++){
    j = i+k; 
    if (j >= n)
      j -= n;
    ts[i] = s[j];
    tcnt[i] = cnt[j];
  }
  for(i=0;i<n;i++){ 
    s[i] = ts[i]; 
    cnt[i] = tcnt[i];
  }
  myfree(ts); myfree(tcnt);
}
void get_filtered_sarray(s,cnt,n,rs,rcnt,rn,nsd,tmin,tmax,minspikes)
     int **s,*cnt,n,***rs,**rcnt,*rn; 
     float nsd;
     int tmin,tmax,minspikes; 
{
  int i,k;
  int **ts,*tcnt,tn,*flag;
  int pflag = 0;
  float mean,sdev;
  flag = (int *)myalloc(n*sizeof(int)); 
  for(i=0;i<n;i++)
    flag[i] = 1;
  if (nsd >= 0.0){ 
    mean_sdev_iarray(cnt,n,&mean,&sdev);
    for(i=0;i<n;i++)
      if (nsd < fabs(((float)cnt[i] - mean)/sdev)){
	flag[i] = 0;
	if (pflag) printf("  (NSD) i = %d\n",i);
      }
  }
  if (tmin >= 0){
    for(i=0;i<n;i++)
      if (cnt[i] > 0){
	if (s[i][0] > tmin){
	  flag[i] = 0;
	  if (pflag) printf("  (TMIN) i = %d\n",i);
	}
      }else
	flag[i] = 0;
  }
  if (tmax >= 0){
    for(i=0;i<n;i++)
      if (cnt[i] > 0){
	if (s[i][cnt[i]-1] < tmax){
	  flag[i] = 0;
	  if (pflag) printf("  (TMAX) i = %d\n",i);
	}
      }else
	flag[i] = 0;
  }
  if (minspikes >= 0){
    for(i=0;i<n;i++)
      if (cnt[i] < minspikes){
	flag[i] = 0;
	if (pflag) printf("  (MINSPIKES) i = %d\n",i);
      }
  }
  tn = sum_iarray(flag,n,0,n);
  ts = (int **)myalloc(tn*sizeof(int *));
  tcnt = (int *)myalloc(tn*sizeof(int));
  k = 0;
  for(i=0;i<n;i++)
    if (flag[i]){
      ts[k] = copy_iarray(s[i],cnt[i]);
      tcnt[k] = cnt[i];
      k += 1;
    }
  myfree(flag);
  *rs = ts; *rcnt = tcnt; *rn = tn;
}
void get_sc_filtered_double_sarray(s1,cnt1,s2,cnt2,n,rs1,rcnt1,rs2,rcnt2,rn,
				   minspikes)
     int **s1,*cnt1,**s2,*cnt2,n,***rs1,**rcnt1,***rs2,**rcnt2,*rn;
     int minspikes;
{
  int i,k;
  int **ts1,*tcnt1,**ts2,*tcnt2,tn,*flag;
  flag = (int *)myalloc(n*sizeof(int)); 
  for(i=0;i<n;i++)
    flag[i] = 1;
  for(i=0;i<n;i++)
    if ((cnt1[i] < minspikes)||(cnt2[i] < minspikes))
      flag[i] = 0;
  tn = sum_iarray(flag,n,0,n);
  ts1 = (int **)myalloc(tn*sizeof(int *));
  ts2 = (int **)myalloc(tn*sizeof(int *));
  tcnt1 = (int *)myalloc(tn*sizeof(int));
  tcnt2 = (int *)myalloc(tn*sizeof(int));
  k = 0;
  for(i=0;i<n;i++)
    if (flag[i]){
      ts1[k] = copy_iarray(s1[i],cnt1[i]);
      ts2[k] = copy_iarray(s2[i],cnt2[i]);
      tcnt1[k] = cnt1[i];
      tcnt2[k] = cnt2[i];
      k += 1;
    }
  myfree(flag);
  *rs1 = ts1; *rcnt1 = tcnt1; *rs2 = ts2; *rcnt2 = tcnt2; *rn = tn;
}
int check_trial_condition_su(coherency,direction,series_type,value)
     int coherency,direction;
     char series_type[];
     int value;
{
  int cflag,dflag,allflag;
  int match;
  cflag = dflag = allflag = 0;
  if (strcmp(series_type,"cseries")==0)
    cflag = 1;
  else if (strcmp(series_type,"dseries")==0)
    dflag = 1;
  else if (strcmp(series_type,"single")==0)
    allflag = 1;
  if ((cflag && (value == coherency))||
      (dflag && (value == direction))|| allflag)
    match = 1;
  else
    match = 0;
  return match;
}
void get_sarray_from_events(etrial,n,spikes,count)
     struct event_trial *etrial;
     int n; 
     int ***spikes,**count;
{
  int i,j;
  int **s,*cnt;
  cnt = (int *)myalloc(n*sizeof(int));
  s = (int **)myalloc(n*sizeof(int *));
  for (i=0;i<n;i++){
    cnt[i] = etrial[i].nev;
    s[i] = (int *)myalloc(etrial[i].nev*sizeof(int));
    for(j=0;j<cnt[i];j++)
      s[i][j] = (etrial[i].ev[j].t[0] +
		 etrial[i].ev[j].t[etrial[i].ev[j].n-1])/2;
  }
  *spikes = s;
  *count = cnt;
}
float *su_vardur_rate_sarray(s,cnt,n,start,vdur,sampling)
     int **s,*cnt,n;
     int start;
     int *vdur;         
     float sampling;
{
  int i;
  float *data;
  data = (float *)myalloc(n*sizeof(float));
  for(i=0;i<n;i++){
    data[i] = (float)count_spikes(s[i],cnt[i],start,vdur[i]) /
              (float)(vdur[i]) * sampling;
  }
  return data;
}
float mean_spike_count_sarray(s,cnt,n,start,period,sampling,rsdev)
     int **s,*cnt,n;
     int start,period;
     float sampling,*rsdev;
{
  int i;
  float avg_sc,*data;
  data = (float *)myalloc(n*sizeof(float));
  avg_sc = 0.0;
  for (i=0;i<n;i++)
    data[i] = (float)count_spikes(s[i],cnt[i],start,period);
  mean_sdev_farray(data,n,&avg_sc,rsdev);
  myfree(data);
  return avg_sc;
}
float mean_spike_count_sarray_window(s,cnt,n,start,period,winwidth,rn,rsdev)
     int **s,*cnt,n,start,period,winwidth,*rn;
     float *rsdev;
{
  int i,j,k;
  int nwin,tn;
  float *data,mean;
  tn = period/winwidth;
  nwin = n * tn;
  data = (float *)myalloc(nwin*sizeof(float));
  k = 0;
  for (i=0;i<n;i++)
    for (j=0;j<tn;j++){
      data[k] = (float)count_spikes(s[i],cnt[i],start+j*winwidth,winwidth);
      k += 1;
    }
  mean_sdev_farray(data,nwin,&mean,rsdev);
  myfree(data);
  *rn = nwin;
  return mean;
}
float mean_spike_rate_sarray(s,cnt,n,start,period,sampling,sdev)
     int **s,*cnt,n; 
     int start,period;
     float sampling,*sdev;
{
  int i;
  float avg_sr;
  float *data;
  data = (float *)myalloc(n*sizeof(float));
  avg_sr = 0.0;
  for (i=0;i<n;i++){
    data[i] = ((float)count_spikes(s[i],cnt[i],start,period) /
	       (float)period * sampling);
  }
  mean_sdev_farray(data,n,&avg_sr,sdev);
  myfree(data);
  return avg_sr;
}
float *get_norm_spike_count_sarray(s,cnt,n,start,period,sampling)
     int **s,*cnt,n;
     int start,period;
     float sampling;
{
  int i;
  float *data;
  data = (float *)myalloc(n*sizeof(float));
  for(i=0;i<n;i++)
    data[i] = (float)count_spikes(s[i],cnt[i],start,period);
  z_score_farray(data,n);
  return data;
}
void get_cycle_trig_avg_trial(data,n,start,ncyc,period,sampling,ravg,rn)
     int *data,n,start,ncyc;
     float period,sampling,**ravg;
     int *rn;
{
  int i,k;
  int m,navg,nignore,nuse;
  float t,duration,*avg;
  navg = (int)period;
  avg = get_zero_farray(navg);
  duration = (float)ncyc * period;
  nignore = nuse = 0;
  for(i=0;i<n;i++){
    t = (float)(data[i] - start);
    if ((t>=0.0)&&(t<duration)){
      m = (int)(t/period); 
      k = (int)(t-(float)m*period);
      if ((k < 0)||(k >= navg))
	nignore += 1;
      else{
	avg[k] += 1.0;
	nuse += 1;
      }
    }
  }
  if (nignore > 0)
    printf("  GET_CYCLE_TRIG_AVG_TRIAL: ignored %d spikes, used %d.\n",nignore,
	   nuse);
  multiply_farray(avg,navg,sampling/(float)ncyc);
  *ravg = avg; *rn = navg;
}
void get_fourier_harmonic_sarray(s,cnt,n,start,fdur,period,sampling,order,
                                 ramplmean,ramplsd,rphmean,rphsd,rampl)
     int **s,*cnt,n,start;
     float fdur,period,sampling;
     int order;
     float *ramplmean,*ramplsd,*rphmean,*rphsd;
     float **rampl;
{
  int i,j;
  float *ampl,*phase,t,todd,teven,fc,r,theta,*d;
  ampl = (float *)myalloc(n*sizeof(float));
  phase = (float *)myalloc(n*sizeof(float));
  fc = 2.0*M_PI / (period/(float)order);
  for(i=0;i<n;i++){ 
    todd = teven = 0.0;
    for(j=0;j<cnt[i];j++){ 
      t = (float)(s[i][j]-start);
      if ((t > 0.0) && (t < fdur)){
        t *= fc;
        todd += sin(t);
        teven += cos(t);
      }
    }
    ampl[i] = 2.0*sqrt(todd*todd + teven*teven) / (fdur/sampling);
    phase[i] = (float)atan2(todd,teven) * 180.0/M_PI;
  }
  mean_sdev_farray(ampl,n,ramplmean,ramplsd);
  vector_average_farray(ampl,phase,n,2,2,&r,&theta);
  d = get_farray(n);
  for(i=0;i<n;i++){
    d[i] = get_signed_circular_diff(phase[i]+180.0,theta+180.0,360.0);
  }
  mean_sdev_farray(d,n,rphmean,rphsd);
  myfree(d);
  *rphmean += theta;
  *rampl = ampl; 
  myfree(phase);
}
void get_fourier_harmonic_sarray_old(s,cnt,n,start,ncyc,period,sampling,order,
				     ramplmean,ramplsd,rphmean,rphsd)
     int **s,*cnt,n,start,ncyc;
     float period,sampling;
     int order;
     float *ramplmean,*ramplsd,*rphmean,*rphsd;
{
  int i;
  int nav;
  float *ampl,*phase,*avg;
  ampl = (float *)myalloc(n*sizeof(float));
  phase = (float *)myalloc(n*sizeof(float));
  for(i=0;i<n;i++){
    get_cycle_trig_avg_trial(s[i],cnt[i],start,ncyc,period,sampling,&avg,&nav);
    get_fourier_harmonic_farray(avg,nav,order,period,&ampl[i],&phase[i]);
    myfree(avg);
  }
  mean_sdev_farray(ampl,n,ramplmean,ramplsd);
  mean_sdev_farray(phase,n,rphmean,rphsd);
  myfree(ampl); myfree(phase);
}
void increment_psth_float_bin_trial(data,n,start,period,binsize,psth,nbin)
     int *data,n;
     float start,period,binsize;
     float *psth;
     int nbin;
{
  int i,k;
  float t;
  for(i=0;i<n;i++){
    t = (float)data[i] - start;
    k = (int)(t/binsize);
    if ((t>=0.0)&&(k >= 0)&&(k<nbin))
      psth[k] += 1.0;
  }
}
void psth_float_bin_trial(data,n,start,period,binsize,rpsth,rn)
     int *data,n;
     float start,period,binsize;
     float **rpsth;
     int *rn;
{
  int i,k;
  int nbin;
  float t,*psth;
  nbin = (int)(period/binsize);
  psth = get_zero_farray(nbin);
  for(i=0;i<n;i++){
    t = (float)data[i] - start;
    k = (int)(t/binsize);
    if ((t>=0.0)&&(k >= 0)&&(k<nbin))
      psth[k] += 1.0;
  }
  *rpsth = psth;
  *rn = nbin;
}
void psth_float_bin_sarray(s,cnt,n,start,period,binsize,rpsth,rn)
     int **s,*cnt,n;
     float start,period,binsize;
     float **rpsth;
     int *rn;
{
  int i,j,k;
  int nbin;
  float t,*psth;
  nbin = (int)(period/binsize);
  psth = get_zero_farray(nbin);
  for(i=0;i<n;i++) 
    for(j=0;j<cnt[i];j++){ 
      t = (float)s[i][j] - start;
      k = (int)(t/binsize);
      if ((t>=0.0)&&(k >= 0)&&(k<nbin))
	psth[k] += 1.0;
    }
  *rpsth = psth;
  *rn = nbin;
}
void su_psth_float_bin_sarray_vdur(s,cnt,n,start,vdur,binsize,rpsth,rn,rpsthn)
     int **s,*cnt,n;
     float start;
     int *vdur;           
     float binsize;
     float **rpsth;       
     int *rn;             
     int **rpsthn;        
{
  int i,j,k,ii;
  int nbin,period,*psthn,max_bin_index;
  float t,*psth;
  period = max_of_iarray(vdur,n);
  nbin = (int)(period/binsize);
  psth  = get_zero_farray(nbin);
  psthn = get_zero_iarray(nbin);  
  for(i=0;i<n;i++){ 
    max_bin_index = (int)(vdur[i]/binsize);
    for(j=0;j<cnt[i];j++){ 
      t = (float)s[i][j] - start;
      k = (int)(t/binsize);
      if ((t >= 0.0) && (k >= 0) && (k < max_bin_index)){
	psth[k] += 1.0;
      }
    }
    for(ii=0;ii<max_bin_index;ii++)  
      psthn[ii] += 1;
  }
  *rpsth = psth;
  *rn = nbin;
  *rpsthn = psthn;
}
int *simple_pst_sarray(s,cnt,n,start,period)
     int **s,*cnt,n,start,period;
{
  int i,j;
  int *pst,t;
  pst = get_zero_iarray(period);
  for (i=0;i<n;i++)
    for (j=0;j<cnt[i];j++){
      t = s[i][j] - start;
      if ((t>=0)&&(t<period))
	pst[t] += 1;
    }
  return pst;
}
float *psth_prob_sarray(s,cnt,n,start,period)
     int **s,*cnt,n,start,period;
{
  int *ipsth;
  float *psth;
  ipsth = simple_pst_sarray(s,cnt,n,start,period); 
  psth = i2farray(ipsth,period); 
  myfree(ipsth);
  multiply_farray(psth,period,1.0/(float)n); 
  return psth;
}
int *simple_psth_bin_sarray(s,cnt,n,start,period,binsize,npsth)
     int **s,*cnt,n,start,period,binsize,*npsth;
{
  int i,j;
  int *psth,t,b;
  *npsth = (period+binsize-1)/binsize;
  psth = get_zero_iarray(*npsth);
  for (i=0;i<n;i++)
    for (j=0;j<cnt[i];j++){
      t = s[i][j] - start;
      if ((t>=0)&&(t<period)){
	b = t / binsize;
	if (b > *npsth) exit_error("SIMPLE_PSTH_BIN_SARRAY","b too big");
	psth[b] += 1.0;
      }
    }
  return psth;
}
float *adaptive_pst_sarray(spikes,count,n,start,period,sampling,ncrit)
     int **spikes,*count,n; 
     int start,period;      
     float sampling;        
     int ncrit;             
{
  int i,j;
  int t,total,*sum;
  int done,a,b;
  float *pst;
  pst = (float *)myalloc(period*sizeof(float));
  sum = get_zero_iarray(period);
  for (i=0;i<n;i++)
    for (j=0;j<count[i];j++){
      t = spikes[i][j] - start;
      if ((t>=0)&&(t<period))
	sum[t] += 1;
    }
  for (i=0;i<period;i++){
    a = b = i;
    done = 0;
    total = sum[a];
    while (!done)
      if (total >= ncrit){
	done = 1;
	pst[i] = (float)total/((float)((b-a+1)*n)/sampling);
      }else{
	a -= 1;
	b += 1;
	if ((a<0) || (b>=period)){
	  done = 1;
	  pst[i] = -1.0;
	}else
	  total += (sum[a] + sum[b]);
      }
  }
  myfree(sum);
  return pst;
}
void spikeu_sarray_condition_sarray(s1,cnt1,s2,cnt2,n,tn,toff,trad,rs,rcnt)
     int **s1;       
     int *cnt1;      
     int **s2;       
     int *cnt2;      
     int n;          
     int tn;         
     int toff;       
     int trad;       
     int ***rs;      
     int **rcnt;     
{
  int i,j,k;
  int k0,k1,t;
  int **x,**s3,*cnt3,*ts;
  printf("  SPIKEU_SARRAY_CONDITION_SARRAY\n");
  printf("    toff = %d\n",toff);
  printf("    trad = %d\n",trad);
  if (trad < 0)
    exit_error("SPIKEU_SARRAY_CONDITION_SARRAY","trad < 0");
  x = get_zero_2d_iarray(n,tn);
  for(i=0;i<n;i++){  
    for (j=0;j<cnt1[i];j++){  
      t = s1[i][j] + toff;
      k0 = t - trad;
      if (k0 < 0)
	k0 = 0;
      k1 = t + trad;
      if (k1 >= tn)
	k1 = tn-1;
      for(k=k0;k<=k1;k++){
	x[i][k] = 1;
      }
    }
  }
  s3 = (int **)myalloc(n*sizeof(int *));
  cnt3 = (int *)myalloc(n*sizeof(int));
  ts = (int *)myalloc(tn*sizeof(int));
  for(i=0;i<n;i++){  
    k = 0;
    for (j=0;j<cnt2[i];j++){  
      t = s2[i][j];
      if ((t >= 0) && (t < tn)){
	if (x[i][t] == 1){
	  ts[k] = t;
	  k += 1;
	}
      }
    }
    s3[i] = copy_iarray(ts,k);
    cnt3[i] = k;
  }
  free_2d_iarray(x,n);
  myfree(ts);
  *rs = s3;
  *rcnt = cnt3;
}
float *get_first_spike_distrib_sarray(s,cnt,n,start,period,rmu,rsd,rrel)
     int **s,*cnt,n;
     int start,period;
     float *rmu,*rsd,*rrel;
{
  int i,k;
  int count,t;
  float *psth,*tdata;
  psth = get_zero_farray(period);
  tdata = (float *)myalloc(n*sizeof(float));
  count = 0;
  for(i=0;i<n;i++){ 
    k = get_first_spike_in_window(s[i],cnt[i],start,period);
    t = k-start;
    if (t >= 0){  
      psth[t] += 1.0;
      tdata[count] = (float)k;
      count += 1;
    }
  }
  mean_sdev_farray(tdata,count,rmu,rsd);
  myfree(tdata);
  *rrel = (float)count/(float)n;
  return psth;
}
int *simple_isi_sarray(spikes,count,n,start,period,max_isi,nn,ntot)
     int **spikes,*count,n; 
     int start,period;      
     int max_isi;           
     int *nn,*ntot; 
{
  int i,j;
  int interval;
  int *isi,nisi;
  nisi = max_isi+1;
  isi = get_zero_iarray(nisi);
  *ntot = 0;
  for (i=0;i<n;i++)
    for (j=0;j<(count[i]-1);j++){
      if ((spikes[i][j] >= start)&&(spikes[i][j+1] < (start+period))){
	interval = spikes[i][j+1] - spikes[i][j];
	if (interval < nisi){
	  isi[interval]  += 1;
	  *ntot += 1;
	}
      }
    }
  *nn = nisi;
  return isi;
}
void get_longest_isi_sarray(s,cnt,n,start,period,k,tisi,pisi)
     int **s,*cnt,n,start,period,k;
     int **tisi,**pisi; 
{
  int i,j;
  int *isi,*t,*p,nisi,ntot,count,min_count,min_isi,tt,ti;
  isi = simple_isi_sarray(s,cnt,n,start,period,period,&nisi,&ntot);
  min_isi = nisi-1; 
  count = isi[min_isi];
  while((min_isi>0)&&(count<k)){
    min_isi -= 1;
    count += isi[min_isi];
  }
  if (count<k)
    exit_error("GET_LONGEST_ISI_SARRAY","Not enough ISIs");
  min_count = k-(count-isi[min_isi]);
  t = (int *)myalloc(k*sizeof(int));
  p = (int *)myalloc(k*sizeof(int));
  ti = 0;
  for (i=0;i<n;i++){ 
    for (j=0;j<(cnt[i]-1);j++){ 
      if ((s[i][j] >= start) && (s[i][j+1] <= start+period)){
	tt = s[i][j+1] - s[i][j];
	if (tt < 0)
	  exit_error("GET_LONGEST_ISI_SARRAY","Spikes out of order");
	if (tt > min_isi){
	  t[ti] = s[i][j];
	  p[ti] = tt;
	  ti += 1;
	}else if ((tt == min_isi) && (min_count > 0)){
	  t[ti] = s[i][j];
	  p[ti] = tt;
	  ti += 1;
	  min_count -= 1;
	}
      }
    }
  }
  myfree(isi);
  *tisi = t; *pisi = p;
} 
int get_fraction_isi_sarray(s,cnt,n,start,period,f)
     int **s,*cnt,n,start,period;
     float f;
{
  int k;
  int *isi,nisi,ntot,total;
  isi = simple_isi_sarray(s,cnt,n,start,period,period,&nisi,&ntot);
  k = 0;
  total = isi[k];
  while((float)total/(float)ntot < f){
    k += 1;
    total += isi[k];
  }
  myfree(isi);
  return k;
} 
int get_time_fraction_isi_sarray(s,cnt,n,start,period,f)
     int **s,*cnt,n,start,period;
     float f;
{
  int i,k;
  int *isi,nisi,ntot;
  float *fisi,total;
  isi = simple_isi_sarray(s,cnt,n,start,period,period,&nisi,&ntot);
  fisi = i2farray(isi,nisi);
  for(i=1;i<nisi;i++)
    fisi[i] *= (float)i;
  norm_area_farray(fisi,nisi,1.0);
  k = 0;
  total = fisi[k];
  while(total < f){
    k += 1;
    total += fisi[k];
  }
  myfree(isi); myfree(fisi);
  return k;
}
void power_burst_metric(pwr,m,n,dx,rp,rcat)
     float **pwr;
     int m,n;
     float dx,*rp;
     char **rcat;
{
  int i,j;
  float *mean,*t,fb,fnb,*rb,*rnb,p;
  int bstart,bstop,width; 
  int imax,imin,imin0;
  int nbstart;
  bstart = 5;  
  nbstart = 3;
  bstop = 13;
  width = 7;
  mean_2d_farray(pwr,m,n,&mean);
  t = get_zero_farray(n); 
  for(i=0;i<(n-width+1);i++)
    for(j=0;j<width;j++)
      t[i] += mean[i+j];
  imax = bstart + max_coord_farray(t+bstart,bstop-bstart+1);
  imin = (imax-1) + min_coord_farray(t+(imax-1),(n-width)-(imax-1)+1);
  imin0 = nbstart + min_coord_farray(t+nbstart,bstop-nbstart+1);
  printf("  imin0= %d  imin= %d  imax= %d\n",imin0,imin,imax);
  rb = (float *)myalloc(m*sizeof(float));
  rnb = (float *)myalloc(m*sizeof(float));
  fb = 1.0;
  printf("********** WYETH - what shoud fnb start at? (Using 1)\n");
  fnb = 1.0;
  for(i=0;i<m;i++){
    rb[i] = sum_farray(pwr[i],n,imax,width) / sum_farray(pwr[i],n,imin,width);
    rnb[i] = sum_farray(pwr[i],n,imin0,width) / (float)width;
    printf("  %f %f\n",rb[i],rnb[i]);
    if (rb[i] > 1.0)
      fb += 1.0;
    if (rnb[i] < 1.0)
      fnb += 1.0;
  }
  fb /= (float)m;
  fnb /= (float)m;
  if (fb >= 0.9){
    printf("  Burst\n");
    *rcat = strdup("b");
    p = mean_farray(rb,m);
  }else if (fnb >= 0.9){
    printf("  Non-burst\n");
    *rcat = strdup("nb");
    p = mean_farray(rnb,m);
  }else if (fb >= 0.5){
    printf("  Mixed, burst\n");
    *rcat = strdup("mb");
    p = mean_farray(rb,m);
  }else{
    printf("  Mixed, non-burst\n");
    *rcat = strdup("mnb");
    p = mean_farray(rnb,m);
  }
  printf("  p = %.4f\n",p);
  *rp = p; 
  myfree(mean); myfree(t); myfree(rb); myfree(rnb);
}
int get_tuning_time(s1,cnt1,n1,s2,cnt2,n2,start,period,k,sigma,signif)
     int **s1,*cnt1,n1,**s2,*cnt2,n2; 
     int start,period,k;
     float sigma,signif; 
{
  int i;
  int *psth1,*psth2,nsig,ttime;
  float *f1,*f2,t,*sig,*smsig;
  printf("  GET_TUNING_TIME\n");
  psth1 = simple_pst_sarray(s1,cnt1,n1,start,period);
  psth2 = simple_pst_sarray(s2,cnt2,n2,start,period);
  f1 = i2farray(psth1,period);
  f2 = i2farray(psth2,period);
  for(i=0;i<period;i+=k){     
    f1[i] += 10.0/(float)n1;  
    f2[i] += 10.0/(float)n1;  
  }
  nsig = period - k + 1;
  sig = (float *)myalloc(nsig*sizeof(float));
  for(i=0;i<nsig;i++)
    tutest(f1-1+i,k,f2-1+i,k,&t,&sig[i]);
  smsig = smooth_with_gaussian(sig,nsig,sigma,0.01);
  ttime = -1;
  for(i=0;i<nsig;i++)
    if ((ttime == -1)&&(smsig[i] < signif))
      ttime = i;
  return ttime;
}
int get_response_onset(s,cnt,n,start,period,bstart,bperiod,k,sigma,signif)
     int **s,*cnt,n,start,period,bstart,bperiod,k;
     float sigma,signif; 
{
  int i,j;
  int *psth,nsig,ttime,r;
  float *fpsth,*sig,*smsig,bg;
  printf("  GET_RESPONSE_ONSET\n");
  psth = simple_pst_sarray(s,cnt,n,start,period);
  fpsth = i2farray(psth,period);
  bg = 0.0;
  for(i=0;i<bperiod;i++)
    bg += fpsth[bstart+i];
  bg += 10.0/(float)n; 
  bg /= (float)(bperiod*n);
  nsig = period - k + 1;
  sig = (float *)myalloc(nsig*sizeof(float));
  for(i=0;i<nsig;i++){
    r = 0;
    for(j=0;j<k;j++)
      r += psth[i+j];
    sig[i] = binomial_signif(r,n*k,bg);
  }
  smsig = smooth_with_gaussian(sig,nsig,sigma,0.01);
  ttime = -1;
  for(i=0;i<nsig;i++)
    if ((ttime == -1)&&(smsig[i] < signif))
      ttime = i;
  return ttime;
}
float *shifted_difference_floats(data1,data2,n)
     float *data1,*data2;
     int n;
{
  int i,j;
  int start,stop;
  float *dist;
  int dn;
  dn = 2*n-1;
  dist = get_zero_farray(dn);
  for(i=-(n-1);i<=(n-1);i++){
    start = 0;
    stop = n-1;
    if (i>0)
      start = i;
    if (i<0)
      stop = n+i-1;
    for(j=start;j<=stop;j++)
      dist[i+n-1] += fabs(data1[j]-data2[j-i]);
    dist[i+n-1] /= (float)(stop-start+1);
  }
  return dist;
}
float *distance_trial(spikes1,n1,spikes2,n2,start,period,sigma)
     int *spikes1,n1,*spikes2,n2;
     int start,period;
     float sigma;
{
  float *s1,*s2;   
  float *sm1,*sm2; 
  float *dist;
  printf("  SPIKES_DISTANCE\n");
  s1 = expand_spike_array(spikes1,n1,start,period);
  s2 = expand_spike_array(spikes2,n2,start,period);
  sm1 = smooth_with_gaussian(s1,period,sigma,0.01);
  sm2 = smooth_with_gaussian(s2,period,sigma,0.01);
  myfree(s1);
  myfree(s2);
  dist = shifted_difference_floats(sm1,sm2,period);
  myfree(sm1);
  myfree(sm2);
  return dist;
}
float log_prob_spikes(data,n,prob,duration)
     int *data,n;
     float *prob;
     int duration;
{
  int i;
  float p;
  p = 0.0;
  for (i=0;i<n;i++)
    p += log(prob[data[i]]);
  return p;
}
float *log_prob_sarray(spikes,count,n,start,period)
     int **spikes,*count,n; 
     int start,period;      
{
  int i,j;
  int t;
  float *prob,*logp;
  exit_error("LOG_PROB_SARRAY","No longer maintained");
  prob = (float *)myalloc(period*sizeof(float));
  logp = (float *)myalloc(n*sizeof(float));
  for (i=0;i<n;i++){
    for (j=0;j<period;j++)      
      prob[j] = 0.0; 
    for (j=0;j<count[i];j++){   
      t = spikes[i][j] - start;
      if ((t>=0)&&(t<period))
	prob[t] -= 1.0;
    }
    for (j=0;j<period;j++){      
      prob[j] /= (float)(n-1);
      if (prob[j] == 0.0)         
	prob[j] = 1.0/(2.0*(float)(n-1));
      if (prob[j] == 1.0)
	prob[j] = 1.0 - 1.0/(2.0*(float)(n-1));
    }
    logp[i] = log_prob_spikes(spikes[i],count[i],prob,period);
  }
  return logp;
}
float *compare_spike_to_pst(s,cnt,pst,n,width_list,nwidth)
     int *s,cnt;
     float *pst;
     int n,*width_list,nwidth;
{
  int i,j,k;
  int width,mid,pos;
  float *sum;
  float mean,sdev;
  double dsum,height;
  mean_sdev_farray(pst,n,&mean,&sdev);
  sum = get_zero_farray(nwidth);
  for(i=0;i<nwidth;i++){ 
    width = width_list[i];
    mid = (width - 1)/2;
    height = 1.0/(float)width;
    dsum = 0.0;
    for(j=0;j<cnt;j++) 
      for(k=0;k<width;k++){
	pos = s[j] - mid + k; 
	if ((pos >= 0) && (pos < n))
	  dsum += height * pst[pos];
	else
	  dsum += height * mean;
      }
    sum[i] = (float)dsum;
  }
  return sum;
}
float avg_isi_trial(data,n,start,period)
     int *data,n;
     int start,period;
{
  int i;
  int count;
  float mean;
  count = 0;
  mean = 0.0;
  for(i=0;i<(n-1);i++)
    if ((data[i] >= start) && (data[i+1] < (start+period))){
      mean += (float)(data[i+1] - data[i]);
      count += 1;
    }
  mean /= (float)count;
  return mean;
}
float *get_trial_intervals(data,n,start,period,pmin,pmax)
     int *data,n;
     int start,period;
     int *pmin,*pmax;
{
  int i,j;
  float *interval,t;
  int t1,t2,min,max;
  interval = (float *)myalloc(period*sizeof(float));
  for(i=0;i<period;i++)
    interval[i] = -0.0001;
  min = period;
  max = -1;
  for (i=0;i<n-1;i++){
    t1 = data[i] - start;
    t2 = data[i+1] - start;
    t = (float)(t2-t1);
    if (t == 0.0) exit_error("GET_TRIAL_INTERVALS","Zero length interval");
    for(j=t1;j<t2;j++)
      if ((j>=0)&&(j<period)){
	interval[j] = t;
	if (j < min)
	  min = j;
	if (j > max)
	  max = j;
      }
  }
  *pmin = min;
  *pmax = max;
  return interval;
}
void interval_threshold(data,n,thresh,flag)
     float *data;
     int n;
     float thresh;
     int flag;
{
  int i,k;
  float mean;
  k = 0;          
  mean = 0.0;
  for(i=0;i<n;i++)
    if (data[i] < thresh){
      mean += data[i];
      k += 1;
    }
  mean /= (float)k;
  for(i=0;i<n;i++)
    if (flag == 1){
      if (data[i] >= thresh)
	data[i] = mean;
    }else{
      if (data[i] < thresh)
	data[i] = mean;
    }
}
void get_intervals_xy_plot(data,n,start,period,rx,ry,rn)
     int *data,n;
     int start,period;
     int **rx,**ry,*rn;
{
  int i,k;
  int *x,*y,maxn,pmin,pmax;
  float *interv;
  maxn = 3*n; 
  x = (int *)myalloc(maxn*sizeof(int));
  y = (int *)myalloc(maxn*sizeof(int));
  interv = get_trial_intervals(data,n,start,period,&pmin,&pmax);
  k = 0;
  x[k] = start;
  y[k] = interv[k];
  k = 1;
  for(i=1;i<period;i++){ 
    if (interv[i] != interv[i-1]){
      x[k] = i-1+start;
      y[k] = (int)(interv[i-1] + 0.5);
      k += 1;
      x[k] = i+start;
      y[k] = (int)(interv[i] + 0.5);
      k += 1;
    }
  }
  x[k] = start+period-1;
  y[k] = (int)(interv[period-1] + 0.5);
  k += 1;
  myfree(interv);
  *rx = x; *ry = y; *rn = k;
}
void isi_mean_sdev_skew_trial(data,n,start,period,rm1,rm2,rm3,rnused)
     int *data,n;
     int start,period;
     float *rm1,*rm2,*rm3;
     int *rnused;
{
  int i,k;
  int spp,t0,t1;
  float *idata;
  if (!is_nondecreasing_iarray(data,n))
    exit_error("ISI_MOMENTS_TRIAL","Interval list not non-decreasing");
  idata = (float *)myalloc(n*sizeof(float));
  spp = start + period;
  k = 0;
  for(i=0;i<(n-1);i++){ 
    t0 = data[i];
    t1 = data[i+1];
    if ((t0 >= start) && (t1 < spp)){
      idata[k] = (float)(t1-t0);
      k += 1;
    }
  }
  *rnused = k;
  if (k >= 2)
    mean_sdev_skew_farray(idata,k,rm1,rm2,rm3);
  else
    *rm1 = *rm2 = *rm3 = -1.0;
  myfree(idata);
}
float *avg_rate_intervals_trial(data,n,start,period,sampling)
     int *data,n;
     int start,period;
     float sampling;
{
  int i,j;
  float *avg,rate;
  int pos;
  int t1,t2;
  avg = get_zero_farray(period);  
  pos = 0;
  rate = 0.0;
  for (i=0;i<n-1;i++){
    t1 = data[i] - start;
    t2 = data[i+1] - start;
    if ((t2 >= 0)&&(t2 <= period)){
      if ((t2-t1) == 0)
	exit_error("AVG_RATE_INTERVALS_SARRAY","Zero interval");
      rate = sampling/(float)(t2-t1);
      for (j=pos;j<t2;j++)
	avg[j] += rate; 
      pos = t2;
    }
  }
  for (j=pos;j<period;j++)
    avg[j] += rate; 
  pos = period;
  return avg;
}
float *avg_rate_intervals_sarray(spikes,count,n,start,period,sampling)
     int **spikes,*count,n; 
     int start,period;      
     float sampling;
{
  int i,j,k;
  float *avg,rate;
  int pos,t1,t2;
  avg = get_zero_farray(period);  
  for (i=0;i<n;i++){
    pos = 0;
    rate = 0.0;
    for (j=0;j<count[i]-1;j++){
      t1 = spikes[i][j] - start;
      t2 = spikes[i][j+1] - start;
      if ((t2 >= 0)&&(t2 <= period)){
	if ((t2-t1) == 0)
	  exit_error("AVG_RATE_INTERVALS_SARRAY","Zero interval");
	rate = sampling/(float)(t2-t1);
	for (k=pos;k<t2;k++)
	  avg[k] += rate; 
	pos = t2;
      }
    }
    for (k=pos;k<period;k++)
      avg[k] += rate; 
    pos = period;
  }
  for (i=0;i<period;i++)
    avg[i] /= (float)n;
  return avg;
}
float *interval_cv_sarray(s,cnt,n,period)
     int **s,*cnt,n; 
     int period; 
{
  int i,j;
  int pos;
  float *cv;
  float *data,mean,sdev;
  int *ptr; 
  printf("INTERVAL_CV_SARRAY\n");
  printf("  period = %d\n",period);
  data = (float *)myalloc(n*sizeof(float));
  cv = (float *)myalloc(period*sizeof(float));
  ptr = (int *)myalloc(n*sizeof(int));
  for (i=0;i<n;i++)
    ptr[i] = -1;
  for(i=0;i<period;i++){
    pos = 0;
    for(j=0;j<n;j++)
      if ((ptr[j]+1) < cnt[j]){ 
	if (i >= s[j][ptr[j]+1]) 
	  ptr[j] = ptr[j] + 1;
	if ((ptr[j]>=0) && ((ptr[j]+1) < cnt[j])){
	  data[pos] = 1.0/(float)(s[j][ptr[j]+1] - s[j][ptr[j]]);
	  pos += 1;
	}
      }
    mean_sdev_farray(data,pos,&mean,&sdev);
    cv[i] = sdev/mean;
  }
  myfree(data);
  myfree(ptr);
  return cv;
}
void avg_trial_cv_sarray(s,cnt,n,start,period,cv_mean,cv_sdev)
     int **s,*cnt,n;
     int start,period;
     float *cv_mean,*cv_sdev;
{
  int i,j;
  int max,ipos,cpos,count,t,spp;
  float *isi_data,*cv_data;
  float mean,sdev;
  max = max_of_iarray(cnt,n);
  isi_data = (float *)myalloc(max*sizeof(float));
  cv_data = (float *)myalloc(n*sizeof(float));
  spp = start + period;
  cpos = 0;
  count = 0;
  for(i=0;i<n;i++){
    ipos = 0;
    for(j=0;j<(cnt[i]-1);j++){
      t = s[i][j];
      if ((t >= start) && (t < spp)){
	isi_data[ipos] = (float)(s[i][j+1]-t);
	ipos += 1;
      }
    }
    if (ipos > 2){
      mean_sdev_farray(isi_data,ipos,&mean,&sdev);
      cv_data[cpos] = sdev/mean;
      cpos += 1;
    }else
      count += 1;
  }
  if (count > 0)
    printf("  *** %d trials not included:  less than 3 ISIs.\n",count);
  mean_sdev_farray(cv_data,cpos,cv_mean,cv_sdev);
  myfree(isi_data);
  myfree(cv_data);
}
void avg_trial_cv2_sarray(s,cnt,n,start,period,cv_mean,cv_sdev)
     int **s,*cnt,n;
     int start,period;
     float *cv_mean,*cv_sdev;
{
  int i,j,k;
  int max,nisi,count,t,spp;
  float *isi_data,*cv_data,*cv;
  float mean,sdev;
  max = max_of_iarray(cnt,n);
  isi_data = (float *)myalloc(max*sizeof(float));
  cv_data = (float *)myalloc(max*sizeof(float));
  cv = (float *)myalloc(n*sizeof(float));
  spp = start + period;
  k = 0;
  count = 0;
  for(i=0;i<n;i++){
    nisi = 0;
    for(j=0;j<(cnt[i]-1);j++){
      t = s[i][j];
      if ((t >= start) && (t < spp)){
	isi_data[nisi] = (float)(s[i][j+1]-t);
	nisi += 1;
      }
    }
    if (nisi > 2){
      for(j=0;j<(nisi-1);j++)
	cv_data[j] = 2.0*fabs(isi_data[j]-isi_data[j+1])/
	  (float)(isi_data[j]+isi_data[j+1]);
      mean_sdev_farray(cv_data,(nisi-1),&mean,&sdev);
      cv[k] = mean;
      k += 1;
    }else
      count += 1;
  }
  if (count > 0)
    printf("  *** %d trials not included:  less than 3 ISIs.\n",count);
  mean_sdev_farray(cv,k,cv_mean,cv_sdev);
  myfree(isi_data); myfree(cv_data); myfree(cv);
}
void get_gap_sarray(s,cnt,n,min,max,rs,rcnt)
     int **s,*cnt,n,min,max;
     int ***rs,**rcnt;
{
  int i,j,k;
  int **sg,*cntg,*stemp,tn,t;
  sg = (int **)myalloc(n*sizeof(int *));
  cntg = (int *)myalloc(n*sizeof(int));
  tn = max_of_iarray(cnt,n);
  stemp = (int *)myalloc(tn*sizeof(int));
  for (i=0;i<n;i++){ 
    k = 0;
    for (j=0;j<(cnt[i]-1);j++){
      t = s[i][j+1] - s[i][j];
      if ((t>=min)&&(t<=max)){
	stemp[k] = s[i][j];
	k += 1;
      }
    }
    sg[i] = copy_iarray(stemp,k);
    cntg[i] = k;
  }
  myfree(stemp);
  *rs = sg; *rcnt = cntg;
} 
void first_last_count_events_trial(s,n,d,start,duration,first,last,count)
     int *s,n,d;
     int *first,*last,*count;
{
  int i;
  int previous,t;
  *count = 0;
  *first = -1; 
  *last = -1;
  previous = start-1;
  for(i=0;i<n;i++){ 
    t = s[i];
    if ((t >= start) && (t < (start+duration)) && (t > (d+previous)) &&
	(*first < 0))
      *first = i;
    else
      previous = t;
  }
  if (*first >= 0){
    previous = start+duration;
    for(i=(n-1);i>=0;i--){ 
      t = s[i];
      if ((t >= start) && (t < (start+duration)) && (t < (previous-d)) &&
	  (*last < 0))
	*last = i;
      else
	previous = t;
    }
    if (*first <= *last){
      *count = 1;
      for(i=*first;i<*last;i++)
	if ((s[i+1] - s[i]) > d)
	  *count += 1;
    }
  }
}
struct event_trial *get_events_sarray(s,cnt,n,start,duration,d)
     int **s,*cnt,n,start,duration;
     int d; 
{
  int i,j,k;
  int ecount;     
  int scount;     
  int end_burst;  
  int first,last,count;
  struct event_trial *etrial;
  etrial = (struct event_trial *)myalloc(n*sizeof(struct event_trial));
  for (i=0;i<n;i++){ 
    first_last_count_events_trial(s[i],cnt[i],d,start,duration,
				  &first,&last,&count);
    etrial[i].nev = count;
    etrial[i].ev = (struct event_struct *)
      myalloc(count*sizeof(struct event_struct));
    if (count > 0){
      k = first;
      ecount = 0;
      scount = 0;
      while (k<=last){ 
	scount += 1;
	k+=1;
	end_burst = 0;
	while (!end_burst) 
	  if (k>last) 
	    end_burst = 1;
	  else if (s[i][k] > (d+s[i][k-1])) 
	    end_burst = 1;
	  else{ 
	    k += 1;
	    scount += 1;
	  }
	etrial[i].ev[ecount].n = scount;
	etrial[i].ev[ecount].t = (int *)myalloc(scount*sizeof(int));
	for(j=0;j<scount;j++)
	  etrial[i].ev[ecount].t[j] = s[i][k-scount+j];
	ecount += 1;
	scount = 0;
      }
    }
  }
  return etrial;
}
int count_events(data,n,start,period,d)
     int *data,n,start,period,d;
{
  int i,k;
  int count,done;
  k = 0;
  done = 0;
  while (!done){
    if (k >= n)
      done = 1;
    else if ((data[k] >= start) && (data[k] < start+period))
      done = 1;
    else
      k += 1;
  }
  count = 0;
  if (k < n){
    for (i=k;i<(n-1);i++) 
      if (data[i+1] < (start+period))
	if (data[i+1] - data[i] > d)
	  count += 1;
    count += 1;
  }
  return count;
}
float count_weighted_events(data,n,start,period,d,option)
     int *data,n,start,period,d,option;
{
  int i,k;
  int done,spikes;
  float weight;
  k = 0;
  done = 0;
  while (!done){
    if (k >= n)
      done = 1;
    else if ((data[k] >= start) && (data[k] < start+period))
      done = 1;
    else
      k += 1;
  }
  weight = 0.0;
  if (k < n){
    spikes = 1;
    for (i=k;i<(n-1);i++) 
      if (data[i+1] < (start+period)){
	if (data[i+1] - data[i] > d){
	  if ((option==-1)&&(spikes > 1))
	    weight += 1.0;
	  else if (option==0)
	    weight += (float)spikes;
	  else if (option==1)
	    weight += 1.0;
	  else if (option==2)
	    weight += (float)(spikes*spikes);
	  else if (option==3)
	    weight += sqrt((float)spikes);
	  spikes = 1;
	}else
	  spikes += 1; 
      }
    if ((option==-1)&&(spikes > 1)) 
      weight += 1.0;
    else if (option==0)
      weight += (float)spikes;
    else if (option==1)
      weight += 1.0;
    else if (option==2)
      weight += (float)(spikes*spikes);
    else if (option==3)
      weight += sqrt((float)spikes);
  }
  return weight;
}
void make_poisson_float_spikes(data,n,start,duration,sampling,lambda,rmu,
			       rsigma,k,seed)
     float **data;
     int *n;
     int start,duration;    
     float sampling;        
     float lambda;          
     float rmu,rsigma;      
     int k;                 
     int *seed;             
{  
  int i;
  float *parray,ftime,grefract,rmus,rsigs;
  int spike_count,max_spikes;
  if (lambda == 0.0){
    *n = 0;
    *data = NULL;
    return;
  }else if (lambda < 0.0){
    exit_error("MAKE_POISSON_FLOAT_SPIKES","Lambda is negative");
  }
  rmus = rmu*sampling;      
  rsigs = rsigma*sampling;  
  max_spikes = 10*duration; 
  parray = (float *)myalloc(max_spikes*sizeof(float));
  spike_count = 0;
  ftime = 0.0;
  for (i=0;i<k;i++)
    ftime += su_expdev(lambda/sampling,seed);
  while (ftime < (float)duration){
    if (spike_count >= max_spikes){
      printf("  lambda = %f\n",lambda);
      printf("  duration= %d,  max_spikes = %d  spike_count = %d\n",
	     duration,max_spikes,spike_count);
      exit_error("MAKE_POISSON_FLOAT_SPIKES","too many spikes");
    }
    parray[spike_count] = (float)start + ftime;
    spike_count += 1;
    for (i=0;i<k;i++){
      ftime += su_expdev(lambda/sampling,seed);
      if ((rsigma > 0.0) && (rmu > 0.0)){
	grefract = su_gaussian_float(rmus,rsigs,seed);
	while (grefract < 0.0)
	  grefract = su_gaussian_float(rmus,rsigs,seed);
	ftime += grefract;
      }
    }
  }
  *n = spike_count;
  *data = parray;
}
void make_poisson_float_spikes_new(data,n,start,duration,sampling,lambda,rmu,
				   rsigma,k,seed)
     float **data;            
     int *n;                  
     float start,duration;    
     float sampling;          
     float lambda,rmu,rsigma; 
     int k;                   
     int *seed;               
{  
  int i;
  float *parray,*ts,ftime,grefract;
  int spike_count,max_spikes;
  if (lambda == 0.0){
    *n = 0;
    *data = NULL;
    return;
  }else if (lambda < 0.0)
    exit_error("MAKE_POISSON_FLOAT_SPIKES_NEW","Lambda is negative");
  max_spikes = (int)(10.0 * lambda * duration/sampling); 
  parray = (float *)myalloc(max_spikes*sizeof(float));
  spike_count = 0;
  ftime = 0.0;
  for (i=0;i<k;i++)
    ftime += su_expdev(lambda/sampling,seed);
  while (ftime < duration){
    if (spike_count >= max_spikes){
      printf("  *** MAKE_POISSON_FLOAT_SPIKES:  too many spikes.  Exiting.\n");
      exit(0);
    }
    parray[spike_count] = start + ftime;
    spike_count += 1;
    for (i=0;i<k;i++){
      ftime += su_expdev(lambda/sampling,seed);
      if ((rsigma > 0.0) && (rmu > 0.0)){
	grefract = su_gaussian_float(rmu*sampling,rsigma*sampling,seed);
	while (grefract < 0.0)
	  grefract = su_gaussian_float(rmu*sampling,rsigma*sampling,seed);
	ftime += grefract;
      }
    }
  }
  *n = spike_count;
  *data = copy_farray(parray,spike_count);
  myfree(parray);
}
int make_poisson_spikes_int_refract(data,period,lambda,sampling,refract,k,
				    seed)
     int *data,period;  
     float lambda;      
     float sampling;    
     int refract,k;
     int *seed;
{  
  int i;
  int time,n;
  float ftime;
  if (*seed < 0){
    *seed = -*seed;
    spike_util_ran2(seed);
  }
  ftime = 0.0;
  for (i=0;i<k;i++)
    ftime += sampling*su_expdev(lambda,seed);
  time = my_rint(ftime);
  n = 0;
  while (time < period) {
    data[time] = 1;
    n += 1;
    while (my_rint(ftime) <= time+refract)
      for (i=0;i<k;i++)
	ftime += sampling*su_expdev(lambda,seed);
    time = my_rint(ftime);
  }
  return n;
}
int make_gaussian_spikes(data,n,mu,sigma,seed)
     int *data,n;
     float mu,sigma;
     int *seed;
{
  int i;
  int count,time,rnum;
  if (*seed < 0){
    *seed = -*seed;
    spike_util_ran2(seed);
  }
  count = 0;
  for (i=0;i<n;i++) {
    if (data[i] == 1)
      count -= 1;
    data[i] = 0;
  }
  time = my_rint(spike_util_ran2(seed)*mu);
  while (time < n) {
    data[time] = 1;
    count += 1;
    rnum = gaussian_int(mu,sigma,seed);
    while (rnum <= 0)
      rnum = gaussian_int(mu,sigma,seed);
    time += rnum;
  }
  return count;
}
int make_uniform_spikes(data,n,dens,seed)
     int *data,n;
     float dens; 
     int *seed;
{
  int i;
  int count;
  if (*seed < 0){
    *seed = -*seed;
    spike_util_ran2(seed);
  }
  count = 0;
  for (i=0;i<n;i++) {
    if (data[i] == 1)
      count -= 1;
    data[i] = 0;
  }
  for (i=0;i<n;i++)
    if (spike_util_ran2(seed) < dens){
      data[i] = 1;
      count += 1;
    }
  return count;
}
int spike_to_gaussian_burst(data,n,width_mu,width_sigma,mu,sigma,seed)
     int *data,n;
     float width_mu,width_sigma; 
     float mu,sigma;             
     int *seed;
{
  int i;
  int start,*old_data,count,burst_width;
  if (*seed < 0){
    *seed = -*seed;
    spike_util_ran2(seed);
  }
  old_data = (int *)myalloc(n*sizeof(int));
  for (i=0;i<n;i++) {
    old_data[i] = data[i];
    data[i] = 0;
  }
  count = 0;
  for (i=0;i<n;i++)
    if (old_data[i] == 1) {
      burst_width = gaussian_int(width_mu,width_sigma,seed);
      if (burst_width > 0) {
	start = i - burst_width / 2;
	if (start < 0) {burst_width += start; start = 0;}
	if ((start+burst_width) > n) burst_width = n-start;
	count += make_gaussian_spikes(data+start,burst_width,mu,sigma,
				      seed);
      }
    }
  myfree(old_data);
  return count;
}
int spike_to_uniform_burst(data,n,dens,width_mu,width_sigma,seed)
     int *data,n;
     float dens; 
     float width_mu,width_sigma;  
     int *seed;
{
  int i;
  int start,*old_data,count,burst_width;
  if (*seed < 0){
    *seed = -*seed;
    spike_util_ran2(seed);
  }
  old_data = (int *)myalloc(n*sizeof(int));
  for (i=0;i<n;i++) {
    old_data[i] = data[i];
    data[i] = 0;
  }
  count = 0;
  for (i=0;i<n;i++)
    if (old_data[i] == 1) {
      burst_width = gaussian_int(width_mu,width_sigma,seed);
      if (burst_width > 0) {
	start = i - burst_width / 2;
	if (start < 0) {burst_width += start; start = 0;}
	if ((start+burst_width) > n) burst_width = n-start;
	count += make_uniform_spikes(data+start,burst_width,dens,seed);
      }
    }
  return count;
}
void spike_util_random_select_spikes(s,n,p,seed,rs,rn)
     float *s;      
     int n;
     float p;       
     int seed;
     float **rs;    
     int *rn;
{
  int i,k;
  float *t;
  if (seed > 0)
    seed *= -1;
  t = (float *)myalloc(n*sizeof(float));
  k = 0;
  for(i=0;i<n;i++){
    if (spike_util_ran2(&seed) < p){
      t[k] = s[i];
      k += 1;
    }
  }
  *rn = k;
  *rs = copy_farray(t,k);
  myfree(t);
}
void make_random_sarray_from_isi(pisi,nisi,rs,rcnt,n,period,seed1,seed2)
     float *pisi;
     int nisi,***rs,**rcnt,n,period;
     int seed1,seed2;
{
  int i,j,k;
  int *isilist,imax,**s,*cnt,kisi,t,*tdata;
  isilist = (int *)myalloc(10000*sizeof(int));
  imax = 0;
  for(i=0;i<nisi;i++){
    k = (int)(10000.0 * pisi[i]);
    for(j=0;j<k;j++){
      isilist[imax] = i;
      imax += 1;
      if (imax > 10000)
	exit_error("MAKE_RANDOM_SARRAY_FROM_ISI","imax TOO LARGE");
    }
  }
  shuffle_iarray(isilist,imax,3,seed1);
  if (seed2 > 0) seed2 = -seed2;
  spike_util_ran2(&seed2);
  cnt = (int *)myalloc(n*sizeof(int));
  s = (int **)myalloc(n*sizeof(int *));
  tdata = (int *)myalloc(period*sizeof(int));
  for(i=0;i<n;i++){
    k = 0;
    t = 0;
    while(t<period){
      kisi = (int)(spike_util_ran2(&seed2) * (float)imax);
      if ((kisi < 0) || (kisi >= imax))
	exit_error("MAKE_RANDOM_SARRAY_FROM_ISI","kisi OUT OF RANGE");
      t += isilist[kisi];
      if (t < period){
	tdata[k] = t;
	k += 1;
      }
    }
    s[i] = copy_iarray(tdata,k);
    cnt[i] = k;
  }
  myfree(tdata);
  *rs = s;
  *rcnt = cnt;
}
void make_random_spikes_from_prob(prob,period,spikes,n,seed)
     float *prob;
     int period,**spikes,*n;
     int *seed;
{
  int i,k;
  int *tdata;
  tdata = (int *)myalloc(period*sizeof(int));
  k = 0;
  for(i=0;i<period;i++){
    if (prob[i] > spike_util_ran2(seed)){
      tdata[k] = i;
      k += 1;
    }
  }
  *spikes = copy_iarray(tdata,k);
  *n = k;
  myfree(tdata);
}
void make_random_spikes_from_prob_refract(prob,rsig,rmax,period,spikes,n,seed)
     float *prob,rsig,rmax;
     int period,**spikes,*n;
     int *seed;
{
  int i,k;
  int *tdata,nr,c,tr;
  float *r,p;
  discrete_gaussian(rsig,0.01,&r,&nr); 
  c = (nr-1)/2; 
  if (rmax > 1.0)
    rmax = 1.0;
  make_max_const_farray(r,nr,rmax);
  tdata = (int *)myalloc(period*sizeof(int));
  k = 0;
  tr = nr; 
  for(i=0;i<period;i++){
    p = prob[i];
    if ((c+tr) < nr)
      p *= 1.0 - r[c+tr];
    if (p > spike_util_ran2(seed)){
      tdata[k] = i;
      k += 1;
      tr = 0;
    }else
      tr += 1;
  }
  *spikes = copy_iarray(tdata,k);
  *n = k;
  myfree(tdata); myfree(r);
}
void make_random_sarray_from_prob(prob,period,spikes,count,n,seed)
     float *prob;
     int period,***spikes,**count,n;
     int seed;
{
  int i,j,k;
  int **s,*cnt,*tdata;
  if (seed > 0)
    seed = -seed;
  spike_util_ran2(&seed);
  cnt = (int *)myalloc(n*sizeof(int));
  s = (int **)myalloc(n*sizeof(int *));
  tdata = (int *)myalloc(period*sizeof(int));
  for(i=0;i<n;i++){
    k = 0;
    for(j=0;j<period;j++)
      if (prob[j] > spike_util_ran2(&seed)){
	tdata[k] = j;
	k += 1;
      }
    s[i] = copy_iarray(tdata,k);
    cnt[i] = k;
  }
  myfree(tdata);
  *spikes = s;
  *count = cnt;
}
void make_random_sarray_from_sarray(s,cnt,n,period,spikes,count,nrec,seed)
     int **s,*cnt,n,period,***spikes,**count,nrec;
     int seed;
{
  int i;
  int *pst;
  float *prob;
  pst = simple_pst_sarray(s,cnt,n,0,period);
  prob = (float *)myalloc(period*sizeof(float));
  for(i=0;i<period;i++)
    prob[i] = (float)pst[i]/(float)n;
  make_random_sarray_from_prob(prob,period,spikes,count,nrec,seed);
  myfree(pst); myfree(prob);
}
void make_poisson_float_spikes_from_prob(prob,n,sampling,pseed,rs,rn)
     float *prob;    
     int n;          
     float sampling; 
     int *pseed;     
     float **rs;     
     int *rn;        
{
  int i,j,k;
  int fn,*warp,t;
  float mu,lambda,*fs,*fint;
  fint = (float *)myalloc(n*sizeof(float));
  fint[0] = prob[0];
  for(i=1;i<n;i++)
    fint[i] = fint[i-1] + prob[i];
  multiply_farray(fint,n,1.0/fint[n-1]);
  j = 0;
  warp = (int *)myalloc(n*sizeof(int));
  for(i=0;i<n;i++){
    while (fint[j] <= ((float)i/(float)n))
      j += 1;
    warp[i] = j;
  }
  myfree(fint);
  mu = mean_farray(prob,n); 
  lambda = mu * sampling;
  k = 1;
  make_poisson_float_spikes(&fs,&fn,0,n,sampling,lambda,0.0,0.0,k,pseed);
  for(j=0;j<fn;j++){
    t = my_rint(fs[j]);
    if (t == n)  
      t = n-1;
    else if (t > n)
      exit_error("MAKE_POISSON_FLOAT_SPIKES_FROM_PROB","t too large");
    fs[j] = (float)warp[t];
  }
  myfree(warp);
  *rs = fs;
  *rn = fn;
}
void make_poisson_f_spikes_from_pr_refr(prob,n,sampling,pseed,refmu,refsig,
					rs,rn)
     float *prob;    
     int n;          
     float sampling; 
     int *pseed;     
     float refmu;    
     float refsig;   
     float **rs;     
     int *rn;        
{
  int i,j,k;
  int fn,*warp,t;
  float mu,lambda,*fs,*fint;
  fint = (float *)myalloc(n*sizeof(float));
  fint[0] = prob[0];
  for(i=1;i<n;i++)
    fint[i] = fint[i-1] + prob[i];
  multiply_farray(fint,n,1.0/fint[n-1]);
  j = 0;
  warp = (int *)myalloc(n*sizeof(int));
  for(i=0;i<n;i++){
    while (fint[j] < ((float)i/(float)n))
      j += 1;
    warp[i] = j;
  }
  myfree(fint);
  mu = mean_farray(prob,n); 
  lambda = mu * sampling;
  k = 1;
  make_poisson_float_spikes(&fs,&fn,0,n,sampling,lambda,refmu,refsig,k,pseed);
  for(j=0;j<fn;j++){
    t = my_rint(fs[j]);
    if (t == n)  
      t = n-1;
    else if (t > n)
      exit_error("MAKE_POISSON_FLOAT_SPIKES_FROM_PROB","t too large");
    fs[j] = (float)warp[t];
  }
  myfree(warp);
  *rs = fs;
  *rn = fn;
}
void make_poisson_spikes_from_prob(prob,n,sampling,pseed,rs,rn)
     float *prob;    
     int n;          
     float sampling; 
     int *pseed;     
     int **rs,*rn;   
{
  int *s,fn;
  float *fs;
  make_poisson_float_spikes_from_prob(prob,n,sampling,pseed,&fs,&fn);
  s = f2iarray(fs,fn);
  myfree(fs);
  *rs = s;
  *rn = fn;
}
void make_poisson_sarray_from_prob(prob,period,spikes,count,n,seed)
     float *prob;
     int period,***spikes,**count,n;
     int seed;
{
  int i,j,k;
  int **s,*cnt,*warp,t;
  float mu,lambda,sampling,*fst,*fint;
  fint = (float *)myalloc(period*sizeof(float));
  fint[0] = prob[0];
  for(i=1;i<period;i++)
    fint[i] = fint[i-1] + prob[i];
  multiply_farray(fint,period,1.0/fint[period-1]);
  j = 0;
  warp = (int *)myalloc(period*sizeof(int));
  for(i=0;i<period;i++){
    while (fint[j] < ((float)i/(float)period))
      j += 1;
    warp[i] = j;
  }
  myfree(fint);
  if (seed > 0)
    seed = -seed;
  spike_util_ran2(&seed);
  cnt = (int *)myalloc(n*sizeof(int));
  s = (int **)myalloc(n*sizeof(int *));
  k = 1;
  sampling = 1000.0;
  mu = mean_farray(prob,period); 
  lambda = mu * sampling;
  for(i=0;i<n;i++){
    make_poisson_float_spikes(&fst,&cnt[i],0,period,sampling,lambda,0.0,
			      0.0,k,&seed);
    for(j=0;j<cnt[i];j++){
      t = my_rint(fst[j]);
      fst[j] = (float)warp[t];
    }
    s[i] = f2iarray(fst,cnt[i]);
    myfree(fst);
  }
  myfree(warp);
  *spikes = s;
  *count = cnt;
}
void make_integrate_fire_spikes_from_farray(d,period,tau,vthresh,vthrel,eltau,
					    vreset,nmu,nsd,pseed,saveflag,
					    outfile,rs,rn)
     float *d; 
     int period;
     float tau,vthresh,vthrel,eltau,vreset,nmu,nsd;
     int saveflag;
     char outfile[];
     int *pseed,**rs,*rn;
{
  int i;
  int *t,*s,n;
  float v,vrest,dtau,deltau,thr;
  float *sav_v,*sav_thr;
  if (vthresh <= 0.0)
    exit_error("MAKE_INTEGRATE_FIRE_SPIKES_FROM_FARRAY","Vthresh <= 0.0");
  if (vreset >= vthresh)
    exit_error("MAKE_INTEGRATE_FIRE_SPIKES_FROM_FARRAY","Vreset >= Vthresh");
  if (vthrel < vthresh)
    exit_error("MAKE_INTEGRATE_FIRE_SPIKES_FROM_FARRAY","Vthrel < Vthresh");
  if (saveflag > 0){
    sav_v = (float *)myalloc(period*sizeof(float));
    sav_thr = (float *)myalloc(period*sizeof(float));
  }else
    sav_v = sav_thr = NULL; 
  if (tau <= 0.0)
    dtau = 1.0;
  else
    dtau = 1.0 - exp(-1.0/tau);
  if (eltau <= 0.0)
    deltau = 1.0;
  else
    deltau = 1.0 - exp(-1.0/eltau);
  if (*pseed > 0)
    *pseed *= -1;
  t = (int *)myalloc(period*sizeof(int));
  vrest = 0.0;
  v = vrest;
  n = 0;
  thr = vthresh;
  for(i=0;i<period;i++){
    v += d[i]; 
    if (nsd > 0.0) 
      v += su_gasdev(pseed)*nsd + nmu;
    if (saveflag > 0){
      sav_v[i] = v;
      sav_thr[i] = thr;
    }
    if (v > thr){
      t[n] = i;
      n += 1;
      v = vreset;
      thr = vthrel;
    }else{
      v -= dtau*(v-vrest);
      thr -= deltau*(thr-vthresh);
    }
  }
  s = copy_iarray(t,n);
  myfree(t);
  if (saveflag > 0){
    if (saveflag == 1)
      write_farray_plot(outfile,sav_v,period);
    else
      append_farray_plot(outfile,"v",sav_v,period,1);
    append_farray_plot(outfile,"vthr",sav_thr,period,1);
    myfree(sav_v); myfree(sav_thr);
  }
  *rs = s; *rn = n;
}
float *spikeu_poisson_g(tn,sampling,rate,seed,mask,maskn,rs,rsn)
     int tn;          
     float sampling;  
     float rate;      
     int seed;        
     float *mask;     
     int maskn;       
     float **rs;      
     int *rsn;        
{
  int n,*s;
  float *fs,*xs,*gt;
  if (rate > 0.0){
    if (seed > 0)
      seed = -seed;
    make_poisson_float_spikes(&fs,&n,0,tn,sampling,rate,0.0,0.0,1,&seed);
    s = f2iarray(fs,n);  
    xs = expand_spike_array(s,n,0,tn);  
    gt = convolve_with_mask_causal(xs,tn,mask,maskn);  
    myfree(s);
    if (rs != NULL){  
      *rs = fs;
      *rsn = n;
    }else{            
      myfree(fs);
    }
    myfree(xs);
  }else{
    gt = get_zero_farray(tn);
    if (rs != NULL){
      *rs = NULL;
      *rsn = 0;
    }
  }
  return gt;
}
float *spikeu_poisson_g_prob(prob,tn,sampling,seed,mask,maskn,rs,rsn)
     float *prob;     
     int tn;          
     float sampling;  
     int seed;        
     float *mask;     
     int maskn;       
     float **rs;      
     int *rsn;        
{
  int n,*s;
  float *fs,*xs,*gt;
  if (tn <= 0)
    exit_error("SPIKEU_POISSON_G_PROB","Data length error");
  if (maskn <= 0)
    exit_error("SPIKEU_POISSON_G_PROB","Mask length error");
  if (seed > 0)
    seed = -seed;
  make_poisson_float_spikes_from_prob(prob,tn,sampling,&seed,&fs,&n);
  s = f2iarray(fs,n);  
  xs = expand_spike_array(s,n,0,tn);  
  gt = convolve_with_mask_causal(xs,tn,mask,maskn);  
  if (rs != NULL){  
    *rs = fs;
    *rsn = n;
  }else{            
    myfree(fs);
  }
  myfree(xs);
  myfree(s);
  return gt;
}
float roc_sarrays(s1,cnt1,n1,s2,cnt2,n2,start,period)
     int **s1,*cnt1,n1,**s2,*cnt2,n2,start,period;
{
  int i;
  float *c1,*c2,roc;
  c1 = (float *)myalloc(n1*sizeof(float));
  c2 = (float *)myalloc(n2*sizeof(float));
  for(i=0;i<n1;i++)
    c1[i] = (float)count_spikes(s1[i],cnt1[i],start,period);
  for(i=0;i<n2;i++)
    c2[i] = (float)count_spikes(s2[i],cnt2[i],start,period);
  roc = roc_farrays(c1,c2,n1,n2);
  myfree(c1); myfree(c2);
  return roc;
}
void pick_cross_spikes(s1,cnt1,s2,cnt2,n,start,period,lag,width,rs,rcnt)
     int **s1,*cnt1,**s2,*cnt2,n,start,period;
     int lag,width; 
     int ***rs,**rcnt; 
{
  int i,j;
  int **s,*cnt,*ts,tcnt,t;
  s = (int **)myalloc(n*sizeof(int *));
  cnt = (int *)myalloc(n*sizeof(int));
  ts = (int *)myalloc(period*sizeof(int));
  for(i=0;i<n;i++){
    cnt[i] = 0;
    for(j=0;j<cnt1[i];j++){ 
      t = s1[i][j];
      tcnt = count_spikes(s2[i],cnt2[i],t+lag,width);
      if (tcnt > 0){
	ts[cnt[i]] = t;
	cnt[i] += 1;
      }
    }
    s[i] = extract_spikes(ts,cnt[i],start,period,&t,0);
    if (t!=cnt[i])
      exit_error("PICK_CROSS_SPIKES","Counts not equal");
  }
  myfree(ts);
  *rs = s;
  *rcnt = cnt;
}
void get_stationarity_sarray(s,cnt,n,min,max)
     int **s,*cnt,n;
     float *min,*max;
{
  float *fcnt,*smooth,sigma;
  if (n < 2)
    exit_error("GET_STATIONARITY_SARRAY","Too few trials");
  sigma = 10.0; 
  fcnt = i2farray(cnt,n);
  smooth = smooth_with_gaussian(fcnt,n,sigma,0.01);
  *min = min_of_farray(smooth,n);
  *max = max_of_farray(smooth,n);
  myfree(fcnt); myfree(smooth);
}
void spike_word_recode_sarray(s,cnt,n,start,period,nlet,dt,binflag,shiftflag,
			      rswi,rwpt)
     int **s,*cnt,n,start,period;
     int nlet;  
     int dt;    
     int binflag; 
     int shiftflag; 
     int ***rswi;   
     int *rwpt;     
{
  int i,j,k,l;
  int *x,wpt,wlen,*wd,base0,base,wi;
  int **swi;
  if (binflag != 1)
    exit_error("SPIKE_WORD_RECODE_SARRAY","Must set binflag to 1");
  base0 = 1;
  for(i=0;i<nlet;i++)
    base0 *= 2;
  wlen = nlet * dt;         
  wpt = period - wlen + 1;  
  swi = get_zero_2d_iarray(n,wpt);
  wd = (int *)myalloc(nlet*sizeof(int));
  for(i=0;i<n;i++){ 
    x = expand_spike_int_array(s[i],cnt[i],start,period);
    for(j=0;j<wpt;j++){ 
      for(k=0;k<nlet;k++){ 
	wd[k] = 0;
	for(l=0;l<dt;l++){
	  wd[k] += x[j+k*dt+l];
	}
      }
      if (binflag == 1){
	if (shiftflag == 0){
	  for(k=0;k<nlet;k++)
	    if (wd[k] > 1)
	      wd[k] = 1;
	}else{
	  exit_error("SPIKE_WORD_RECODE_SARRAY","Must set shiftflag to 0");
	}
      }
      wi = 0; 
      base = base0/2;
      for(k=0;k<nlet;k++){
	wi += wd[k] * base;
	base /= 2;
      }
      if (wi >= base0){
	exit_error(" **** HERE HERE","wi > base0");
      }
      swi[i][j] = wi;
    }
    myfree(x);
  }
  myfree(wd);
  *rswi = swi; *rwpt = wpt;
}
void spike_word_distrib_sarray(s,cnt,n,start,period,nlet,dt,binflag,shiftflag,
			       rp,rpn)
     int **s,*cnt,n,start,period;
     int nlet;  
     int dt;    
     int binflag; 
     int shiftflag; 
     float **rp;
     int *rpn;
{
  int i,j,k;
  int *x,xn,ntrunc,nshift,tot_trunc,tot_shift,np,base,wi,nw;
  float *p;
  if (binflag != 1)
    exit_error("SPIKE_WORD_DISTRIB_SARRAY","Must set binflag to 1");
  np = 1;
  for(i=0;i<nlet;i++)
    np *= 2;
  p = get_zero_farray(np);
  tot_trunc = tot_shift = 0;
  for(i=0;i<n;i++){
     x = histogram_expand_sarray(s[i],cnt[i],start,period,dt,binflag,
       shiftflag,&xn,&ntrunc,&nshift);
    tot_trunc += ntrunc;
    tot_shift += nshift;
    nw = xn - (nlet - 1);
    for(j=0;j<nw;j++){ 
      wi = 0; 
      base = np/2;
      for(k=0;k<nlet;k++){
	wi += x[j+k] * base;
	base /= 2;
      }
      if (wi >= np){
	exit_error(" **** HERE HERE","wi > np");
      }
      p[wi] += 1.0;
    }
    myfree(x);
  }
  if (binflag == 1){
    printf("      Total bins truncated:  %d\n",tot_trunc);
    if (shiftflag == 1)
      printf("      Total spikes shifted:  %d\n",tot_shift);
  }
  *rp = p; *rpn = np;
}
void stim_condition_word_distrib_sarray(s,cnt,n,start,period,shiftflag,binflag,
					stim,stimn,stimt0,stimdt,
					t0,tn,sord,swn,rord,rwn,rdt,
					r_pr,r_sn,r_rn)
     int **s,*cnt,n,start,period;  
     int shiftflag,binflag;
     int **stim,*stimn;  
     int stimt0; 
     float stimdt; 
     int t0,tn;  
     int sord;   
     int swn;    
     int rord;   
     int rwn;    
     int rdt;    
     float ****r_pr;  
     int *r_sn,*r_rn; 
{
  int i,j,k;
  int si,ri,sn,rn,wpt,tw0,**rwi,rwin;
  float ***pr;
  printf("  STIM_CONDITION_WORD_DISTRIB_SARRAY\n");
  printf("    %d trials, start %d, period %d\n",n,start,period);
  printf("    Stimulus onset relative to spike t0:  %d\n",stimt0);
  printf("    Stimulus frame exact dt:  %f\n",stimdt);
  spike_word_recode_sarray(s,cnt,n,start,period,rwn,rdt,binflag,shiftflag,
			   &rwi,&rwin);
  if (rord != 2)
    exit_error("STIM_CONDITION_WORD_DISTRIB_SARRAY","Response must be binary");
  sn = 1; 
  for(i=0;i<swn;i++)
    sn *= sord;
  rn = 1; 
  for(i=0;i<rwn;i++)
    rn *= rord;
  pr = get_zero_3d_farray(tn,sn,rn);
  printf("    Probability matrix is  %d t  x  %d stim  x  %d resp\n",tn,sn,rn);
  for(i=0;i<n;i++){ 
    wpt = stimn[i];
    for(j=0;j<wpt;j++){ 
      si = stim[i][j]; 
      if ((si < 0)|| (si >= sn))
	exit_error("STIM_CONDITION_WORD_DISTRIB_SARRAY","si out of range");
      tw0 = stimt0 + my_rint((float)j*stimdt); 
      tw0 += t0; 
      for(k=0;k<tn;k++){ 
	if (tw0+k < rwin){ 
	  ri = rwi[i][tw0+k];
	  if ((ri < 0)|| (ri >= rn))
	    exit_error("STIM_CONDITION_WORD_DISTRIB_SARRAY","ri out of range");
	  pr[k][si][ri] += 1.0; 
	}
      }
    }
  }
  free_2d_iarray(rwi,n);
  *r_pr = pr; *r_sn = sn; *r_rn = rn;
}
void spike_util_sta_r_v_stim_sarray(s,cnt,n,start,period,sampling,
				    stim,stimn,stimt0,stscale,
				    sta,stan,toff,stn,normstim,normr,
				    outfile,nbin,
				    mindata,dump_nbin,dump_stscale,write_n,
				    var_flag,sigsm,cond,
				    rxx,ryy,rnn)
     int **s,*cnt,n;     
     int start,period;   
     float sampling;  
     float **stim;  
     int *stimn;    
     int stimt0;    
     float stscale; 
     float *sta;    
     int stan;      
     int toff;      
     int stn;       
     int normstim;  
     int normr;     
     char outfile[]; 
     int nbin;       
     int mindata;    
     int dump_nbin;  
     float dump_stscale;  
     int write_n;    
     int var_flag;   
     float sigsm;    
     struct spike_util_cond_struct *cond;    
     float **rxx,**ryy;  
     int *rnn;           
{
  int i,j,k,t;
  int sn,*rvsn,t1,st0,maxi,nmax,cflag,tc0,ccn0,ccn1,t0;
  float min,max,x,y,sr,srsd,mean,sd,*sx,*smooth;
  float *conv,*star,*rvs,**rvs2d,**rvs2dx,*rx,*xx,*yy,*yn,*zstim;
  char name[SLEN],tstr[SLEN],tstr1[SLEN];
  int run_flag,ndump;
  ndump = 200;
  printf("    SPIKE_UTIL_STA_R_V_STIM_SARRAY\n");
  printf("      Stimulus scaling:  %f\n",stscale);
  printf("      STA is %d long\n",stan);
  printf("      Stim is %d long\n",stimn[0]);
  printf("      %d trials\n",n);
  printf("      stimt0 = %d\n",stimt0);
  rvs = get_zero_farray(nbin);
  rvsn = get_zero_iarray(nbin);
  if (var_flag > 0){
    nmax = stimn[0] * n;  
    rvs2d = get_2d_farray(nbin,nmax);
    rvs2dx = get_2d_farray(nbin,nmax);
  }
  star = copy_farray(sta,stan);
  reverse_farray(star,stan);
  if (normstim == 3){  
    norm_area_farray(star,stan,1.0);
  }
  t0 = start - stimt0;
  t1 = t0 + period;
  if (t0 < 0){
    t0 = 0;  
  }
  ccn0 = ccn1 = 0;
  cond->nsuc = cond->nfail = 0;
  for(i=0;i<n;i++){ 
    sn = stimn[i];
    if (t1 > (sn-40)){ 
      printf("  t1 = %d  sn-40 = %d\n",t1,sn-40);
      printf("  *** Decrease 'rvs_period', currently %d\n",period);
      exit_error("SPIKE_UTIL_STA_R_V_STIM_SARRAY","Too close to stim end");
    }
    conv = convolve_with_mask_causal(stim[i],sn,star,stan);
    if (normstim == 1){
      printf("    z-scoring stimulus %d\n",i);
      z_score_farray(conv,sn);
      if (cond->ctype != NULL){
	if (strcmp(cond->data,"stim")==0){
	  zstim = copy_farray(stim[i],sn);
	  printf(" WYETH HERE STIM STIM STIM STIM\n");
	  z_score_farray(zstim,sn);
	}
      }
    }else if (normstim == 2){
      z_score_farray(conv,sn);
      multiply_farray(conv,sn,stscale);
      mean_sdev_farray(conv,sn,&mean,&sd);
      printf("WYETH HERE Normstim = 2 ======>  mean = %f  SD = %f\n",mean,sd);
      if (cond->ctype != NULL){
	if (strcmp(cond->data,"stim")==0){
	  zstim = copy_farray(stim[i],sn);
	  z_score_farray(zstim,sn);
	}
      }
    }else
      multiply_farray(conv,sn,stscale);
    if (sigsm > 0.0){
      sx = expand_spike_array(s[i],cnt[i],start,period);
      smooth = smooth_with_gaussian(sx,period,sigsm,0.01);
      myfree(sx);
      sx = smooth;
      multiply_farray(sx,period,sampling);
    }
    if (i==0){
      get_min_max_farray(conv,sn,&min,&max);
      printf("      min max = %f %f\n",min,max);
      min *= 1.1;
      max *= 1.1;
    }
    run_flag = 0; 
    for(t=t0;t<t1;t++){  
      x = conv[t];
      st0 = t + stimt0 + toff;
      if (sigsm > 0.0)
	y = sx[st0];    
      else
	y = mean_spike_rate(s[i],cnt[i],st0,stn,sampling); 
      k = (int)((x-min)/(max-min) * (float)nbin);
      if (cond->ctype != NULL){
	tc0 = t + cond->win0;
	if ((tc0 >= 0) && (tc0+cond->winn <= sn)){
	  if (strcmp(cond->data,"stim")==0){
	    cflag = varcond_check_condition(cond,tc0,zstim,sn,outfile);
	  }else{
	    cflag = varcond_check_condition(cond,tc0,conv,sn,outfile);
	  }
	  if (cflag){
	    ccn1 += 1;
	    cond->nsuc += 1;
	  }else{
	    ccn0 += 1;
	    cond->nfail += 1;
	  }
	}else
	  cflag = 0;
      }
      if (k == -1)
	k = 0;
      else if (k == nbin)
	k = nbin-1;
      if (cflag == 0){
	;  
      }else if ((k < 0) || (k >= nbin)){
	printf("*** k = %d  (outside of bin range)\n",k);
      }else{
	rvs[k] += y;
	if (var_flag > 0){
	  rvs2d[k][rvsn[k]] = y;
	  rvs2dx[k][rvsn[k]] = x;
	}
	rvsn[k] += 1;
	if ((dump_nbin > 0) && (stscale == dump_stscale) &&
	    (k >= (nbin-dump_nbin))){
	  if (run_flag == 0){
	    printf("  Start of run, trial %d, k %d, time %d\n",i,k,st0);
	    run_flag = 1;     
	    append_farray_plot_offset("zzz.rvs.dump.pl","stim",
				      &(stim[i][t-160]),ndump,-160);
	    append_xplot_spike_sampling("zzz.rvs.dump.pl","spikes",s[i],cnt[i],
					t-160,ndump+160,2,sampling,t);
	  }
	}else{
	  run_flag = 0;
	}
      }
    }
    if (sigsm > 0.0)
      myfree(sx);
    myfree(conv);
  }
  if (cond->ctype != NULL){
    printf("      Condition satisfied: %7d\n",ccn1);
    printf("      Condition failed:    %7d\n",ccn0);
    printf("        Success rate =  %.2f %%\n",100.0*
	   (float)ccn1/((float)(ccn1+ccn0)));
    cond->fsuc = (float)cond->nsuc/((float)(cond->nsuc+cond->nfail));
  }
  sr = mean_spike_rate_sarray(s,cnt,n,start,period,sampling,&srsd);
  rx = (float *)myalloc(nbin*sizeof(float));
  for(i=0;i<nbin;i++){
    rx[i] = min + (max-min)/(float)nbin * (float)(i+0.5);
    if (rvsn[i] > 0){
      rvs[i] /= (float)rvsn[i];
      if (normr == 1)
	rvs[i] /= sr;
    }else
      rvs[i] = -1.0;
  }
  xx = (float *)myalloc(nbin*sizeof(float));
  yy = (float *)myalloc(nbin*sizeof(float));
  yn = (float *)myalloc(nbin*sizeof(float));
  k = 0;
  for(i=0;i<nbin;i++){
    if (rvsn[i] >= mindata){
      xx[k] = rx[i];
      if (rvs[i] > 0.0001)  
	yy[k] = rvs[i];
      else
	yy[k] = 0.0001;
      yn[k] = (float)rvsn[i];
      k += 1;
      maxi = i;
    }
  }
  printf("      Bin %d is highest with > %d entries.\n",maxi,mindata);
  sprintf(name,"scale_%.2f",stscale);
  if (var_flag == 1){        
    sprintf(tstr,"%s %d\n",name,nbin);
    append_string_to_file(outfile,tstr);
    for(i=0;i<nbin;i++){
      mean_sdev_farray(rvs2d[i],rvsn[i],&mean,&sd);
      sprintf(tstr,"%3d %5d %6.2f %8.4f\n",i,rvsn[i],mean,sd);
      append_string_to_file(outfile,tstr);
    }
  }else if ((var_flag == 4)||(var_flag == 5)){  
    FILE *fout,*fopen();
    if ((fout = fopen(outfile,"a"))==NULL)
      exit_error("SPIKE_UTIL_STA_R_V_STIM_SARRAY","Cannot open file");
    fprintf(fout,"/newplot\n");
    fprintf(fout,"/plotname %s\n",name);
    if (var_flag == 4){
      for(i=0;i<nbin;i++){
	for(j=0;j<rvsn[i];j++){
	  fprintf(fout,"%d %.2f\n",i,rvs2d[i][j]);
	}
      }
    }else{
      for(i=0;i<nbin;i++){
	for(j=0;j<rvsn[i];j++){
	  fprintf(fout,"%.2e %.2f\n",rvs2dx[i][j],rvs2d[i][j]);
	}
      }
    }
    fclose(fout);
  }else{
    append_farray_xy_plot(outfile,xx,yy,k,name);
    if (write_n == 1){
      sprintf(tstr,"N_%.2f",stscale);
      append_farray_xy_plot(outfile,xx,yn,k,tstr);
    }
  }
  if (var_flag > 0){
    free_2d_farray(rvs2d,nbin);
    free_2d_farray(rvs2dx,nbin);
  }
  myfree(rx); myfree(yn);
  *rxx = xx; *ryy = yy; *rnn = k;
}
void spike_util_sta_r_v_stim_bayes_sarray(s,cnt,n,start,period,sampling,
					  stim,stimn,stimt0,stscale,
					  sta,stan,toff,stn,t0,normstim,normr,
					  outfile,nbin,mindata,
					  rxx,ryy,rnn)
     int **s,*cnt,n,start,period;  
     float sampling;  
     float **stim;  
     int *stimn;    
     int stimt0;    
     float stscale; 
     float *sta;    
     int stan;      
     int toff;      
     int stn;       
     int t0;        
     int normstim;  
     int normr;     
     char outfile[]; 
     int nbin;       
     int mindata;    
     float **rxx,**ryy;  
     int *rnn;           
{
  int i,j,k,t;
  int sn,*rvsn,t1,psn,psgsn;
  float min,max,x,y,sr,srsd,mean,sdev;
  float *conv,*star,*ps,*psgs,*rvs,*rx,*xx,*yy;
  char name[SLEN];
  printf("    SPIKE_UTIL_STA_R_V_STIM_BAYES_SARRAY\n");
  printf("      Stimulus scaling:  %f\n",stscale);
  printf("      STA is %d long\n",stan);
  printf("      Stim is %d long\n",stimn[0]);
  printf("      %d trials\n",n);
  printf("      stimt0 = %d\n",stimt0);
  ps   = get_zero_farray(nbin);  
  psn  = 0;
  psgs = get_zero_farray(nbin);  
  psgsn = 0;
  rvs  = get_zero_farray(nbin);
  rvsn = get_zero_iarray(nbin);
  star = copy_farray(sta,stan);
  reverse_farray(star,stan);
  norm_area_farray(star,stan,1.0);
  for(i=0;i<n;i++){ 
    sn = stimn[i];
    t1 = sn - 40;  
    conv = convolve_with_mask_causal(stim[i],sn,star,stan);
    if (normstim == 1)
      z_score_farray(conv,sn);
    else
      multiply_farray(conv,sn,stscale);
    if (i==0){
      get_min_max_farray(conv,sn,&min,&max);
      printf("  min max = %f %f\n",min,max);
      min *= 1.1;
      max *= 1.1;
    }
    for(t=t0;t<t1;t++){
      x = conv[t];
      k = (int)((x-min)/(max-min) * (float)nbin);
      if ((k < 0) || (k >= nbin)){
	printf("k = %d\n",k);
      }else{
	ps[k] += 1.0;
	psn += 1;
      }
    }
    for(j=0;j<cnt[i];j++){
      t = s[i][j] - stimt0 - toff;
      if ((t >= t0) && (t < t1)){
	x = conv[t];
	k = (int)((x-min)/(max-min) * (float)nbin);
	if ((k < 0) || (k >= nbin)){
	  printf("k = %d\n",k);
	}else{
	  psgs[k] += 1.0;
	  psgsn += 1;
	}
      }
    }
    myfree(conv);
  }
  mean_sdev_freq_hist_farray(ps,nbin,&mean,&sdev);
  printf("  mean = %f  sdev = %f\n",mean,sdev);
  multiply_farray(ps,nbin,1.0/(float)psn); 
  append_farray_plot(outfile,"ps",ps,nbin,1);
  multiply_farray(psgs,nbin,1.0/(float)psgsn); 
  append_farray_plot(outfile,"psgs",psgs,nbin,1);
  sr = mean_spike_rate_sarray(s,cnt,n,start,period,sampling,&srsd);
  rx = (float *)myalloc(nbin*sizeof(float));
  for(i=0;i<nbin;i++){
    if (ps[i] > 0.0){
      rvs[i] = sr * psgs[i] / ps[i];
      if (normr == 1)  
	rvs[i] /= sr;
    }else
      rvs[i] = -1.0;
    rx[i] = min + (max-min)/(float)nbin * (float)(i+0.5);
  }
  xx = (float *)myalloc(nbin*sizeof(float));
  yy = (float *)myalloc(nbin*sizeof(float));
  k = 0;
  for(i=0;i<nbin;i++){
    if (ps[i] > 0.0){
      xx[k] = rx[i];
      yy[k] = rvs[i];
      k += 1;
    }
  }
  sprintf(name,"scale_%.2f",stscale);
  append_farray_xy_plot(outfile,xx,yy,k,name);
  myfree(rx); myfree(rvs); myfree(ps); myfree(psgs);
  *rxx = xx; *ryy = yy; *rnn = k;
}
