/*C  (C) Copyright  Wyeth Bair 1993-2011                                     */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "my_util.h"
#include "nr_util.h"
#include "carray_util.h"
#include "iarray_util.h"
#include "farray_util.h"
#include "func_util.h"
#include "myrand_util.h"
#include "data_util.h"
#include "stmh.h"
#include "stm_pasushape.h"   
int glob_tot = 0;
int stm_get_tsamp(tscale,stim_samp)     
     double tscale;     
     double stim_samp;  
{
  int tsamp;
  tsamp = my_rint(1.0/(tscale*stim_samp)); 
  if (tsamp <= 0)
    tsamp = 1; 
  return tsamp;
}
int stm_get_ti0(tn,tscale,st0,tsamp)
     int tn;         
     double tscale;  
     double st0;     
     int tsamp;      
{
  int t0i,t0_patti;
  double sim_dur_sec;
  double dt_patt_sec;
  sim_dur_sec = tn * tscale;  
  if (st0 < 0.0)
    exit_error("STM_GET_TI0","Stimulus start time, 'st0', is negative");
  if (st0 >= sim_dur_sec)
    exit_error("STM_GET_TI0","Stimulus start time exceeds simulation time");
  dt_patt_sec = tsamp * tscale;
  t0_patti = my_rint(st0 / dt_patt_sec);
  t0i = t0_patti * tsamp;
  return t0i;
}
int stm_get_t1i(tn,tscale,st0,stn,tsamp)
     int tn;         
     double tscale;  
     double st0;     
     double stn;     
     int tsamp;      
{
  int t0i,t1i,dur_raw_ideal;
  t0i = stm_get_ti0(tn,tscale,st0,tsamp);
  dur_raw_ideal = my_rint(stn / tscale);
  t1i = t0i + dur_raw_ideal;
  if (t1i >= tn)
    t1i = tn - 1;
  return t1i;
}
int stm_get_n(tn,tscale,st0,stn,dwell,tsamp)
     int tn;         
     double tscale;  
     double st0;     
     double stn;     
     int dwell;      
     int tsamp;      
{
  int n;
  int t0i,t1i,dur_raw;
  t0i = stm_get_ti0(tn,tscale,st0,tsamp);
  t1i = stm_get_t1i(tn,tscale,st0,stn,tsamp);
  dur_raw = t1i - t0i + 1;
  n = (int)(dur_raw / (tsamp * dwell));
  if (n <= 0)
    exit_error("STIM_GET_N","Number of unique stimulus frames is <= 0");
  return n;
}
int stm_util_count_new_frames(zn,tscale,t0,tn,tsamp,dwell)
     int zn;
     float tscale,t0,tn;
     int tsamp;             
     int dwell;
{
  int i;
  int cnt,dn;
  float t;
  cnt = 0;
  dn = dwell;
  for(i=1;i<=zn;i++){
    t = (float)(i-1)*tscale;
    if (((i-1)%tsamp) == 0){
      if ((t >= t0) && (t < t0+tn)){
	if (dn == dwell){
	  cnt += 1;
	  dn = 0;
	}
	dn += 1;
      }
    }
  }
  return cnt;
}
void stm_get_ran_seq(mylogf,distrib,ord,seed,fps,stn,dt,v1,v2,v3,rs,rsn)
     char *mylogf;  
     int distrib;   
     int ord;       
     int seed;      
     float fps;     
     float stn;     
     int dt;        
     float v1,v2;   
     float v3;      
     float **rs;    
     int *rsn;      
{
  int i;
  int *seqi,seqn;
  float *seqf;
  if (distrib == 2){
    seqn = (int)(stn * fps / (float)dt);
    printf("    seqn     %d\n",seqn);
    seqi = myrand_get_std_bin_seq(seqn,seed);
  }else if (distrib == 3){
    seqn = (int)(stn * fps / (float)dt);
    seqi = myrand_get_std_tern_seq(seqn,seed);
  }else if (distrib == 7){
    seqn = 1;
    for(i=0;i<ord;i++)
      seqn *= 2;
    seqi = myrand_get_std_mseq(seqn,seed,ord);
  }else
    mylog_exit(mylogf,"STM_GET_RAN_SEQ  Unknown distrib value");
  seqf = (float *)myalloc(seqn*sizeof(float));
  if (distrib == 3){
    for(i=0;i<seqn;i++){
      if (seqi[i] == -1)   
	seqf[i] = v1;
      else if (seqi[i] == 1)   
	seqf[i] = v2;
      else
	seqf[i] = v3;   
    }
  }else{
    for(i=0;i<seqn;i++){
      if (seqi[i] == 1)   
	seqf[i] = v1;
      else
	seqf[i] = v2;
    }
  }
  myfree(seqi);
  *rs  = seqf;  
  *rsn = seqn;  
}
void stm_get_ran_seq_2d(mylogf,n,distrib,seed,fps,stn,dt,v1,v2,v3,rs,rsn)
     char *mylogf;  
     int n;         
     int distrib;   
     int seed;      
     float fps;     
     float stn;     
     int dt;        
     float v1,v2;   
     float v3;      
     float ***rs;   
     int *rsn;      
{
  int i,j,k;
  int *seqi,seqn;
  float **seqf;
  seqn = 1 + (int)(stn * fps / (float)dt);
  if (distrib == 2){
    seqi = myrand_get_std_bin_seq(n*seqn,seed);
  }else if (distrib == 3){
    seqi = myrand_get_std_tern_seq(n*seqn,seed);
  }else{
    mylog_exit(mylogf,"STM_GET_RAN_SEQ_2D  Unknown distrib value");
  }
  seqf = get_2d_farray(seqn,n);
  if (distrib == 3){
    k = 0;
    for(i=0;i<seqn;i++){
      for(j=0;j<n;j++){
	if (seqi[k] == -1)   
	  seqf[i][j] = v1;
	else if (seqi[k] == 1)   
	  seqf[i][j] = v2;
	else
	  seqf[i][j] = v3;   
	k += 1;
      }
    }
  }else{
    k = 0;
    for(i=0;i<seqn;i++){
      for(j=0;j<n;j++){
	if (seqi[k] == 1)   
	  seqf[i][j] = v1;
	else
	  seqf[i][j] = v2;
	k += 1;
      }
    }
  }
  myfree(seqi);
  *rs  = seqf;  
  *rsn = seqn;  
}
void stm_shape_fill_recursive(data,xn,yn,i,j,sval)
     float **data;       
     int xn,yn;          
     int i,j;            
     float sval;         
{
  glob_tot += 1;
  if (glob_tot > 180000){
    printf("  *** The shape is too large in number of pixels\n");
    exit_error("STM_SHAPE_FILL_RECURSIVE","Too many recursive calls");
  }
  data[i][j] = sval;
  if (i+1 < xn)
    if (data[i+1][j] != sval)
      stm_shape_fill_recursive(data,xn,yn,i+1,j,sval);
  if (i > 0)
    if (data[i-1][j] != sval)
      stm_shape_fill_recursive(data,xn,yn,i-1,j,sval);
  if (j+1 < yn)
    if (data[i][j+1] != sval)
      stm_shape_fill_recursive(data,xn,yn,i,j+1,sval);
  if (j > 0)
    if (data[i][j-1] != sval)
      stm_shape_fill_recursive(data,xn,yn,i,j-1,sval);
}
void stm_shape_fill_ring_recur(data,xn,yn,i,j,sval)
     float **data;       
     int xn,yn;          
     int i,j;            
     float sval;         
{
  glob_tot += 1;
  if (glob_tot > 180000)
    return;
	data[i][j] = sval;
	if (i+1 < xn)
	  if (data[i+1][j] != sval)
	    stm_shape_fill_recursive(data,xn,yn,i+1,j,sval);
	if (i > 0)
	  if (data[i-1][j] != sval)
	    stm_shape_fill_recursive(data,xn,yn,i-1,j,sval);
	if (j+1 < yn)
	  if (data[i][j+1] != sval)
	    stm_shape_fill_recursive(data,xn,yn,i,j+1,sval);
	if (j > 0)
	  if (data[i][j-1] != sval)
	    stm_shape_fill_recursive(data,xn,yn,i,j-1,sval);
}
void stm_shape_fill_nonrecur(data,xn,yn,i,j,sval)
     float **data;       
     int xn,yn;          
     int i,j;            
     float sval;         
{
  int done;
  done = 0;
  while(done == 0){
    if ((i >= 0) && (i < xn)){
      if ((j >= 0) && (j < yn)){
	if (data[i][j] != sval){
	  data[i][j] = sval;
	  stm_shape_fill_recursive(data,xn,yn,i+1,j,sval);
	  stm_shape_fill_recursive(data,xn,yn,i-1,j,sval);
	  stm_shape_fill_recursive(data,xn,yn,i,j+1,sval);
	  stm_shape_fill_recursive(data,xn,yn,i,j-1,sval);
	}
      }
    }
  }
}
void stm_shape_draw_contour(data,xn,yn,sx,sy,sn,size,xc,yc,sval)
     float **data;       
     int xn,yn;          
     float *sx,*sy;      
     int sn;             
     float size;         
     float xc,yc;        
     float sval;         
{
  int i;
  int xi,yi;
  float fx,fy;
  for(i=0;i<sn;i++){  
    fx = sx[i]*size + xc;  
    fy = sy[i]*size + yc;  
    xi =  my_rint(fx);     
    yi =  my_rint(fy);
    if ((xi >= 0) && (xi < xn) && (yi >= 0) && (yi < yn))
      data[xi][yi] = sval;
  }
}
void stm_shape_draw_contour_antialias(data,xn,yn,sx,sy,sn,size,xc,yc,sval,
				      aa_sd)
     float **data;       
     int xn,yn;          
     float *sx,*sy;      
     int sn;             
     float size;         
     float xc,yc;        
     float sval;         
     float aa_sd;        
{
  int i,j,k;
  int aa_rad,xi,yi,x0,x1,y0,y1;
  float fx,fy,gc,dx,dy,gx,gy,g;
  aa_rad = my_rint(aa_sd * 2.0);    
  if (aa_rad < 1)
    aa_rad = 1;
  gc = -1.0/(2.0 * aa_sd * aa_sd);  
  for(i=0;i<sn;i++){  
    fx = sx[i]*size + xc;  
    fy = sy[i]*size + yc;  
    xi =  my_rint(fx);  
    yi =  my_rint(fy);
    x0 = xi - aa_rad;
    x1 = xi + aa_rad;
    if (x0 < 0)
      x0 = 0;
    if (x1 >= xn)
      x1 = xn-1;
    y0 = yi - aa_rad;
    y1 = yi + aa_rad;
    if (y0 < 0)
      y0 = 0;
    if (y1 >= yn)
      y1 = yn-1;
    for(j=x0;j<=x1;j++){
      dx = (float)j - fx;  
      gx = exp(gc * dx*dx);  
      for(k=y0;k<=y1;k++){
	dy = (float)k - fy;  
	gy = exp(gc * dy*dy);   
	g = gx * gy * sval;    
	if (g > data[j][k]){  
	  data[j][k] = g;     
	}
      }
    }
  }
}
void stm_shape_draw_fill(mylogf,data,xn,yn,sx,sy,sn,size,xc,yc,sval,bgval,
			 x0,y0,aa_sd)
     char *mylogf;       
     float **data;       
     int xn,yn;          
     float *sx,*sy;      
     int sn;             
     float size;         
     float xc,yc;        
     float sval;         
     float bgval;        
     float x0,y0;        
     float aa_sd;        
{
  int i,j;
  int xin,yin,cnt,si,sj,xci,yci;
  float txc,tyc,**td;
  xin = my_rint(x0*size + xc);
  yin = my_rint(y0*size + yc);
  if ((yin < 0) || (xin < 0)  || (xin >= xn) ||  (yin >= yn)){
    xci = (int)(xc - (xn-1)/2.0);          
    yci = (int)(yc - (yn-1)/2.0);
    txc = xc - (float)xci;  
    tyc = yc - (float)yci;
    xin = my_rint(x0*size + txc);
    yin = my_rint(y0*size + tyc);
    td = get_const_2d_farray(xn,yn,bgval);
    stm_shape_draw_contour(td,xn,yn,sx,sy,sn,size,txc,tyc,sval);
    glob_tot = 0;
    stm_shape_fill_recursive(td,xn,yn,xin,yin,sval);
    for(i=0;i<xn;i++){
      si = i + xci;
      if ((si >= 0) && (si < xn)){
	for(j=0;j<yn;j++){
	  sj = j + yci;
	  if ((sj >= 0) && (sj < yn)){
	    data[si][sj] = td[i][j];
	  }
	}
      }
    }
    free_2d_farray(td,xn);
  }else{
    stm_shape_draw_contour(data,xn,yn,sx,sy,sn,size,xc,yc,sval);
    glob_tot = 0;
    stm_shape_fill_recursive(data,xn,yn,xin,yin,sval);
  }
  for(i=0;i<xn;i++){
    for(j=0;j<yn;j++){
      if (data[i][j] == bgval){
	cnt = 0;
	if (i > 0)
	  if (data[i-1][j] == sval)
	    cnt += 1;
	if (i < (xn-1))
	  if (data[i+1][j] == sval)
	    cnt += 1;
	if (j > 0)
	  if (data[i][j-1] == sval)
	    cnt += 1;
	if (j < (yn-1))
	  if (data[i][j+1] == sval)
	    cnt += 1;
	if (cnt == 4)
	  data[i][j] = sval;  
      }
    }
  }
  if (aa_sd > 0.0){  
    stm_shape_draw_contour_antialias(data,xn,yn,sx,sy,sn,size,xc,yc,sval,
				     aa_sd);
  }
}
float **stm_noise_get_grid(xn,yn,gw,gh,pixw,pixh,x0,y0,gdata,bgval)
     int xn,yn;     
     int gw,gh;     
     int pixw,pixh; 
     int x0,y0;     
     float **gdata; 
     float bgval;
{
  int i,j;
  int i0,i1,j0,j1,gdi,gdj,gdi0,gdj0,bx0,by0,iwrap,jwrap,jwrap0;
  float **g,gval;
  g = get_const_2d_farray(xn,yn,bgval);
  i0 = x0;
  i1 = x0 + gw*pixw - 1;
  if (i0 < 0){
    i0 = 0;
    iwrap = (i0-x0) % pixw;  
    gdi0  = (i0-x0) / pixw;
  }else{
    gdi0 = 0;
    iwrap = 0;
  }
  if (i1 >= xn)
    i1 = xn-1;
  j0 = y0;
  j1 = y0 + gh*pixh - 1;
  if (j0 < 0){
    j0 = 0;
    jwrap0 = (j0-y0) % pixh;
    gdj0   = (j0-y0) / pixh;
  }else{
    gdj0 = 0;
    jwrap0 = 0;
  }
  if (j1 >= yn)
    j1 = yn-1;
  gdi = gdi0;
  for(i=i0;i<=i1;i++){
    gdj = gdj0;
    gval = gdata[gdi][gdj];
    jwrap = jwrap0;
    for(j=j0;j<=j1;j++){
      g[i][j] = gval;
      jwrap += 1;
      if ((jwrap == pixh) && (j < j1)){  
	jwrap = 0;
	gdj += 1;  
	gval = gdata[gdi][gdj];
      }
    }
    iwrap += 1;
    if (iwrap == pixw){
      iwrap = 0;
      gdi += 1;  
    }
  }
  return g;
}
float **stm_vanhat_get_patch(seed,xn,yn,imagedir,normtype)
     int seed;        
     int xn,yn;       
     char *imagedir;  
     char *normtype;  
{
  int i,j;
  int image_n,image_w,image_h,xi,yi,ii,**idata,*idt,done;
  float **fdata,dmin,dmax,fc,*fdt,fval;
  image_n = 100;
  image_w = 1536;
  image_h = 1024;
  if (seed > 0)
    seed = -seed;
  fdata = get_2d_farray(xn,yn);
  done = 0;
  while(done == 0){  
    xi = (int)((image_w - xn + 1) * myrand_util_ran2(&seed));  
    yi = (int)((image_h - yn + 1) * myrand_util_ran2(&seed));  
    ii = 1 + (int)(image_n * myrand_util_ran2(&seed));         
    idata = read_vanhateren(imagedir,ii,xi,yi,xn,yn);
    dmax = dmin = (float)idata[0][0];
    for(i=0;i<xn;i++){
      idt = idata[i];
      fdt = fdata[i];
      for(j=0;j<yn;j++){
	fdt[j] = fval = (float)idt[j];
	if (fval > dmax)
	  dmax = fval;
	else if (fval < dmin)
	  dmin = fval;
      }
    }
    free_2d_iarray(idata,xn);
    if (dmax > dmin)
      done = 1;
    else{
      ; 
    }
  }
  if (strcmp(normtype,"-1 to 1")==0){
    fc = 2.0 / (dmax - dmin);
    for(i=0;i<xn;i++){
      fdt = fdata[i];
      for(j=0;j<yn;j++){
	fdt[j] = -1.0 + fc * (fdt[j] - dmin);
      }
    }
  }else{
    exit_error("STM_VANHAT_GET_PATCH","Unknown 'normtype'");
  }
  return fdata;
}
float *stm_ranstep_get_stim(mylogf,sf,jsz,jzero,jn,seed,stn,fps,dt,iflag,rseqn)
     char *mylogf;  
     float sf;      
     float jsz;     
     int jzero;     
     int jn;        
     int seed;      
     float stn;     
     float fps;     
     int dt;        
     int iflag;     
     int *rseqn;    
{
  int i;
  int nstep,seqn;
  int *seqi;
  float jfrac,*seqf;
  if (jsz > 0.0){
    jfrac = jsz * sf;      
    if (jzero == 1){
      nstep = jn*2 + 1;    
    }else{
      nstep = jn*2;        
    }
  }else{  
    jsz = 0.0;  
    jfrac = 1.0/(float)jn;  
    nstep = jn;
  }
  seqn = (int)(stn * fps / (float)dt);
  seqi = myrand_get_std_unif_int_seq(seqn,seed,nstep); 
  if (iflag == 1)
    seqf = i2farray(seqi,seqn);
  else{
    seqf = (float *)myalloc(seqn*sizeof(float));
    if (jsz == 0.0){
      for(i=0;i<seqn;i++)
	seqf[i] = (float)seqi[i]/(float)jn;        
    }else if (jzero == 1){
      for(i=0;i<seqn;i++)
	seqf[i] = jfrac * (float)(seqi[i] - jn);   
    }else{
      for(i=0;i<seqn;i++){
	if (seqi[i] < jn)
	  seqf[i] = jfrac * (float)(seqi[i]-jn);   
	else
	  seqf[i] = jfrac * (float)(seqi[i]+1-jn); 
      }
    }
  }
  *rseqn = seqn;
  return seqf;
}
float **stm_barnoise_get_template(mylogf,xn,yn,nbar,w,h,theta,bord,xc,yc,hoff)
     char *mylogf;  
     int xn,yn;     
     int nbar;      
     float w,h;     
     float theta;   
     float bord;    
     float xc,yc;   
     float hoff;    
{
  int i,j;
  int txi;
  float **t,a,b,c,d,thetar,xctrf,yctrf,xx,yy,x,y,ymin,ymax;
  double x0,dx,tx,marg,xtra,bord_frac;
  while(theta >= 360.0)
    theta -= 360.0;
  while(theta < 0.0)
    theta += 360.0;
  thetar = -theta * M_PI/180.0;
  a =  cos(thetar); b = sin(thetar); 
  c = -sin(thetar); d = cos(thetar);
  t = get_2d_farray(xn,yn);
  xctrf = ((float)xn-1.0)/2.0;  
  yctrf = ((float)yn-1.0)/2.0;
  ymin = -(float)h/2.0;  
  ymax =  (float)h/2.0;  
  dx = (double)w;
  x0 = -(double)(nbar+2)*dx/2.0;
  bord_frac = bord/w;
  for(i=0;i<xn;i++){
    xx = (float)i - xctrf;   
    xx -= xc;
    for(j=0;j<yn;j++){
      yy = (float)j - yctrf; 
      yy -= yc;
      x = a*xx + c*yy;  
      y = b*xx + d*yy;
      x -= hoff;  
      if ((y > ymax) || (y < ymin)){  
	t[i][j] = 0.0;  
      }else{
	tx = (x - x0)/dx;  
	txi = (int)tx;     
	marg = tx - (double)txi;  
	if (marg < bord_frac)
	  xtra = -0.5 * (1.0 - marg/bord_frac);  
	else if (marg > (1.0-bord_frac))
	  xtra = 0.5 * (marg - (1.0-bord_frac))/bord_frac;
	else
	  xtra = 0.0;  
	if (tx < 0.5){  
	  t[i][j] = 0.0;  
	}else if (tx >= ((double)nbar+1.5)){
	  t[i][j] = 0.0;  
	}else{
	  t[i][j] = (float)txi + (float)xtra;
	}
      }
    }
  }
  return t;
}
void stm_barray_get_bar_centers(mylogf,nbar,npos,theta,barw,barl,bar_gap,
				 pos_gap,rcx,rcy)
     char *mylogf;  
     int nbar;      
     int npos;      
     float theta;   
     float barw;    
     float barl;    
     float bar_gap; 
     float pos_gap; 
     float ***rcx,***rcy; 
{
  int i,j;
  float a,b,c,d,x0,y0,x,y,**cx,**cy;
  theta += 90.0;  
  while (theta >= 360.0)
    theta -= 360.0;
  while (theta < 0.0)
    theta += 360.0;
  a =  cos(M_PI/180.0*theta); b = sin(M_PI/180.0*theta);
  c = -sin(M_PI/180.0*theta); d = cos(M_PI/180.0*theta);
  cx = get_2d_farray(nbar,npos);
  cy = get_2d_farray(nbar,npos);
  x0 = -((float)(nbar-1) * (barl + bar_gap))/2.0;
  y0 = -((float)(npos-1) * (barw + pos_gap))/2.0;
  x = x0;
  y = y0;
  for(i=0;i<nbar;i++){
    for(j=0;j<npos;j++){
      cx[i][j] = a*x + c*y; 
      cy[i][j] = b*x + d*y; 
      y += barw + pos_gap;
    }
    x += barl + bar_gap;
    y = y0;
  }
  *rcx = cx;
  *rcy = cy;
}
int **stm_barray_get_seq(mylogf,n,nbar,npos,seqflag,seed)
     char *mylogf;  
     int n;         
     int nbar;      
     int npos;      
     int seqflag;   
     int seed;      
{
  int i,j,k;
  int **seqi;
  float fran;
  seqi = get_zero_2d_iarray(n,nbar);
  if (seed > 0)
    seed = -seed;
  if (seqflag == 0){  
    fran = (float)(npos + 1);
    for(i=0;i<n;i++){
      for(j=0;j<nbar;j++){
	seqi[i][j] = (int)(fran * myrand_util_ran2(&seed));
      }
    }
  }else if (seqflag == 1){  
    fran = (float)(npos * nbar + 1);
    for(i=0;i<n;i++){
      k = (int)(fran * myrand_util_ran2(&seed));
      j = k/npos;
      if (j < nbar){  
	seqi[i][j] = (k % npos) + 1;
      }
    }
  }else if (seqflag == 2){  
    fran = (float)(npos + 1);
    for(i=0;i<n;i++){
      for(j=0;j<(nbar-1);j++){
	seqi[i][j] = (int)(fran * myrand_util_ran2(&seed));
      }
      seqi[i][nbar-1] = seqi[i][0];
    }
  }else if (seqflag == 99){  
    for(i=0;i<n;i++){
      for(j=0;j<nbar;j++){
	seqi[i][j] = 1;
      }
    }
  }else{
    mylog_exit(mylogf,"STM_BARRAY_GET_SEQ  Unknown seqflag");
  }
  return seqi;
}
void stm_stat1d_get_seq(mylogf,nblock,seqflag,seed,tscale,rndur,rton,rtoff,rx)
     char *mylogf;  
     int nblock;    
     int seqflag;   
     int seed;      
     float tscale;  
     int *rndur;    
     int **rton;    
     int **rtoff;   
     int ***rx;     
{
  int i;
  int ndur,*ton,*toff,*seedlist,**indx;
  mylog(mylogf,"  STM_STAT1D_GET_SEQ\n");
  if (seqflag == 0){
    ndur = 6;
    ton = (int *)myalloc(ndur*sizeof(int));
    toff = (int *)myalloc(ndur*sizeof(int));
    ton[0] = my_rint( 1.0/tscale);  toff[0] = my_rint(16.0/tscale);
    ton[1] = my_rint( 2.0/tscale);  toff[1] = my_rint( 2.0/tscale);
    ton[2] = my_rint( 4.0/tscale);  toff[2] = my_rint( 4.0/tscale);
    ton[3] = my_rint( 8.0/tscale);  toff[3] = my_rint( 8.0/tscale);
    ton[4] = my_rint(16.0/tscale);  toff[4] = my_rint( 8.0/tscale);
    ton[5] = my_rint(32.0/tscale);  toff[5] = my_rint(16.0/tscale); 
  }else{
    mylog_exit(mylogf,"STM_STAT1D_GET_SEQ  Unknown seqflag");
  }
  seedlist = get_seeds(seed,1000000,nblock);
  indx = (int **)myalloc(nblock*sizeof(int *));
  for(i=0;i<nblock;i++)
    indx[i] = get_shuffle_index(ndur,3,seedlist[i]);
  myfree(seedlist);
  *rndur = ndur;
  *rton  = ton;
  *rtoff = toff;
  *rx    = indx;
}
float **stm_dirmod_get_test_pat(mylogf,n,pdur,t_etf,t_pat,n_anti,fps,sflag,
				degflag)
     char *mylogf;
     int n;          
     int pdur;       
     float t_etf;    
     int t_pat;      
     int n_anti;     
     float fps;      
     int sflag;      
     int degflag;    
{
  int i,j;
  float **pat,x,frac;
  char ggstr[SLEN];
  frac = t_etf/fps;  
  if (degflag)
    frac *= 360.0;   
  if (sflag == 1)
    frac *= -1.0;
  pat = get_const_2d_farray(n,pdur,-frac);  
 if (t_pat == 0){
    for(i=0;i<n;i++){
      if (i%2 == 0){
	;                       
      }else{
	for(j=0;j<pdur;j++)
	  pat[i][j] = frac;     
      }
    }
  }else if (t_pat == 1){
    for(i=0;i<n;i++){
      if (i%2 == 0){
	;                       
      }else{
	for(j=0;j<n_anti;j++){
	  if (j < pdur){
	    pat[i][j] = frac;   
	  }
	}
      }
    }
  }else{
    mylog_exit(mylogf,"STM_DIRMOD_GET_TEST_PAT  Unknown pattern type.\n");
  }
  return pat;
}
void stm_dirmod_test_seq(mylogf,t_dur,t_int,t_etf,t_pat,n_anti,fps,sflag,
			 degflag,rseqf,rseqn)
     char *mylogf;
     int t_dur;      
     int t_int;      
     float t_etf;    
     int t_pat;      
     int n_anti;     
     float fps;      
     int sflag;      
     int degflag;    
     float **rseqf;
     int *rseqn;
{
  int i,j,k;
  int n,nt,nn,ti;
  float *seqf,**pat;
  char ggstr[SLEN];
  n = *rseqn;        
  nt = n / t_int;    
  nn = n + nt*t_dur; 
  pat = stm_dirmod_get_test_pat(mylogf,nt,t_dur,t_etf,t_pat,n_anti,fps,sflag,
				degflag);
  seqf = (float *)myalloc(nn*sizeof(float));
  k = 0;
  for(i=0;i<nt;i++){
    for(j=0;j<t_int;j++){
      ti = i*t_int + j;
      if (ti >= n)
	mylog_exit(mylogf,"STM_DIRMOD_TEST_SEQ  Should not happen\n");
      seqf[k] = (*rseqf)[ti];   
      k += 1;
    }
    for(j=0;j<t_dur;j++){       
      seqf[k] = pat[i][j];
      k += 1;
    }
  }
  ti += 1;
  while (ti < n){ 
    seqf[k] = (*rseqf)[ti];
    ti += 1;
    k += 1;
    if (k > nn)
      mylog_exit(mylogf,"STM_DIRMOD_TEST_SEQ  Should not happen for k\n");
  }
  free_2d_farray(pat,nt);
  myfree(*rseqf); 
  *rseqf = seqf;
  *rseqn = nn;
}
void stm_wy4st_get_seq(mylogf,n,seed,ststr,rst)
     char *mylogf;
     int n;         
     int seed;      
     char *ststr;   
     int **rst;     
{
  int i,k;
  int *st,bit_s,bit_c;
  char ggstr[SLEN];
  st = get_zero_iarray(n);
  if (ststr != NULL){ 
    if (strlen(ststr) != n)
      mylog_exit(mylogf,"STM_WY4ST_GET_SEQ - state string length is not n");
    for(i=0;i<n;i++){
      st[i] = (int)(ststr[i]) - (int)'0';
    }
  }else if ((seed >= 0) && (seed < 256)){  
    for(i=0;i<n;i++){
      bit_s = (seed >> (i*2  )) % 2;
      bit_c = (seed >> (i*2+1)) % 2;
      k = n-1-i;
      if (bit_c == 0){
	if (bit_s == 0)
	  st[k] = 0;      
	else
	  st[k] = 1;      
      }else{
	if (bit_s == 0)
	  st[k] = 2;      
	else
	  st[k] = 3;      
      }
    }
  }else{
    st = myrand_get_std_unif_int_seq(n,seed,4);
  }
  *rst = st;
}
struct stmh_gpatch *stm_gpatch_get(x0,y0,dir,sf,tf,sdp,sdo,ph0,amp)
     float x0,y0;
     float dir;
     float sf;
     float tf;
     float sdp;
     float sdo;
     float ph0;
     float amp;
{
  struct stmh_gpatch *gp;
  gp = (struct stmh_gpatch *)myalloc(sizeof(struct stmh_gpatch));
  gp->ptype   = 1;
  gp->x0      = x0;
  gp->y0      = y0;
  gp->dir     = dir;
  gp->sf      = sf;
  gp->tf      = tf;
  gp->sd_par  = sdp;
  gp->sd_orth = sdo;
  gp->ph0     = ph0;
  gp->amp     = amp;
  return gp;
}
void stm_gpatch_config_simp(gxn,gyn,boxw,boxh,dir1,dir2,seed,
			    n,dirbal,rcx,rcy,rdir,rph1,rph2)
     int gxn,gyn;       
     float boxw,boxh;   
     float dir1,dir2;   
     int seed;          
     int n;             
     float dirbal;      
     float **rcx;       
     float **rcy;       
     float **rdir;      
     float **rph1;      
     float **rph2;      
{
  int i,j,k;
  int **flag,cnt,nless;
  float *cx,*cy,*dir,*ph1,*ph2,gx0,gy0,x,y,dir_more,dir_less;
  cx  = (float *)myalloc(n*sizeof(float));
  cy  = (float *)myalloc(n*sizeof(float));
  dir = (float *)myalloc(n*sizeof(float));
  ph1 = (float *)myalloc(n*sizeof(float));
  ph2 = (float *)myalloc(n*sizeof(float));
  if (seed > 0)
    seed *= -1;
  if (n > (gxn * gyn)/2){
    flag = get_const_2d_iarray(gxn,gyn,1);
    cnt = gxn*gyn;  
    while(cnt > n){  
      i = (int)((float)gxn * myrand_util_ran2(&seed));
      j = (int)((float)gyn * myrand_util_ran2(&seed));
      if (flag[i][j] == 1){
	flag[i][j] = 0;
	cnt -= 1;
      }
    }
  }else{
    flag = get_const_2d_iarray(gxn,gyn,0);
    cnt = 0;
    while(cnt < n){  
      i = (int)((float)gxn * myrand_util_ran2(&seed));
      j = (int)((float)gyn * myrand_util_ran2(&seed));
      if (flag[i][j] == 0){
	flag[i][j] = 1;
	cnt += 1;
      }
    }
  }
  gx0 = -((float)(gxn - 1) * boxw) / 2.0; 
  gy0 = -((float)(gyn - 1) * boxh) / 2.0; 
  x = gx0;
  y = gy0;
  k = 0;
  for(i=0;i<gxn;i++){
    for(j=0;j<gyn;j++){
      if (flag[i][j] == 1){
	cx[k] = x;
	cy[k] = y;
	ph1[k] = 360.0 * myrand_util_ran2(&seed);
	ph2[k] = 360.0 * myrand_util_ran2(&seed);
	k += 1;
      }
      y += boxh;
    }
    x += boxw;
    y = gy0;
  }
  if ((dirbal < 0.0) || (dirbal > 1.0))
    exit_error("STM_GPATCH_CONFIG_SIMP","Bad 'dir_balance', must be [0..1]");
  if (dirbal <= 0.5){
    dir_more = dir2;
    dir_less = dir1;
    nless = my_rint(dirbal * (float)n);
  }else{
    dir_more = dir1;
    dir_less = dir2;
    nless = n - my_rint(dirbal * (float)n);
  }
  for(i=0;i<n;i++)
    dir[i] = dir_more;  
  cnt = 0;   
  while(cnt < nless){  
    i = (int)((float)n * myrand_util_ran2(&seed));
    if (dir[i] != dir_less){
      dir[i] = dir_less;
      cnt += 1;
    }
  }
  free_2d_iarray(flag,gxn);
  *rcx  = cx;
  *rcy  = cy;
  *rdir = dir;
  *rph1 = ph1;
  *rph2 = ph2;
}
void stm_get_coherence_dot_movie(sscale,framerate,coher,speed,theta,sizex,
				 sizey,nframes,dpf,dt,seed,rdotx,rdoty,rdotn)
     float sscale,framerate,coher,speed,theta,sizex,sizey;
     int nframes,dpf,dt,seed;
     float ***rdotx,***rdoty;
     int **rdotn;
{
  int i,j,k;
  int *dotn,n;
  float **dotx,**doty,x,y,dx,dy;
  if (seed > 0)
    seed = -seed;
  dotn = (int *)myalloc(nframes*sizeof(int));
  dotx = (float **)myalloc(nframes*sizeof(float *));
  doty = (float **)myalloc(nframes*sizeof(float *));
  for(i=0;i<nframes;i++){
    dotn[i] = dpf;
    dotx[i] = (float *)myalloc(dpf*sizeof(float));
    doty[i] = (float *)myalloc(dpf*sizeof(float));
  }
  dy = speed * (float)dt/framerate / sscale * sin(theta/180.0*M_PI);
  dx = speed * (float)dt/framerate / sscale * cos(theta/180.0*M_PI);
  for (i=0;i<nframes;i++){
    n = 0; 
    k = i - dt; 
    if (k >= 0){ 
      for (j=0;j<dpf;j++) 
	if (myrand_util_ran2(&seed) < coher){ 
	  x = dotx[k][j] + dx;
	  y = doty[k][j] + dy;
	  if (x < 0.0) x += sizex; else if (x >= sizex) x -= sizex;
	  if (y < 0.0) y += sizey; else if (y >= sizey) y -= sizey;
	  dotx[i][n] = x; 
	  doty[i][n] = y;
	  n += 1;
	}
    }
    while (n < dpf){ 
      dotx[i][n] = myrand_util_ran2(&seed)*sizex;
      doty[i][n] = myrand_util_ran2(&seed)*sizey;
      n += 1;
    }
  }
  *rdotn = dotn; *rdotx = dotx; *rdoty = doty;
}
void stm_mask_dot_movie(dotx,doty,dotn,nframes,cx,cy,w,h,sscale,aptype)
     float **dotx,**doty;    
     int *dotn,nframes;
     float cx,cy,w,h;        
     float sscale;           
     int aptype;
{
  int i,j;
  int n,dflag;
  float x,y,r2pix,tx,ty,apcx,apcy,wo2,ho2;
  wo2 = w/(2.0*sscale);   
  ho2 = h/(2.0*sscale);   
  r2pix = wo2*wo2;        
  apcx = cx/sscale;
  apcy = cy/sscale;
  for (i=0;i<nframes;i++){
    n = 0; 
    for (j=0;j<dotn[i];j++){
      tx = dotx[i][j];
      ty = doty[i][j];
      x = tx - apcx;      
      y = ty - apcy;
      dflag = 0;
      if (aptype == 0){ 
	dflag = 1;
      }else if (aptype == 1){ 
	if ((x*x + y*y) <= r2pix)
	  dflag = 1;
      }else if (aptype == 2){ 
	if ((x >= -wo2) && (x <= wo2) && (y >= -ho2) && (y <= ho2))
	  dflag = 1;
      }else{
	printf("*** Aperture type %d not implemented yet\n",aptype);
	dflag = 1;
      }
      if (dflag){
	dotx[i][n] = tx; 
	doty[i][n] = ty;
	n += 1;
      }
    }
    dotn[i] = n;
  }
}
void stm_dot_movie_shift(dotx,doty,dotn,nframes,dx,dy)
     float **dotx,**doty;   
     int *dotn;             
     int nframes;           
     float dx,dy;           
{
  int i,j;
  int nd;
  for(i=0;i<nframes;i++){
    nd = dotn[i];  
    for(j=0;j<nd;j++){
      dotx[i][j] += dx;
      doty[i][j] += dy;
    }
  }
}
void stm_dot_movie_copy(dotx,doty,dota,dotn,nframes,rdotx,rdoty,rdota,rdotn)
     float **dotx,**doty;      
     float **dota;             
     int *dotn;                
     int nframes;              
     float ***rdotx,***rdoty;  
     float ***rdota;           
     int **rdotn;              
{
  int i;
  int nd,*cdn;
  float **cdx,**cdy,**cda;
  cdx = (float **)myalloc(nframes*sizeof(float *));
  cdy = (float **)myalloc(nframes*sizeof(float *));
  if (dota != NULL)
    cda = (float **)myalloc(nframes*sizeof(float *));
  cdn = (int *)myalloc(nframes*sizeof(int));
  for(i=0;i<nframes;i++){
    nd = dotn[i];  
    cdx[i] = copy_farray(dotx[i],nd);
    cdy[i] = copy_farray(doty[i],nd);
    if (dota != NULL)
      cda[i] = copy_farray(dota[i],nd);
    cdn[i] = nd;
  }
  *rdotx = cdx;
  *rdoty = cdy;
  if (dota != NULL)
    *rdota = cda;
  *rdotn = cdn;
}
struct stmh_dotmov *stm_dotmov_init(nframes,dpf,aflag,zflag,fflag)
     int nframes;  
     int dpf;      
     int aflag;    
     int zflag;    
     int fflag;    
{
  int i;
  struct stmh_dotmov *dm;
  dm = (struct stmh_dotmov *)myalloc(sizeof(struct stmh_dotmov));
  dm->nframes = nframes;
  dm->cnt =    (int *)myalloc(nframes*sizeof(int));
  dm->x   = (float **)myalloc(nframes*sizeof(float *));
  dm->y   = (float **)myalloc(nframes*sizeof(float *));
  if (aflag == 1)
    dm->a   = (float **)myalloc(nframes*sizeof(float *));
  else
    dm->a = NULL;
  if (zflag == 1)
    dm->z   = (float **)myalloc(nframes*sizeof(float *));
  else
    dm->z = NULL;
  if (fflag == 1)
    dm->flag = (int **)myalloc(nframes*sizeof(int *));
  else
    dm->flag = NULL;
  for(i=0;i<nframes;i++){
    dm->cnt[i] = dpf;
    dm->x[i] = (float *)myalloc(dpf*sizeof(float));
    dm->y[i] = (float *)myalloc(dpf*sizeof(float));
    if (aflag == 1)
      dm->a[i] = (float *)myalloc(dpf*sizeof(float));
    if (zflag == 1)
      dm->z[i] = (float *)myalloc(dpf*sizeof(float));
    if (fflag == 1)
      dm->flag[i] = get_zero_iarray(dpf);
  }
  return dm;
}
void stm_dotmov_get_parts(dm,fflag,aflag,rx,ry,ra,rcnt)
     struct stmh_dotmov *dm;
     int fflag;    
     int aflag;    
     float ***rx;  
     float ***ry;  
     float ***ra;  
     int  **rcnt;  
{
  int i,j,k;
  int n,*dotn,ntot,nval;
  float **dotx,**doty,**dota;
  n = dm->nframes;
  dotn = get_zero_iarray(n);
  dotx = (float **)myalloc(n*sizeof(float *));
  doty = (float **)myalloc(n*sizeof(float *));
  if (aflag == 1)
    dota = (float **)myalloc(n*sizeof(float *));
  else
    dota = NULL;
  if (fflag == 1){
    for(i=0;i<n;i++){  
      ntot = dm->cnt[i];  
      nval = sum_iarray(dm->flag[i],ntot,0,ntot);  
      dotx[i] = (float *)myalloc(nval*sizeof(float));
      doty[i] = (float *)myalloc(nval*sizeof(float));
      if (aflag == 1)
	dota[i] = (float *)myalloc(nval*sizeof(float));
      k = 0;
      for(j=0;j<ntot;j++){
	if (dm->flag[i][j] == 1){  
	  dotx[i][k] = dm->x[i][j];   
	  doty[i][k] = dm->y[i][j];
	  if (aflag == 1)
	    dota[i][k] = dm->a[i][j];
	  k += 1;
	}
      }
      if (k != nval)
	exit_error("STM_DOTMOV_GET_PARTS","WYETH - this should not happen");
      dotn[i] = nval;
    }
  }else{
    exit_error("STM_DOTMOV_GET_PARTS","WYETH - not implemented yet");
  }
  *rx   = dotx;
  *ry   = doty;
  *ra   = dota;
  *rcnt = dotn;
}
struct stmh_dotmov *stm_dotmov_get_rand_xya(szx,szy,nframes,ndots,seed,z,bipol)
     float szx,szy;  
     int nframes;    
     int ndots;      
     int seed;       
     float z;        
     int bipol;      
{
  int i,j;
  struct stmh_dotmov *dm;
  if (seed > 0)
    seed = -seed;
  dm = stm_dotmov_init(nframes,ndots,1,1,1);  
  for(i=0;i<nframes;i++){
    for(j=0;j<ndots;j++){
      dm->x[i][j] = szx * myrand_util_ran2(&seed);
      dm->y[i][j] = szy * myrand_util_ran2(&seed);
      dm->z[i][j] = z;
      dm->a[i][j] = 1.0;   
      if (bipol == 1){
	if (myrand_util_ran2(&seed) < 0.5)
	  dm->a[i][j] = -1.0;  
      }
      dm->flag[i][j] = 1;  
    }
  }
  return dm;
}
struct stmh_dotmov *stm_dotmov_get_deep_f0(szx,szy,ndots,pseed,z0,z1,bipol)
     float szx,szy;  
     int ndots;      
     int *pseed;     
     float z0,z1;    
     int bipol;      
{
  int i;
  float dz,z;
  struct stmh_dotmov *dm;
  dm = stm_dotmov_init(1,ndots,1,1,0);  
  dz = (z1 - z0) / (float)ndots;
  z = z0 + dz;  
  for(i=0;i<ndots;i++){
    dm->x[0][i] = szx * myrand_util_ran2(pseed);
    dm->y[0][i] = szy * myrand_util_ran2(pseed);
    dm->z[0][i] = z;
    dm->a[0][i] = 1.0;  
    if (bipol == 1){
      if (myrand_util_ran2(pseed) < 0.5)
	dm->a[0][i] = -1.0;  
    }
    z += dz;
  }
  return dm;
}
void stm_dotmov_get_deep(sscale,sizex,sizey,framerate,speed,maxshift,denv,
			 nframes,dpf,dt,seed,tseq,tsn,bipol,dotcorr,rdml,rdmr)
     float sscale;        
     float sizex,sizey;   
     float framerate;     
     float speed;         
     float maxshift;      
     int denv;            
     int nframes;         
     int dpf;             
     int dt;              
     int seed;            
     float *tseq;         
     int tsn;             
     int bipol;           
     int dotcorr;         
     struct stmh_dotmov **rdml,**rdmr;  
{
  int i,j,k;
  int dpf2;
  float z,z0,z1,ppf,xpz,dx,*xref,*yref,xoff,aa,fz;
  struct stmh_dotmov *dm0,*dml,*dmr,*dm0r;
  if ((tsn * dt) < nframes)
    exit_error("STM_DOTMOV_GET_DEEP","Motion sequence too short");
  dpf2 = 2*dpf;  
  if (seed > 0)
    seed = -seed;
  z0 = -2.0;
  z1 =  2.0;
  dm0 = stm_dotmov_get_deep_f0(sizex,sizey,dpf2,&seed,z0,z1,bipol);
  if (dotcorr == 0){
    dm0r = stm_dotmov_get_deep_f0(sizex,sizey,dpf2,&seed,z0,z1,bipol);
  }
  ppf = speed / (framerate * sscale) ;  
  xpz = (maxshift/2.0) / sscale;  
  dml = stm_dotmov_init(nframes,dpf2,1,1,1);  
  dmr = stm_dotmov_init(nframes,dpf2,1,1,1);  
  xref = dm0->x[0];  
  yref = dm0->y[0];  
  k = 0;
  for(i=0;i<nframes;i++){
    dx = ppf*tseq[k];  
    for(j=0;j<dpf2;j++){  
      if (i==0)
	z = dm0->z[0][j];    
      else
	z = dml->z[i-1][j];  
      z += dx / xpz;
      while (z > 2.0)
	z -= 4.0;
      while (z <= -2.0)
	z += 4.0;
      xoff = z * xpz;  
      dml->x[i][j] = xref[j] + xoff;
      dml->y[i][j] = yref[j];
      dml->z[i][j] = z;
      if (dotcorr == 0){
	dmr->x[i][j] = dm0r->x[0][j] - xoff;
	dmr->y[i][j] = dm0r->y[0][j];
      }else{
	dmr->x[i][j] = xref[j] - xoff;
	dmr->y[i][j] = yref[j];
      }
      dmr->z[i][j] = z;
      aa = dm0->a[0][j];
      fz = fabs(z);
      if (fz > 1.0)
	aa = 0.0;
      else{
	if (denv == 1)
	  aa *= 1.0 - fz;         
	else if (denv == 2)
	  aa *= sqrt(1.0 - fz);   
	else if (denv == 3)
	  aa *= cos(z*M_PI*0.5);  
	else if (denv == 4)
	  aa *= sqrt(1.0 - z*z);  
      }
      dml->a[i][j] = aa;
      dmr->a[i][j] = aa;
      if ((z > -1.0) && (z <= 1.0)){
	dml->flag[i][j] = 1;  
	dmr->flag[i][j] = 1;
      }else{
	dml->flag[i][j] = 0;  
	dmr->flag[i][j] = 0;
      }
    }
    if (((i+1) % dt) == 0)  
      k += 1;
  }
  *rdml = dml;
  *rdmr = dmr;
}
void stm_dotmov_get_planes(sscale,sizex,sizey,framerate,speed,maxshift,denv,
			   nframes,dpf,nplanes,dt,seed,tseq,tsn,bipol,rdml,
			   rdmr)
     float sscale;        
     float sizex,sizey;   
     float framerate;     
     float speed;         
     float maxshift;      
     int denv;            
     int nframes;         
     int dpf;             
     int nplanes;         
     int dt;              
     int seed;            
     float *tseq;         
     int tsn;             
     int bipol;           
     struct stmh_dotmov **rdml,**rdmr;  
{
  int i,j,k;
  float z,ppf,xpz,dx,*xref,*yref,xoff,aa,fz;
  struct stmh_dotmov *dml,*dmr;
  if ((tsn * dt) < nframes)
    exit_error("STM_DOTMOV_GET_PLANES","Motion sequence too short");
  z = 0.0;
  dml = stm_dotmov_get_rand_xya(sizex,sizey,nframes,dpf,seed,z,bipol);
  if (nplanes == 2){
    for(i=0;i<dpf/2;i++)
      dml->z[0][i] = 1.0;  
  }
  ppf = speed / (framerate * sscale) ;  
  xpz = maxshift / sscale;  
  dmr = stm_dotmov_init(nframes,dpf,1,1,1);
  k = 0;
  for(i=0;i<nframes;i++){
    xref = dml->x[i];  
    yref = dml->y[i];  
    dx = ppf*tseq[k];  
    for(j=0;j<dpf;j++){  
      if (i == 0)
	z = dml->z[i][j];    
      else
	z = dml->z[i-1][j];  
      z += dx / xpz;
      while (z > 1.0)
	z -= 2.0;
      while (z <= -1.0)
	z += 2.0;
      xoff = z * xpz;  
      dmr->x[i][j] = xref[j] - xoff;
      dmr->y[i][j] = yref[j];
      dmr->z[i][j] = z;
      dml->z[i][j] = z;  
      aa = dml->a[i][j];
      fz = fabs(z);
      if (fz > 1.0)
	aa = 0.0;    
      else{
	if (denv == 1)
	  aa *= 1.0 - fz;         
	else if (denv == 2)
	  aa *= sqrt(1.0 - fz);   
	else if (denv == 3)
	  aa *= cos(z*M_PI*0.5);  
	else if (denv == 4)
	  aa *= sqrt(1.0 - z*z);  
      }
      dml->a[i][j] = aa;
      dmr->a[i][j] = aa;
      dml->flag[i][j] = 1;  
      dmr->flag[i][j] = 1;
    }
    if (((i+1) % dt) == 0)  
      k += 1;
  }
  *rdml = dml;
  *rdmr = dmr;
}
void stm_star_get_trig(stn,fps,sampling,dt_ad,dt_t0,dt_t1,state0,ttype,
		       rtrig,rntrig)
     float stn;       
     float fps;       
     float sampling;  
     int dt_ad;       
     int dt_t1;       
     int dt_t0;       
     int state0;      
     char *ttype;     
     int **rtrig;     
     int *rntrig;     
{
  int i,k;
  int frpcyc,ncyc,*t,first_flag;
  double dtsec,t0s;
  frpcyc = dt_ad + dt_t0 + dt_t1;  
  ncyc = (int)(stn*fps) / frpcyc;  
  dtsec = (double)frpcyc / (double)fps;  
  t = (int *)myalloc(ncyc*sizeof(int));  
  k = 0;
  first_flag = state0;
  for(i=0;i<ncyc;i++){
    t0s = (double)i*dtsec;  
    if (strcmp(ttype,"ad")==0){        
      t[k] = my_rint(sampling * t0s);
      k += 1;
    }else if (strcmp(ttype,"ad0")==0){
      if (first_flag == 0){
	t[k] = my_rint(sampling * t0s);
	k += 1;
      }
    }else if (strcmp(ttype,"ad1")==0){
      if (first_flag == 1){
	t[k] = my_rint(sampling * t0s);
	k += 1;
      }
    }else{
      printf("  trig_type  %s\n",ttype);
      exit_error("STM_STAR_GET_TRIG","Unknown trig_type");
    }
    first_flag = 1 - first_flag;  
  }
  *rntrig = k;   
  *rtrig = t;    
}
void stm_pasupathy_shape_idrot_list(mylogf,stype,rn,rid,rrot)
     char *mylogf;  
     char *stype;   
     int  *rn;      
     int **rid;     
     int **rrot;    
{
  int i,j,k;
  int n,nr,ddeg,*shid,*rot;
  if (strcmp(stype,"pc370")==0){
    n = 370;               
    shid = (int *)myalloc(n*sizeof(int));  
    rot  = (int *)myalloc(n*sizeof(int));  
    k = 0;
    for(i=0;i<51;i++){
      nr = pasu_shape_nrot[i];  
      ddeg = 360 / nr;
      for(j=0;j<nr;j++){
	shid[k] = i+1;
	rot[k] = j*ddeg;
	k += 1;
      }
    }
    if (k != 370)
      exit_error("STM_PASUPATHY_SHAPE_IDROT_LIST","Should not happen");
  }else{
    printf("  stype:  %s\n",stype);
    exit_error("STM_PASUPATHY_SHAPE_IDROT_LIST","Bad 'stype'");
  }
  *rn = n;
  *rid = shid;
  *rrot = rot;
}
void stm_pasupathy_shape_get_control(mylogf,shid,rx,ry,rn)
     char *mylogf;  
     int shid;      
     float **rx;    
     float **ry;    
     int    *rn;    
{
  int i;
  int n;
  float *t,*x,*y;
  if ((shid < 1) || (shid > 51)){
    mylogx(mylogf,"STM_PASUPATHY_SHAPE_GET_CONTROL",
	   "Shape ID must be [1..51]");
  }
  if      (shid ==  1)  t = pasu_shape_01;
  else if (shid ==  2)  t = pasu_shape_02;
  else if (shid ==  3)  t = pasu_shape_03;
  else if (shid ==  4)  t = pasu_shape_04;
  else if (shid ==  5)  t = pasu_shape_05;
  else if (shid ==  6)  t = pasu_shape_06;
  else if (shid ==  7)  t = pasu_shape_07;
  else if (shid ==  8)  t = pasu_shape_08;
  else if (shid ==  9)  t = pasu_shape_09;
  else if (shid == 10)  t = pasu_shape_10;
  else if (shid == 11)  t = pasu_shape_11;
  else if (shid == 12)  t = pasu_shape_12;
  else if (shid == 13)  t = pasu_shape_13;
  else if (shid == 14)  t = pasu_shape_14;
  else if (shid == 15)  t = pasu_shape_15;
  else if (shid == 16)  t = pasu_shape_16;
  else if (shid == 17)  t = pasu_shape_17;
  else if (shid == 18)  t = pasu_shape_18;
  else if (shid == 19)  t = pasu_shape_19;
  else if (shid == 20)  t = pasu_shape_20;
  else if (shid == 21)  t = pasu_shape_21;
  else if (shid == 22)  t = pasu_shape_22;
  else if (shid == 23)  t = pasu_shape_23;
  else if (shid == 24)  t = pasu_shape_24;
  else if (shid == 25)  t = pasu_shape_25;
  else if (shid == 26)  t = pasu_shape_26;
  else if (shid == 27)  t = pasu_shape_27;
  else if (shid == 28)  t = pasu_shape_28;
  else if (shid == 29)  t = pasu_shape_29;
  else if (shid == 30)  t = pasu_shape_30;
  else if (shid == 31)  t = pasu_shape_31;
  else if (shid == 32)  t = pasu_shape_32;
  else if (shid == 33)  t = pasu_shape_33;
  else if (shid == 34)  t = pasu_shape_34;
  else if (shid == 35)  t = pasu_shape_35;
  else if (shid == 36)  t = pasu_shape_36;
  else if (shid == 37)  t = pasu_shape_37;
  else if (shid == 38)  t = pasu_shape_38;
  else if (shid == 39)  t = pasu_shape_39;
  else if (shid == 40)  t = pasu_shape_40;
  else if (shid == 41)  t = pasu_shape_41;
  else if (shid == 42)  t = pasu_shape_42;
  else if (shid == 43)  t = pasu_shape_43;
  else if (shid == 44)  t = pasu_shape_44;
  else if (shid == 45)  t = pasu_shape_45;
  else if (shid == 46)  t = pasu_shape_46;
  else if (shid == 47)  t = pasu_shape_47;
  else if (shid == 48)  t = pasu_shape_48;
  else if (shid == 49)  t = pasu_shape_49;
  else if (shid == 50)  t = pasu_shape_50;
  else if (shid == 51)  t = pasu_shape_51;
  n = pasu_shape_n[shid-1] / 2;  
  x = get_farray(n);
  y = get_farray(n);
  for(i=0;i<n;i++){
    x[i] = t[i*2];
    y[i] = t[i*2+1];
  }
  *rn = n;
  *rx = x;
  *ry = y;
}
void stm_pasupathy_yosh_get_control(mylogf,shid,rx,ry,rn)
     char *mylogf;  
     int shid;      
     float **rx;    
     float **ry;    
     int    *rn;    
{
  int i;
  int n;
  float *t,*x,*y;
  if ((shid < 1) || (shid > 43)){
    mylogx(mylogf,"STM_PASUPATHY_YOSH_GET_CONTROL",
	   "Shape ID must be [1..43]");
  }
  if      (shid ==  1)  t =                pasu_shape_32; 
  else if (shid ==  2)  t =                pasu_shape_33; 
  else if (shid ==  3)  t =                pasu_shape_34; 
  else if (shid ==  4)  t =                pasu_shape_35; 
  else if (shid ==  5)  t =                pasu_shape_36; 
  else if (shid ==  6)  t = yosh_shape_05;
  else if (shid ==  7)  t = yosh_shape_06;
  else if (shid ==  8)  t =                pasu_shape_24; 
  else if (shid ==  9)  t =                pasu_shape_25; 
  else if (shid == 10)  t =                pasu_shape_26; 
  else if (shid == 11)  t =                pasu_shape_27; 
  else if (shid == 12)  t =                pasu_shape_28; 
  else if (shid == 13)  t =                pasu_shape_29; 
  else if (shid == 14)  t =                pasu_shape_30; 
  else if (shid == 15)  t =                pasu_shape_31; 
  else if (shid == 16)  t = yosh_shape_15; 
  else if (shid == 17)  t = yosh_shape_16;
  else if (shid == 18)  t = yosh_shape_17;
  else if (shid == 19)  t = yosh_shape_18;
  else if (shid == 20)  t = yosh_shape_19;
  else if (shid == 21)  t = yosh_shape_20; 
  else if (shid == 22)  t =                pasu_shape_19; 
  else if (shid == 23)  t =                pasu_shape_16; 
  else if (shid == 24)  t =                pasu_shape_17; 
  else if (shid == 25)  t =                pasu_shape_18; 
  else if (shid == 26)  t =                pasu_shape_21; 
  else if (shid == 27)  t =                pasu_shape_22; 
  else if (shid == 28)  t =                pasu_shape_23; 
  else if (shid == 29)  t =                pasu_shape_50; 
  else if (shid == 30)  t =                pasu_shape_38; 
  else if (shid == 31)  t = yosh_shape_30;
  else if (shid == 32)  t =                pasu_shape_41; 
  else if (shid == 33)  t = yosh_shape_32;
  else if (shid == 34)  t = yosh_shape_33;
  else if (shid == 35)  t = yosh_shape_34;
  else if (shid == 36)  t =                pasu_shape_05; 
  else if (shid == 37)  t =                pasu_shape_06; 
  else if (shid == 38)  t =                pasu_shape_07; 
  else if (shid == 39)  t =                pasu_shape_10; 
  else if (shid == 40)  t =                pasu_shape_11; 
  else if (shid == 41)  t =                pasu_shape_14; 
  else if (shid == 42)  t =                pasu_shape_15; 
  else if (shid == 43)  t = yosh_shape_42;
  n = yosh_shape_n[shid-1] / 2;  
  x = get_farray(n);
  y = get_farray(n);
  for(i=0;i<n;i++){
    x[i] = t[i*2];
    y[i] = t[i*2+1];
  }
  *rn = n;
  *rx = x;
  *ry = y;
}
void stm_pasupathy_yasm_get_control(mylogf,shid,rx,ry,rn)
     char *mylogf;  
     int shid;      
     float **rx;    
     float **ry;    
     int    *rn;    
{
  int i;
  int n;
  float *t,*x,*y;
  if ((shid < 1) || (shid > 13)){
    mylogx(mylogf,"STM_PASUPATHY_YASM_GET_CONTROL",
	   "Shape ID must be [1..13]");
  }
  if      (shid ==  1)  t = yasm_shape_01;
  else if (shid ==  2)  t = yasm_shape_02;
  else if (shid ==  3)  t = yasm_shape_03;
  else if (shid ==  4)  t = yasm_shape_04;
  else if (shid ==  5)  t = yasm_shape_05;
  else if (shid ==  6)  t = yasm_shape_06;
  else if (shid ==  7)  t = yasm_shape_07;
  else if (shid ==  8)  t = yasm_shape_08;
  else if (shid ==  9)  t = yasm_shape_09;
  else if (shid == 10)  t = yasm_shape_10;
  else if (shid == 11)  t = yasm_shape_11;
  else if (shid == 12)  t = yasm_shape_12;
  else if (shid == 13)  t = yasm_shape_13;
  n = yasm_shape_n[shid-1] / 2;  
  x = get_farray(n);
  y = get_farray(n);
  for(i=0;i<n;i++){
    x[i] = t[i*2];
    y[i] = t[i*2+1];
  }
  *rn = n;
  *rx = x;
  *ry = y;
}
void stm_pasupathy_shape_compare(sset1,shid1,sset2,shid2)
     char *sset1;   
     int shid1;     
     char *sset2;   
     int shid2;     
{
  int i;
  int sn,sn2,flag;
  float *sx,*sy,*sx2,*sy2;
  if (strcmp(sset1,"yasm")==0)
    stm_pasupathy_yasm_get_control(NULL,shid1,&sx,&sy,&sn);
  else if (strcmp(sset1,"yosh")==0)
    stm_pasupathy_yosh_get_control(NULL,shid1,&sx,&sy,&sn);
  else
    stm_pasupathy_shape_get_control(NULL,shid1,&sx,&sy,&sn);
  if (strcmp(sset2,"yasm")==0)
    stm_pasupathy_yasm_get_control(NULL,shid2,&sx2,&sy2,&sn2);
  else if (strcmp(sset2,"yosh")==0)
    stm_pasupathy_yosh_get_control(NULL,shid2,&sx2,&sy2,&sn2);
  else
    stm_pasupathy_shape_get_control(NULL,shid2,&sx2,&sy2,&sn2);
  flag = 0;
  if (sn != sn2){
    printf("    ==>  Point counts differ:  %d %d\n",sn,sn2);
    flag = 1;
  }else{
    for(i=0;i<sn;i++){
      if (sx[i] != sx2[i]){
	printf("    ==>  x-coord %d differs:  %f %f\n",i,sx[i],sx2[i]);
	flag = 1;
      }
      if (sy[i] != sy2[i]){
	printf("    ==>  y-coord %d differs:  %f %f\n",i,sy[i],sy2[i]);
	flag = 1;
      }
    }
  }
  if (flag == 1)
    printf("    _____________ %s %d   %s %d ___________\n",sset1,shid1,
	   sset2,shid2);
  else
    printf("    ==>  SHAPES ARE THE SAME\n");
}
void stm_pasupathy_shape_get_inside_point(shape_set,shid,theta,rx0,ry0)
     char *shape_set; 
     int shid;        
     float theta;     
     float *rx0;      
     float *ry0;      
{
  float xx[1],yy[1];
  if (strcmp(shape_set,"pc2001")==0){
    if ((shid >= 8) && (shid <= 10)){
      *rx0 = 0.4;
      *ry0 = 0.4;
    }else if (shid == 11){  
      *rx0 = -0.4;
      *ry0 =  0.4;
    }else if (shid == 12){  
      *rx0 =  0.0;
      *ry0 =  0.2;
    }else if (shid == 14){  
      *rx0 =  0.0;
      *ry0 =  0.2;
    }else if (shid == 15){  
      *rx0 =  0.0;
      *ry0 =  0.2;
    }else{
      *rx0 = 0.0;
      *ry0 = 0.0;
    }
  }else if (strcmp(shape_set,"yosh")==0){
    if (shid == 39){
      *rx0 = 0.4;
      *ry0 = 0.4;
    }else if (shid == 40){  
      *rx0 = -0.4;
      *ry0 =  0.4;
    }else if (shid == 41){  
      *rx0 =  0.0;
      *ry0 =  0.2;
    }else if (shid == 42){  
      *rx0 =  0.0;
      *ry0 =  0.2;
    }else{
      *rx0 = 0.0;
      *ry0 = 0.0;
    }
  }else if (strcmp(shape_set,"yasm")==0){
    *rx0 = 0.0;
    *ry0 = -0.2;
  }else{
    *rx0 = 0.0;
    *ry0 = 0.0;
  }
  if (theta != 0.0){
    xx[0] = *rx0;
    yy[0] = *ry0;
    farray_rotate_coords(theta,xx,yy,1);
    *rx0 = xx[0];
    *ry0 = yy[0];
  }
}
double stm_pasupathy_spline_eqn(p1,p2,p3,p4,t)
     double p1,p2,p3,p4;  
     double t;            
{
  double t3,t2,v;
  t2 = t*t;
  t3 = t2*t;
  v = (p1*(    -t3 + 3.0*t2 - 3.0*t + 1.0) +
       p2*( 3.0*t3 - 6.0*t2         + 4.0) +
       p3*(-3.0*t3 + 3.0*t2 + 3.0*t + 1.0) +
       p4*      t3)/6.0;
  return v;
}
void stm_pasupathy_shape_get_coords(mylogf,shape_set,shid,theta,normflag,
				    px,rx,ry,rn)
     char *mylogf;  
     char *shape_set;  
     int shid;      
     float theta;   
     int normflag;  
     int px;        
     float **rx;    
     float **ry;    
     int    *rn;    
{
  int i,j,k;
  int sn,fn;
  float *sx,*sy,*tsx,*tsy,*fx,*fy,tmin,tmax;
  double x1,x2,x3,x4,y1,y2,y3,y4,t;
  if (strcmp(shape_set,"yasm")==0)
    stm_pasupathy_yasm_get_control(mylogf,shid,&sx,&sy,&sn);
  else if (strcmp(shape_set,"yosh")==0)
    stm_pasupathy_yosh_get_control(mylogf,shid,&sx,&sy,&sn);
  else
    stm_pasupathy_shape_get_control(mylogf,shid,&sx,&sy,&sn);
  if (normflag == 1){
    get_min_max_farray(sx,sn,&tmin,&tmax);
    add_const_farray(sx,sn,-(tmax+tmin)/2.0);
    get_min_max_farray(sy,sn,&tmin,&tmax);
    add_const_farray(sy,sn,-(tmax+tmin)/2.0);
  }
  farray_rotate_coords(theta,sx,sy,sn);
  tsx = get_farray(sn+2);
  tsy = get_farray(sn+2);
  for(i=0;i<sn;i++){
    tsx[i] = sx[i];
    tsy[i] = sy[i];
  }
  tsx[sn  ] = sx[1];  
  tsx[sn+1] = sx[2];
  tsy[sn  ] = sy[1];
  tsy[sn+1] = sy[2];
  fn = px * (sn-1);  
  fx = get_farray(fn);
  fy = get_farray(fn);
  k = 0;
  for(i=0;i<(sn-1);i++){
    x1 = tsx[i];
    x2 = tsx[i+1];
    x3 = tsx[i+2];
    x4 = tsx[i+3];
    y1 = tsy[i];
    y2 = tsy[i+1];
    y3 = tsy[i+2];
    y4 = tsy[i+3];
    for(j=0;j<px;j++){
      t = (double)j/(double)px;
      fx[k] = stm_pasupathy_spline_eqn(x1,x2,x3,x4,t);
      fy[k] = stm_pasupathy_spline_eqn(y1,y2,y3,y4,t);
      k += 1;
    }
  }
  *rx = fx;
  *ry = fy;
  *rn = fn;
  myfree(tsx);
  myfree(tsy);
  myfree(sx);
  myfree(sy);
}
void stm_pasupathy_shape(mylogf,data,xn,yn,xc,yc,shape_set,shid,theta,size,
			 fgval,bgval,aa_sd,fill_flag,norm_flag)
     char *mylogf;       
     float **data;       
     int xn,yn;          
     float xc,yc;        
     char *shape_set;    
     int shid;           
     float theta;        
     float size;         
     float fgval;        
     float bgval;        
     float aa_sd;        
     int fill_flag;      
     int norm_flag;      
{
  int sn,normflag,px;
  float *sx,*sy,sz_scale,x0,y0;
  char ggstr[SLEN];
  if (0){
    printf("WYETH DEBUG ==== CALLING COMPARE\n");
    stm_pasupathy_shape_compare("yosh",1,"pc2001",32);
    stm_pasupathy_shape_compare("yosh",2,"pc2001",33);
    stm_pasupathy_shape_compare("yosh",3,"pc2001",34);
    stm_pasupathy_shape_compare("yosh",4,"pc2001",35);
    stm_pasupathy_shape_compare("yosh",5,"pc2001",36);
    stm_pasupathy_shape_compare("yosh",8,"pc2001",24);
    stm_pasupathy_shape_compare("yosh",9,"pc2001",25);
    stm_pasupathy_shape_compare("yosh",10,"pc2001",26);
    stm_pasupathy_shape_compare("yosh",11,"pc2001",27);
    stm_pasupathy_shape_compare("yosh",12,"pc2001",28);
    stm_pasupathy_shape_compare("yosh",13,"pc2001",29);
    stm_pasupathy_shape_compare("yosh",14,"pc2001",30);
    stm_pasupathy_shape_compare("yosh",15,"pc2001",31);
    stm_pasupathy_shape_compare("yosh",16,"pc2001",40);
    stm_pasupathy_shape_compare("yosh",21,"pc2001",43);
    stm_pasupathy_shape_compare("yosh",22,"pc2001",19);
    stm_pasupathy_shape_compare("yosh",23,"pc2001",16);
    stm_pasupathy_shape_compare("yosh",24,"pc2001",17);
    stm_pasupathy_shape_compare("yosh",25,"pc2001",18);
    stm_pasupathy_shape_compare("yosh",26,"pc2001",21);
    stm_pasupathy_shape_compare("yosh",27,"pc2001",22);
    stm_pasupathy_shape_compare("yosh",28,"pc2001",23);
    stm_pasupathy_shape_compare("yosh",29,"pc2001",50);
    stm_pasupathy_shape_compare("yosh",30,"pc2001",38);
    stm_pasupathy_shape_compare("yosh",32,"pc2001",41);
    stm_pasupathy_shape_compare("yosh",36,"pc2001",5);
    stm_pasupathy_shape_compare("yosh",37,"pc2001",6);
    stm_pasupathy_shape_compare("yosh",38,"pc2001",7);
    stm_pasupathy_shape_compare("yosh",39,"pc2001",10);
    stm_pasupathy_shape_compare("yosh",40,"pc2001",11);
    stm_pasupathy_shape_compare("yosh",41,"pc2001",14);
    stm_pasupathy_shape_compare("yosh",42,"pc2001",15);
    exit(0);
  }
  if (strcmp(shape_set,"yasm")==0){
    sz_scale = size / (1.7 * 1.85);  
    px = (int)(size / 4);  
  }else{
    sz_scale = size / (pasu_shape_02[5] * 1.85);  
    px = (int)(size / 2);  
  }
  normflag = 0;  
  stm_pasupathy_shape_get_coords(mylogf,shape_set,shid,theta,normflag,px,
				 &sx,&sy,&sn);
  stm_pasupathy_shape_get_inside_point(shape_set,shid,theta,&x0,&y0);
  if (fill_flag == 1){
    stm_shape_draw_fill(mylogf,data,xn,yn,sx,sy,sn,sz_scale,xc,yc,fgval,bgval,
			x0,y0,aa_sd);
  }else{
    stm_shape_draw_contour_antialias(data,xn,yn,sx,sy,sn,sz_scale,xc,yc,fgval,
				     aa_sd);
  }
  if (norm_flag == 0){
    ; 
  }else if (norm_flag == 1){
    if (bgval != 0.0)
      mylog(mylogf,"*** STM_PASUPATHY_SHAPE: norm interacts with bgval");
    norm_area_2d_farray(data,xn,yn,1.0);  
  }else{
    exit_error("STM_PASUPATHY_SHAPE","Bad value for 'norm_flag'");
  }
  myfree(sx);
  myfree(sy);
}
