/*C  (C) Copyright  Wyeth Bair 1993-2010                                     */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "my_util.h"
#include "nr_util.h"
#include "carray_util.h"
#include "iarray_util.h"
#include "farray_util.h"
#include "func_util.h"
#include "myrand_util.h"
#include "data_util.h"
#include "spike_util.h"
#include "paramfile_util.h"
#include "sig_util.h"
#include "mod_util.h"
#include "pop_cell_util.h"
#include "mod_conn_util.h"
#include "mod_mon_util.h"
#include "mod.h" 
#include "ifc.h" 
#include "paramfile.h"
#define HMAX 0.0005       
int global_dump_flag = 0;
int global_dump_flag2 = 0;
int globseed;
#define IM1 2147483563
#define IM2 2147483399
#define AM (1.0/IM1)
#define IMM1 (IM1-1)
#define IA1 40014
#define IA2 40692
#define IQ1 53668
#define IQ2 52774
#define IR1 12211
#define IR2 3791
#define NTAB 32
#define NDIV (1+IMM1/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)
/*C  (C) Copr. 1986-92 Numerical Recipes Software #.3.                       */
float pop_util_ran2(idum)
     int *idum; 
{
  int j;
  int k; 
  static int idum2=123456789; 
  static int iy=0; 
  static int iv[NTAB]; 
  float temp;
  if (*idum <= 0) {
    if (-(*idum) < 1) *idum=1;
    else *idum = -(*idum);
    idum2=(*idum);
    for (j=NTAB+7;j>=0;j--) {
      k=(*idum)/IQ1;
      *idum=IA1*(*idum-k*IQ1)-k*IR1;
      if (*idum < 0) *idum += IM1;
      if (j < NTAB) iv[j] = *idum;
    }
    iy=iv[0];
  }
  k=(*idum)/IQ1;
  *idum=IA1*(*idum-k*IQ1)-k*IR1;
  if (*idum < 0) *idum += IM1;
  k=idum2/IQ2;
  idum2=IA2*(idum2-k*IQ2)-k*IR2;
  if (idum2 < 0) idum2 += IM2;
  j=iy/NDIV;
  iy=iv[j]-idum2;
  iv[j] = *idum;
  if (iy < 1) iy += IMM1;
  if ((temp=AM*iy) > RNMX) return RNMX;
  else return temp;
}
#undef IM1
#undef IM2
#undef AM
#undef IMM1
#undef IA1
#undef IA2
#undef IQ1
#undef IQ2
#undef IR1
#undef IR2
#undef NTAB
#undef NDIV
#undef EPS
#undef RNMX
/*C  (C) Copr. 1986-92 Numerical Recipes Software #.3.                       */
float pop_util_gasdev(idum)
     int *idum;
{
  static int iset=0;
  static float gset;
  float fac,rsq,v1,v2;
  if ((iset == 0) || (*idum <= 0)){  
    do {
      v1=2.0*pop_util_ran2(idum)-1.0;
      v2=2.0*pop_util_ran2(idum)-1.0;
      rsq=v1*v1+v2*v2;
    } while (rsq >= 1.0 || rsq == 0.0);
    fac=sqrt(-2.0*log(rsq)/rsq);
    gset=v1*fac;
    iset=1;
    return v2*fac;
  }else{
    iset=0;
    return gset;
  }
}
float *pop_util_get_mask_diff_exp(m,name,rnm)
     struct model_struct *m;          
     char name[];                     
     int *rnm;                        
{
  int nm;
  float t1,t2,amp,*mask;
  char tstr[SLEN];
  sprintf(tstr,"%s_tau_f",name);
  t1 = paramfile_get_float_param_or_exit(m->ppl,tstr);  
  sprintf(tstr,"%s_tau_r",name);
  t2 = paramfile_get_float_param_or_exit(m->ppl,tstr);
  sprintf(tstr,"%s_amp",name);
  amp = paramfile_get_float_param_or_exit(m->ppl,tstr);
  nm = (int)(t1 * 7.0);        
  mask = diff_exp_farray(0.0,amp,t1,t2,nm);
  *rnm = nm;
  return mask;
}
float *pop_util_get_mask_diff_exp_mech(mech,rnm)
     struct pop_mech *mech;  
     int *rnm;               
{
  int nm;
  float *mask;
  nm = (int)(mech->tau_f * 7.0);    
  mask = diff_exp_farray(0.0,mech->amp,mech->tau_f,mech->tau_r,nm);
  *rnm = nm;
  return mask;
}
float *popu_mech_onode_get_mask_alpha(o,sampling,rn)
     struct onode *o;    
     float sampling;     
     int *rn;            
{
  int maskn;
  float tau,amp,alpha,*mask;
  char *mtype;
  mtype = onode_getpar_chr_exit(o,"type");
  tau   = onode_getpar_flt_exit(o,"tau");
  amp   = onode_getpar_flt_exit(o,"amp");
  if (strcmp(mtype,"psg_alpha")!=0)
    exit_error("POPU_MECH_ONODE_GET_MASK_ALPHA","Wrong mask type");
  maskn = (int)(tau * sampling * 10.0);
  alpha = 1.0/(tau * sampling); 
  mask = alpha_farray(0.0,alpha,1.0,maskn);
  make_max_const_farray(mask,maskn,amp);
  myfree(mtype);
  *rn = maskn;
  return mask;
}
void pop_util_onode_get_pop_xn_yn_zn(o,name,rxn,ryn,rzn)
     struct onode *o;    
     char *name;
     int *rxn,*ryn,*rzn;
{
  int axn,ayn,azn;
  char *aname;
  struct onode *po,*geo,*ao;
  po = onode_get_node_type_item_val(o,"pop","name",name);
  if (po == NULL){
    *rxn = *ryn = *rzn = -1;
    return;
  }
  geo = onode_child_get_unique(po,"geometry");
  if (onode_test_ostr(po,"area")){
    aname = onode_getpar_chr_exit(po,"area");
    ao = onode_get_node_type_item_val(o,"area","name",aname);
    myfree(aname);
    axn = onode_getpar_int_exit(ao,"xn");
    ayn = onode_getpar_int_exit(ao,"yn");
    azn = onode_getpar_int_dflt(ao,"zn",1);
    *rxn = onode_getpar_int_dflt(geo,"xn",axn);
    *ryn = onode_getpar_int_dflt(geo,"yn",ayn);
  }else{
    *rxn = onode_getpar_int_exit(geo,"xn");
    *ryn = onode_getpar_int_exit(geo,"yn");
    azn = 1;
  }
  *rzn = onode_getpar_int_dflt(geo,"zn",azn);
}
struct pop_area *pop_util_get_area_by_name(mpt,name)
     struct pop_top *mpt;
     char *name;
{
  int i;
  struct pop_area *pa;
  pa = NULL;
  for(i=0;i<mpt->narea;i++){
    if (strcmp(mpt->area[i]->name,name)==0){
      if (pa == NULL)
	pa = mpt->area[i];
      else
	exit_error("POP_UTIL_GET_AREA_BY_NAME","Duplicate area names");
    }
  }
  return pa;
}
struct pop_layer *pop_util_get_layer_by_name(mpt,name)
     struct pop_top *mpt;
     char *name;
{
  int i;
  struct pop_layer *pl;
  pl = NULL;
  for(i=0;i<mpt->nlay;i++){
    if (strcmp(mpt->lay[i]->name,name)==0){
      if (pl == NULL)
	pl = mpt->lay[i];
      else
	exit_error("POP_UTIL_GET_LAYER_BY_NAME","Duplicate population names");
    }
  }
  return pl;
}
int popu_get_layer_index_by_name(mpt,name)
     struct pop_top *mpt;
     char *name;
{
  int i,k;
  k = -1;
  for(i=0;i<mpt->nlay;i++){
    if (strcmp(mpt->lay[i]->name,name)==0){
      if (k == -1)
	k = i;
      else
	exit_error("POPU_GET_LAYER_INDEX_BY_NAME","Duplicate population name");
    }
  }
  return k;
}
struct pop_layer *popu_get_lay_named(mpt,o,parname,callstr)
     struct pop_top *mpt;   
     struct onode *o;       
     char *parname;         
     char *callstr;         
{
  char *lay_name;
  struct pop_layer *lay;
  lay_name = onode_getpar_chr_exit(o,parname);
  lay = pop_util_get_layer_by_name(mpt,lay_name);
  if (lay == NULL){
    mylog_cval(mpt->logf,"*** Parameter Name",parname);
    mylog_cval(mpt->logf,"*** Caller String",callstr);
    mylog_cval(mpt->logf,"*** Layer Name",lay_name);
    mylog_exit(mpt->logf,"POPU_GET_LAY_NAMED  Layer not found.");
  }
  myfree(lay_name);
  return lay;
}
struct pop_mech *popu_get_mech_named(mpt,lay,o,parname,callstr)
     struct pop_top *mpt;   
     struct pop_layer *lay; 
     struct onode *o;       
     char *parname;         
     char *callstr;         
{
  char *mech_name;
  struct pop_mech *mech;
  mech_name = onode_getpar_chr_exit(o,parname);
  mech = pop_cell_get_mech_by_name(lay,mech_name);
  if (mech == NULL){
    mylog_cval(mpt->logf,"Parameter Name",parname);
    mylog_cval(mpt->logf,"Caller String",callstr);
    mylog_cval(mpt->logf,"Mech Name",mech_name);
    mylog_exit(mpt->logf,"POPU_GET_MECH_NAMED  Named mechanism not found.");
  }
  myfree(mech_name);
  return mech;
}
struct pop_cell *popu_get_c_named(mpt,o,parname,callstr)
     struct pop_top *mpt;   
     struct onode *o;       
     char *parname;         
     char *callstr;         
{
  int x,y,z;
  char *lay_name,*cell_name;
  struct pop_layer *lay;
  struct pop_cell *c;
  cell_name = onode_getpar_chr_exit(o,parname);
  pop_cell_parse_unit_name(cell_name,&lay_name,&x,&y,&z);
  lay = pop_util_get_layer_by_name(mpt,lay_name);
  if (lay == NULL){
    mylog_cval(mpt->logf,"*** Parameter Name",parname);
    mylog_cval(mpt->logf,"*** Caller String",callstr);
    mylog_cval(mpt->logf,"*** Layer Name",lay_name);
    mylog_exit(mpt->logf,"POPU_GET_C_NAMED  Layer not found.");
  }
  c = &(lay->c[x][y][z]);
  if (c == NULL){
    mylog_cval(mpt->logf,"*** cell name",cell_name);
    mylog_exit(mpt->logf,"POPU_GET_C_NAMED  cell not found");
  }
  myfree(lay_name);
  myfree(cell_name);
  return c;
}
void pop_util_get_xn_yn_zn_by_name(mpt,name,rxn,ryn,rzn)
     struct pop_top *mpt;
     char *name;
     int *rxn,*ryn,*rzn;
{
  struct pop_layer *pl;
  pl = pop_util_get_layer_by_name(mpt,name);
  *rxn = pl->xn;
  *ryn = pl->yn;
  *rzn = pl->zn;
}
struct pop_map *pop_util_get_map_by_name(mpt,name)
     struct pop_top *mpt;
     char *name;
{
  int i;
  struct pop_map *pm;
  char *tname;
  pm = NULL;
  for(i=0;i<mpt->nmap;i++){
    tname = onode_getpar_chr_ptr_exit(mpt->map[i]->o,"name");
    if (strcmp(tname,name)==0){
      if (pm == NULL)
	pm = mpt->map[i];
      else
	exit_error("POP_UTIL_GET_MAP_BY_NAME","Duplicate map names");
    }
  }
  return pm;
}
void pop_util_map_make(mpt,k)
     struct pop_top *mpt;        
     int k;
{
  char *mylogf,ggstr[SLEN],*aname,*mtype;
  struct onode *mo;       
  struct pop_map *mp;     
  struct pop_area *area;
  mylogf = mpt->logf;
  mylog(mylogf,"  POP_UTIL_MAP_MAKE\n");
  if ((k < 0) || (k >= mpt->nmap))
    mylogx(mylogf,"POP_UTIL_MAP_MAKE","Map index out of bounds");
  mo = mpt->map[k]->o;
  mp = mpt->map[k];
  if (onode_item(mo,"area")==1){
    aname = onode_getpar_chr_ptr_exit(mo,"area");
    area = pop_util_get_area_by_name(mpt,aname); 
    if (area == NULL){
      sprintf(ggstr,"Could not find 'area' %s for map.\n",aname);
      mylogx(mylogf,ggstr);
    }
    mtype = onode_getpar_chr_ptr_exit(mo,"type");
    if (strcmp(mtype,"ori")==0){
      mod_conn_onode_make_map_ori(mylogf,mp,area->xn,area->yn);
    }else if (strcmp(mtype,"ocdom")==0){
      mod_conn_onode_make_map_od(mylogf,mp,area->xn,area->yn);
    }else if (strcmp(mtype,"sf")==0){
      mod_conn_onode_make_map_sf(mylogf,mp,area->xn,area->yn);
    }else if (strcmp(mtype,"rfx")==0){
      mod_conn_onode_make_map_rfx(mylogf,mp,area->xn,area->yn,area->xf);
    }else{
      sprintf(ggstr,"POP_UTIL_MAP_MAKE:  Unknown map type,  %s\n",mtype);
      mylogx(mylogf,ggstr);
    }
  }else{
    sprintf(ggstr,"*** WARNING: Map should contain 'area' parameter.\n");
    mylog(mylogf,ggstr);
  }
}
float **popu_warp_apply(map,xn,yn,wx,wy)
     float **map;      
     int xn,yn;        
     float **wx,**wy;  
{
  int i,j;
  int xi,yi;
  float **wmap;
  wmap = get_2d_farray(xn,yn);
  for (i=0;i<xn;i++){
    for(j=0;j<yn;j++){
      xi = my_rint(wx[i][j]);      
      yi = my_rint(wy[i][j]);
      while(xi < 0)	xi += xn;  
      while(xi >= xn)   xi -= xn;  
      while(yi < 0)     yi += yn;
      while(yi >= yn)   yi -= yn;
      wmap[i][j] = map[xi][yi];
    }
  }
  return wmap;
}
float popu_warp_remap(y,n,x,dx)
     float *y;    
     int n;       
     float x;     
     float dx;    
{
  int i;
  float v,r;
  i = (int)(x/dx);   
  if (x < 0.0){
    v = y[0]; 
  }else if (i >= (n-1)){  
    v = y[n-1]*x/((n-1)*dx);
  }else{
    r = x - ((float)i*dx);  
    v = y[i] + (r/dx)*(y[i+1]-y[i]);  
  }
  return v;
}
void popu_warp_map_coords(xn,yn,wpar,nwarps,fns,n,xfact,rwx,rwy)
     int xn,yn;       
     float **wpar;    
     int nwarps;      
     float **fns;     
     int n;           
     float xfact;     
     float ***rwx,***rwy;  
{
  int i,j,w;
  int xi,yi;
  float cx,cy,dx,dy,rr,x,y,r,theta,**wx,**wy;
  wx = get_2d_farray(xn,yn);
  wy = get_2d_farray(xn,yn);
  for (i=0;i<xn;i++){
    for(j=0;j<yn;j++){
      x = (float)i;
      y = (float)j;
      for(w=0;w<nwarps;w++){  
	cx = wpar[w][4];  
	cy = wpar[w][5];
	dx = x - cx;  
	dy = y - cy;
	r = sqrt(dx*dx + dy*dy);
	theta = atan2f(dy,dx);
        rr = popu_warp_remap(fns[w],n,r,xfact); 
	x = cx + rr * cos(theta); 
	y = cy + rr * sin(theta);
      }
      wx[i][j] = x; 
      wy[i][j] = y;
    }
  }
  *rwx = wx;  
  *rwy = wy;
}
float **popu_warp_get_map_arrays(nwarps,n,params,xfact,dumpfile)
     int nwarps;        
     int n;             
     float **params;    
     float xfact;       
     char *dumpfile;    
{
  int i,j;
  float a,sf,ph,sigma,**fns,x,ff,*t;
  char tstr[SLEN];
  if (dumpfile != NULL)
    remove_file(dumpfile);
  t = (float *)myalloc(n*sizeof(float));
  fns = get_2d_farray(nwarps,n);
  for (i=0;i<nwarps;i++){
    a     = params[i][0];
    sf    = params[i][1];   
    ph    = params[i][2];   
    sigma = params[i][3];
    ff = sf * 2.0 * M_PI;   
    fns[i][0] = 0.0;
    for (j=1;j<n;j++){
      x = (float)j * xfact;
      t[j] = a*cos(ff*x + ph) * exp(-x*x/(2.0*sigma*sigma));
      fns[i][j] = fns[i][j-1] + 1.0 + t[j];
    }
    if (dumpfile != NULL){
      sprintf(tstr,"warp_%d",i);
      append_farray_plot("zz.fmap.pl",tstr,fns[i],n,1);
      sprintf(tstr,"gabor_%d",i);
      append_farray_plot("zz.fmap.pl",tstr,t,n,1);
    }
  }
  myfree(t);
  return fns;
}
float **popu_warp_gen_params(mylogf,wo,xn,yn,rnw,rnpar,rfmlen,rxfact)
     char *mylogf;       
     struct onode *wo;   
     int xn,yn;          
     int *rnw;           
     int *rnpar;         
     int *rfmlen;        
     float *rxfact;      
{
  int i;
  int npar,nw,seed;
  int xctype;  
  float a,sigma,ph,sf,cx,cy,**params,marg,a_max,a_min,sd_min,sd_max;
  char ggstr[SLEN];
  nw     = onode_getpar_int_exit(wo,"n");     
  seed   = onode_getpar_int_exit(wo,"seed");  
  a_min  = onode_getpar_flt_exit(wo,"ampl_min");
  a_max  = onode_getpar_flt_exit(wo,"ampl_max");
  sd_min = onode_getpar_flt_exit(wo,"sd_min");
  sd_max = onode_getpar_flt_exit(wo,"sd_max");
  marg   = onode_getpar_int_exit(wo,"margin");  
  xctype = onode_getpar_int_exit(wo,"expcon_type");
  *rfmlen = (int)(1 + sqrt(xn*xn + yn*yn));  
  *rxfact = 1.0;
  npar = 6;
  params = get_2d_farray(nw,npar);
  if (seed > 0)
    seed = -seed;
  sprintf(ggstr,"                a      sf    ph     sigma     cx      cy\n");
  mylog(mylogf,ggstr);
  for (i=0;i<nw;i++){
    a     =  a_min + ( a_max -  a_min)*myrand_util_ran2(&seed);
    sigma = sd_min + (sd_max - sd_min)*myrand_util_ran2(&seed);
    if (xctype == 0)        
      ph = 0.0;
    else if (xctype == 1)   
      ph = M_PI;
    else if (xctype == 2){  
      if (i%2 == 0)
	ph = M_PI;
      else
	ph = 0.0;
    }else if (xctype == 3){  
      if (myrand_util_ran2(&seed) < 0.5){
	ph = 0.0;
      }else{
	ph = M_PI;
      }
    }
    cx = marg + (float)(xn-2*marg)*myrand_util_ran2(&seed);
    cy = marg + (float)(yn-2*marg)*myrand_util_ran2(&seed);
    sf = 1.0 / (3.2 * sigma);  
    params[i][0] = a;
    params[i][1] = sf;
    params[i][2] = ph;
    params[i][3] = sigma;
    params[i][4] = cx;
    params[i][5] = cy;
    sprintf(ggstr,"    Warp[%2d]  %.4f %.4f %.2f %9.4f %7.2f %7.2f\n",i,
	    params[i][0],params[i][1],params[i][2],params[i][3],params[i][4],
	    params[i][5]);
    mylog(mylogf,ggstr);
  }
  *rnw = nw;
  *rnpar = npar;
  return params;
}
float **popu_warp_map_simple(nx,ny,linesep)
     int nx,ny;    
     int linesep;  
{
  int i,j;
  float **map;
  map = get_2d_farray(ny,nx);
  for (i=0;i<nx;i++){
    for (j=0;j<ny;j++){
      if ((i % linesep == 0) || (j % linesep == 0))
	map[j][i] = 1.0;
      else
	map[j][i] = 0.0;
    }
  }
  return map;
}
void popu_warp_map(mpt,wo)
     struct pop_top *mpt;    
     struct onode *wo;       
{
  int i;
  int xn,yn,nw,fmlen,npar;
  float xfact,**wpar,**fmap,**wx,**wy,**wmap;
  char *mylogf,ggstr[SLEN],*dumpfile,*txtfile;
  struct pop_map *mp;     
  mylogf = mpt->logf;
  mylog(mylogf,"  POPU_WARP_MAP\n");
  if (mpt->nmap <= 0)
    return;
  xn = mpt->map[0]->xn;
  yn = mpt->map[0]->yn;
  wpar = popu_warp_gen_params(mylogf,wo,xn,yn,&nw,&npar,&fmlen,&xfact);
  dumpfile = NULL;
  fmap = popu_warp_get_map_arrays(nw,fmlen,wpar,xfact,dumpfile);
  popu_warp_map_coords(xn,yn,wpar,nw,fmap,fmlen,xfact,&wx,&wy);
  for(i=0;i<mpt->nmap;i++){
    mp = mpt->map[i];  
    if ((mp->xn != xn) || (mp->yn != yn))
      mylogx(mylogf,"POPU_WARP_MAP  All maps are not the same size");
    wmap = popu_warp_apply(mp->map2,xn,yn,wx,wy); 
    free_2d_farray(mp->map2,xn);  
    mp->map2 = wmap;
    sprintf(ggstr,"    Warping map %d  xn,yn  %d %d\n",i,xn,yn);
    mylog(mylogf,ggstr);
    txtfile = onode_getpar_chr_dflt(mp->o,"write_as_text_warp",NULL);
    if (txtfile != NULL){
      mod_conn_write_map_text(txtfile,mp->map2,xn,yn,1);  
      myfree(txtfile);
    }
  }
  free_2d_farray(fmap,nw);
  free_2d_farray(wpar,nw);
  free_2d_farray(wx,xn);
  free_2d_farray(wy,xn);
}
int pop_util_check_lclass(ltype,lclass)
     char *ltype;
     char *lclass;
{
  int flag;
  flag = 0;
  if (strcmp(lclass,"lgn0")==0){
    if (strcmp(ltype,"lgn")==0)
      flag = 1;
  }else if (strcmp(lclass,"rgci")==0){
    if (strcmp(ltype,"retina0_rgc0")==0)
      flag = 1;
  }else{
    exit_error("POP_UTIL_CHECK_LCLASS","Unknown layer class");
  }
  return flag;
}
int pop_util_lclass_is_lgn0(lay)
     struct pop_layer *lay;
{
  int flag;
  char *lt;
  lt = lay->laytype;
  if (lt == NULL)
    exit_error("POP_UTIL_LCLASS_IS_LGN0","Null layer type");
  flag = pop_util_check_lclass(lt,"lgn0");
  return flag;
}
int pop_util_lclass_is_rgci(lay)
     struct pop_layer *lay;
{
  int flag;
  char *lt;
  flag = 0;
  lt = lay->laytype;
  if (lt == NULL)
    exit_error("POP_UTIL_LCLASS_IS_RGCI","Null layer type");
  if (strcmp(lt,"retina0_gc0")==0){
    if (lay->geomt == NULL)
      exit_error("POP_UTIL_LCLASS_IS_RGCI","geomt is NULL");
    else
      if (strcmp(lay->geomt,"irregular")==0){
	flag = 1;
      }
  }
  return flag;
}
struct pop_cell ***popu_make_init_cells(xn,yn,zn,pl)
     int xn,yn,zn;          
     struct pop_layer *pl;  
{
  int i,j,k;
  struct pop_cell ***c,*cp;
  c = (struct pop_cell ***)myalloc(xn*sizeof(struct pop_cell **));
  for(i=0;i<xn;i++){
    c[i] = (struct pop_cell **)myalloc(yn*sizeof(struct pop_cell *));
    for(j=0;j<yn;j++){
      c[i][j] = (struct pop_cell *)myalloc(zn*sizeof(struct pop_cell));
      for(k=0;k<zn;k++){
	cp = &(c[i][j][k]);
	pop_cell_init_cell(cp);
	cp->rfx = (float)i;
	cp->rfy = (float)j;
	cp->layx = i;  
	cp->layy = j;  
	cp->layz = k;  
	cp->pl = pl;  
      }
    }
  }
  return c;
}
struct pop_layer *popu_make_pop_layer(pname)
     char *pname;
{
  struct pop_layer *pl;
  pl = (struct pop_layer *)myalloc(sizeof(struct pop_layer));
  if (pname == NULL)
    pl->name = NULL;
  else
    pl->name = strdup(pname);
  pl->runflag = 0;
  pl->lgn_in_flag  = 0;  
  pl->lgn_in_n     = 0;  
  pl->lgn_gain     = 0;  
  pl->lgn_gain_tau = 0.010;
  pl->lgn_gain_ca  = 1.0;
  pl->lgn_gain_cb  = 1.0;
  pl->laytype = NULL;
  pl->icon = NULL;
  pl->area = NULL;
  pl->ninlist = 0;
  pl->inlist = NULL;
  pl->geomt = NULL;
  pl->xn = -1;
  pl->yn = -1;
  pl->zn = -1;
  pl->x0  = pl->y0  = 0.0;
  pl->xf  = pl->yf  = 0.0;
  pl->umx = pl->umy = 0.0;
  pl->oddxoff = 0;
  pl->irr_n = 0;
  pl->irr_x = NULL;
  pl->irr_y = NULL;
  pl->irr_id = NULL;
  pl->save_name_list = NULL;
  pl->layseed_n = -1;
  pl->layseed = NULL;
  pl->nmech = 0;
  pl->mech = NULL;
  pl->c = NULL;
  pl->attrib_fn = 0;
  pl->attrib_fname = NULL;
  pl->ifcp = NULL;
  pl->nmda_p = NULL;
  pl->csav_n = 0;
  pl->csav_datid = NULL;
  pl->csav_source = NULL;
  pl->csav_samp = NULL;
  pl->cnt = NULL;
  pl->s = NULL;
  pl->rl = NULL;
  pl->rr = NULL;
  pl->rgl = NULL;      
  pl->rgr = NULL;
  return pl;
}
struct pop_layer *popu_make_layer_irreg(pname,n,zn,cx,cy,rfx,rfy)
     char *pname;      
     int n;            
     int zn;           
     float *cx,*cy;    
     float *rfx,*rfy;  
{
  int i,j,k;
  struct pop_layer *pl;
  struct pop_cell ***c,*cp;
  pl = popu_make_pop_layer(pname);
  pl->laytype = strdup("irreg");
  pl->xn = n;
  pl->yn = 1;
  pl->zn = zn;
  c = (struct pop_cell ***)myalloc(pl->xn*sizeof(struct pop_cell **));
  for(i=0;i<pl->xn;i++){
    c[i] = (struct pop_cell **)myalloc(pl->yn*sizeof(struct pop_cell *));
    for(j=0;j<pl->yn;j++){
      c[i][j] = (struct pop_cell *)myalloc(pl->zn*sizeof(struct pop_cell));
      for(k=0;k<pl->zn;k++){
	cp = &(c[i][j][k]);
	pop_cell_init_cell(cp);
	cp->cx = cx[i];
	cp->cy = cy[i];
	cp->rfx = rfx[i];
	cp->rfy = rfy[i];
	cp->layx = i;  
	cp->layy = j;  
	cp->layz = k;  
      }
    }
  }
  pl->c = c;
  return pl;
}
struct pop_layer *popu_make_layer_cart(pname,xn,yn,zn)
     char *pname;        
     int xn,yn,zn;       
{
  struct pop_layer *pl;
  pl = popu_make_pop_layer(pname);
  pl->laytype = strdup("default");
  pl->xn = xn;
  pl->yn = yn;
  pl->zn = zn;
  pl->c = popu_make_init_cells(xn,yn,zn,pl);
  return pl;
}
struct pop_layer *popu_make_layer_template(name,xn,yn,zn,pa,lpre,fx,fy,fw,fn)
     char *name;
     int xn,yn,zn;
     struct pop_area *pa;
     struct pop_layer *lpre;
     int   **fx;  
     int   **fy;  
     float **fw;  
     int    *fn;  
{
  int i;
  int zi,stype,inindex,pxn,pyn;
  float w,tdelay;
  struct pop_layer *pl;
  struct onode *ot,*o;
  struct pop_cell *post,*pre;
  struct pop_syn *tsyn;
  pl = popu_make_layer_cart(name,xn,yn,zn);
  pl->geomt = strdup("default");
  pl->area = pa;
  pl->x0 = pl->y0 = 0;
  pl->xf = pl->yf = 1.0;
  pl->ninlist = 1;
  pl->inlist = (struct onode **)myalloc(sizeof(struct onode *));
  ot = paramfile_onode_create_onode();
  ot->otype = strdup("input");
  o = paramfile_onode_get_init_onode("item",NULL,"pop_origin",lpre->name,
				     1,NULL,NULL,1);
  onode_insert_child_at_end(ot,o);
  o = paramfile_onode_get_init_onode("item",NULL,"type","template",
				     1,NULL,NULL,1);
  onode_insert_child_at_end(ot,o);
  pl->inlist[0] = ot;
  stype = 1;
  tdelay = 0.0;
  inindex = 0;
  w = 1.0;
  pxn = lpre->xn;
  pyn = lpre->yn;
  for(zi=0;zi<zn;zi++){
    post = &(pl->c[0][0][zi]);
    for(i=0;i<fn[zi];i++){
      pre = &(lpre->c[fx[zi][i]][fy[zi][i]][0]);
      tsyn = pop_cell_add_synapse(pre,post,stype,fw[zi][i],tdelay,inindex);
    }
  }
  return pl;
}
struct pop_top *popu_make_pop_top(logf,sscale,tscale,xn,yn,tn)
     char *logf;
     float sscale;
     float tscale;
     int xn,yn,tn;
{
  struct pop_top *mpt;
  mpt = (struct pop_top *)myalloc(sizeof(struct pop_top));
  if  (logf == NULL){
    mpt->logf = NULL;
  }else
    mpt->logf = strdup(logf);
  mpt->tscale = tscale;
  mpt->sscale = sscale;
  mpt->xn = xn;
  mpt->yn = yn;
  mpt->tn = tn;
  mpt->x1 = 0.0;
  mpt->x2 = mpt->tscale * (float)mpt->tn;
  mpt->out_prefix = NULL;
  mpt->binoc = 0;     
  mpt->retflag = 0;   
  mpt->narea = 0;
  mpt->area = NULL;
  mpt->nlay = 0;
  mpt->lay = NULL;
  mpt->nmap = 0;
  mpt->map = NULL;
  mpt->conn_rw = NULL;
  mpt->conn_read_dir = NULL;
  mpt->conn_write_dir = NULL;
  return mpt;
}
struct pop_area *popu_make_area_default(name,xn,yn,sscale,um)
     char *name;
     int xn,yn;
     float sscale;
     float um;
{
  struct pop_area *pa;
  pa = (struct pop_area *)myalloc(sizeof(struct pop_area));
  pa->name = strdup(name);
  pa->x0   = 0;
  pa->y0   = 0;
  pa->xf   = 1.0;
  pa->yf   = 1.0;
  pa->xn   = xn;
  pa->yn   = yn;
  pa->umx  = um;
  pa->umy  = um;
  pa->sscale = sscale;
  return pa;
}
struct pop_area *popu_make_area(logf,ao,sscale)
     char *logf;
     struct onode *ao;  
     float sscale;      
{
  struct pop_area *pa;
  pa = (struct pop_area *)myalloc(sizeof(struct pop_area));
  pa->name = onode_getpar_chr_exit(ao,"name");
  pa->x0   = onode_getpar_flt_exit(ao,"x0");
  pa->y0   = onode_getpar_flt_exit(ao,"y0");
  pa->xf   = onode_getpar_flt_dflt(ao,"xf",1.0);
  pa->yf   = onode_getpar_flt_dflt(ao,"yf",1.0);
  pa->xn   = onode_getpar_int_exit(ao,"xn");
  pa->yn   = onode_getpar_int_exit(ao,"yn");
  pa->umx  = onode_getpar_flt_dflt(ao,"umx",0.0);
  pa->umy  = onode_getpar_flt_dflt(ao,"umy",pa->umx);
  pa->sscale = sscale;
  return pa;
}
void popu_init_geom(logf,o,po,pl,retflag)
     char *logf;
     struct onode *o;    
     struct onode *po;   
     struct pop_layer *pl;
     int retflag;             
{
  struct onode *geo;
  geo = onode_child_get_unique(po,"geometry");
  pl->geomt = onode_getpar_chr_dflt(geo,"type","default");
  if (pop_util_lclass_is_lgn0(pl) ||  
      ((strcmp(pl->laytype,"retina0_gc0")==0) && (retflag==100))){
    pl->xn = onode_getpar_int_exit(o,"xn"); 
    pl->yn = onode_getpar_int_exit(o,"yn");
  }else if (pl->area != NULL){
    pl->xn = onode_getpar_int_dflt(geo,"xn",pl->area->xn); 
    pl->yn = onode_getpar_int_dflt(geo,"yn",pl->area->yn);
  }else{
    pl->xn = onode_getpar_int_exit(geo,"xn");
    pl->yn = onode_getpar_int_exit(geo,"yn");
  }
  pl->zn = onode_getpar_int_dflt(geo,"zn",1);
  pl->x0      = onode_getpar_flt_dflt(geo,"x0",0.0);
  pl->y0      = onode_getpar_flt_dflt(geo,"y0",0.0);
  pl->xf      = onode_getpar_flt_dflt(geo,"xf",1.0);
  pl->yf      = onode_getpar_flt_dflt(geo,"yf",1.0);
  pl->oddxoff = onode_getpar_flt_dflt(geo,"oddxoff",0.0);
}
struct pop_layer *popu_input_get_origin(mpt,lay,iin)
     struct pop_top *mpt;     
     struct pop_layer *lay;   
     int iin;                 
{
  struct pop_layer *l;
  struct onode *o;
  o = lay->inlist[iin];
  l = popu_get_lay_named(mpt,o,"pop_origin","POPU_INPUT_GET_ORIGIN");
  return l;
}
int popu_input_is_paired(lay,iin)
     struct pop_layer *lay;
     int iin;
{
  int flag;
  struct onode *o;
  flag = 0;
  if (lay->inlist != NULL){
    o = onode_child_get_unique(lay->inlist[iin],"input_pair");
    if (o != NULL)
      flag = 1;
    o = onode_get_item_child(lay->inlist[iin],"syn_int");
    if (o != NULL)
      flag = 1;
  }
  return flag;
}
int popu_input_is_bg(lay,iin)
     struct pop_layer *lay;
     int iin;
{
  int flag;
  struct onode *o;
  flag = 0;
  o = lay->inlist[iin];
  flag = onode_test_chr(o,"type","bg");
  if (o != NULL)
    flag = 1;
  return flag;
}
struct pop_layer *popu_input_get_paired_lay(mpt,lay,iin)
     struct pop_top *mpt;     
     struct pop_layer *lay;
     int iin;
{
  struct onode *po,*io;
  struct pop_layer *lay2;
  lay2 = NULL;
  if (popu_input_is_paired(lay,iin)){
    io = lay->inlist[iin];   
    po = onode_child_get_unique(io,"input_pair");  
    if (po == NULL){
      lay2 = popu_get_lay_named(mpt,io,"pop_origin",
				"POPU_INPUT_GET_PAIRED_LAY");
    }else{
      if (onode_get_item_child(po,"pop_origin_2") != NULL)
	lay2 = popu_get_lay_named(mpt,po,"pop_origin_2",
				  "POPU_INPUT_GET_PAIRED_LAY");
      else 
	lay2 = popu_get_lay_named(mpt,io,"pop_origin",
				  "POPU_INPUT_GET_PAIRED_LAY");
    }
  }  
  return lay2;
}
float *pop_util_get_nmda_shape(alph1,alph2,alph3,amp1,amp2,amp3,n)
     float alph1,alph2,alph3,amp1,amp2,amp3;
     int n;
{
  float *g,*gt;
  g  = alpha_farray(0.0,alph1,amp1,n);
  gt = alpha_farray(0.0,alph2,amp2,n);
  add_to_farray(g,gt,n);
  myfree(gt);
  gt = alpha_farray(0.0,alph3,amp3,n);
  add_to_farray(g,gt,n);
  myfree(gt);
  return g;
}
void pop_util_compute_nmda_table(ppmv,v1,v2,rg,rn)
     int ppmv;    
     int v1,v2;   
     float **rg;  
     int *rn;     
{
  int i;
  int nmv,n;
  float *g,*x,C,a1,a2,b1,b2,A,B1,B2,v,t1,t2;
  nmv = v2 - v1 + 1; 
  n = nmv * ppmv;  
  g = get_zero_farray(n);
  x = get_zero_farray(n);
  for(i=0;i<n;i++){
    v = v1 + (float)i/(float)ppmv; 
    C = 100.0; 
    a1 = exp(-0.016 * v - 2.91);
    a2 = C * exp(-0.045 * v - 6.97); 
    b1 = exp(0.009 * v + 1.22);
    b2 = exp(0.017 * v + 0.96);
    A = exp(-2.847);
    B1 = exp(-0.693);
    B2 = exp(-3.101);
    t1 = (a1 + a2)*(a1*B1 + a2*B2);
    t2 = A*a1*(b1 + B1) + A*a2*(b2 + B2);
    g[i] = 1.0/(1.0 + t1/t2);
    x[i] = v;
  }
  myfree(x);
  *rg = g; *rn = n;
}
void pop_util_nmda_match_intcurr_waveform(myid,mylogf,gwav,n,gov,gov_n,gov_v0,
					  gov_ppmv,nmda_frac,v,v_rev,
					  alph1,alph2,alph3,ramp1,ramp2,ramp3,
					  outprefix,name)
     int myid;         
     char *mylogf;     
     float *gwav;      
     int n;            
     float *gov;       
     int gov_n;        
     float gov_v0;     
     int gov_ppmv;     
     float nmda_frac;  
     float v;          
     float v_rev;      
     float alph1,alph2,alph3;
     float *ramp1,*ramp2,*ramp3;   
     char outprefix[];
     char name[];
{
  int k;
  int nnmda;
  float *gnmda,ampa_curr,nmda_curr,ff,rescale;
  char outfile[SLEN],tstr[SLEN];
  mylog(mylogf,"  POP_UTIL_NMDA_MATCH_INTCURR_WAVEFORM\n");
  sprintf(tstr,"    Matching at %.2f mV\n",v);
  mylog(mylogf,tstr);
  nnmda = (int)(1.0/alph3 * 10.0);
  gnmda = pop_util_get_nmda_shape(alph1,alph2,alph3,*ramp1,*ramp2,*ramp3,
				  nnmda);
  ampa_curr = sum_farray(gwav,n,0,n) * (v_rev - v);
  k = (v - gov_v0) * gov_ppmv;
  if ((k < 0) || (k >= gov_n))
    mylog_exit(mylogf,
	       "POP_UTIL_NMDA_MATCH_INTCURR_WAVEFORM  NMDA table index");
  ff = gov[k];
  sprintf(tstr,"    NMDA voltage-dependent factor at %.2f mV is %f\n",v,ff);
  mylog(mylogf,tstr);
  nmda_curr = ff * sum_farray(gnmda,nnmda,0,nnmda) * (v_rev - v);
  sprintf(tstr,"    Integrated current AMPA = %f\n",ampa_curr);
  mylog(mylogf,tstr);
  sprintf(tstr,"    Integrated current NMDA = %f\n",nmda_curr);
  mylog(mylogf,tstr);
  sprintf(tstr,"    NMDA fraction target is %f\n",nmda_frac);
  mylog(mylogf,tstr);
  rescale = ampa_curr/nmda_curr * nmda_frac/(1.0 - nmda_frac);
  sprintf(tstr,"      rescaling NMDA amplitude by %f\n",rescale);
  mylog(mylogf,tstr);
  *ramp1 *= rescale;
  *ramp2 *= rescale;
  *ramp3 *= rescale;
  myfree(gnmda);
  gnmda = pop_util_get_nmda_shape(alph1,alph2,alph3,*ramp1,*ramp2,*ramp3,
				  nnmda);
  nmda_curr = ff * sum_farray(gnmda,nnmda,0,nnmda) * (v_rev - v);
  sprintf(tstr,"    Integrated current NMDA = %f\n",nmda_curr);
  mylog(mylogf,tstr);
  sprintf(tstr,"    NMDA / (AMPA + NMDA) = %f\n",
	  nmda_curr/(ampa_curr+nmda_curr));
  mylog(mylogf,tstr);
  multiply_farray(gnmda,nnmda,ff);
  myfree(gnmda);
}
void pop_util_compute_intcurr_ampa_nmda_layer(myid,mylogf,lay,v,outprefix)
     int myid;
     char *mylogf;
     struct pop_layer *lay;
     float v;
     char outprefix[];
{
  int nampa;
  float *gampa;
  char tstr[SLEN];
  struct pop_cell *c;
  struct nmda_param *tp;
  mylog(mylogf,"  POP_UTIL_COMPUTE_INTCURR_AMPA_NMDA_LAYER\n");
  c = &(lay->c[0][0][0]);
  tp = lay->nmda_p;
  nampa = (int)(c->ex1tau*1000.0 * 10.0); 
  gampa = alpha_farray(0.0,1.0/(c->ex1tau*1000.0),c->ex1amp,nampa);
  sprintf(tstr,"      AMPA alpha-func tau= %.4f  ampl= %.4f\n",c->ex1tau,
	  c->ex1amp);
  mylog(mylogf,tstr);
  pop_util_nmda_match_intcurr_waveform(myid,mylogf,gampa,nampa,tp->nmda_gv,
				       tp->nmda_n,
				       tp->nmda_v0f,tp->nmda_ppmv,
				       tp->nmda_frac,v,c->cifcp->v_ex,
				       tp->nmda_alpha_1,tp->nmda_alpha_2,
				       tp->nmda_alpha_3,&(tp->nmda_amp_1),
				       &(tp->nmda_amp_2),&(tp->nmda_amp_3),
				       outprefix,"Layer");
  myfree(gampa);
}
void pop_util_compute_intcurr_ampa_nmda_lgn(myid,mylogf,lay,m,v,outprefix)
     int myid;
     char *mylogf;
     struct pop_layer *lay;
     struct model_struct *m;  
     float v;
     char outprefix[];
{
  int gtn;
  float *gt;
  char name[SLEN];
  struct pop_cell *c;
  struct pop_mech *mech;
  struct nmda_param *tp;
  mylog(mylogf,"  POP_UTIL_COMPUTE_INTCURR_AMPA_NMDA_LGN\n");
  c = &(lay->c[0][0][0]);
  tp = lay->nmda_p;
  if (m->ppl != NULL){
    strcpy(name,"lgn2ctx");
    gt = pop_util_get_mask_diff_exp(m,name,&gtn);
  }else{
    mech = pop_cell_get_mech_by_name(lay,"lgn_ex");
    gt = pop_util_get_mask_diff_exp_mech(mech,&gtn);
  }
  pop_util_nmda_match_intcurr_waveform(myid,mylogf,gt,gtn,tp->nmda_gv,
				       tp->nmda_n,
				       tp->nmda_v0f,tp->nmda_ppmv,
				       tp->lgn_nmda_frac,v,c->cifcp->v_ex,
				       tp->nmda_alpha_1,tp->nmda_alpha_2,
				       tp->nmda_alpha_3,
				       &(tp->lgn_nmda_amp_1),
				       &(tp->lgn_nmda_amp_2),
				       &(tp->lgn_nmda_amp_3),
				       outprefix,"LGN");
  myfree(gt);
}
void pop_util_nmda_prep_constants_layer(lay)
     struct pop_layer *lay;
{
  float alph1,alph2,alph3;
  struct nmda_param *tp;
  tp = lay->nmda_p;
  alph1 = 1000.0 * tp->nmda_alpha_1; 
  alph2 = 1000.0 * tp->nmda_alpha_2; 
  alph3 = 1000.0 * tp->nmda_alpha_3; 
  tp->nmda1_aetau = tp->nmda_amp_1 * M_E * alph1;
  tp->nmda2_aetau = tp->nmda_amp_2 * M_E * alph2;
  tp->nmda3_aetau = tp->nmda_amp_3 * M_E * alph3;
  tp->nmda1_tau2inv = alph1 * alph1;
  tp->nmda2_tau2inv = alph2 * alph2;
  tp->nmda3_tau2inv = alph3 * alph3;
  tp->nmda1_totau = 2.0 * alph1;
  tp->nmda2_totau = 2.0 * alph2;
  tp->nmda3_totau = 2.0 * alph3;
}
int popu_flag_resp(r,lay,cmap)
     struct response_struct *r;     
     struct pop_layer *lay;         
     int ***cmap;                   
{
  int i,j;
  int xi,yi,zi,new,tot;
  new = tot = 0;
  for(i=0;i<r->n;i++){
    if (strcmp(r->plname[i],lay->name)==0){
      xi = r->xi[i];
      yi = r->yi[i];
      zi = r->zi[i];
      if (cmap[xi][yi][zi] == 0){
	new += 1;
	cmap[xi][yi][zi] = 1;
      }
      tot += 1;
    }
  }
  return new;
}
void pop_util_response_link(r,laylist,nlay,mylogf)
     struct response_struct *r;     
     struct pop_layer **laylist;    
     int nlay;                      
     char *mylogf;
{
  int i;
  int nmdaflag;
  struct pop_cell *c;
  struct pop_syn *syn;
  struct pop_layer *pl;
  char tstr[SLEN];
  mylog(mylogf,"  POP_UTIL_RESPONSE_LINK\n");
  for(i=0;i<r->n;i++){ 
    if (r->rformat[i] == 0){
      sprintf(tstr,"*** POP_UTIL_RESPONSE_LINK  Format 0 not impl'd yet");
      mylog_exit(mylogf,tstr);
    }else if (r->rformat[i] == 10){
      syn = pop_cell_laylist_get_syn_pointer(laylist,nlay,r->plname[i],
					     r->xi[i],r->yi[i],r->zi[i],
					     r->plname1[i],
					     r->xi1[i],r->yi1[i],r->zi1[i],0);
      if (syn == NULL){
	sprintf(tstr,"*** POP_UTIL_RESPONSE_LINK  No synapse found for %s\n",
		r->nd_name[i]);
	mylog_exit(mylogf,tstr);
      }
      pop_cell_syn_set_save(mylogf,syn,r->datid[i]);
    }else{
      c = pop_cell_laylist_get_cell_pointer(laylist,nlay,r->plname[i],
					    r->xi[i],r->yi[i],r->zi[i]);
      if (c == NULL){
	sprintf(tstr,"*** POP_UTIL_RESPONSE_LINK  No cell found for %s\n",
		r->nd_name[i]);
	mylog_exit(mylogf,tstr);
      }
      if (c->pl->ifcp != NULL){  
	if (c->pl->ifcp->g_tran > 0.0)
	  nmdaflag = 1;
	else
	  nmdaflag = 0;
      }
      pop_cell_set_save(mylogf,c,r->datid[i],nmdaflag);
    }
  }
}
void pop_util_response_handover(m,r,laylist,nlay,mylogf)
     struct model_struct *m;        
     struct response_struct *r;     
     struct pop_layer **laylist;    
     int nlay;                      
     char *mylogf;
{
  int i;
  int flag,din,dn,csi,csav_flag;
  float *gtot,*di,*d;
  struct pop_layer *pl;
  struct pop_cell *c;
  struct pop_syn *syn;
  char tstr[SLEN];
  for(i=0;i<r->n;i++){ 
    if (r->rformat[i] == 0){
      sprintf(tstr,"*** POP_UTIL_RESPONSE_HANDOVER  Format 0 not impl'd yet");
      mylog_exit(mylogf,tstr);
    }else if (r->rformat[i] == 10){
      syn = pop_cell_laylist_get_syn_pointer(laylist,nlay,r->plname[i],
					     r->xi[i],r->yi[i],r->zi[i],
					     r->plname1[i],
					     r->xi1[i],r->yi1[i],r->zi1[i],0);
      if (syn == NULL){
	sprintf(tstr,"*** POP_UTIL_RESPONSE_HANDOVER  No syn found for %s\n",
		r->nd_name[i]);
	mylog_exit(mylogf,tstr);
      }
    }else{
      c = pop_cell_laylist_get_cell_pointer(laylist,nlay,r->plname[i],
					    r->xi[i],r->yi[i],r->zi[i]);
      if (c == NULL){
	sprintf(tstr,"*** POP_UTIL_RESPONSE_HANDOVER  No cell found for %s\n",
		r->nd_name[i]);
	mylog_exit(mylogf,tstr);
      }
    }
    if (strcmp(r->datid[i],"spikes")==0){
      if (!c->savs)
	mylog_exit(mylogf,"*** POP_UTIL_RESPONSE_HANDOVER Spikes not saved\n");
      mod_util_resp_store_s(r,i,c->s,c->ns,1,mylogf);
    }else if (strcmp(r->datid[i],"lgn_gx")==0){
      if (!c->savg)
	mylog_exit(mylogf,"*** POP_UTIL_RESPONSE_HANDOVER g not saved\n");
      if (r->samp[i] != c->samp0) 
	mylog_exit(mylogf,"*** POP_UTIL_RESPONSE_HANDOVER sampl mismatch\n");
      if (c->pl->nmda_p != NULL){
	gtot = add_farrays(c->gtx0,c->gtn0,c->n0);
	mod_util_resp_store_f(r,i,gtot,c->n0,0,mylogf); 
      }else{
	mod_util_resp_store_f(r,i,c->gtx0,c->n0,1,mylogf); 
      }
    }else if (strcmp(r->datid[i],"lgn_gain")==0){
      if (c->ggain != NULL)
	mod_util_resp_store_f(r,i,c->ggain,c->n0,1,mylogf); 
      else
	mylog_exit(mylogf,"*** POP_UTIL_RESPONSE_HANDOVER no gain signal\n");
    }else if (strcmp(r->datid[i],"lgn_ga")==0){
      if (!c->savg)
	mylog_exit(mylogf,"*** POP_UTIL_RESPONSE_HANDOVER g not saved\n");
      if (r->samp[i] != c->samp0) 
	mylog_exit(mylogf,"*** POP_UTIL_RESPONSE_HANDOVER sampl mismatch\n");
      mod_util_resp_store_f(r,i,c->gtx0,c->n0,1,mylogf);
    }else if (strcmp(r->datid[i],"lgn_ia")==0){
      if ((!c->savg) || (!c->savv))
	mylog_exit(mylogf,"*** POP_UTIL_RESPONSE_HANDOVER g or v not saved\n");
      if (r->samp[i] != c->samp0) 
	mylog_exit(mylogf,"*** POP_UTIL_RESPONSE_HANDOVER sampl mismatch\n");
      di = mod_util_resp_derive_current(c->gtx0,c->samp0,(float *)NULL,c->n0,
					1.0,c->vm,c->vmt,c->vn,
					c->cifcp->v_th_x,
					c->cifcp->v_ex,
					mylogf,&din);
      mod_util_resp_store_f(r,i,di,c->n0,1,mylogf);
      myfree(di);
    }else if (strcmp(r->datid[i],"in_ii")==0){
      if ((!c->savg) || (!c->savv))
	mylog_exit(mylogf,"*** POP_UTIL_RESPONSE_HANDOVER g or v not saved\n");
      di = mod_util_resp_derive_current(c->gtin1,0.0,c->gtex1t,c->gtex1n,
					0.001,c->vm,c->vmt,c->vn,
					c->cifcp->v_th_x,
					c->cifcp->v_in,
					mylogf,&din);
      mod_util_resp_store_f_reg(r,i,c->gtex1t,di,din,c->samp0,mylogf);
      myfree(di);
    }else if (strcmp(r->datid[i],"ex_ix")==0){
      if ((!c->savg) || (!c->savv))
	mylog_exit(mylogf,"*** POP_UTIL_RESPONSE_HANDOVER g or v not saved\n");
      di = mod_util_resp_derive_current(c->gtex1,0.0,c->gtex1t,c->gtex1n,
					0.001,c->vm,c->vmt,c->vn,
					c->cifcp->v_th_x,
					c->cifcp->v_ex,
					mylogf,&din);
      mod_util_resp_store_f_reg(r,i,c->gtex1t,di,din,c->samp0,mylogf);
      myfree(di);
    }else if (strcmp(r->datid[i],"lgn_gn")==0){
      if (!c->savg)
	mylog_exit(mylogf,"*** POP_UTIL_RESPONSE_HANDOVER g not saved\n");
      if (r->samp[i] != c->samp0) 
	mylog_exit(mylogf,"*** POP_UTIL_RESPONSE_HANDOVER sampl mismatch\n");
      if (c->pl->nmda_p == NULL){
	mylog_exit(mylogf,"*** POP_UTIL_RESPONSE_HANDOVER  No NMDA\n");
      }
      mod_util_resp_store_f(r,i,c->gtn0,c->n0,1,mylogf);
    }else if (strcmp(r->datid[i],"ex_gx")==0){
      if (!c->savg)
	mylog_exit(mylogf,"*** POP_UTIL_RESPONSE_HANDOVER g not saved\n");
      mod_util_resp_store_f_reg(r,i,c->gtex1t,c->gtex1,c->gtex1n,c->samp0,
				mylogf);
    }else if (strcmp(r->datid[i],"in_gi")==0){
      if (!c->savg)
	mylog_exit(mylogf,"*** POP_UTIL_RESPONSE_HANDOVER g not saved\n");
      mod_util_resp_store_f_reg(r,i,c->gtex1t,c->gtin1,c->gtex1n,c->samp0,
				mylogf);
    }else if (strcmp(r->datid[i],"gad")==0){
      if (!c->savg)
	mylog_exit(mylogf,"*** POP_UTIL_RESPONSE_HANDOVER g not saved\n");
      if (r->samp[i] != c->samp0) 
	mylog_exit(mylogf,"*** POP_UTIL_RESPONSE_HANDOVER sampl mismatch\n");
      if (c->gta0 == NULL)  
	mylog_exit(mylogf,"*** POP_UTIL_RESPONSE_HANDOVER gta0 is NULL\n");
      mod_util_resp_store_f(r,i,c->gta0,c->n0,1,mylogf);
    }else if ((strcmp(r->datid[i],"vm")==0)||
	      (strcmp(r->datid[i],"rate")==0)){ 
      if (strcmp(r->rtype[i],"f")!=0)
	mylog_exit(mylogf,"*** POP_UTIL_RESPONSE_HANDOVER Vm not 'f'\n");
      if (!c->savv)
	mylog_exit(mylogf,"*** POP_UTIL_RESPONSE_HANDOVER Vm not saved\n");
      if (c->vm == NULL)
	mylog_exit(mylogf,"*** POP_UTIL_RESPONSE_HANDOVER  vm is NULL\n");
      if (c->vmt != NULL)
	mod_util_resp_store_f_reg(r,i,c->vmt,c->vm,c->vn,1.0,mylogf);
      else{ 
	mod_util_resp_store_f(r,i,c->vm,c->vn,1,mylogf);
      }
    }else if (strcmp(r->datid[i],"gx")==0){   
      mod_util_resp_store_f(r,i,c->gtx0,c->n0,1,mylogf); 
    }else if (strcmp(r->datid[i],"gi")==0){   
      mod_util_resp_store_f(r,i,c->gti0,c->n0,1,mylogf); 
    }else if (strcmp(r->datid[i],"vd")==0){
      if (!c->savd)
	mylog_exit(mylogf,"*** POP_UTIL_RESPONSE_HANDOVER Vd not saved\n");
      mod_util_resp_store_f_reg(r,i,c->vmt,c->vmd,c->vn,1.0,mylogf);
    }else if ((strcmp(r->datid[i],"si_mask")==0) ||
	      (strcmp(r->datid[i],"si_mask1")==0) ||
	      (strcmp(r->datid[i],"si_mask2")==0)){
      d = pop_cell_si_get_data_ptr(mylogf,syn,r->datid[i],&dn);
      mod_util_resp_store_f(r,i,d,dn,1,mylogf);
    }else{
      csav_flag = 0;
      if (c->csav_cnt != NULL){  
	csi = search_2d_carray(c->pl->csav_datid,r->datid[i],c->pl->csav_n);
	if (csi >= 0){  
	  d  = c->csav_f[csi];   
	  dn = c->csav_cnt[csi];
	  if (d != NULL){
	    if (r->samp[i] != c->pl->csav_samp[csi])
	      mylogx(mylogf,"POP_UTIL_RESPONSE_HANDOVER",
		     "Sampling mismatch for csav\n");
	    mod_util_resp_store_f(r,i,d,dn,1,mylogf);  
	    csav_flag = 1;
	  }else{
	    mylogx(mylogf,"POP_UTIL_RESPONSE_HANDOVER","data is NULL");
	  }
	}else{
	  ; 
	}
      }
      if (csav_flag == 0){
	sprintf(tstr,"  datid:  %s\n",r->datid[i]);
	mylog(mylogf,tstr);
	mylogx(mylogf,"POP_UTIL_RESPONSE_HANDOVER","Unknown data ID");
      }
    }
  }
}
void pop_util_cell_spike_inc(c,m)
     struct pop_cell *c;
     int m; 
{
  int i;
  int n,nn;
  float *s,*t;
  n = c->maxscnt;
  nn = m*n;
  printf("  Spike storage limit being increased from %d to %d\n",n,nn);
  s = (float *)myalloc(nn*sizeof(float));
  t = c->s;
  for(i=0;i<n;i++)
    s[i] = t[i];
  myfree(c->s);
  c->s = s;
  c->maxscnt = nn;
}
void pop_util_add_spike_to_cb(syn,t,amp)
     struct pop_syn *syn;  
     float t;              
     float amp;            
{
  int j,k;
  int n;
  struct pop_circbuf *cb;
  if (syn->stype == 1){
    cb = syn->post->ex1s;
  }else if (syn->stype == 2){
    cb = syn->post->in1s;
  }else{
    cb = NULL;
    exit_error("POP_UTIL_ADD_SPIKE_TO_CB","Unknown synapse type");
  }
  n = cb->n;
  if (t < cb->t){
    printf("syn->post  %s\n",syn->post->name);
    printf("syn->pre   %s\n",syn->pre->name);
    printf("t = %f\n",t);
    printf("cb->t = %f\n",cb->t);
    exit_error("POP_UTIL_ADD_SPIKE_TO_CB","Spike arrives in past");
  }
  k = (int)((t - cb->t)/cb->dt); 
  if (k >= n){
    printf("  cb->t:              %f \n",cb->t);
    printf("  cb->dt:             %f \n",cb->dt);
    printf("  cb->n:              %d \n",cb->n);
    printf("  Time to add spike:  %f \n",t);
    exit_error("POP_UTIL_ADD_SPIKE_TO_CB","Spike arrives in future");
  }
  j = cb->i + k; 
  if (j >= n)
    j -= n;
  cb->d[j] += amp;  
}
float pop_util_circbuf_val_at_t(cb,t)
     struct pop_circbuf *cb;  
     double t;                
{
  int cbn,nmask,npast,dti,ti0;
  float fval;
  double dt;
  cbn  = cb->n;         
  nmask = cb->maskn;    
  npast = cbn - nmask;  
  dt  = t - cb->t;   
  if (dt < 0.0){
    dti = (int)(-0.5 + dt/cb->dt);  
    if (dti < -npast){
      printf("  negative dt  %.6f   npast = %d\n",dt,npast);
      exit_error("POP_UTIL_CIRCBUF_VAL_AT_T","Negative dt, not allowed");
    }
  }else{
    dti = (int)(0.5 + dt/cb->dt);  
  }
  if (dti >= nmask){  
    fval = 0.0;  
  }else{
    ti0 = cb->i + dti;
    if (ti0 >= cbn)  
      ti0 -= cbn;
    if (ti0 < 0)     
      ti0 += cbn;
    fval = cb->d[ti0];
  }
  return fval;
}
float pop_util_circbuf_val_at_t_interp(cb,t)
     struct pop_circbuf *cb;  
     double t;                
{
  int cbn,nmask,npast,dti,ti0,ti1;
  float fval0,fval1,fval,tfrac;
  double dt,cbdt;
  cbn  = cb->n;         
  nmask = cb->maskn;    
  npast = cbn - nmask;  
  cbdt = cb->dt;        
  dt = t - cb->t;   
  if (dt < 0.0){
    exit_error("POP_UTIL_CIRCBUF_VAL_AT_T_INTERP","Negative Not imp'd");
    dti = (int)(-0.5 + dt/cbdt);  
    if (dti < -npast){
      printf("  negative dt  %.6f   npast = %d\n",dt,npast);
      exit_error("POP_UTIL_CIRCBUF_VAL_AT_T_INTERP","Negative dt not allowed");
    }
  }else{
    dti = (int)(dt/cbdt);  
    tfrac = (dt - dti*cbdt)/cbdt;      
  }
  if (dti >= nmask){  
    fval = 0.0;  
  }else{
    ti0 = cb->i + dti;
    if (ti0 >= cbn)  
      ti0 -= cbn;
    if (ti0 < 0)     
      ti0 += cbn;
    fval0 = cb->d[ti0];
    if (dti == (nmask-1)){  
      fval1 = 0.0;
    }else{
      ti1 = ti0 + 1;
      if (ti1 >= cbn)  
	ti1 = 0;
      fval1 = cb->d[ti1];
    }
    fval = fval0 + tfrac * (fval1-fval0);  
  }
  return fval;
}
void pop_util_circbuf_mask_update_t(cb,t,mask,maskn,sav,rsavti)
     struct pop_circbuf *cb;  
     double t;                
     float *mask;             
     int    maskn;            
     float *sav;              
     int   *rsavti;           
{
  int i,k;
  int cbn,cbi,dti,ti0,ti1;
  int erase_i0,erase_i1;
  float *cbd;
  double cbt,cbdt,dt;
  int savi;
  if (cb == NULL)
    exit_error("POP_UTIL_CIRCBUF_MASK_UPDATE_T","Cirbuf is NULL");
  cbn  = cb->n;    
  cbi  = cb->i;    
  cbt  = cb->t;    
  cbdt = cb->dt;   
  cbd  = cb->d;    
  dt  = t - cbt;   
  if (dt < 0.0){
    printf("  t %f   cbt %f\n",t,cbt);
    exit_error("POP_UTIL_CIRCBUF_MASK_UPDATE_T","Negative dt, not allowed");
  }
  dti = (int)(0.5 + dt/cbdt);  
  *rsavti = -1;  
  if (sav != NULL){
    savi = (int)(0.5 + t/cbdt);
  }
  if (dti >= cbn){
    ti0 = 0;  
    for(i=0;i<maskn;i++){     
      cbd[i] = mask[i];
      if (sav != NULL) sav[savi+i] = cbd[i];
    }
    for(i=maskn;i<cbn;i++)   
      cbd[i] = 0.0;
    if (sav != NULL)
      *rsavti = savi + maskn-1;
  }else{
    ti0 = cbi + dti;
    if (ti0 >= cbn)  
      ti0 -= cbn;
    erase_i0 = cbi + maskn;
    if (erase_i0 >= cbn)
      erase_i0 -= cbn;
    erase_i1 = erase_i0 + dti - 1;
    if (erase_i1 >= cbn)
      erase_i1 -= cbn;
    if (erase_i0 <= erase_i1){
      for(i=erase_i0;i<=erase_i1;i++)   
	cbd[i] = 0.0;
    }else{
      for(i=erase_i0;i<cbn;i++)   
	cbd[i] = 0.0;
      for(i=0;i<=erase_i1;i++)    
	cbd[i] = 0.0;
    }
    ti1 = ti0 + maskn - 1;
    if (ti1 >= cbn)
      ti1 -= cbn;
    k = 0;
    if (ti0 <= ti1){
      for(i=ti0;i<=ti1;i++){   
	cbd[i] += mask[k];
	if (sav != NULL) sav[savi+k] = cbd[i];
	k += 1;
      }
    }else{
      for(i=ti0;i<cbn;i++){   
	cbd[i] += mask[k];
	if (sav != NULL) sav[savi+k] = cbd[i];
	k += 1;
      }
      for(i=0;i<=ti1;i++){    
	cbd[i] += mask[k];
	if (sav != NULL) sav[savi+k] = cbd[i];
	k += 1;
      }
    }
    if (sav != NULL){
      if (k > 0)
	*rsavti = savi + k-1;
    }
  }
  cb->t = cbt + (double)dti*cbdt;   
  cb->i = ti0;
}
void pop_util_process_sid_ds02(syn,t,amp)
     struct pop_syn *syn;  
     float t;              
     float amp;            
{
  int i,j,k;
  int ci,n,maskn,dsavn,savti;
  float w,*mask,*dsav;
  struct pop_mech *msi;
  struct pop_circbuf *cb;
  ci = syn->si->ci;               
  msi = syn->si->siu->s001->msi;  
  cb = syn->si->siu->s001->wt;    
  n = cb->n;                      
  k = (int)((t - cb->t)/cb->dt); 
  if (k < 0){
    printf("*** *** WARNING:\n");
    printf("  Pre-syn cell: %s\n",syn->pre->name);
    printf("  Post-syn cell: %s\n",syn->post->name);
    printf("  tlast: %f  alast: %f  tdelay: %f\n",syn->tlast,syn->alast,
	   syn->tdelay);
    printf("  ci = %d\n",ci);
    printf("  k = %d  cb->dt %f   cb->i %d\n",k,cb->dt,cb->i);
    printf("*** *** WARNING  spike is too early:  t=%f  cb->t = %f\n",t,cb->t);
    printf("  Spike set to current time\n");
    printf("CONTINUING...\n");
    k = 0;  
  }
  if (k >= n)
    j = -1;    
  else{
    j = cb->i + k; 
    if (j >= n)
      j -= n;
  }
  if (ci == 0){ 
    if (msi->si_comp == 2){  
      if (j < 0){
	w = 1.0;   
      }else{
	w = 1.0 - cb->d[j];  
	if (w < 0.0)
	  w = 0.0;           
      }
    }else{  
      if (j < 0){
	w = 0.0;     
      }else{
	w = cb->d[j];  
	if (w < 0.0)
	  w = 0.0;           
      }
    }
    pop_util_add_spike_to_cb(syn,t,w*amp);
  }else if (ci == 1){ 
    mask  = msi->mask1;  
    maskn = msi->mask1n;
    if (syn->si->siu->s001->sisv != NULL){
      dsav = pop_cell_si_get_data_ptr(NULL,syn,"si_mask",&dsavn);
    }else
      dsav = NULL;
    pop_util_circbuf_mask_update_t(cb,t,mask,maskn,dsav,&savti);
    if (savti != -1){
      if (dsav != NULL){  
	pop_cell_si_sav_set_n(NULL,syn,"si_mask",savti);
      }
    }
  }else
    exit_error("POP_UTIL_PROCESS_SID_DS02","Bad ci value");
}
void pop_util_process_sid_symmask(syn,t,amp)
     struct pop_syn *syn;  
     float t;              
     float amp;            
{
  int i,k;
  int ti,ci,di,n,maskn,dsavn,jw,jm;
  float w,*mask,*d,*dsav;
  struct pop_mech *msi;
  struct pop_circbuf *cbm,*cbw;
  ci = syn->si->ci;               
  msi = syn->si->siu->s002->msi;  
  if (ci == 0){
    cbm = syn->si->siu->s002->wt1;   
    cbw = syn->si->siu->s002->wt2;   
  }else if (ci == 1){
    cbm = syn->si->siu->s002->wt2;   
    cbw = syn->si->siu->s002->wt1;   
  }else{
    exit_error("POP_UTIL_PROCESS_SID_SYMMASK","Bad ci value");
  }
  n = cbm->n;                        
  k = (int)((t - cbm->t)/cbm->dt); 
  if (k < 0){
    printf("k = %d  cbm->dt %f\n",k,cbm->dt);
    printf("*** WARNING  spike is too early:  t=%f  cbm->t = %f\n",t,cbm->t);
    exit_error("POP_UTIL_PROCESS_SID_SYMMASK","Spike in past");
    k = 0;  
  }
  if (k >= n)
    jm = -1;    
  else{
    jm = cbm->i + k; 
    if (jm >= n)
      jm -= n;
  }
  k = (int)((t - cbw->t)/cbw->dt); 
  if (k < 0){
    printf("*** WARNING  spike is too early:  t=%f  cbw->t = %f\n",t,cbw->t);
    exit_error("POP_UTIL_PROCESS_SID_SYMMASK","Spike in past");
    k = 0;  
  }
  if (k >= n)
    jw = -1;    
  else{
    jw = cbw->i + k; 
    if (jw >= n)
      jw -= n;
  }
  if (msi->si_comp == 2){  
    if (jw < 0){
      w = 1.0;   
    }else{
      w = 1.0 - cbw->d[jw];  
      if (w < 0.0)
	w = 0.0;           
    }
  }else{  
    if (jw < 0){
      w = 0.0;     
    }else{
      w = cbw->d[jw];  
      if (w < 0.0)
	w = 0.0;           
    }
  }
  pop_util_add_spike_to_cb(syn,t,w*amp); 
  mask  = msi->mask1;  
  maskn = msi->mask1n;
  d = cbm->d;
  if (syn->si->siu->s002->sisv != NULL){
    if (ci == 0){
      dsav = pop_cell_si_get_data_ptr(NULL,syn,"si_mask1",&dsavn);
    }else
      dsav = pop_cell_si_get_data_ptr(NULL,syn,"si_mask2",&dsavn);
    ti = (int)(t/cbm->dt); 
  }else
    dsav = NULL;
  if (jm < 0){
    for(i=0;i<maskn;i++){
      d[i] = mask[i];
      if (dsav != NULL){ 
	dsav[ti + i] = d[i];   
      }
    }
    while(i < n){
      d[i] = 0.0;  
      if (dsav != NULL){ 
	dsav[ti + i] = d[i];   
      }
      i += 1;
    }
    cbm->i = 0;   
    cbm->t = t;   
  }else{
    i = cbm->i; 
    while(i != jm){
      d[i] = 0.0; 
      i += 1;
      if (i >= n)
	i -= n;
    }
    di = jm;
    for(i=0;i<maskn;i++){
      d[di] += mask[i];
      if (dsav != NULL){  
	dsav[ti + i] = d[di];  
      }
      di += 1;
      if (di >= n)
	di -= n;
    }
    cbm->i = jm;
    cbm->t = t;
  }
  if (dsav != NULL){
    if (ci == 0)
      pop_cell_si_sav_set_n(NULL,syn,"si_mask1",ti+maskn);
    else
      pop_cell_si_sav_set_n(NULL,syn,"si_mask2",ti+maskn);
  }
}
void pop_util_process_sid_mult(syn,t,amp)
     struct pop_syn *syn;  
     float t;              
     float amp;            
{
  int ci,maskn,dsavn,savti;
  float *mask,*dsav;
  struct pop_mech *msi;
  struct pop_circbuf *cbm;
  ci = syn->si->ci;               
  msi = syn->si->siu->s002->msi;  
  if (ci == 0){
    cbm = syn->si->siu->s002->wt1;   
  }else if (ci == 1){
    cbm = syn->si->siu->s002->wt2;   
  }else{
    exit_error("POP_UTIL_PROCESS_SID_MULT","Bad ci value");
  }
  mask  = msi->mask1;  
  maskn = msi->mask1n;
  if (syn->si->siu->s002->sisv != NULL){
    if (ci == 0){
      dsav = pop_cell_si_get_data_ptr(NULL,syn,"si_mask1",&dsavn);
    }else{
      dsav = pop_cell_si_get_data_ptr(NULL,syn,"si_mask2",&dsavn);
    }
  }else{
    dsav = NULL;
  }
  pop_util_circbuf_mask_update_t(cbm,t,mask,maskn,dsav,&savti);
  if (savti != -1){
    if (dsav != NULL){  
      if (ci == 0)
	pop_cell_si_sav_set_n(NULL,syn,"si_mask1",savti);
      else
	pop_cell_si_sav_set_n(NULL,syn,"si_mask2",savti);
    }
  }
}
void pop_util_process_si001(syn,msi,tgen,amp)
     struct pop_syn *syn;
     struct pop_mech *msi;
     float tgen;           
     float amp;            
{
  if (strcmp(msi->type,"si_ds02")==0){
    pop_util_process_sid_ds02(syn,tgen,amp);
  }else if ((strcmp(msi->type,"si_pair_rfdist")==0) ||
	    (strcmp(msi->type,"si_nonlin")==0)){
    if (onode_test_chr(msi->o,"nonlin","mask")){
      pop_util_process_sid_ds02(syn,tgen,amp);
    }else if (onode_test_chr(msi->o,"nonlin","symmask")){
      pop_util_process_sid_symmask(syn,tgen,amp);
    }else if (onode_test_chr(msi->o,"nonlin","mult")||
	      onode_test_chr(msi->o,"nonlin","divide")){
      pop_util_process_sid_mult(syn,tgen,amp);
    }else{
      printf("***  nonlin value not imp'd yet\n");
      exit(0);
    }
  }else{
    exit_error("POP_UTIL_PROCESS_SI001","Unknown SID name");
  }
}
void pop_util_process_si_spike(syn,tgen,amp)
     struct pop_syn *syn;
     float tgen;           
     float amp;            
{
  struct pop_mech *msi;
  if (syn->si->siui == 1){
    msi = syn->si->siu->s001->msi;
    pop_util_process_si001(syn,msi,tgen,amp);
  }else if (syn->si->siui == 2){
    msi = syn->si->siu->s002->msi;
    pop_util_process_si001(syn,msi,tgen,amp);  
  }else{
    exit_error("POP_UTIL_PROCESS_SI_SPIKE","Unknown SI Union Index");
  }
}
void pop_util_process_spike_postsyn(syn,tgen,sdf,sdtau)
     struct pop_syn *syn;
     float tgen;           
     float sdf,sdtau;      
{
  float t,w,dt,anew;
  t = tgen + 0.0015;  
  t += syn->tdelay;   
  if (sdf < 1.0){
    dt = t - syn->tlast;
    anew = 1.0 - (1.0 - syn->alast*sdf) * exp(-dt/sdtau);
    if (( anew < 0.0) || (anew > 1.0)){
      printf("anew %f  dt %f  (%f - %f)\n",anew,dt,t,syn->tlast);
    }
    syn->tlast = t;
    syn->alast = anew;
  }else{
    anew = 1.0;
  }
  w = syn->w; 
  if (syn->si != NULL){
    pop_util_process_si_spike(syn,t,(anew*w));
  }else{
    pop_util_add_spike_to_cb(syn,t,(anew*w));
  }
}
void pop_util_pair_get_params(o,rdir,rdist,rjitp,rjito,rseed)
     struct onode *o;
     float *rdir,*rdist,*rjitp,*rjito;
     int *rseed;
{
  int seed;
  float dir,dist,jitp,jito;
  dir   = onode_getpar_flt_exit(o,"offset_direction");   
  dist  = onode_getpar_flt_exit(o,"offset_distance");    
  jitp  = onode_getpar_flt_exit(o,"offset_jitter_para"); 
  jito  = onode_getpar_flt_exit(o,"offset_jitter_orth"); 
  seed  = onode_getpar_int_exit(o,"offset_jitter_seed"); 
}
void pop_util_mech_config_mask(mylogf,m)
     char *mylogf;
     struct pop_mech *m;           
{
  int i;
  int wtn,masktype,maskn,npast;
  float *mask,mc,tt,sigma,wdt,dtms,dt,maskamp,alpha;
  char *mask_comp,*nonlin;
  mylog(mylogf,"  POP_UTIL_MECH_CONFIG_MASK\n");
  dt       = onode_getpar_flt_exit(m->o,"dt");
  masktype = onode_getpar_int_exit(m->o,"masktype");
  maskamp  = onode_getpar_flt_dflt(m->o,"mask_amp",1.0);
  wdt = 0.001; 
  dtms = my_rint(dt/wdt);  
  npast = 0.002/wdt;  
  if (masktype == 1)
    wtn = (int)(2.0*dtms);   
  else if (masktype == 2)
    wtn = (int)(4.0*dtms);
  else if (masktype == 3)
    wtn = (int)(2.0*dtms);
  else if (masktype == 4)
    wtn = (int)(3.0*dtms);
  else if (masktype == 5)
    wtn = (int)(7.0*dtms);
  wtn += npast;
  m->wdt = wdt;
  m->wtn = wtn;
  maskn = wtn - npast; 
  if (masktype == 1){ 
    mask = (float *)myalloc(maskn*sizeof(float));
    mc = 2.0*M_PI * 3.0/(4.0*dtms);
    for(i=0;i<maskn;i++){
      tt = (float)i;
      mask[i] = -maskamp*sin(mc*tt);
    }
  }else if (masktype == 2){ 
    sigma = dtms/sqrt(2.0);
    mask = maxwell_farray(0.0,sigma,1.0,maskn);
    make_max_const_farray(mask,maskn,maskamp);
  }else if (masktype == 3){ 
    mask = (float *)myalloc(maskn*sizeof(float));
    mc = M_PI /(2.0*dtms);
    for(i=0;i<maskn;i++){
      tt = (float)i;
      mask[i] = maskamp * sin(mc*tt);
    }
  }else if (masktype == 4){ 
    mask = (float *)myalloc(maskn*sizeof(float));
    for(i=0;i<maskn;i++){
      tt = (float)i;
      mask[i] = maskamp * func_gaussian_one(tt,0.0,dtms);
    }
  }else if (masktype == 5){ 
    alpha = 1.0/dtms; 
    mask = alpha_farray(0.0,alpha,1.0,maskn);
    make_max_const_farray(mask,maskn,maskamp);
  }else
    mylog_exit(mylogf,"POP_UTIL_MECH_CONFIG_MASK  Bad masktype value.");
  m->mask1n = maskn;
  m->mask1 = mask;
  nonlin = onode_getpar_chr_dflt(m->o,"nonlin","NULL"); 
  if (strcmp(nonlin,"mult")==0){
    m->si_comp = 1;
  }else if (strcmp(nonlin,"divide")==0){
    m->si_comp = 2;
  }else{
    mask_comp  = onode_getpar_chr_dflt(m->o,"mask_comp","prod");
    if (strcmp(mask_comp,"prod")==0)
      m->si_comp = 1;
    else if (strcmp(mask_comp,"prod1m")==0){
      m->si_comp = 2;
    }else
      exit_error("POP_UTIL_MECH_CONFIG_MASK","Unknown 'mask_comp' value");
    myfree(mask_comp);
  }
  myfree(nonlin);
  m->a = onode_getpar_flt_dflt(m->o,"c_a",0.0); 
  m->b = onode_getpar_flt_dflt(m->o,"c_b",1.0); 
}
struct input_pair *pop_util_input_pair(mpt,lay,io,inindex)
     struct pop_top *mpt;     
     struct pop_layer *lay;   
     struct onode *io;        
     int inindex;             
{
  char *lf;
  struct onode *od,*oip;
  struct input_pair *p;
  lf = mpt->logf;
  mylog(lf,"  POP_UTIL_INPUT_PAIR\n");
  oip = onode_child_get_unique(io,"input_pair");  
  if (oip == NULL)
    mylogx(lf,"POP_UTIL_INPUT_PAIR","Cannot find <input_pair>");
  p = (struct input_pair *)myalloc(sizeof(struct input_pair));
  p->pair_seed = -1;   
  p->tseed = -1;       
  p->in_index = inindex;
  p->distrib = NULL;
  p->z1 = NULL;
  od = onode_child_get_unique(oip,"pair_distrib");
  if (od != NULL){
    p->distrib = onode_getpar_chr_exit(od,"type");
    if ((strcmp(p->distrib,"rf_dist")==0) ||
	(strcmp(p->distrib,"rf_dist_phase")==0)){
      p->rf_offset_dir         = onode_getpar_flt_exit(od,"direction");
      p->rf_offset_dist        = onode_getpar_flt_exit(od,"distance");
      p->rf_offset_sd_par  = onode_getpar_flt_exit(od,"jitter_para");
      p->rf_offset_sd_orth = onode_getpar_flt_exit(od,"jitter_orth");
      p->pair_seed        = onode_getpar_int_exit(od,"jitter_seed");
      if (strcmp(p->distrib,"rf_dist_phase")==0){
	p->rf_ph_shift  = onode_getpar_flt_dflt(od,"ph_shift",90.0);
	p->rf_ph_dev    = onode_getpar_flt_dflt(od,"ph_dev",5.0);
      }else{
	p->rf_ph_shift  =  0.0;
	p->rf_ph_dev    = -1.0;
      }
    }else
      mylogx(lf,"POP_UTIL_INPUT_PAIR","Unknown pair_distrib type");
  }else{
    mylogx(lf,"POP_UTIL_INPUT_PAIR","Cannot find <pair_distrib>");
  }
  p->pop0 = popu_get_lay_named(mpt,io,"pop_origin","POP_UTIL_INPUT PAIR pop0");
  if (onode_test_ostr(oip,"pop_origin_2")){
    p->pop1 = popu_get_lay_named(mpt,oip,"pop_origin_2",
				 "POP_UTIL_INPUT PAIR pop1");
  }else{
    p->pop1 = p->pop0;
  }
  p->z1 = onode_getpar_chr_dflt(oip,"pop_origin_2_z","all");
  if (onode_test_ostr(oip,"syn_int")){
    if (onode_test_chr(oip,"syn_int","NULL") == 0){
      p->msi = popu_get_mech_named(mpt,lay,oip,"syn_int","MOD_POP_INPUT");
    }else
      p->msi = NULL;
  }
  return p;
}
void pop_util_make_cell_style_ds01(mylogf,mppl,laylist,nlay,c)
     char *mylogf;
     struct param_pair_list *mppl; 
     struct pop_layer **laylist;   
     int nlay;                     
     struct pop_cell *c;           
{
  int i,j;
  struct pop_layer *pl;
  mylog(mylogf,"  POP_UTIL_MAKE_CELL_STYLE_DS01\n");
  if (c->subclass != NULL)
    myfree(c->subclass);
  c->subclass = strdup("ds01");  
  for(i=0;i<nlay;i++){
    pl = laylist[i];
    pop_cell_remove_all_syn_from_layer_to_cell(mylogf,pl,c);
  }
  {
    int *cx,*cy,seed,seed0,nsamp,maxrep,nposs,evflag,xn,yn,*tcx,*tcy,tcn;
    int *seeds;
    float sd_orth,sd_par,sdo,sdp,eps,sscale,xc,yc,orideg;
    char tstr[SLEN],conname[SLEN];
    float dr,sd_noise;
    struct pop_lgn_in *ln;
    int tcn0,*tcx0,*tcy0,tcn1,*tcx1,*tcy1;
    strcpy(conname,"pop_ex_conn0"); 
    sprintf(tstr,"%s_sd_orth",conname);
    sd_orth = paramfile_get_float_param_or_exit(mppl,tstr);
    sprintf(tstr,"%s_sd_par",conname);
    sd_par = paramfile_get_float_param_or_exit(mppl,tstr);
    sprintf(tstr,"%s_seed",conname);
    seed = paramfile_get_int_param_or_exit(mppl,tstr);
    sprintf(tstr,"%s_eps",conname);
    eps = paramfile_get_float_param_or_exit(mppl,tstr);
    sprintf(tstr,"%s_nsamp",conname);
    nsamp = paramfile_get_int_param_or_exit(mppl,tstr);
    sprintf(tstr,"%s_maxrep",conname);
    maxrep = paramfile_get_int_param_or_exit(mppl,tstr);
    sscale = paramfile_get_float_param_or_exit(mppl,"sscale");
    sdo = sd_orth / sscale;  
    sdp = sd_par / sscale;  
    i = c->layx;
    j = c->layy;
    xc = c->pl->x0 + (float)i*c->pl->xf;     
    yc = c->pl->y0 + (float)j*c->pl->yf;
    if (c->lgn_n > 0)
      ln = c->lgnin[0];
    else
      exit_error("POP_UTIL_MAKE_CELL_STYLE_DS01","No LGN inputs");
    xn = ln->lay->xn;  
    yn = ln->lay->yn;
    printf(" USE ATTRIB TO GET ORI\n");
    printf(" USE ATTRIB TO GET ORI\n");
    seeds = get_seeds(seed,100000,4);
    evflag = 1; 
    nposs = mod_conn_gauss_02(mylogf,xn,yn,xc,yc,sdo,sdp,orideg,seeds[0],
			      maxrep,eps,nsamp/2,evflag,&tcx1,&tcy1,&tcn1);
    nposs = mod_conn_gauss_02(mylogf,xn,yn,xc,yc,sdo,sdp,orideg,seeds[1],
			      maxrep,eps,nsamp/2,evflag,&tcx0,&tcy0,&tcn0);
    exit_error("POP_UTIL_MAKE_CELL_STYLE_DS01","WYETH - BROKEN");
    {
      struct pop_mech *mechr;  
      pop_cell_add_lgn_input(c,ln->lay,mechr,tcx0,tcy0,tcn0,tcx1,tcy1,tcn1,
			     nposs,-1.0);
    }
    dr = 4.0;
    sd_noise = 0.0;
    mod_conn_pick_paired_connections(tcx1,tcy1,tcn1,xn,yn,orideg,dr,
				     sd_noise,seeds[2],&tcx,&tcy);
    append_iarray(&(ln->cx1),ln->cn1,tcx,ln->cn1);
    append_iarray(&(ln->cy1),ln->cn1,tcy,ln->cn1);
    myfree(tcx); myfree(tcy);
    ln->cn1 = 2*ln->cn1;
    mod_conn_pick_paired_connections(tcx0,tcy0,tcn0,xn,yn,orideg,dr,
				     sd_noise,seeds[3],&tcx,&tcy);
    append_iarray(&(ln->cx0),ln->cn0,tcx,ln->cn0);
    append_iarray(&(ln->cy0),ln->cn0,tcy,ln->cn0);
    myfree(tcx); myfree(tcy);
    ln->cn0 = 2*ln->cn0;
    myfree(seeds);
  }
}
void pop_util_customize(m,mpt)
     struct model_struct *m;     
     struct pop_top *mpt;        
{
  int i,k;
  int x,y,z,layi,nlay,sii,cflag;
  char tstr[SLEN],*layname,*mylogf,*ctype;
  float dsdph,dsdev;
  struct param_pair *tp;
  struct pop_layer *l0,*l1;
  struct pop_cell *c0,*c1;
  struct pop_mech *msi;
  struct pop_layer **laylist;
  struct onode *t;
  mylogf = mpt->logf;
  mylog(mylogf,"  POP_UTIL_CUSTOMIZE\n");
  laylist = mpt->lay; 
  nlay = mpt->nlay;   
  cflag = 0;  
  t = onode_get_next_type(m->o->o,"customize");
  while(t != NULL){
    cflag = 1;   
    ctype = onode_getpar_chr_exit(t,"type");  
    sprintf(tstr,"    %s\n",ctype);
    mylog(mylogf,tstr);
    if (strcmp(ctype,"synapse_delete")==0){
      if (onode_test_ostr(t,"unit_pre")){
	c0 = popu_get_c_named(mpt,t,"unit_pre","POP_UTIL_CUSTOMIZE u_pre");
	if (onode_test_ostr(t,"unit_post")){
	  c1 = popu_get_c_named(mpt,t,"unit_post","POP_UTIL_CUSTOMIZE u_post");
	  if (onode_test_ostr(t,"si_index")){  
	    sii = onode_getpar_int_exit(t,"si_index");
	    printf("sii = %d\n",sii);
	  }
	  exit_error("POP_UTIL_CUSTOMIZE","Not Imp'd yet");
	}else if (onode_test_ostr(t,"layer_post")){
	  l1 = popu_get_lay_named(mpt,t,"layer_post","POP_UTIL_CUSTOMIZE lpo");
	  if (onode_test_ostr(t,"si_index")){  
	    sii = onode_getpar_int_exit(t,"si_index");
	    pop_cell_remove_all_syn_from_cell_to_layer_si(mylogf,c0,l1,sii);
	  }else{
	    pop_cell_remove_all_syn_from_cell_to_layer(mylogf,c0,l1);
	  }
	}
      }else if (onode_test_ostr(t,"layer_pre")){
	l0 = popu_get_lay_named(mpt,t,"layer_pre","POP_UTIL_CUSTOMIZE l_pre");
	if (onode_test_ostr(t,"unit_post")){
	  c1 = popu_get_c_named(mpt,t,"unit_post","POP_UTIL_CUSTOMIZE u_post");
	  exit_error("POP_UTIL_CUSTOMIZE","Not tested yet");
	  pop_cell_remove_all_syn_from_layer_to_cell(mylogf,l0,c1);
	}
      }else{
	mylog_exit(mylogf,"POP_UTIL_CUSTOMIZE  synapse_delete bad pre-syn\n");
      }
    }else if (strcmp(ctype,"synapse_adjust")==0){
      float synw;
      if (onode_test_ostr(t,"unit_pre")){
	c0 = popu_get_c_named(mpt,t,"unit_pre","POP_UTIL_CUSTOMIZE u_pre");
	if (onode_test_ostr(t,"unit_post")){
	  c1 = popu_get_c_named(mpt,t,"unit_post","POP_UTIL_CUSTOMIZE u_post");
	  if (onode_test_ostr(t,"si_index")){  
	    sii = onode_getpar_int_exit(t,"si_index");
	    printf("sii = %d\n",sii);
	  }
	  exit_error("POP_UTIL_CUSTOMIZE","Not Imp'd yet");
	}else if (onode_test_ostr(t,"layer_post")){
	  l1 = popu_get_lay_named(mpt,t,"layer_post","POP_UTIL_CUSTOMIZE lpo");
	  synw = onode_getpar_flt_exit(t,"syn_weight");
	  pop_cell_adjust_all_syn_from_cell_to_layer(mylogf,c0,l1,synw);
	  printf("HERE WYETH w = %f\n",synw);
	}
      }else if (onode_test_ostr(t,"layer_pre")){
	l0 = popu_get_lay_named(mpt,t,"layer_pre","POP_UTIL_CUSTOMIZE l_pre");
	if (onode_test_ostr(t,"unit_post")){
	  c1 = popu_get_c_named(mpt,t,"unit_post","POP_UTIL_CUSTOMIZE u_post");
	  exit_error("POP_UTIL_CUSTOMIZE","Not tested yet");
	  pop_cell_remove_all_syn_from_layer_to_cell(mylogf,l0,c1);
	}
      }else{
	mylog_exit(mylogf,"POP_UTIL_CUSTOMIZE  synapse_delete bad pre-syn\n");
      }
    }else if (strcmp(ctype,"synapse_add")==0){
      float synw,tdelay;
      int syntype,inindex;
      if (onode_test_ostr(t,"unit_pre")){
	c0 = popu_get_c_named(mpt,t,"unit_pre","POP_UTIL_CUSTOMIZE u_pre");
	if (onode_test_ostr(t,"unit_post")){
	  c1 = popu_get_c_named(mpt,t,"unit_post","POP_UTIL_CUSTOMIZE u_post");
	  synw = onode_getpar_flt_exit(t,"syn_weight");
	  syntype = onode_getpar_flt_exit(t,"syn_type");
	  tdelay = onode_getpar_flt_dflt(t,"syn_tdelay",0.0);
	  inindex = -2;  
	  pop_cell_add_synapse(c0,c1,syntype,synw,tdelay,inindex);
	}else if (onode_test_ostr(t,"layer_post")){
	  exit_error("POP_UTIL_CUSTOMIZE","Not Imp'd yet");
	}
      }else if (onode_test_ostr(t,"layer_pre")){
	  exit_error("POP_UTIL_CUSTOMIZE","Not Imp'd yet");
      }else{
	mylog_exit(mylogf,"POP_UTIL_CUSTOMIZE  synapse_add bad pre-syn\n");
      }
    }else{
      mylog_exit(mylogf,"POP_UTIL_CUSTOMIZE  Unknown type of customize\n");
    }
    myfree(ctype);
    t = onode_get_next_type(t->next,"customize");
  }
  if (cflag == 1)
    mylog(mylogf,"    Done customize\n");
  else
    mylog(mylogf,"    No customization\n");
}
void pop_util_add_noise(data,n,seed,ntype,nmu,nsd,ntsd,tscale)
     float *data;    
     int n,seed;
     char ntype[];   
     float nmu,nsd;  
     float ntsd;     
     float tscale;   
{
  int tseed,ns;
  float *sf,*noise,*snoise,rate,sampling,amp;
  if (strcmp(ntype,"gfg")==0){
    noise = gaussian_corr_noise(ntsd,nmu,nsd,n,seed);
    add_to_farray(data,noise,n);
    myfree(noise);
  }else if (strcmp(ntype,"gshot")==0){
    rate = nmu; 
    amp = ntsd;
    tseed = seed;
    sampling = 1.0/tscale;
    make_poisson_float_spikes(&sf,&ns,0,n,sampling,rate,0.0,0.0,1,&tseed);
    noise = expand_spike_farray(sf,ns,0,n);
    snoise = smooth_with_gaussian(noise,n,nsd,0.01);
    multiply_farray(snoise,n,amp * sqrt(2.0*M_PI)*nsd);
    myfree(noise); myfree(sf);
    noise = snoise;
    add_to_farray(data,noise,n);
    myfree(noise);
  }else{
    printf("ntype = %s\n",ntype);
    exit_error("POP_UTIL_ADD_NOISE","Unknown noise type");
  }
}
void pop_util_set_flag_grid_input_layer(mylogf,flag,xn,yn,onflag,lay,lgn_name)
     char *mylogf;
     int **flag,xn,yn;          
     int onflag;                
     struct pop_layer *lay;     
     char *lgn_name;            
{
  int i,j,k,l;
  int n,*xi,*yi;
  struct pop_cell *c;
  struct pop_lgn_in *ln,*lp;
  for(i=0;i<lay->xn;i++){
    for(j=0;j<lay->yn;j++){
      for(k=0;k<lay->zn;k++){ 
	c = &(lay->c[i][j][k]);
	for(l=0;l<c->lgn_n;l++){ 
	  ln = c->lgnin[l];
	  if (strcmp(ln->lay->name,lgn_name)==0){ 
	    if ((ln->lay->xn != xn) || (ln->lay->yn != yn)){ 
	      mylogx(mylogf,"POP_UTIL_SET_FLAG_GRID_INPUT_LAYER","Bad xn,yn");
	    }
	    if (onflag == 0){          
	      n  = ln->cn0;
	      xi = ln->cx0;
	      yi = ln->cy0;
	    }else if (onflag == 1){    
	      n  = ln->cn1;
	      xi = ln->cx1;
	      yi = ln->cy1;
	    }else
	      mylogx(mylogf,"POP_UTIL_SET_FLAG_GRID_INPUT_LAYER","Bad onflag");
	    for(l=0;l<n;l++){
	      flag[xi[l]][yi[l]] = 1;
	    }
	  }
	  if (ln->cpair != NULL){
	    lp = ln->cpair;
	    if (strcmp(lp->lay->name,lgn_name)==0){
	      if ((lp->lay->xn != xn) || (lp->lay->yn != yn)){
		mylogx(mylogf,"POP_UTIL_SET_FLAG_GRID_INPUT_LAYER",
		       "Bad xn,yn");
	      }
	      if (onflag == 0){          
		n  = lp->cn0;
		xi = lp->cx0;
		yi = lp->cy0;
	      }else if (onflag == 1){    
		n  = lp->cn1;
		xi = lp->cx1;
		yi = lp->cy1;
	      }else
		mylogx(mylogf,"POP_UTIL_SET_FLAG_GRID_INPUT_LAYER",
		       "Bad onflag");
	      for(l=0;l<n;l++){
		if (xi[l] != -1)  
		  flag[xi[l]][yi[l]] = 1;
	      }
	    }
	  }
	}
      }
    }
  }
}
float *pop_util_expand_ds01_paired_spikes(cx,cy,cn,xn,yn,s,cnt,tn,cw)
     int *cx,*cy,cn;           
     int xn,yn;                
     float ***s;               
     int **cnt;                
     int tn;                   
     float *cw;                
{
  int i,j,k;
  int xi,yi,t,tcnt;
  float *ts,*x;
  int nmask,npair;
  float *mask,*ma,mc,mu,tt,mcent;
  nmask = 40; 
  mask = (float *)myalloc(nmask*sizeof(float));
  mc = 2.0*M_PI*1.5/(float)nmask;
  mcent = (float)nmask/2.0;
  for(i=0;i<nmask;i++){
    tt = (float)i;
    mask[i] = -2.0*sin(mc*tt);
  }
  x = get_zero_farray(tn);
  ma = (float *)myalloc(tn*sizeof(float));
  npair = cn/2;
  for(i=0;i<npair;i++){  
    xi = cx[i];
    yi = cy[i];
    if ((xi<0)||(xi>=xn)||(yi<0)||(yi>=yn))
      exit_error("POP_UTIL_EXPAND_GRID_SPIKES","invalid grid coordinate");
    for(j=0;j<tn;j++)
      ma[j] = 0.0; 
    ts = s[xi][yi];
    tcnt = cnt[xi][yi];
    for(j=0;j<tcnt;j++){
      t = my_rint(ts[j]);
      if ((t >= 0)&&(t < (tn-nmask))){
	for(k=0;k<nmask;k++){
	  ma[t+k] += mask[k];
	}
      }
    }
    xi = cx[npair+i];
    yi = cy[npair+i];
    if ((xi<0)||(xi>=xn)||(yi<0)||(yi>=yn))
      exit_error("POP_UTIL_EXPAND_GRID_SPIKES","invalid grid coordinate");
    ts = s[xi][yi];
    tcnt = cnt[xi][yi];
    for(j=0;j<tcnt;j++){
      t = my_rint(ts[j]);
      if ((t >= 0)&&(t < tn)){
	if (ma[t] > 0.0)  
	  x[t] += ma[t];  
      }
    }
  }
  myfree(ma);
  return x;
}
float *pop_util_expand_spikes(s,n,tn,w,sdf,sdtau)
     float *s;      
     int n;         
     int tn;        
     float w;       
     float sdf;     
     float sdtau;   
{
  float *x;
  x = get_zero_farray(tn);   
  spikeu_expand_spikes_wsd_accum(s,n,x,tn,w,sdf,sdtau);
  return x;
}
float *pop_util_expand_grid_spikes(cx,cy,cn,xn,yn,s,cnt,tn,cw,sdf,sdtau)
     int *cx,*cy,cn;     
     int xn,yn;          
     float ***s;         
     int **cnt;          
     int tn;             
     float *cw;          
     float sdf;          
     float sdtau;        
{
  int i,j;
  int xi,yi,wflag;
  float *tx,*x,w;
  x = get_zero_farray(tn);
  if (cw == (float *)NULL)
    wflag = 0;
  else
    wflag = 1;
  for(i=0;i<cn;i++){  
    xi = cx[i];
    yi = cy[i];
    if (wflag)
      w = cw[i];
    else
      w = 1.0;
    if ((xi<0)||(xi>=xn)||(yi<0)||(yi>=yn))
      exit_error("POP_UTIL_EXPAND_GRID_SPIKES","invalid grid coordinate");
    tx = pop_util_expand_spikes(s[xi][yi],cnt[xi][yi],tn,w,sdf,sdtau);
    for(j=0;j<tn;j++)
      x[j] += tx[j];
    myfree(tx);
  }
  return x;
}
void pop_util_sum_gt_spk_grid(lay,cx,cy,cn,cw,gw,xn,yn,s,cnt,m,name,mechr,
				     gt,gnt,
				     tn,dumpname,subclass)
     struct pop_layer *lay;    
     int *cx,*cy,cn;           
     float *cw;                
     float gw;                 
     int xn,yn;                
     float ***s;               
     int **cnt;                
     struct model_struct *m;   
     char name[];              
     struct pop_mech *mechr;   
     float *gt;                
     float *gnt;               
     int tn;                   
     char dumpname[];
     char *subclass;           
{
  int nm,nn;
  float *mask,*xsp,*g,*nmda_mask,sdf,sdtau;
  char tstr[SLEN];
  struct pop_mech *mech;
  struct nmda_param *tp;
  if (m->ppl == NULL){
    sdf = mechr->sdf;
    sdtau = mechr->sdtau;
    mask = pop_util_get_mask_diff_exp_mech(mechr,&nm);
  }else{
    sprintf(tstr,"%s_sdf",name);
    sdf = paramfile_get_float_param_or_exit(m->ppl,tstr);
    sprintf(tstr,"%s_sdtau",name);
    sdtau = paramfile_get_float_param_or_exit(m->ppl,tstr);
    mask = pop_util_get_mask_diff_exp(m,name,&nm);
  }
  if (dumpname[0] != '\0')
    append_farray_plot(dumpname,"LGN_EX_PSC",mask,nm,1);
  if (subclass == NULL)
    xsp = pop_util_expand_grid_spikes(cx,cy,cn,xn,yn,s,cnt,tn,cw,sdf,sdtau);
  else if (strcmp(subclass,"ds01")==0)
    xsp = pop_util_expand_ds01_paired_spikes(cx,cy,cn,xn,yn,s,cnt,tn,cw);
  else
    exit_error("POP_UTIL_SUM_GT_SPK_GRID","Unknown subclass");
  if ((gw >= 0.0) && (gw != 1.0)){ 
    multiply_farray(xsp,tn,gw); 
  }
  if (dumpname[0] != '\0')
    append_farray_plot(dumpname,"sum_spikes",xsp,tn,1);
  g = convolve_with_mask_causal(xsp,tn,mask,nm);
  if (dumpname[0] != '\0')
    append_farray_plot(dumpname,"g_conv_AMPA",g,tn,1);
  add_to_farray(gt,g,tn);
  myfree(g);
  myfree(mask);
  tp = lay->nmda_p;
  if (gnt != (float *)NULL){
    nn = (int)(1.0/tp->nmda_alpha_3 * 6.0);
    if (nn > tn){
      nn = tn-1;
    }
    nmda_mask = pop_util_get_nmda_shape(tp->nmda_alpha_1,tp->nmda_alpha_2,
					tp->nmda_alpha_3,tp->lgn_nmda_amp_1,
					tp->lgn_nmda_amp_2,
					tp->lgn_nmda_amp_3,nn);
    if (dumpname[0] != '\0')
      append_farray_plot(dumpname,"NMDA_PSC",nmda_mask,nn,1);
    g = convolve_with_mask_causal(xsp,tn,nmda_mask,nn);
    if (dumpname[0] != '\0')
      append_farray_plot(dumpname,"g_conv_NMDA",g,tn,1);
    add_to_farray(gnt,g,tn);
    myfree(g);
    myfree(nmda_mask);
  }
  myfree(xsp);
}
void pop_util_pair_process(logf,mechp,x1,x2,n,mask,nm,mask_nmda,nn,gt,gnt,
			   dumpname)
     char *logf;              
     struct pop_mech *mechp;  
     float *x1,*x2;           
     int n;                   
     float *mask;             
     int nm;                  
     float *mask_nmda;        
     int nn;                  
     float *gt;               
     float *gnt;              
     char *dumpname;          
{
  float *g1,*g2,*g,*gn;
  g1 = convolve_with_mask_causal(x1,n,mask,nm);
  g2 = convolve_with_mask_causal(x2,n,mask,nm);
  if (strcmp(mechp->type,"multiply")==0){
    g = multiply_farrays(g1,g2,n);
    if (mechp->normv != 1.0)
      multiply_farray(g,n,mechp->normv);  
  }else if (strcmp(mechp->type,"divide")==0){
    multiply_farray(g2,n,mechp->b);
    add_const_farray(g2,n,mechp->a);
    g = divide_farrays(g1,g2,n);
  }else
    mylogx(logf,"POP_UTIL_PAIR_PROCESS","Unknown mech type");
  add_to_farray(gt,g,n); 
  if (dumpname[0] != '\0'){
    append_farray_plot(dumpname,"g1_conv_AMPA",g1,n,1);
    append_farray_plot(dumpname,"g2_conv_AMPA",g2,n,1);
    printf("NORMw = %f\n",mechp->normv);
    append_farray_plot(dumpname,"gMULT",g,n,1);
    append_farray_plot(dumpname,"gt_TOT",gt,n,1);
  }
  if (gnt != (float *)NULL){ 
    gn = convolve_with_mask_causal(g,n,mask_nmda,nn);
    add_to_farray(gnt,gn,n);
    myfree(gn);
  }
  myfree(g1);
  myfree(g2);
  myfree(g);
}
void pop_util_gt_sum_grid_pair(logf,lay,lgn_in1,lgn_in2,cw,xn,yn,
			       gt,gnt,tn,dumpname)
     char *logf;
     struct pop_layer *lay;       
     struct pop_lgn_in *lgn_in1;  
     struct pop_lgn_in *lgn_in2;  
     float *cw;                   
     int xn,yn;                   
     float *gt;                   
     float *gnt;                  
     int tn;                      
     char dumpname[];
{
  int i;
  int nm,nn,cn;
  int *cx_1,*cy_1,*cx_2,*cy_2,xi,yi,**n_1,**n_2,n1,n2;
  float ***s_1,***s_2,*s1,*s2;
  float *mask,*mask_nmda,sdf,sdtau,*x1,*x2,w;
  char tstr[SLEN];
  struct pop_mech *mechr,*mechp;
  struct nmda_param *tp;
  mechr = lgn_in1->mech;  
  mechp = lgn_in2->mech;  
  sdf   = mechr->sdf;
  sdtau = mechr->sdtau;
  mask  = pop_util_get_mask_diff_exp_mech(mechr,&nm);
  if (dumpname[0] != '\0') 
    append_farray_plot(dumpname,"LGN_EX_PSC",mask,nm,1);
  if (gnt != (float *)NULL){
    tp = lay->nmda_p;
    nn = (int)(1.0/tp->nmda_alpha_3 * 6.0);
    if (nn > tn){
      mylog(logf,"*** Shortening NMDA mask");
      nn = tn-1;
    }
    mask_nmda = pop_util_get_nmda_shape(tp->nmda_alpha_1,tp->nmda_alpha_2,
					tp->nmda_alpha_3,tp->lgn_nmda_amp_1,
					tp->lgn_nmda_amp_2,
					tp->lgn_nmda_amp_3,nn);
    if (dumpname[0] != '\0')
      append_farray_plot(dumpname,"NMDA_PSC",mask_nmda,nn,1);
  }else{
    mask_nmda = NULL;
    nn = -1;
  }
  if (strcmp(mechp->type,"multiply")==0){
    mechp->normv = ((double)sum_farray(mask,nm,0,nm) /
		    (double)sum_square_farray(mask,nm));
  }
  cx_1 = lgn_in1->cx1;
  cy_1 = lgn_in1->cy1;
  cx_2 = lgn_in2->cx1;
  cy_2 = lgn_in2->cy1;
  s_1  = lgn_in1->lay->s[1];       
  n_1  = lgn_in1->lay->cnt[1];     
  s_2  = lgn_in2->lay->s[1];       
  n_2  = lgn_in2->lay->cnt[1];     
  cn = lgn_in1->cn1;  
  if (lgn_in2->cn1 != cn)
    mylogx(logf,"POP_UTIL_GT_SUM_GRID_PAIR","Paired connection count error");
  for(i=0;i<cn;i++){  
    if (cw != NULL)   
      w = cw[i];
    else
      w = 1.0;
    xi = cx_1[i];  
    yi = cy_1[i];
    s1 = s_1[xi][yi];       
    n1 = n_1[xi][yi];
    xi = cx_2[i];  
    if (xi >= 0){
      yi = cy_2[i];
      s2 = s_2[xi][yi];    
      n2 = n_2[xi][yi];    
      x1 = pop_util_expand_spikes(s1,n1,tn,w,sdf,sdtau);
      x2 = pop_util_expand_spikes(s2,n2,tn,w,sdf,sdtau);
      if (dumpname[0] != '\0'){
	append_farray_plot(dumpname,"expanded_train_1",x1,tn,1);
	append_farray_plot(dumpname,"expanded_train_2",x2,tn,1);
      }
      pop_util_pair_process(logf,mechp,x1,x2,tn,mask,nm,mask_nmda,nn,gt,gnt,
			    dumpname);
      myfree(x1);
      myfree(x2);
    }
  }
  cx_1 = lgn_in1->cx0;
  cy_1 = lgn_in1->cy0;
  cx_2 = lgn_in2->cx0;
  cy_2 = lgn_in2->cy0;
  s_1  = lgn_in1->lay->s[0];       
  n_1  = lgn_in1->lay->cnt[0];     
  s_2  = lgn_in2->lay->s[0];       
  n_2  = lgn_in2->lay->cnt[0];     
  cn = lgn_in1->cn0;  
  if (lgn_in2->cn0 != cn)
    mylogx(logf,"POP_UTIL_GT_SUM_GRID_PAIR","Paired connection count error");
  for(i=0;i<cn;i++){  
    if (cw != NULL)   
      w = cw[i];
    else
      w = 1.0;
    xi = cx_1[i];
    yi = cy_1[i];
    s1 = s_1[xi][yi];       
    n1 = n_1[xi][yi];
    xi = cx_2[i];  
    if (xi >= 0){
      yi = cy_2[i];
      s2 = s_2[xi][yi];    
      n2 = n_2[xi][yi];    
      x1 = pop_util_expand_spikes(s1,n1,tn,w,sdf,sdtau);
      x2 = pop_util_expand_spikes(s2,n2,tn,w,sdf,sdtau);
      if (dumpname[0] != '\0'){
	append_farray_plot(dumpname,"expanded_train_1",x1,tn,1);
	append_farray_plot(dumpname,"expanded_train_2",x2,tn,1);
      }
      pop_util_pair_process(logf,mechp,x1,x2,tn,mask,nm,mask_nmda,nn,gt,gnt,
			    dumpname);
      myfree(x1);
      myfree(x2);
    }
  }
  myfree(mask);
  if (mask_nmda != NULL)
    myfree(mask_nmda);
}
void pop_util_set_gad_layer(lay,xn,yn,tnsec)
     struct pop_layer *lay;    
     int xn,yn;                
     float tnsec;              
{
  int i,j,k;
  int tn;
  struct pop_cell *c;
  tn = my_rint(tnsec * 1000.0); 
  for(i=0;i<lay->xn;i++){
    for(j=0;j<lay->yn;j++){
      for(k=0;k<lay->zn;k++){ 
	c = &(lay->c[i][j][k]);
	if (c->gta0 != NULL){  
	  zero_farray(c->gta0,c->gta_max); 
	}else{
	  ; 
	}
	if (c->gtx0 == NULL){
	  c->gtx0 = get_zero_farray(tn);    
	  if (lay->lgn_gain == 1)
	    c->ggain = get_zero_farray(tn); 
	  c->gti0 = get_zero_farray(tn);    
	  if (lay->nmda_p != NULL)
	    c->gtn0 = get_zero_farray(tn);
	  c->n0 = tn;
	}else{
	  if (tn != c->n0)
	    exit_error("POP_UTIL_SET_GAD_LAYER","tn changed");
	  zero_farray(c->gtx0,tn);
	  if (lay->lgn_gain == 1)
	    zero_farray(c->ggain,tn);
	  zero_farray(c->gti0,tn);
	  if (lay->nmda_p != NULL)
	    zero_farray(c->gtn0,tn);
	}
      }
    }
  }
}
void popu_sum_gain(logf,ln,gsum,tn,c,tscale)
     char *logf;              
     struct pop_lgn_in *ln;   
     float *gsum;             
     int tn;                  
     struct pop_cell *c;      
     float tscale;            
{
  int i,j,k;
  int xn,yn,n,xi,yi,*x,*y,ndup;
  float ***graw,*tg;
  char gtype[SLEN];
  int x1,x2,y1,y2,xc,yc,ri,ti;
  float r;    
  float sig;  
  float dx2,dy,dist,w,wtot;
  float *tsum;
  strcpy(gtype,"sz1_sum");
  graw = ln->lay->rgl;
  xn   = ln->lay->xn;
  yn   = ln->lay->yn;
  ndup = my_rint(c->samp0 * tscale);
  if ((ndup < 1) || (ndup > 10))
    mylogx(logf,"POPU_SUM_GAIN","ndup out of range");
  if (graw == NULL)
    mylogx(logf,"POPU_SUM_GAIN","LGN 3D gain response is NULL");
  if (gsum == NULL)
    mylogx(logf,"POPU_SUM_GAIN","gsum is NULL");
  if (strcmp(gtype,"input_sum")==0){
    n = ln->cn0;  
    x = ln->cx0;
    y = ln->cy0;
    for(i=0;i<n;i++){  
      xi = x[i];
      yi = y[i];
      tg = graw[xi][yi];
      for(j=0;j<tn;j++){
	k = j/2;
	gsum[j] += tg[k];
      }
    }
    n = ln->cn1;  
    x = ln->cx1;
    y = ln->cy1;
    for(i=0;i<n;i++){  
      xi = x[i];
      yi = y[i];
      tg = graw[xi][yi];
      for(j=0;j<tn;j++){
	k = j/2;
	gsum[j] += tg[k];
      }
    }
  }else if (strcmp(gtype,"sz1_sum")==0){
    xc = my_rint(c->rfx);
    yc = my_rint(c->rfy);
    sig = pop_cell_attrib_get_f(c,"sz1");
    r  = 3.0 * sig;
    ri = my_rint(r);
    x1 = xc - ri;
    x2 = xc + ri;
    if (x1 < 0)
      x1 = 0;
    if (x2 >= xn)
      x2 = xn-1;
    y1 = yc - ri;
    y2 = yc + ri;
    if (y1 < 0)
      y1 = 0;
    if (y2 >= yn)
      y2 = yn-1;
    tsum = get_zero_farray(tn);
    wtot = 0.0;
    for(i=x1;i<=x2;i++){
      dx2 = (float)((i-xc)*(i-xc));
      for(j=y1;j<=y2;j++){
	dy = (float)(j-yc);
	tg = graw[i][j];
	dist = sqrt(dx2 + dy*dy);
	w = (float)func_gaussian_one(dist,0.0,sig);
	wtot += w;
	for(ti=0;ti<tn;ti++){
	  k = ti/2;
	  tsum[ti] += w * tg[k];
	}
      }
    }
    for(ti=0;ti<tn;ti++){
      tsum[ti] /= wtot;      
      gsum[ti] += tsum[ti];  
    }
  }else if (strcmp(gtype,"sz1_sum")==0){
    mylogx(logf,"POPU_SUM_GAIN","Unknown gain type.");
  }
}
void pop_util_set_gt_on_off_input_layer(logf,lay,m,
					xn,yn,tnsec,seed,outfile,binoc_flag)
     char *logf;               
     struct pop_layer *lay;    
     struct model_struct *m;   
     int xn,yn;                
     float tnsec;              
     int seed;                 
     char outfile[];           
     int binoc_flag;           
{
  int i,j,k,gi,ti;
  int nseed,*seedlist,si,tn;
  float *gt,*gnt,*gsum,*cw,tscale,gfrac,gadd,gmult,*sm_gain,tg,ocdom,gw;
  char dumpf[SLEN];
  struct pop_cell *c;
  struct pop_lgn_in *ln;
  struct pop_mech *mechr;
  int nm;
  float sdf,sdtau,*xsp,*mask,*g;
  if (m->ppl == NULL){
    mechr = pop_cell_get_mech_by_name(lay,"lgn_ex");
    if (mechr == NULL)
      exit_error("POP_UTIL_SET_GT_ON_OFF_INPUT_LAYER","Assumed mech missing");
    sdf = mechr->sdf;
    sdtau = mechr->sdtau;
    mask = pop_util_get_mask_diff_exp_mech(mechr,&nm);
  }else{
    exit_error("POP_UTIL_SET_GT_ON_OFF_INPUT_LAYER","Old way");
  }
  cw = (float *)NULL;  
  nseed = lay->xn * lay->yn * lay->zn;
  seedlist = get_seeds(seed,100000,nseed);
  tn = my_rint(tnsec * lay->c[0][0][0].samp0);
  tscale = onode_getpar_flt_exit(m->o,"tscale");
  if (lay->lgn_gain == 1){
    gfrac = exp(- tscale/lay->lgn_gain_tau);
    gadd  = lay->lgn_gain_ca;
    gmult = lay->lgn_gain_cb;
  }
  si = 0; 
  for(i=0;i<lay->xn;i++){
    for(j=0;j<lay->yn;j++){
      for(k=0;k<lay->zn;k++){ 
	c = &(lay->c[i][j][k]);
	popc_precomp_reset(c,lay,tn);
	gt = c->gtx0;     
	gnt = c->gtn0;    
	gsum = c->ggain;  
	if ((i==lay->xn/2) && (j==lay->yn/2) && (outfile[0] != '\0')){
	  strcpy(dumpf,outfile);  
	  printf("  Dumping %s for cell %d %d\n",dumpf,i,j);
	}else
	  dumpf[0] = '\0';  
	if (binoc_flag == 0){
	  if ((c->lgn_n > 0) && (c->lgnin == NULL)){
	    xsp = popc_precomp_get_expanded_spikes(c,tn,sdf,sdtau);
	    g = convolve_with_mask_causal(xsp,tn,mask,nm); 
	    myfree(xsp); 
	    add_to_farray(gt,g,tn);  
	    myfree(g);
	    if (lay->lgn_gain == 1){
	      exit_error("POP_UTIL_SET_GT_ON_OFF_INPUT_LAYER",
			 "Gain not implemented for retina mesh");
	    }
	    if (gnt != NULL){  
	      exit_error("POP_UTIL_SET_GT_ON_OFF_INPUT_LAYER",
			 "NMDA not implemented for retina mesh");
	    }
	  }else{
	    for(gi=0;gi<c->lgn_n;gi++){ 
	      ln = c->lgnin[gi];
	      gw = ln->gw;  
	      if (ln->cpair == NULL){
		pop_util_sum_gt_spk_grid(lay,ln->cx0,ln->cy0,ln->cn0,cw,gw,xn,
					 yn,
					 ln->lay->s[0],ln->lay->cnt[0],m,NULL,
					 ln->mech,gt,gnt,tn,dumpf,c->subclass);
		pop_util_sum_gt_spk_grid(lay,ln->cx1,ln->cy1,ln->cn1,cw,gw,xn,
					 yn,
					 ln->lay->s[1],ln->lay->cnt[1],m,NULL,
					 ln->mech,gt,gnt,tn,dumpf,c->subclass);
	      }else{
		pop_util_gt_sum_grid_pair(logf,lay,ln,ln->cpair,cw,xn,yn,
					  gt,gnt,tn,dumpf);
	      }
	      if (lay->lgn_gain == 1){
		popu_sum_gain(logf,ln,gsum,tn,c,tscale);
	      }
	    }
	  }
	  if (lay->lgn_gain == 1){
	    sm_gain = get_farray(tn); 
	    sm_gain[0] = gadd + gmult*gsum[0];
	    gt[0] /= sm_gain[0];   
	    if (lay->nmda_p != NULL)
	      gnt[0] /= sm_gain[0];  
	    for(ti=1;ti<tn;ti++){
	      tg = gadd + gmult*gsum[ti];
	      sm_gain[ti] = gfrac * sm_gain[ti-1] + (1.0-gfrac)*tg;
	      gt[ti] /= sm_gain[ti];   
	      if (lay->nmda_p != NULL)
		gnt[ti] /= sm_gain[ti];  
	    }
	    myfree(gsum);  
	    c->ggain = sm_gain;
	  }
	}else{  
	  ocdom = pop_cell_attrib_get_f(c,"ocdom");
	  for(gi=0;gi<c->lgn_n;gi++){ 
	    ln = c->lgnin[gi];
	    gw = ln->gw;  
	    if (ln->cpair != NULL){
	      exit_error("POP_UTIL_SET_GT_ON_OFF_INPUT_LAYER",
			 "LGN pairing not coded for binoc yet.");
	    }
	    if (ocdom == -1.0){ 
	      pop_util_sum_gt_spk_grid(lay,ln->cx1,ln->cy1,ln->cn1,cw,gw,xn,yn,
				       ln->lay->s[1],ln->lay->cnt[1],m,NULL,
				       ln->mech,gt,gnt,tn,dumpf,
				       c->subclass); 
	      pop_util_sum_gt_spk_grid(lay,ln->cx0,ln->cy0,ln->cn0,cw,gw,xn,yn,
				       ln->lay->s[0],ln->lay->cnt[0],m,NULL,
				       ln->mech,gt,gnt,tn,dumpf,
				       c->subclass); 
	    }else if (ocdom == 1.0){ 
	      pop_util_sum_gt_spk_grid(lay,ln->cx1,ln->cy1,ln->cn1,cw,gw,xn,yn,
				       ln->lay->s[3],ln->lay->cnt[3],m,NULL,
				       ln->mech,gt,gnt,tn,dumpf,
				       c->subclass); 
	      pop_util_sum_gt_spk_grid(lay,ln->cx0,ln->cy0,ln->cn0,cw,gw,xn,yn,
				       ln->lay->s[2],ln->lay->cnt[2],m,NULL,
				       ln->mech,gt,gnt,tn,dumpf,
				       c->subclass); 
	    }else{
	      printf("  Cell ocdom = %f\n",ocdom);
	      exit_error("POP_UTIL_SET_GT_ON_OFF_INPUT_LAYER",
			 "Ambiguous ocdom");
	    }
	  }
	  if (lay->lgn_gain == 1){
	    exit_error("POP_UTIL_SET_GT_ON_OFF_INPUT_LAYER",
		       "LGN gain not imp'd for binoc yet.");
	  }
	}
	multiply_farray(gt,tn,c->cifcp->gx_scale);
	add_const_farray(gt,tn,c->cifcp->gx_bias);
	pop_util_add_noise(gt,tn,seedlist[si],
			   c->cifcp->gx_noise,
			   c->cifcp->gx_noise_mu,
			   c->cifcp->gx_noise_sd,
			   c->cifcp->gx_noise_tsd,tscale);
	si += 1; 
	if (c->cifcp->grect)
	  half_wave_rectify_farray(gt,tn);
	if (gnt != NULL){
	  multiply_farray(gnt,tn,c->cifcp->gx_scale);
	}
      }
    }
  }
  myfree(mask);
  myfree(seedlist);
}
void pop_util_corr_nearby_spikes(xn,yn,s,cnt,periodms,corr_d,corr_p,tsd,seed)
     int xn,yn;       
     float ***s;      
     int **cnt;       
     float periodms;  
     int corr_d;      
     float corr_p;    
     float tsd;       
     int seed;        
{
  int i,j,k,ii,jj;
  int nn,**skn,tn,d0x,d0y,maxn,n;
  float pkeep,***sk,*ts,*ss,tmax;
  if (seed > 0)
    seed = -seed;
  tmax = periodms - 1.0; 
  nn = corr_d * corr_d - 1;                
  pkeep = 1.0/(corr_p * (float)nn + 1.0);  
  maxn = max_of_2d_iarray(cnt,xn,yn);
  ts = (float *)myalloc(5*maxn*sizeof(float)); 
  sk = (float ***)myalloc(xn*sizeof(float **));
  skn = (int **)myalloc(xn*sizeof(int *));
  for(i=0;i<xn;i++){
    sk[i] = (float **)myalloc(yn*sizeof(float *));
    skn[i] = (int *)myalloc(xn*sizeof(int));
    for(j=0;j<yn;j++){
      ss = s[i][j];
      n = cnt[i][j];
      tn = 0;  
      for(k=0;k<n;k++){
	if (pop_util_ran2(&seed) <= pkeep){  
	  ts[tn] = ss[k];
	  tn += 1;
	}
      }
      sk[i][j] = copy_farray(ts,tn);
      skn[i][j] = tn;
    }
  }
  for(i=0;i<xn;i++){
    for(j=0;j<yn;j++){ 
      tn = 0;  
      d0x = i - (int)((corr_d - 1)/2);
      d0y = j - (int)((corr_d - 1)/2);
      for(ii=d0x;ii<(d0x+corr_d);ii++){
	for(jj=d0y;jj<(d0y+corr_d);jj++){  
	  if ((ii > 0) && (jj > 0) && (ii < (xn-1)) && (jj < (yn-1))){
	    ss = sk[ii][jj];
	    n = skn[ii][jj];
	    if ((ii == i) && (jj == j)){
	      for(k=0;k<n;k++){  
		ts[tn] = ss[k];
		tn += 1;
	      }
	    }else{
	      if (tsd > 0.0){
		for(k=0;k<n;k++){
		  if (pop_util_ran2(&seed) <= corr_p){  
		    ts[tn] = ss[k] + tsd*pop_util_gasdev(&seed); 
		    if ((ts[tn] > 0.0) && (ts[tn] < tmax))
		      tn += 1;
		  }
		}
	      }else{
		for(k=0;k<n;k++){
		  if (pop_util_ran2(&seed) <= corr_p){  
		    ts[tn] = ss[k];
		    tn += 1;
		  }
		}
	      }
	    }
	  }
	}
      }
      if (tn > 1)
	sort_farray(ts,tn);
      myfree(s[i][j]);
      s[i][j] = copy_farray(ts,tn);
      cnt[i][j] = tn;
    }
  }
  myfree(ts);
  free_3d_farray(sk,xn,yn,0); 
  free_2d_iarray(skn,xn);
}
void pop_util_init_bg_spikes(lay,pseed,tn,tscale)
     struct pop_layer *lay;    
     int *pseed;
     int tn;
     float tscale;
{
  int i,j,k;
  int n,s1n,s2n,s3n,done,scx_n,sci_n,pp_n;
  float duration,rmu,rsigma,sampling,f,sr1,*s1,*s2,*s3,cfrac_x,cfrac_i;
  float rate_x,rate_i,rate_sx,rate_si,rate_px,rate_pi,*scx,*sci,*pp_s,*pp_i;
  struct pop_cell *c;
  struct onode *bgo;
  done = 0;
  i = 0;
  bgo = NULL;
  while(done == 0){
    if (onode_test_chr(lay->inlist[i],"type","bg")==1){
      bgo = lay->inlist[i];
      done = 1;
    }else{
      i += 1;
      if (i >= lay->ninlist)
	done = 1;
    }
  }
  rmu = 0.0;
  rsigma = 0.0;
  n = 1;
  sampling = 1.0; 
  duration = (float)tn * tscale * sampling;
  if (*pseed > 0)
    *pseed *= -1;
  scx = sci = NULL;         
  scx_n = sci_n = 0;        
  rate_sx = rate_si = 0.0;  
  rate_px = rate_pi = 0.0;  
  rate_x = rate_i = 0.0;
  if (bgo != NULL){
    cfrac_x = onode_getpar_flt_dflt(bgo,"ex_corr_frac",0.0);
    if (cfrac_x > 0.0){
      rate_x = onode_getpar_flt_exit(bgo,"ex_rate");
      rate_sx = rate_x * cfrac_x; 
      rate_px = rate_x - rate_sx; 
      make_poisson_float_spikes_new(&scx,&scx_n,0.0,duration,sampling,rate_sx,
				    rmu,rsigma,n,pseed);
    }
    cfrac_i = onode_getpar_flt_dflt(bgo,"in_corr_frac",0.0);
    if (cfrac_i > 0.0){
      rate_i = onode_getpar_flt_exit(bgo,"in_rate");
      rate_si = rate_i * cfrac_i; 
      rate_pi = rate_i - rate_si; 
      make_poisson_float_spikes_new(&sci,&sci_n,0.0,duration,sampling,rate_si,
				    rmu,rsigma,n,pseed);
    }
  }
  for(i=0;i<lay->xn;i++){
    for(j=0;j<lay->yn;j++){
      for(k=0;k<lay->zn;k++){ 
	c = &(lay->c[i][j][k]);
	if (c->bg_x_s != NULL){
	  myfree(c->bg_x_s);
	  c->bg_x_s = NULL;
	}
	if (c->bg_i_s != NULL){
	  myfree(c->bg_i_s);
	  c->bg_i_s = NULL;
	}
	c->bg_x_n = -1;
	c->bg_i_n = -1;
      }
    }
  }
  for(i=0;i<lay->xn;i++){
    for(j=0;j<lay->yn;j++){
      for(k=0;k<lay->zn;k++){ 
	c = &(lay->c[i][j][k]);
	if (c->bg_x_rate > 0.0){ 
	  if (c->bg_x_corr_f > 0.0){ 
	    if (c->bg_x_corr_c == NULL){ 
	      make_poisson_float_spikes_new(&pp_s,&pp_n,0.0,
					    duration,sampling,rate_px,
					    rmu,rsigma,n,pseed);
	      c->bg_x_s = merge_spike_arrays_float(pp_s,pp_n,scx,scx_n);
	      c->bg_x_n = pp_n + scx_n;
	      if (pp_s != NULL)
		myfree(pp_s);
	    }else{
	      if (c->bg_x_corr_c->bg_x_n >= 0){ 
		f = c->bg_x_corr_f;
		s1 = c->bg_x_corr_c->bg_x_s;
		s1n = c->bg_x_corr_c->bg_x_n;
		spike_util_random_select_spikes(s1,s1n,f,*pseed,&s2,&s2n);
		sr1 = (1.0 - f) * c->bg_x_rate;
		if (sr1 > 0.0){
		  make_poisson_float_spikes_new(&s3,&s3n,0.0,duration,sampling,
						sr1,rmu,rsigma,n,pseed);
		}else{
		  s3 = NULL;
		  s3n = 0;
		}
		c->bg_x_s = merge_spike_arrays_float(s2,s2n,s3,s3n);
		c->bg_x_n = s2n + s3n;
		myfree(s2);
		if (s3 != NULL)
		  myfree(s3);
	      }else{
		make_poisson_float_spikes_new(&(c->bg_x_s),&(c->bg_x_n),0.0,
					      duration,sampling,c->bg_x_rate,
					      rmu,rsigma,n,pseed);
	      }
	    }
	  }else{
	    make_poisson_float_spikes_new(&(c->bg_x_s),&(c->bg_x_n),0.0,
					  duration,sampling,c->bg_x_rate,
					  rmu,rsigma,n,pseed);
	  }
	  c->bg_x_k = 0; 
	}
	if (c->bg_i_rate > 0.0){
	  if (c->bg_i_corr_f > 0.0){ 
	    if (c->bg_i_corr_c == NULL){ 
	      make_poisson_float_spikes_new(&pp_s,&pp_n,0.0,
					    duration,sampling,rate_pi,
					    rmu,rsigma,n,pseed);
	      c->bg_i_s = merge_spike_arrays_float(pp_s,pp_n,sci,sci_n);
	      c->bg_i_n = pp_n + sci_n;
	      if (pp_s != NULL)
		myfree(pp_s);
	    }else{
	      if (c->bg_i_corr_c->bg_i_n >= 0){ 
		f = c->bg_i_corr_f;
		s1 = c->bg_i_corr_c->bg_i_s;
		s1n = c->bg_i_corr_c->bg_i_n;
		spike_util_random_select_spikes(s1,s1n,f,*pseed,&s2,&s2n);
		sr1 = (1.0 - f) * c->bg_i_rate;
		if (sr1 > 0.0){
		  make_poisson_float_spikes_new(&s3,&s3n,0.0,duration,sampling,
						sr1,rmu,rsigma,n,pseed);
		}else{
		  s3 = NULL;
		  s3n = 0;
		}
		c->bg_i_s = merge_spike_arrays_float(s2,s2n,s3,s3n);
		c->bg_i_n = s2n + s3n;
		myfree(s2);
		if (s3 != NULL)
		  myfree(s3);
	      }else{
		make_poisson_float_spikes_new(&(c->bg_i_s),&(c->bg_i_n),0.0,
					      duration,sampling,c->bg_i_rate,
					      rmu,rsigma,n,pseed);
	      }
	    }
	  }else{
	    make_poisson_float_spikes_new(&(c->bg_i_s),&(c->bg_i_n),0.0,
					  duration,sampling,c->bg_i_rate,
					  rmu,rsigma,n,pseed);
	  }
	  c->bg_i_k = 0; 
	}
      }
    }
  }
  if (scx != NULL)  myfree(scx);
  if (sci != NULL)  myfree(sci);
}
void pop_util_init_bg_rate(laylist,n)
     struct pop_layer **laylist;
     int n;
{
  int i,j,k,l;
  struct pop_cell *c;
  struct pop_layer *lay;
  printf("******************* POP_UTIL_INIT_BG_RATE\n");
  for(l=0;l<n;l++){ 
    lay = laylist[l];
    if (lay->runflag == 1){
      for(i=0;i<lay->xn;i++){
	for(j=0;j<lay->yn;j++){
	  for(k=0;k<lay->zn;k++){ 
	    c = &(lay->c[i][j][k]);
	    if (c->bg_x_rate > 0.0){
	      c->bg_x_rate /= 1000.0;
	    }
	    if (c->bg_i_rate > 0.0){
	      c->bg_i_rate /= 1000.0;
	    }
	  }
	}
      }
    }
  }
}
void pop_input_01(c,t1)
     struct pop_cell *c;   
     float t1;             
{
  int k;
  int gi,gn,done,divide_flag,stype;
  float x_tau,x_aetau,x_tau2inv,x_totau;
  float i_tau,i_aetau,i_tau2inv,i_totau;
  float x_g_new,x_g_old,x_gd_new,x_gd_old,*x_gdata;
  float i_g_new,i_g_old,i_gd_new,i_gd_old,*i_gdata;
  float stor_x,stor_i,tval,dg_ex,dg_in,old_tsav;
  double t,dt,t0,sdt;
  struct pop_circbuf *x_sbuf,*i_sbuf,*x_gbuf,*i_gbuf;
  struct pop_syn *syn_ptr;
  float tot,a,b,c_a,c_b;
  struct pop_circbuf *cb;
  old_tsav = -1.0;
  x_tau = c->ex1tau;  
  x_aetau = c->ex1amp * M_E / x_tau;
  x_tau2inv = 1.0/(x_tau*x_tau);
  x_totau = 2.0/x_tau;
  i_tau = c->in1tau;  
  i_aetau = c->in1amp * M_E / i_tau;
  i_tau2inv = 1.0/(i_tau*i_tau);
  i_totau = 2.0/i_tau;
  x_g_old = c->gex1;     
  x_gd_old = c->gdex1;   
  i_g_old = c->gin1;     
  i_gd_old = c->gdin1;   
  x_sbuf = c->ex1s;      
  i_sbuf = c->in1s;      
  x_gbuf = c->ex1g;      
  i_gbuf = c->in1g;      
  x_gdata = x_gbuf->d;   
  i_gdata = i_gbuf->d;   
  dt = x_gbuf->dt;       
  gn = x_gbuf->n;        
  sdt = x_sbuf->dt;      
  gi = x_gbuf->i;        
  t0 = x_sbuf->t + sdt;  
  t = c->gex1t;          
  if (c->savg){ 
    if (c->gtex1n > 0){  
      stor_x = c->gtex1[c->gtex1n - 1];  
      stor_i = c->gtin1[c->gtex1n - 1];  
    }else
      stor_x = stor_i = -1.0;
  }
  while(t < t1){
    t += dt;
    x_g_new = x_g_old + x_gd_old * dt;
    x_gd_new = x_gd_old - (x_totau * x_gd_old + x_tau2inv * x_g_old) * dt;
    i_g_new = i_g_old + i_gd_old * dt;
    i_gd_new = i_gd_old - (i_totau * i_gd_old + i_tau2inv * i_g_old) * dt;
    if (t >= t0){ 
      k = x_sbuf->i; 
      done = 0;
      while(!done){
	if (c->bg_x_k >= c->bg_x_n){
	  done = 1; 
	}else{
	  if (t >= c->bg_x_s[c->bg_x_k]){
	    x_sbuf->d[k] += c->bg_x_amp;  
	    c->bg_x_k += 1;
	  }else{
	    done = 1;
	  }
	}
      }
      done = 0;
      while(!done){
	if (c->bg_i_k >= c->bg_i_n){
	  done = 1; 
	}else{
	  if (t >= c->bg_i_s[c->bg_i_k]){
	    i_sbuf->d[k] += c->bg_i_amp;  
	    c->bg_i_k += 1;
	  }else{
	    done = 1;
	  }
	}
      }
      x_gd_new += x_sbuf->d[k] * x_aetau;
      x_sbuf->d[k] = 0.0;   
      i_gd_new += i_sbuf->d[k] * i_aetau;
      i_sbuf->d[k] = 0.0;   
      k += 1;               
      if (k >= x_sbuf->n)
	k = 0;
      x_sbuf->i = k;
      t0 += sdt;            
    }
    gi += 1;
    if (gi == gn)
      gi = 0;
    x_gdata[gi] = x_g_new;
    i_gdata[gi] = i_g_new;
    if (c->syn_proc_code == 1){
      tot = 0.0;
      syn_ptr = c->in;
      stype = -1;
      while (syn_ptr != NULL){
	if (syn_ptr->si != NULL){
	  if ((syn_ptr->si->syn_code == 100002) && (syn_ptr->si->ci == 0)){
	    if (stype == -1)
	      stype = syn_ptr->stype;
	    else{
	      if (syn_ptr->stype != stype){
		printf("*** old stype:  %d   now found %d\n",stype,
		       syn_ptr->stype);
		exit_error("POP_INPUT_01","syn_ptr type mismatch");
	      }
	    }
	    c_a = syn_ptr->si->siu->s002->msi->a;
	    c_b = syn_ptr->si->siu->s002->msi->b;
	    if (syn_ptr->si->siu->s002->msi->si_comp == 1)  
	      divide_flag = 0;
	    else
	      divide_flag = 1;  
	    cb = syn_ptr->si->siu->s002->wt1;
	    a = pop_util_circbuf_val_at_t(cb,t);
	    cb = syn_ptr->si->siu->s002->wt2;
	    b = pop_util_circbuf_val_at_t(cb,t);
	    if (divide_flag == 1)
	      tval = a/(c_a + c_b * b);    
	    else
	      tval = a*(c_a + c_b * b);    
	    if (tval < 0.0)
	      tval = 0.0;
	    tot += syn_ptr->w * tval; 
	  }
	}
	syn_ptr = syn_ptr->post_next;
      }
      if (stype == 1)
	x_gdata[gi] += tot;  
      else
	i_gdata[gi] += tot;  
    }
    if (c->savg){
      dg_ex = (x_gdata[gi] - stor_x)/(x_gdata[gi] + 0.001);
      dg_in = (i_gdata[gi] - stor_i)/(i_gdata[gi] + 0.001);
      if ((dg_ex > 0.02) || (dg_ex < -0.02) ||
	  (dg_in > 0.02) || (dg_in < -0.02) || (t - old_tsav > 0.001)){
	c->gtex1[c->gtex1n] = x_gdata[gi];   
	c->gtin1[c->gtex1n] = i_gdata[gi];   
	c->gtex1t[c->gtex1n] = t*1000.0; 
	c->gtex1n += 1;
	if (c->gtex1n  > c->gtex1max){  
	  printf("%d %d\n",c->gtex1n,c->gtex1max);
	  exit_error("POP_INPUT_01","Too many stored values");
	}
	stor_x = x_gdata[gi];  
	stor_i = i_gdata[gi];
	old_tsav = t;
      }
    }
    x_g_old = x_g_new;
    x_gd_old = x_gd_new;
    i_g_old = i_g_new;
    i_gd_old = i_gd_new;
  }
  x_gbuf->i = gi;
  x_gbuf->t = t;  
  c->gex1t = t;
  x_sbuf->t = t0 - sdt;
  c->gex1 = x_g_new;
  c->gdex1 = x_gd_new;
  c->gin1 = i_g_new;
  c->gdin1 = i_gd_new;
  i_sbuf->i = x_sbuf->i;  
  i_sbuf->t = x_sbuf->t;
}
void pop_input_01a(c,t1)
     struct pop_cell *c;   
     float t1;             
{
  int k;
  int gi,gn,done,divide_flag,stype;
  float x_tau,x_aetau,x_tau2inv,x_totau;
  float i_tau,i_aetau,i_tau2inv,i_totau;
  float x_g_new,x_g_old,x_gd_new,x_gd_old,*x_gdata;
  float i_g_new,i_g_old,i_gd_new,i_gd_old,*i_gdata;
  float stor_x,stor_i,tval,dg_ex,dg_in,old_tsav;
  double t,dt,t0,sdt;
  struct pop_circbuf *x_sbuf,*i_sbuf,*x_gbuf,*i_gbuf;
  struct pop_syn *syn_ptr;
  float tot,a,b,c_a,c_b;
  struct pop_circbuf *cb;
  old_tsav = -1.0;
  x_tau = c->ex1tau;  
  x_aetau = c->ex1amp * M_E / x_tau;
  x_tau2inv = 1.0/(x_tau*x_tau);
  x_totau = 2.0/x_tau;
  i_tau = c->in1tau;  
  i_aetau = c->in1amp * M_E / i_tau;
  i_tau2inv = 1.0/(i_tau*i_tau);
  i_totau = 2.0/i_tau;
  x_g_old = c->gex1;     
  x_gd_old = c->gdex1;   
  i_g_old = c->gin1;     
  i_gd_old = c->gdin1;   
  x_sbuf = c->ex1s;      
  i_sbuf = c->in1s;      
  x_gbuf = c->ex1g;      
  i_gbuf = c->in1g;      
  x_gdata = x_gbuf->d;   
  i_gdata = i_gbuf->d;   
  dt = x_gbuf->dt;       
  gn = x_gbuf->n;        
  sdt = x_sbuf->dt;      
  gi = x_gbuf->i;        
  t0 = x_sbuf->t + sdt;  
  t = c->gex1t;          
  if (c->savg){ 
    if (c->gtex1n > 0){  
      stor_x = c->gtex1[c->gtex1n - 1];  
      stor_i = c->gtin1[c->gtex1n - 1];  
    }else
      stor_x = stor_i = -1.0;
  }
  k = x_sbuf->i; 
  if (pop_util_ran2(&globseed) < c->bg_x_rate){
    x_sbuf->d[k] += c->bg_x_amp;  
  }
  if (pop_util_ran2(&globseed) < c->bg_i_rate){
    i_sbuf->d[k] += c->bg_i_amp;  
  }
  while(t < t1){
    t += dt;
    x_g_new = x_g_old + x_gd_old * dt;
    x_gd_new = x_gd_old - (x_totau * x_gd_old + x_tau2inv * x_g_old) * dt;
    i_g_new = i_g_old + i_gd_old * dt;
    i_gd_new = i_gd_old - (i_totau * i_gd_old + i_tau2inv * i_g_old) * dt;
    if (t >= t0){ 
      k = x_sbuf->i; 
      x_gd_new += x_sbuf->d[k] * x_aetau;
      x_sbuf->d[k] = 0.0;   
      i_gd_new += i_sbuf->d[k] * i_aetau;
      i_sbuf->d[k] = 0.0;   
      k += 1;               
      if (k >= x_sbuf->n)
	k = 0;
      x_sbuf->i = k;
      t0 += sdt;            
    }
    gi += 1;
    if (gi == gn)
      gi = 0;
    x_gdata[gi] = x_g_new;
    i_gdata[gi] = i_g_new;
    if (c->syn_proc_code == 1){
      tot = 0.0;
      syn_ptr = c->in;
      stype = -1;
      while (syn_ptr != NULL){
	if (syn_ptr->si != NULL){
	  if ((syn_ptr->si->syn_code == 100002) && (syn_ptr->si->ci == 0)){
	    if (stype == -1)
	      stype = syn_ptr->stype;
	    else{
	      if (syn_ptr->stype != stype){
		printf("*** old stype:  %d   now found %d\n",stype,
		       syn_ptr->stype);
		exit_error("POP_INPUT_01","syn_ptr type mismatch");
	      }
	    }
	    c_a = syn_ptr->si->siu->s002->msi->a;
	    c_b = syn_ptr->si->siu->s002->msi->b;
	    if (syn_ptr->si->siu->s002->msi->si_comp == 1)  
	      divide_flag = 0;
	    else
	      divide_flag = 1;  
	    cb = syn_ptr->si->siu->s002->wt1;
	    a = pop_util_circbuf_val_at_t(cb,t);
	    cb = syn_ptr->si->siu->s002->wt2;
	    b = pop_util_circbuf_val_at_t(cb,t);
	    if (divide_flag == 1)
	      tval = a/(c_a + c_b * b);    
	    else
	      tval = a*(c_a + c_b * b);    
	    if (tval < 0.0)
	      tval = 0.0;
	    tot += syn_ptr->w * tval; 
	  }
	}
	syn_ptr = syn_ptr->post_next;
      }
      if (stype == 1)
	x_gdata[gi] += tot;  
      else
	i_gdata[gi] += tot;  
    }
    if (c->savg){
      dg_ex = (x_gdata[gi] - stor_x)/(x_gdata[gi] + 0.001);
      dg_in = (i_gdata[gi] - stor_i)/(i_gdata[gi] + 0.001);
      if ((dg_ex > 0.02) || (dg_ex < -0.02) ||
	  (dg_in > 0.02) || (dg_in < -0.02) || (t - old_tsav > 0.001)){
	c->gtex1[c->gtex1n] = x_gdata[gi];   
	c->gtin1[c->gtex1n] = i_gdata[gi];   
	c->gtex1t[c->gtex1n] = t*1000.0; 
	c->gtex1n += 1;
	if (c->gtex1n  > c->gtex1max){  
	  printf("%d %d\n",c->gtex1n,c->gtex1max);
	  exit_error("POP_INPUT_01","Too many stored values");
	}
	stor_x = x_gdata[gi];  
	stor_i = i_gdata[gi];
	old_tsav = t;
      }
    }
    x_g_old = x_g_new;
    x_gd_old = x_gd_new;
    i_g_old = i_g_new;
    i_gd_old = i_gd_new;
  }
  x_gbuf->i = gi;
  x_gbuf->t = t;  
  c->gex1t = t;
  x_sbuf->t = t0 - sdt;
  c->gex1 = x_g_new;
  c->gdex1 = x_gd_new;
  c->gin1 = i_g_new;
  c->gdin1 = i_gd_new;
  i_sbuf->i = x_sbuf->i;  
  i_sbuf->t = x_sbuf->t;
}
void pop_input_02(c,t1,lay)
     struct pop_cell *c;
     float t1;             
     struct pop_layer *lay;
{
  int k;
  int gi,gn,done;
  float ex_tau,ex_aetau,ex_tau2inv,ex_totau;
  float in_tau,in_aetau,in_tau2inv,in_totau;
  float n1_aetau,n1_tau2inv,n1_totau;
  float n2_aetau,n2_tau2inv,n2_totau;
  float n3_aetau,n3_tau2inv,n3_totau;
  float ex_g_new,ex_g_old,ex_gd_new,ex_gd_old,*x_gdata;
  float in_g_new,in_g_old,in_gd_new,in_gd_old,*i_gdata;
  float n1_g_new,n1_g_old,n1_gd_new,n1_gd_old,*n_gdata;
  float n2_g_new,n2_g_old,n2_gd_new,n2_gd_old;
  float n3_g_new,n3_g_old,n3_gd_new,n3_gd_old;
  float nm_g_new;
  float stor_x,stor_i,stor_n,stor_eps;
  float xsc;
  double t,dt,t0,sdt;
  struct pop_circbuf *x_sbuf,*i_sbuf;
  stor_eps = 0.1;       
  n1_aetau   = lay->nmda_p->nmda1_aetau;
  n1_tau2inv = lay->nmda_p->nmda1_tau2inv;
  n1_totau   = lay->nmda_p->nmda1_totau;
  n2_aetau   = lay->nmda_p->nmda2_aetau;
  n2_tau2inv = lay->nmda_p->nmda2_tau2inv;
  n2_totau   = lay->nmda_p->nmda2_totau;
  n3_aetau   = lay->nmda_p->nmda3_aetau;
  n3_tau2inv = lay->nmda_p->nmda3_tau2inv;
  n3_totau   = lay->nmda_p->nmda3_totau;
  ex_tau = c->ex1tau;  
  ex_aetau = c->ex1amp * M_E / ex_tau;
  ex_tau2inv = 1.0/(ex_tau*ex_tau);
  ex_totau = 2.0/ex_tau;
  in_tau = c->in1tau;  
  in_aetau = c->in1amp * M_E / in_tau;
  in_tau2inv = 1.0/(in_tau*in_tau);
  in_totau = 2.0/in_tau;
  n1_g_old  = c->gex1_nmda1;  
  n2_g_old  = c->gex1_nmda2;  
  n3_g_old  = c->gex1_nmda3;  
  n1_gd_old = c->gdex1_nmda1; 
  n2_gd_old = c->gdex1_nmda2; 
  n3_gd_old = c->gdex1_nmda3; 
  ex_g_old  = c->gex1;       
  ex_gd_old = c->gdex1;      
  in_g_old  = c->gin1;       
  in_gd_old = c->gdin1;      
  x_sbuf = c->ex1s; 
  i_sbuf = c->in1s; 
  n_gdata = c->ex1g_nmda->d; 
  x_gdata = c->ex1g->d;
  i_gdata = c->in1g->d;   
  dt = c->ex1g->dt;       
  gn = c->ex1g->n;        
  sdt = x_sbuf->dt;       
  gi = c->ex1g->i;        
  t0 = x_sbuf->t + sdt;   
  t = c->gex1t;           
  if (c->savg){
    if (c->gtex1n > 0){
      stor_x = c->gtex1[c->gtex1n - 1];
      stor_i = c->gtin1[c->gtex1n - 1];
      stor_n = c->gtex1_nmda[c->gtex1n - 1];
    }else
      stor_x = stor_i = stor_n = -1.0; 
  }
  while(t < t1){
    t += dt;
    n1_g_new = n1_g_old + n1_gd_old * dt;
    n2_g_new = n2_g_old + n2_gd_old * dt;
    n3_g_new = n3_g_old + n3_gd_old * dt;
    ex_g_new = ex_g_old + ex_gd_old * dt;
    in_g_new = in_g_old + in_gd_old * dt;
    n1_gd_new = n1_gd_old - (n1_totau * n1_gd_old + n1_tau2inv * n1_g_old) *dt;
    n2_gd_new = n2_gd_old - (n2_totau * n2_gd_old + n2_tau2inv * n2_g_old) *dt;
    n3_gd_new = n3_gd_old - (n3_totau * n3_gd_old + n3_tau2inv * n3_g_old) *dt;
    ex_gd_new = ex_gd_old - (ex_totau * ex_gd_old + ex_tau2inv * ex_g_old) *dt;
    in_gd_new = in_gd_old - (in_totau * in_gd_old + in_tau2inv * in_g_old) *dt;
    if (t >= t0){ 
      k = x_sbuf->i; 
      done = 0;
      while(!done){
	if (c->bg_x_k >= c->bg_x_n){
	  done = 1; 
	}else{
	  if (t >= c->bg_x_s[c->bg_x_k]){
	    x_sbuf->d[k] += c->bg_x_amp;  
	    c->bg_x_k += 1;
	  }else{
	    done = 1;
	  }
	}
      }
      done = 0;  
      while(!done){
	if (c->bg_i_k >= c->bg_i_n){
	  done = 1; 
	}else{
	  if (t >= c->bg_i_s[c->bg_i_k]){
	    i_sbuf->d[k] += c->bg_i_amp;  
	    c->bg_i_k += 1;
	  }else{
	    done = 1;
	  }
	}
      }
      xsc = x_sbuf->d[k];
      n1_gd_new += xsc * n1_aetau;  
      n2_gd_new += xsc * n2_aetau;
      n3_gd_new += xsc * n3_aetau;
      ex_gd_new += xsc * ex_aetau;  
      x_sbuf->d[k] = 0.0;   
      in_gd_new += i_sbuf->d[k] * in_aetau;
      i_sbuf->d[k] = 0.0;   
      k += 1;               
      if (k >= x_sbuf->n)
	k = 0;
      x_sbuf->i = k;
      t0 += sdt;                
    }
    if (c->savg){
      nm_g_new = n1_g_new + n2_g_new + n3_g_new;
      if (((ex_g_new - stor_x) > stor_eps)||((ex_g_new - stor_x) < -stor_eps)||
	  ((in_g_new - stor_i) > stor_eps)||((in_g_new - stor_i) < -stor_eps)||
	  ((nm_g_new - stor_n) > stor_eps)||((nm_g_new - stor_n) < -stor_eps)){
	c->gtex1[c->gtex1n]      = ex_g_new;
	c->gtex1_nmda[c->gtex1n] = nm_g_new;
	c->gtin1[c->gtex1n]      = in_g_new;
	c->gtex1t[c->gtex1n] = t*1000.0; 
	c->gtex1n += 1;
	if (c->gtex1n  > c->gtex1max)
	  exit_error("POP_INPUT_02","Too many stored values");
	stor_x = ex_g_new;
	stor_i = in_g_new;
	stor_n = nm_g_new;
      }
    }
    gi += 1;
    if (gi == gn)
      gi = 0;
    n_gdata[gi] = n1_g_new + n2_g_new + n3_g_new;
    x_gdata[gi] = ex_g_new;
    i_gdata[gi] = in_g_new;
    n1_g_old = n1_g_new;
    n2_g_old = n2_g_new;
    n3_g_old = n3_g_new;
    n1_gd_old = n1_gd_new;
    n2_gd_old = n2_gd_new;
    n3_gd_old = n3_gd_new;
    ex_g_old = ex_g_new;
    ex_gd_old = ex_gd_new;
    in_g_old = in_g_new;
    in_gd_old = in_gd_new;
  }
  c->ex1g->i = gi;
  c->ex1g->t = t;  
  c->gex1t = t;
  x_sbuf->t = t0 - sdt;
  c->gex1 = ex_g_new;
  c->gdex1 = ex_gd_new;
  c->gin1 = in_g_new;
  c->gdin1 = in_gd_new;
  i_sbuf->i = x_sbuf->i;  
  i_sbuf->t = x_sbuf->t;
}
void pop_util_deriv_01(x,v,dydx,c)
     float x;      
     float v;      
     float *dydx;  
     struct pop_cell *c;  
{
  int k;
  int n01,i0,i1;
  float dt,t,g_ex,g_in,g_ad,tt,gexd,gind;
  struct ifc_param *p;
  struct pop_circbuf *gbuf,*i_gbuf;
  gbuf   = c->ex1g;
  i_gbuf = c->in1g;
  if ((x > gbuf->t) || (x < (gbuf->t - (gbuf->n-1)*gbuf->dt))){
    printf("DERIV:  x=%f gbuf->t=%f  gbuf->i=%d\n",x,gbuf->t,gbuf->i);
    exit_error("POP_UTIL_DERIV_01","x not covered by gbuf");
  }
  tt = (gbuf->t - x)/gbuf->dt;
  k = (int)tt;
  dt = (float)(k+1) - tt;  
  i1 = gbuf->i - k;
  if (i1 > 0)
    i0 = i1 - 1;
  else if (i1 == 0)
    i0 = gbuf->n - 1;
  else{
    i1 += gbuf->n;
    i0 = i1 - 1;
  }
  gexd =   gbuf->d[i0] + dt * (  gbuf->d[i1] -   gbuf->d[i0]);
  gind = i_gbuf->d[i0] + dt * (i_gbuf->d[i1] - i_gbuf->d[i0]);
  if (c->ad1g != NULL){
    g_ad = pop_util_circbuf_val_at_t_interp(c->ad1g,x);
  }else{
    g_ad = 0.0;
  }
  p = c->cifcp;
  n01 = c->n0 - 1;  
  if (n01 <= 0){ 
    g_ex = 0.0;
    g_in = 0.0;
  }else{
    t = x * c->samp0;   
    k = (int)t;     
    if (k < n01){ 
      dt = t - (float)k;
      g_ex = c->gtx0[k] +  dt * (c->gtx0[k+1]-c->gtx0[k]);
      g_in = c->gti0[k] +  dt * (c->gti0[k+1]-c->gti0[k]);
    }else{
      g_ex = c->gtx0[n01];
      g_in = c->gti0[n01];
    }
  }
  g_ex += gexd;
  g_in += gind;
  *dydx = 1000.0/p->c_x * (g_ex        * (p->v_ex     - v) +
			   g_in        * (p->v_in     - v) +
			   p->g_leak_x * (p->v_leak_x - v) +
			   g_ad        * (p->v_ad     - v));
}
void pop_util_deriv_02(float x, float *y, float *dydx, struct pop_cell *c,
		       struct pop_layer *lay)
{
  int k;
  int n01,i0,i1;
  float dt,t,g_ex,g_nm,g_in,g_ad,tt,gexd,gnmd,gind,vs,vd,gt,rfact,ff;
  struct ifc_param *p;
  struct nmda_param *tp;
  struct pop_circbuf *gbuf,*i_gbuf,*n_gbuf;
  if (global_dump_flag)
    printf("GLOBAL DUMP--------------------------------\n");
  vs = y[1]; 
  vd = y[2]; 
  gbuf = c->ex1g;
  n_gbuf = c->ex1g_nmda;
  i_gbuf = c->in1g;
  if ((x > gbuf->t) || (x < (gbuf->t - (gbuf->n-1)*gbuf->dt))){
    printf("DERIV:  x=%f gbuf->t=%f  gbuf->i=%d\n",x,gbuf->t,gbuf->i);
    exit_error("POP_UTIL_DERIV_01","x not covered by gbuf");
  }
  tt = (gbuf->t - x)/gbuf->dt;
  k = (int)tt;
  dt = (float)(k+1) - tt;  
  i1 = gbuf->i - k;
  if (i1 > 0)
    i0 = i1 - 1;
  else if (i1 == 0)
    i0 = gbuf->n - 1;
  else{
    i1 += gbuf->n;
    i0 = i1 - 1;
  }
  gexd =   gbuf->d[i0] + dt * (  gbuf->d[i1] -   gbuf->d[i0]);
  gnmd = n_gbuf->d[i0] + dt * (n_gbuf->d[i1] - n_gbuf->d[i0]);
  gind = i_gbuf->d[i0] + dt * (i_gbuf->d[i1] - i_gbuf->d[i0]);
  if (c->ad1g != NULL){
    g_ad = pop_util_circbuf_val_at_t_interp(c->ad1g,x);
  }else{
    g_ad = 0.0;
  }
  n01 = c->n0 - 1;  
  if (n01 == 0){ 
    g_ex = 0.0;
    g_nm = 0.0;
    g_in = 0.0;
  }else{
    t = x * c->samp0;   
    k = (int)t;     
    if (k < n01){ 
      dt = t - (float)k;
      g_ex = c->gtx0[k] +  dt * (c->gtx0[k+1]-c->gtx0[k]);
      g_nm = c->gtn0[k] +  dt * (c->gtn0[k+1]-c->gtn0[k]);
      g_in = c->gti0[k] +  dt * (c->gti0[k+1]-c->gti0[k]);
    }else{
      g_ex = c->gtx0[n01];
      g_nm = c->gtn0[n01];
      g_in = c->gti0[n01];
    }
  }
  tp = lay->nmda_p;
  k = (vd - tp->nmda_v0f) * tp->nmda_ppmv;
  if ((k < 0) || (k >= tp->nmda_n)){
    printf("vd = %f  vs = %f\n",vd,vs);
    printf("k = %d   tp->nmda_n = %d\n",k,tp->nmda_n);
    exit_error("POP_UTIL_DERIV_O2","NMDA table index");
  }
  ff = tp->nmda_gv[k];
  g_ex += gexd + ff*(gnmd + g_nm); 
  g_in += gind;
  p = c->cifcp;
  gt = p->g_tran;
  dydx[1] = 1000.0/p->c_x * (g_ex        * (p->v_ex     - vs) +
			     g_in        * (p->v_in     - vs) +
			     p->g_leak_x * (p->v_leak_x - vs) +
			     g_ad        * (p->v_ad     - vs) +
			     gt          * (vd - vs));
  if (x < c->reft)
    if (dydx[1] > 0.0){ 
      rfact = 1.0 - (c->reft - x)/p->trefr_x_s;
      dydx[1] *= rfact;  
      if (global_dump_flag){
	printf("  Refraction rfact = %f\n",rfact);
	printf("    x < reft  -->   %f  < %f\n",x,c->reft);
      }
    }
  dydx[2] = 1000.0/p->c_x * (g_ex        * (p->v_ex     - vd) +
			     g_in        * (p->v_in     - vd) +
			     p->g_leak_x * (p->v_leak_x - vd) +
			     gt          * (vs - vd));
  if (global_dump_flag){
    printf("  gex=%f gin=%f gad=%f glk=%f gt=%f\n",g_ex,g_in,g_ad,p->g_leak_x,
	   gt);
    printf("  1000.0/c_x  =  %f\n",1000.0/p->c_x);
    printf("  v_ex - vs  =  %f\n",p->v_ex     - vs);
    printf("  v_in - vs  =  %f\n",p->v_in     - vs);
    printf("  v_lk - vs  =  %f\n",p->v_leak_x - vs);
    printf("  v_ad - vs  =  %f\n",p->v_ad     - vs);
    printf("  vd   - vs  =  %f\n",vd          - vs);
    printf("  v_ex - vd  =  %f\n",p->v_ex     - vd);
    printf("  v_in - vd  =  %f\n",p->v_in     - vd);
    printf("  v_lk - vd  =  %f\n",p->v_leak_x - vd);
    printf("  vs   - vd  =  %f\n",vs          - vd);
    printf("  old vs = %f\n",vs);
    printf("  old vd = %f\n",vd);
    printf("  soma dydx = %f\n",dydx[1]);
    printf("  dend dydx = %f\n",dydx[2]);
    printf("  x (time) = %f\n",x);
  }
}
/*C  (C) Copr. 1986-92 Numerical Recipes Software #.3.                       */
void pop_rkck_01(float y, float dydx, float x, float h, float *yout,
		 float *yerr, struct pop_cell *c)
{
  static float a2=0.2,a3=0.3,a4=0.6,a5=1.0,a6=0.875,b21=0.2,
    b31=3.0/40.0,b32=9.0/40.0,b41=0.3,b42 = -0.9,b43=1.2,
    b51 = -11.0/54.0, b52=2.5,b53 = -70.0/27.0,b54=35.0/27.0,
    b61=1631.0/55296.0,b62=175.0/512.0,b63=575.0/13824.0,
    b64=44275.0/110592.0,b65=253.0/4096.0,c1=37.0/378.0,
    c3=250.0/621.0,c4=125.0/594.0,c6=512.0/1771.0,
    dc5 = -277.0/14336.0;
  float dc1=c1-2825.0/27648.0,dc3=c3-18575.0/48384.0,
    dc4=c4-13525.0/55296.0,dc6=c6-0.25;
  float ak2,ak3,ak4,ak5,ak6,ytemp;
  ytemp=y+b21*h*dydx;
  pop_util_deriv_01(x+a2*h,ytemp,&ak2,c);
  ytemp=y+h*(b31*dydx+b32*ak2);
  pop_util_deriv_01(x+a3*h,ytemp,&ak3,c);
  ytemp=y+h*(b41*dydx+b42*ak2+b43*ak3);
  pop_util_deriv_01(x+a4*h,ytemp,&ak4,c);
  ytemp=y+h*(b51*dydx+b52*ak2+b53*ak3+b54*ak4);
  pop_util_deriv_01(x+a5*h,ytemp,&ak5,c);
  ytemp=y+h*(b61*dydx+b62*ak2+b63*ak3+b64*ak4+b65*ak5);
  pop_util_deriv_01(x+a6*h,ytemp,&ak6,c);
  *yout=y+h*(c1*dydx+c3*ak3+c4*ak4+c6*ak6);
  *yerr=h*(dc1*dydx+dc3*ak3+dc4*ak4+dc5*ak5+dc6*ak6);
}
/*C  (C) Copr. 1986-92 Numerical Recipes Software #.3.                       */
void pop_rkck_02(float y[], float dydx[], int n,float x, float h, float yout[],
		 float yerr[], struct pop_cell *cp, struct pop_layer *lay,
		 void (*derivs)(float, float [], float [], struct pop_cell *,
				struct pop_layer *))
{
  int i;
  static float a2=0.2,a3=0.3,a4=0.6,a5=1.0,a6=0.875,b21=0.2,
    b31=3.0/40.0,b32=9.0/40.0,b41=0.3,b42 = -0.9,b43=1.2,
    b51 = -11.0/54.0, b52=2.5,b53 = -70.0/27.0,b54=35.0/27.0,
    b61=1631.0/55296.0,b62=175.0/512.0,b63=575.0/13824.0,
    b64=44275.0/110592.0,b65=253.0/4096.0,c1=37.0/378.0,
    c3=250.0/621.0,c4=125.0/594.0,c6=512.0/1771.0,
    dc5 = -277.0/14336.0;
  float dc1=c1-2825.0/27648.0,dc3=c3-18575.0/48384.0,
    dc4=c4-13525.0/55296.0,dc6=c6-0.25;
  float *ak2,*ak3,*ak4,*ak5,*ak6,*ytemp;
  ak2=vector(1,n);
  ak3=vector(1,n);
  ak4=vector(1,n);
  ak5=vector(1,n);
  ak6=vector(1,n);
  ytemp=vector(1,n);
  for (i=1;i<=n;i++)
    ytemp[i]=y[i]+b21*h*dydx[i];
  (*derivs)(x+a2*h,ytemp,ak2,cp,lay);  
  for (i=1;i<=n;i++)
    ytemp[i]=y[i]+h*(b31*dydx[i]+b32*ak2[i]);
  (*derivs)(x+a3*h,ytemp,ak3,cp,lay);
  for (i=1;i<=n;i++)
    ytemp[i]=y[i]+h*(b41*dydx[i]+b42*ak2[i]+b43*ak3[i]);
  (*derivs)(x+a4*h,ytemp,ak4,cp,lay);
  for (i=1;i<=n;i++)
    ytemp[i]=y[i]+h*(b51*dydx[i]+b52*ak2[i]+b53*ak3[i]+b54*ak4[i]);
  (*derivs)(x+a5*h,ytemp,ak5,cp,lay);
  for (i=1;i<=n;i++)
    ytemp[i]=y[i]+h*(b61*dydx[i]+b62*ak2[i]+b63*ak3[i]+b64*ak4[i]+b65*ak5[i]);
  (*derivs)(x+a6*h,ytemp,ak6,cp,lay);
  for (i=1;i<=n;i++)
    yout[i]=y[i]+h*(c1*dydx[i]+c3*ak3[i]+c4*ak4[i]+c6*ak6[i]);
  for (i=1;i<=n;i++)
    yerr[i]=h*(dc1*dydx[i]+dc3*ak3[i]+dc4*ak4[i]+dc5*ak5[i]+dc6*ak6[i]);
  free_vector(ytemp,1,n);
  free_vector(ak6,1,n);
  free_vector(ak5,1,n);
  free_vector(ak4,1,n);
  free_vector(ak3,1,n);
  free_vector(ak2,1,n);
}
#define SAFETY 0.9
#define PGROW -0.2
#define PSHRNK -0.25
#define ERRCON 1.89e-4
/*C  (C) Copr. 1986-92 Numerical Recipes Software #.3.                       */
void pop_rkqs_01(float *y,       
		 float dydx,     
		 float *x,       
		 float htry,     
		 float eps,      
		 float yscal,    
		 float *hdid,    
		 float *hnext,   
		 struct pop_cell *c)  
{
  float errmax,h,xnew,yerr,ytemp;
  h=htry;
  for (;;){
    pop_rkck_01(*y,dydx,*x,h,&ytemp,&yerr,c);
    errmax = fabs(yerr/yscal);
    errmax /= eps;
    if (errmax > 1.0){
      h = SAFETY*h*pow(errmax,PSHRNK);
      if (h < 0.1*h){  
	h *= 0.1;
	printf("POP_UTIL *** HERE h = %f\n",h);
	printf("THIS SHOULD NEVER HAPPEN\n");
	exit(0);
      }
      xnew = (*x)+h;
      if (xnew == *x)
	nrerror("stepsize underflow in pop_rkqs_01");
      continue;
    }else{
      if (errmax > ERRCON)
	*hnext = SAFETY*h*pow(errmax,PGROW);
      else
	*hnext=5.0*h;
      *x += (*hdid=h);
      *y=ytemp;
      break;
    }
  }
}
#undef SAFETY
#undef PGROW
#undef PSHRNK
#undef ERRCON
#define SAFETY 0.9
#define PGROW -0.2
#define PSHRNK -0.25
#define ERRCON 1.89e-4
/*C  (C) Copr. 1986-92 Numerical Recipes Software #.3.                       */
void pop_rkqs_02(float y[], float dydx[], int n, float *x, float htry,
		 float eps, float yscal[], float *hdid, float *hnext,
		 struct pop_cell *cp, struct pop_layer *lay,
		 void (*derivs)(float, float [], float [], struct pop_cell *,
				struct pop_layer *))
{
  void pop_rkck_02(float y[], float dydx[], int n, float x, float h,
		   float yout[],float yerr[], struct pop_cell *cp,
		   struct pop_layer *lay,
		   void (*derivs)(float, float [], float [], struct pop_cell *,
				  struct pop_layer *));
  int i;
  float errmax,h,xnew,*yerr,*ytemp,wyy;
  yerr = vector(1,n);
  ytemp = vector(1,n);
  h=htry;
  for (;;){
    pop_rkck_02(y,dydx,n,*x,h,ytemp,yerr,cp,lay,derivs);
    errmax=0.0;
    for (i=1;i<=n;i++){
      wyy = fabs(yerr[i]/yscal[i]);
      if (wyy > errmax)
	errmax = wyy;
    }
    errmax /= eps;
    if (errmax > 1.0){
      h = SAFETY*h*pow(errmax,PSHRNK);
      if (h < 0.1*h){  
	h *= 0.1;
	printf("POP_UTIL *** HERE h = %f\n",h);
	printf("THIS SHOULD NEVER HAPPEN\n");
	exit(0);
      }
      xnew = (*x)+h;
      if (xnew == *x)
	nrerror("stepsize underflow in rkqs");
      continue;
    }else{
      if (errmax > ERRCON)
	*hnext = SAFETY*h*pow(errmax,PGROW);
      else
	*hnext=5.0*h;
      *x += (*hdid=h);
      for (i=1;i<=n;i++)
	y[i]=ytemp[i];
      break;
    }
  }
  free_vector(ytemp,1,n);
  free_vector(yerr,1,n);
}
#undef SAFETY
#undef PGROW
#undef PSHRNK
#undef ERRCON
void pop_process_spike_01(c,x,y,yd)
     struct pop_cell *c;
     float x;             
     float y;             
     float yd;            
{
  int j;
  int toffset,savti;
  struct pop_syn *tsyn;
  if (c->savs){
    if (c->ns >= c->maxscnt)      
      pop_util_cell_spike_inc(c,2);
    c->s[c->ns] = x*1000.0;  
    c->ns += 1;
  }
  if (c->savv){
    if ((c->vn+2) >= c->vnmax) 
      exit_error("POP_PROCESS_SPIKE_01","Vm storage exceeded");
    c->vmt[c->vn] = x; 
    c->vm[c->vn] = y;  
    if (c->savd){
      c->vmd[c->vn] = yd;    
      c->vmd[c->vn+1] = yd;  
      c->vmd[c->vn+2] = yd;  
    }
    c->vn += 1;
    c->vmt[c->vn] = x;
    c->vm[c->vn] = c->cifcp->v_spike;   
    c->vn += 1;
    c->vmt[c->vn] = x;
    c->vm[c->vn] = c->cifcp->v_reset_x; 
    c->vn += 1;
    c->xsav = x;
    c->reft = x + c->cifcp->trefr_x_s;  
  }
  if (c->nad > 0){
    pop_util_circbuf_mask_update_t(c->ad1g,x,c->gta_pulse,c->nad,c->gta0,
				   &savti);
  }
  tsyn = c->out;
  while(tsyn != NULL){  
    pop_util_process_spike_postsyn(tsyn,x,c->input_ex_sdf,c->input_ex_sdtau);
    tsyn = tsyn->pre_next; 
  }
}
void pop_step_poiss_01(c,x1,x2)
     struct pop_cell *c;
     float x1,x2;         
{
  int k;
  int n01,i1;
  float x,dt,t,g_ex,g_in,g_ad,tt,gexd,gind,prob,rate;
  struct poiss_param *p;
  struct pop_circbuf *gbuf,*i_gbuf;
  x = c->x; 
  if (c->savv && (c->vn < c->vnmax) && ((x - c->xsav) > c->dxsav)){
    c->vmt[c->vn] = x;  
    c->vm[c->vn] = c->y;
    c->vn += 1;
    c->xsav = x;
  }
  gbuf   = c->ex1g;
  i_gbuf = c->in1g;
  if ((x > gbuf->t) || (x < (gbuf->t - (gbuf->n-1)*gbuf->dt))){
    printf("x=%f gbuf->t=%f  gbuf->i=%d\n",x,gbuf->t,gbuf->i);
    exit_error("POP_STEP_POISS_01","x not covered by gbuf");
  }
  tt = (gbuf->t - x)/gbuf->dt;
  k = (int)(0.5 + tt);  
  i1 = gbuf->i - k;
  if (i1 < 0)
    i1 += gbuf->n;
  gexd =   gbuf->d[i1];
  gind = i_gbuf->d[i1];
  if (c->ad1g != NULL){
    g_ad = pop_util_circbuf_val_at_t_interp(c->ad1g,x);
  }else{
    g_ad = 0.0;
  }
  n01 = c->n0 - 1;  
  if (n01 <= 0){  
    g_ex = 0.0;
    g_in = 0.0;
  }else{ 
    t = x * c->samp0;   
    k = (int)t;     
    if (k < n01){ 
      dt = t - (float)k;
      g_ex = c->gtx0[k] +  dt * (c->gtx0[k+1]-c->gtx0[k]);
      g_in = c->gti0[k] +  dt * (c->gti0[k+1]-c->gti0[k]);
    }else{
      g_ex = c->gtx0[n01];
      g_in = c->gti0[n01];
    }
  }
  g_ex += gexd;
  g_in += gind;
  p = c->cpoissp;
  rate = p->offset1 + p->scale*(g_ex - p->scale_in*g_in - p->scale_ad*g_ad
				+ p->offset0);
  prob = rate * HMAX;
  if (prob > 1.0){
    prob = 1.0;
  }else if (prob < 0.0){
    prob = 0.0;
  }
  x += HMAX;  
  c->y = prob / HMAX;  
  c->nok += 1;  
  if (prob > 0.0){
    if (pop_util_ran2(&globseed) < prob){
      pop_process_spike_01(c,x,c->y,0.0); 
      x += c->cifcp->trefr_x/1000.0;    
    }
  }
  c->x = x;     
  c->h = HMAX;  
}
#define TINY 1.0e-30
/*C  (C) Copr. 1986-92 Numerical Recipes Software #.3.                       */
void pop_step_01(c,x1,x2)
     struct pop_cell *c;
     float x1,x2;         
{
  float x,y,dydx,yscal,h,hdid,hnext;
  x = c->x; 
  y = c->y;
  h = c->h;
  pop_util_deriv_01(x,y,&dydx,c); 
  yscal = fabs(y)+fabs(dydx*h)+TINY; 
  if (c->savv && (c->vn < c->vnmax) && ((x - c->xsav) > c->dxsav)){
    c->vmt[c->vn] = x;  
    c->vm[c->vn] = y;
    c->vn += 1;
    c->xsav = x;
  }
  if ((x+h) > x2) 
    h = x2-x;
  pop_rkqs_01(&y,dydx,&x,h,c->eps,yscal,&hdid,&hnext,c); 
  if (hnext > HMAX)
    hnext = HMAX;
  if (hdid == h)
    c->nok += 1;
  else
    c->nbad += 1;
  if (y >= c->cifcp->v_th_x){
    pop_process_spike_01(c,x,y,0.0); 
    y = c->cifcp->v_reset_x;          
    x += c->cifcp->trefr_x/1000.0;    
    hnext = c->h1;                
  }
  if (fabs(hnext) <= c->hmin){
    printf("hnext = %e\n",hnext);
    nrerror("Step size too small in odeint");
  }
  h = hnext;
  c->x = x; 
  c->y = y;
  c->h = h;
}
#undef TINY
#define TINY 1.0e-30
void pop_step_02(lay,c,x1,x2)
     struct pop_layer *lay; 
     struct pop_cell *c;
     float x1,x2;  
{
  int nvar;
  float x,yscal[3],h,hdid,hnext;
  float y[3];
  float dydx[3];
  nvar = 2;
  x = c->x; 
  y[1] = c->y;  
  y[2] = c->yd; 
  h = c->h;
  pop_util_deriv_02(x,y,dydx,c,lay); 
  yscal[1] = fabs(y[1])+fabs(dydx[1]*h)+TINY;
  yscal[2] = fabs(y[2])+fabs(dydx[2]*h)+TINY;
  if (c->savv && (c->vn < c->vnmax) && ((x - c->xsav) > c->dxsav)){
    c->vmt[c->vn] = x;  
    c->vm[c->vn] = y[1];
    if (c->savd)
      c->vmd[c->vn] = y[2];
    c->vn += 1;
    c->xsav = x;
  }
  if ((x+h) > x2) 
    h = x2-x;
  pop_rkqs_02(y,dydx,nvar,&x,h,c->eps,yscal,&hdid,&hnext,c,lay,
	      pop_util_deriv_02);
  if (hnext > HMAX)
    hnext = HMAX;
  if (hdid == h)
    c->nok += 1;
  else
    c->nbad += 1;
  if (y[1] >= c->cifcp->v_th_x){
    pop_process_spike_01(c,x,y[1],y[2]); 
    y[1] = c->cifcp->v_reset_x;       
  }
  if (fabs(hnext) <= c->hmin){
    printf("hnext = %e\n",hnext);
    nrerror("Step size too small in odeint");
  }
  h = hnext;
  c->x = x; 
  c->y = y[1];
  c->yd = y[2];
  c->h = h;
}
#undef TINY
void pop_util_init_cell_trial(c,x1,x2)
     struct pop_cell *c;
     float x1,x2;
{
  int i;
  int nstor;
  c->y = c->ystart;
  if (c->cifcp->g_tran > 0.0)
    c->yd = c->ystart;
  c->x = x1;
  c->h = c->h1;
  c->reft = -1.0; 
  c->gex1t = 0.0; 
  c->gex1  = 0.0;  
  c->gdex1 = 0.0;
  c->gin1  = 0.0;  
  c->gdin1 = 0.0;
  if (c->savv){
    c->xsav = x1 - 2.0*c->dxsav; 
    c->vn = 0;
    nstor = (x2-x1)/(c->dxsav/2.0); 
    if ((c->vm != NULL) && (c->vnmax != nstor)){ 
      myfree(c->vm); c->vm = NULL;
      myfree(c->vmt); c->vmt = NULL;
      if (c->savd){
	myfree(c->vmd);
	c->vmd = NULL;
      }
    }
    if (c->vm == NULL){
      c->vnmax = nstor;
      c->vm  = (float *)myalloc(c->vnmax * sizeof(float));
      c->vmt = (float *)myalloc(c->vnmax * sizeof(float));
      if (c->savd)
	c->vmd  = (float *)myalloc(c->vnmax * sizeof(float));
    }else{
      if (c->vnmax != nstor){
      }
    }
  }
  if (c->sava){  
    c->gta_n = 0;  
    c->gta_max = my_rint((x2 - x1)*c->samp0);
    c->gta_max += c->nad + 1;  
    if (c->gta0 == NULL){  
      c->gta0 = (float *)myalloc(c->gta_max*sizeof(float));
    }
    for(i=0;i<c->gta_max;i++)
      c->gta0[i] = 0.0;
  }
  if (c->savs){
    c->ns = 0;
    if (c->s == NULL) 
      c->s = (float *)myalloc(c->maxscnt*sizeof(float));
  }
  if (c->savg){
    c->gtex1[0] = c->gex1;   
    c->gtex1t[0] = c->gex1t;
    c->gtin1[0] = c->gin1;   
    c->gtex1n = 1;           
  }
  pop_cell_zero_circbuf(c->ex1s);  
  pop_cell_zero_circbuf(c->in1s);
  pop_cell_zero_circbuf(c->ex1g);
  pop_cell_zero_circbuf(c->in1g);
  if (c->ad1g != NULL)
    pop_cell_zero_circbuf(c->ad1g);  
  pop_cell_reset_all_out_synapses(c);
}
void pop_util_run_layer_list(myid,mylogf,m,r,s,laylist,n,x1,x2,seed,flag)
     int myid;
     char *mylogf;
     struct model_struct *m;        
     struct response_struct *r;     
     struct stim_struct *s;         
     struct pop_layer **laylist;
     int n;
     float x1,x2;                   
     int seed;
     int flag;                      
{
  int i,j,k,l,mm,ti;
  int nstep,mon_flag,pflag;
  float dt,dtmon,tmon;
  double t;
  char tstr[SLEN];
  struct pop_cell *c;
  struct pop_layer *lay;
  pflag = 0;
  mylog(mylogf,"  POP_UTIL_RUN_LAYER_LIST\n");
  if (seed > 0)
    globseed = -seed;
  else
    globseed = seed;
  for(l=0;l<n;l++){ 
    lay = laylist[l];
    if (lay->runflag == 1){
      for(i=0;i<lay->xn;i++){
	for(j=0;j<lay->yn;j++){
	  for(k=0;k<lay->zn;k++){ 
	    c = &(lay->c[i][j][k]);
	    pop_util_init_cell_trial(c,x1,x2);
	  }
	}
      }
    }
  }
  if (pflag) printf("  pop_util_run_layer_list   001\n");
  mon_flag = paramfile_get_int_param_default(r->ppl,"mon_flag",0);
  dtmon = paramfile_get_float_param_default(r->ppl,"mon_dt",0.010);
  tmon  = 0.0;    
  if (pflag) printf("  pop_util_run_layer_list   002\n");
  dt = 0.001; 
  nstep = (x2-x1)/dt;
  t = 0.0;
  sprintf(tstr,"    Run for %d time steps, stepsize %.5f sec\n",nstep,dt);
  mylog(mylogf,tstr);
  for (ti=0;ti<nstep;ti++){ 
    t += (double)dt;
    tmon += (double)dt;
    for(l=0;l<n;l++){ 
      lay = laylist[l];
      if (lay->runflag == 1){
	for(i=0;i<lay->xn;i++){
	  for(j=0;j<lay->yn;j++){
	    for(k=0;k<lay->zn;k++){ 
	      c = &(lay->c[i][j][k]);
	      if (c->cifcp->g_tran == 0.0){
		if (flag == -1){  
		  pop_input_01a(c,t+HMAX);   
		}else{
		  pop_input_01(c,t+HMAX);    
		}
	      }else
		pop_input_02(c,t+HMAX,lay); 
	    }
	  }
	}
      }
    }
    if (pflag) printf("  pop_util_run_layer_list   003\n");
    for(l=0;l<n;l++){ 
      lay = laylist[l];
      if (lay->runflag == 1){
	for(i=0;i<lay->xn;i++){
	  for(j=0;j<lay->yn;j++){
	    for(k=0;k<lay->zn;k++){ 
	      c = &(lay->c[i][j][k]);
	      mm = 0;
	      while(c->x < t){ 
		if (c->cifcp->g_tran == 0.0){
		  if (c->cpoissp != NULL){
		    pop_step_poiss_01(c,x1,x2);
		  }else
		    pop_step_01(c,x1,x2); 
		}else{
		  if (mm > 195){
		    global_dump_flag = 1;
		    append_farray_xy_plot("z.dmp",c->vmt,c->vm,c->vn,"v");
		    append_farray_xy_plot("z.dmp",c->vmt,c->vmd,c->vn,"vd");
		    append_farray_plot("z.dmp.pl","gtx0",c->gtx0,c->n0,1);
		    append_farray_plot("z.dmp.pl","gtn0",c->gtn0,c->n0,1);
		    append_farray_plot("z.dmp.pl","gti0",c->gti0,c->n0,1);
		    if (c->gta0 != NULL){
		      append_farray_plot("z.dmp.pl","gta0",c->gta0,c->n0,1);
		    }
		    printf("WDBHere %s  x,y  %d %d\n",c->name,c->layx,c->layy);
		    exit(0);
		  }
		  pop_step_02(lay,c,x1,x2); 
		}
		mm += 1;
		if (mm > 200){ 
		  printf("l,i,j,k = %d %d %d %d   mm = %d\n",l,i,j,k,mm);
		  mylog_exit(mylogf,
		    "POP_UTIL_RUN_LAYER_LIST  Too many steps in interval\n");
		}
	      }
	    }
	  }
	}
      }
    }
    if (pflag) printf("  pop_util_run_layer_list   004  ti = %d\n",ti);
    if (tmon > dtmon){
      if ((myid == -1) && (mon_flag)){
	mod_mon_monitor(m,r,s,laylist,n,(float)t);
      }
      tmon = 0.0;
    }
    if (pflag) printf("  pop_util_run_layer_list   005\n");
  }
}
void popu_mar_write_onode_items(fout,o,bcode,vers)
     FILE *fout;
     struct onode *o;   
     int bcode;         
     int vers;          
{
  int nitem;
  struct onode *t,*s;
  fwrite((char *)&bcode,sizeof(int),1,fout);  
  nitem = onode_count_otype(o,"item");        
  fwrite((char *)&nitem,sizeof(int),1,fout);
  s = NULL;
  t = onode_get_next_type(o->o,"item");
  while(t != NULL){
    carray_nchar_write(t->name,fout);        
    carray_nchar_write(t->val,fout);         
    t = onode_get_next_type(t->next,"item");
  }
  bcode *= -1;
  fwrite((char *)&bcode,sizeof(int),1,fout);  
}
void popu_mar_write_area_name(fout,mpt,lay,vers)
     FILE *fout;
     struct pop_top *mpt;
     struct pop_layer *lay;
     int vers;
{
  int bcode;
  bcode = 10010;  
  fwrite((char *)&bcode,sizeof(int),1,fout);  
  carray_nchar_write(lay->area->name,fout);   
  bcode *= -1;
  fwrite((char *)&bcode,sizeof(int),1,fout);  
}
void popu_mar_write_area(fout,mpt,area,vers)
     FILE *fout;
     struct pop_top *mpt;
     struct pop_area *area;
     int vers;
{
  int bcode;
  bcode = 10001;
  fwrite((char *)&bcode,sizeof(int),1,fout);  
  printf("      Area %s  (%d,%d)\n",area->name,area->xn,area->yn);
  carray_nchar_write(area->name,fout);   
  fwrite((char *)&(area->x0),sizeof(float),1,fout);
  fwrite((char *)&(area->y0),sizeof(float),1,fout);
  fwrite((char *)&(area->xf),sizeof(float),1,fout);
  fwrite((char *)&(area->yf),sizeof(float),1,fout);
  fwrite((char *)&(area->xn),sizeof(int),1,fout);
  fwrite((char *)&(area->yn),sizeof(int),1,fout);
  fwrite((char *)&(area->umx),sizeof(float),1,fout);
  fwrite((char *)&(area->umy),sizeof(float),1,fout);
  fwrite((char *)&(area->sscale),sizeof(float),1,fout);
  bcode *= -1;
  fwrite((char *)&bcode,sizeof(int),1,fout);  
}
void popu_mar_write_attrib(fout,lay,vers)
     FILE *fout;
     struct pop_layer *lay;
     int vers;
{
  int i,j,k;
  int ai,*vatt,attn;
  float fval;
  struct pop_cell *c;
  attn = lay->attrib_fn;                    
  fwrite((char *)&attn,sizeof(int),1,fout);
  for(i=0;i<attn;i++){
    carray_nchar_write(lay->attrib_fname[i],fout);     
  }
  vatt = popc_attrib_layer_variation(lay);  
  for(i=0;i<attn;i++){
    fwrite((char *)&(vatt[i]),sizeof(int),1,fout);
  }
  for(ai=0;ai<attn;ai++){
    if (vatt[ai] == 0){
      fval = lay->c[0][0][0].attrib_f[ai];
      fwrite((char *)&fval,sizeof(float),1,fout);
    }else if (vatt[ai] == 1){
      for(i=0;i<lay->xn;i++){
	for(j=0;j<lay->yn;j++){
	  c = &(lay->c[i][j][0]);
	  fwrite((char *)&(c->attrib_f[ai]),sizeof(float),1,fout);
	}
      }
    }else if (vatt[ai] == 2){
      for(i=0;i<lay->zn;i++){
	c = &(lay->c[0][0][i]);
	fwrite((char *)&(c->attrib_f[ai]),sizeof(float),1,fout);
      }
    }else if (vatt[ai] == 3){
      for(i=0;i<lay->xn;i++){
	for(j=0;j<lay->yn;j++){
	  for(k=0;k<lay->zn;k++){
	    c = &(lay->c[i][j][k]);
	    fwrite((char *)&(c->attrib_f[ai]),sizeof(float),1,fout);
	  }
	}
      }
    }else{
      exit_error("POPU_MAR_WRITE_ATTRIB","Unknown variation code");
    }
  }
}
void popu_mar_write_lay(fout,mpt,lay,vers)
     FILE *fout;
     struct pop_top *mpt;
     struct pop_layer *lay;
     int vers;
{
  int i;
  int bcode;
  int pflag = 0;
  bcode = 20001;
  fwrite((char *)&bcode,sizeof(int),1,fout);  
  if (pflag == 1) printf("HERE __ 1\n");
  carray_nchar_write(lay->name,fout);        
  if (pflag == 1) printf("HERE __ 1.a\n");
  if (vers > 1)
    carray_nchar_write(lay->laytype,fout);   
  if (pflag == 1) printf("HERE __ 1.b\n");
  if (vers > 3)
    carray_nchar_write(lay->geomt,fout);     
  if (pflag == 1) printf("HERE __ 1.c\n");
  fwrite((char *)&(lay->xn),sizeof(int),1,fout);
  fwrite((char *)&(lay->yn),sizeof(int),1,fout);
  fwrite((char *)&(lay->zn),sizeof(int),1,fout);
  if (pflag == 1) printf("HERE __ 1.b\n");
  fwrite((char *)&(lay->x0),sizeof(float),1,fout);
  fwrite((char *)&(lay->y0),sizeof(float),1,fout);
  fwrite((char *)&(lay->xf),sizeof(float),1,fout);
  fwrite((char *)&(lay->yf),sizeof(float),1,fout);
  fwrite((char *)&(lay->oddxoff),sizeof(float),1,fout);
  if (pflag == 1) printf("HERE __ 2\n");
  if (lay->geomt != NULL){
    if (strcmp(lay->geomt,"irregular")==0){
      fwrite((char *)&(lay->irr_n),sizeof(int),1,fout);
      for(i=0;i<lay->irr_n;i++){
	fwrite((char *)&(lay->irr_x[i]),sizeof(float),1,fout);
	fwrite((char *)&(lay->irr_y[i]),sizeof(float),1,fout);
	fwrite((char *)&(lay->irr_id[i]),sizeof(int),1,fout);
      }
    }
  }
  if (pflag == 1) printf("HERE __ 3\n");
  fwrite((char *)&(lay->ninlist),sizeof(int),1,fout);  
  if (vers > 2){
    popu_mar_write_attrib(fout,lay,vers);
  }
  if (lay->icon != NULL){
    popu_mar_write_onode_items(fout,lay->icon->o,20100,vers);
  }
  if (lay->area != NULL){
    popu_mar_write_area_name(fout,mpt,lay,vers);
  }
  bcode *= -1;
  fwrite((char *)&bcode,sizeof(int),1,fout);  
}
int popu_mar_write_syn_reg(fout,vers,laypre,laypost,c,nbyte_pre,iin)
     FILE *fout;
     int vers;
     struct pop_layer *laypre;
     struct pop_layer *laypost;
     struct pop_cell *c;
     int nbyte_pre;
     int iin;
{
  int cnt,ci_pre,xs,ys;
  short si;
  float w;
  struct pop_syn *syn;
  struct pop_cell *cpre;
  xs = laypre->yn * laypre->zn; 
  ys = laypre->zn;
  cnt = 0;
  syn = c->in;
  while(syn != NULL){
    if (syn->inindx == iin)
      cnt += 1;
    syn = syn->post_next;
  }
  fwrite((char *)&cnt,sizeof(int),1,fout);
  syn = c->in;
  while(syn != NULL){
    if (syn->inindx == iin){
      cpre = syn->pre;  
      ci_pre = xs*cpre->layx + ys*cpre->layy + cpre->layz; 
      if (nbyte_pre == 2){
	si = (short)ci_pre;
	fwrite((char *)&si,sizeof(short),1,fout);
      }else{  
	fwrite((char *)&ci_pre,sizeof(int),1,fout);
      }
      fwrite((char *)&syn->w,sizeof(float),1,fout);
    }
    syn = syn->post_next;
  }
  return cnt;
}
void popu_mar_write_syn_reg_pair(fout,vers,laypre,laypre2,laypost,c,
				 nbyte_pre,nbyte_pre2,iin)
     FILE *fout;
     int vers;
     struct pop_layer *laypre,*laypre2;  
     struct pop_layer *laypost;
     struct pop_cell *c;
     int nbyte_pre,nbyte_pre2;
     int iin;
{
  int cnt,ci_pre,xs,ys,xs2,ys2;
  short si;
  float w;
  struct pop_syn *syn,*syn1;
  struct pop_cell *cpre;
  xs = laypre->yn * laypre->zn; 
  ys = laypre->zn;
  xs2 = laypre2->yn * laypre2->zn; 
  ys2 = laypre2->zn;
  cnt = 0;
  syn = c->in;
  while(syn != NULL){
    if (syn->inindx == iin){
      if (syn->si == NULL)
	exit_error("POPU_MAR_WRITE_SYN_REG_PAIR","No SI for synapse");
      else if (syn->si->ci == 0)
	cnt += 1;
    }
    syn = syn->post_next;
  }
  fwrite((char *)&cnt,sizeof(int),1,fout);
  syn = c->in;
  while(syn != NULL){
    if ((syn->inindx == iin) && (syn->si->ci == 0)){
      cpre = syn->pre;  
      ci_pre = xs*cpre->layx + ys*cpre->layy + cpre->layz; 
      if (nbyte_pre == 2){
	si = (short)ci_pre;
	fwrite((char *)&si,sizeof(short),1,fout);
      }else{  
	fwrite((char *)&ci_pre,sizeof(int),1,fout);
      }
      syn1 = pop_cell_syn_si_find(NULL,c->in,syn->si->siu,1);
      cpre = syn1->pre;  
      ci_pre = xs2*cpre->layx + ys2*cpre->layy + cpre->layz; 
      if (nbyte_pre2 == 2){
	si = (short)ci_pre;
	fwrite((char *)&si,sizeof(short),1,fout);
      }else{  
	fwrite((char *)&ci_pre,sizeof(int),1,fout);
      }
      fwrite((char *)&syn->w,sizeof(float),1,fout);
    }
    syn = syn->post_next;
  }
}
void popu_mar_write_syn_lgn(fout,vers,laypre,laypost,c,nbyte_pre)
     FILE *fout;
     int vers;
     struct pop_layer *laypre;
     struct pop_layer *laypost;
     struct pop_cell *c;
     int nbyte_pre;
{
  int i;
  int lxi,lyi,lzi,n0,n1,n,ci_pre,xs,ys;
  short si;
  float w;
  struct pop_lgn_in *lin;
  lin = c->lgnin[0];
  n0 = lin->cn0;  
  n1 = lin->cn1;  
  n  = n0 + n1;   
  w = 1.0;        
  xs = laypre->yn * laypre->zn; 
  ys = laypre->zn;
  fwrite((char *)&n,sizeof(int),1,fout);
  lzi = 0; 
  for(i=0;i<n0;i++){
    lxi = lin->cx0[i];
    lyi = lin->cy0[i];
    ci_pre = xs*lxi + ys*lyi + lzi; 
    if (nbyte_pre == 2){
      si = (short)ci_pre;
      fwrite((char *)&si,sizeof(short),1,fout);
    }else{  
      fwrite((char *)&ci_pre,sizeof(int),1,fout);
    }
    fwrite((char *)&w,sizeof(float),1,fout);
  }
  lzi = 1; 
  for(i=0;i<n1;i++){
    lxi = lin->cx1[i];
    lyi = lin->cy1[i];
    ci_pre = xs*lxi + ys*lyi + lzi; 
    if (nbyte_pre == 2){
      si = (short)ci_pre;
      fwrite((char *)&si,sizeof(short),1,fout);
    }else{  
      fwrite((char *)&ci_pre,sizeof(int),1,fout);
    }
    fwrite((char *)&w,sizeof(float),1,fout);
  }
}
void popu_mar_write_input(fout,mpt,lay,iin,vers)
     FILE *fout;
     struct pop_top *mpt;
     struct pop_layer *lay;
     int iin;       
     int vers;
{
  int i,j,k;
  int xn,yn,zn,bcode,ci,nbyte_pre,nbyte_pre2,nbyte_post,lgnflag,sycnt;
  short si;
  struct pop_cell *c;
  struct pop_layer *lpre,*lpre2;
  struct onode *oi;
  printf("    POPU_MAR_WRITE_INPUT\n");
  oi = lay->inlist[iin];  
  if (popu_input_is_paired(lay,iin)){
    bcode = 30002;
    lpre = popu_input_get_origin(mpt,lay,iin);
    printf("      (pair) Pre-syn layer:  %s\n",lpre->name);
    lpre2 = popu_input_get_paired_lay(mpt,lay,iin);
    printf("      (pair) Pre-syn layer 2:  %s\n",lpre2->name);
  }else if (onode_test_chr(oi,"type","bg")){
    bcode = 30003;
    fwrite((char *)&bcode,sizeof(int),1,fout);  
    bcode *= -1;
    fwrite((char *)&bcode,sizeof(int),1,fout);  
    printf("      bg input\n");
    return;
  }else if (onode_test_chr(oi,"type","template")){ 
    bcode = 30004;
    lpre = popu_input_get_origin(mpt,lay,iin);
    lpre2 = NULL;
    printf("      Pre-syn layer:  %s\n",lpre->name);
  }else{
    bcode = 30001;
    lpre = popu_input_get_origin(mpt,lay,iin);
    lpre2 = NULL;
    printf("      Pre-syn layer:  %s\n",lpre->name);
  }
  fwrite((char *)&bcode,sizeof(int),1,fout);  
  if (pop_util_lclass_is_lgn0(lpre))  
    lgnflag = 1;
  else
    lgnflag = 0;
  carray_nchar_write(lpre->name,fout);
  if (lpre2 != NULL)
    carray_nchar_write(lpre2->name,fout);
  carray_nchar_write(lay->name,fout);
  xn = lay->xn;  
  yn = lay->yn;
  zn = lay->zn;
  if (xn*yn*zn <= 65536)
    nbyte_post = 2;     
  else
    nbyte_post = 4;     
  if (lpre->xn*lpre->yn*lpre->zn <= 65536)
    nbyte_pre = 2;
  else
    nbyte_pre = 4;
  if (lpre2 != NULL){
    if (lpre2->xn*lpre2->yn*lpre2->zn <= 65536)
      nbyte_pre2 = 2;
    else
      nbyte_pre2 = 4;
  }
  fwrite((char *)&nbyte_pre,sizeof(int),1,fout);
  if (lpre2 != NULL)
    fwrite((char *)&nbyte_pre2,sizeof(int),1,fout);
  fwrite((char *)&nbyte_post,sizeof(int),1,fout);
  sycnt = 0;
  ci = 0;  
  for(i=0;i<xn;i++){
    for(j=0;j<yn;j++){
      for(k=0;k<zn;k++){    
	c = &(lay->c[i][j][k]);  
	if (nbyte_post == 2){
	  si = (short)ci;
	  fwrite((char *)&si,sizeof(short),1,fout);
	}else{ 
	  fwrite((char *)&ci,sizeof(int),1,fout);
	}
	if (lgnflag){
	  popu_mar_write_syn_lgn(fout,vers,lpre,lay,c,nbyte_pre);
	}else if (lpre2 != NULL){
	  popu_mar_write_syn_reg_pair(fout,vers,lpre,lpre2,lay,c,
				      nbyte_pre,nbyte_pre2,iin);
	}else{
	  sycnt += popu_mar_write_syn_reg(fout,vers,lpre,lay,c,nbyte_pre,iin);
	}
	ci += 1;
      }
    }
  }
  printf("      %d synapses\n",sycnt);
  bcode *= -1;
  fwrite((char *)&bcode,sizeof(int),1,fout);  
}
void popu_mar_write(mpt,outfile,vers)
     struct pop_top *mpt;
     char *outfile;
     int vers;
{
  FILE *fopen(),*fout;
  int i,j,k;
  int fcode,n;
  struct pop_layer *lay;
  int pflag = 0;
  printf("  POPU_MAR_WRITE\n");
  fcode = 15551777;    
  if ((vers <= 0) || (vers > 4))
    mylogx(mpt->logf,"POPU_MAR_WRITE","Unknown version number");
  if ((fout = fopen(outfile,"w")) == NULL){
    printf("  *** File %s\n",outfile);
    mylogx(mpt->logf,"POPU_MAR_WRITE","Cannot open file");
  }
  fwrite((char *)&fcode,sizeof(int),1,fout);  
  fwrite((char *)&vers,sizeof(int),1,fout);   
  if (pflag == 1) printf("HERE 001\n");
  k = 3000;
  fwrite((char *)&k,sizeof(int),1,fout);   
  fwrite((char *)&(mpt->xn),sizeof(int),1,fout);
  fwrite((char *)&(mpt->yn),sizeof(int),1,fout);
  fwrite((char *)&(mpt->sscale),sizeof(float),1,fout);
  fwrite((char *)&(mpt->tscale),sizeof(float),1,fout);
  k = -k;
  fwrite((char *)&k,sizeof(int),1,fout);   
  if (pflag == 1) printf("HERE 002\n");
  printf("    Writing Areas\n");
  k = 10000;
  n = mpt->narea;
  fwrite((char *)&k,sizeof(int),1,fout);   
  fwrite((char *)&n,sizeof(int),1,fout);   
  for(i=0;i<n;i++){
    popu_mar_write_area(fout,mpt,mpt->area[i],vers);
  }
  k = -k;
  fwrite((char *)&k,sizeof(int),1,fout);   
  if (pflag == 1) printf("HERE 003\n");
  k = 20000;
  n = mpt->nlay;
  fwrite((char *)&k,sizeof(int),1,fout);   
  fwrite((char *)&n,sizeof(int),1,fout);   
  for(i=0;i<n;i++){
    popu_mar_write_lay(fout,mpt,mpt->lay[i],vers);
  }
  k = -k;
  fwrite((char *)&k,sizeof(int),1,fout);   
  if (pflag == 1) printf("HERE 004\n");
  k = 30000;
  fwrite((char *)&k,sizeof(int),1,fout);   
  for(i=0;i<mpt->nlay;i++){
    lay = mpt->lay[i];
    for(j=0;j<lay->ninlist;j++){
      popu_mar_write_input(fout,mpt,lay,j,vers);
    }
  }
  k = -k;
  fwrite((char *)&k,sizeof(int),1,fout);   
  if (pflag == 1) printf("HERE 005\n");
  fclose(fout);
  printf("    Wrote %s\n",outfile);
}
int ***popu_rsp_gen_get_map(mpt,smap,lay)
     struct pop_top *mpt;
     int ****smap;
     struct pop_layer *lay;
{
  int k;
  k = popu_get_layer_index_by_name(mpt,lay->name);
  if (k == -1){
    printf("  Layer  %s\n",lay->name);
    exit_error("POPU_RSP_GEN_GET_MAP","Layer not found");
  }
  if (smap[k] == NULL){  
    smap[k] = get_zero_3d_iarray(lay->xn,lay->yn,lay->zn);
  }
  return smap[k];
}
struct pop_layer *popu_rsp_gen_get_lay_ptr(mpt,lname)
     struct pop_top *mpt;
     char *lname;
{
  struct pop_layer *pl;
  pl = pop_cell_get_layer_pointer(mpt->lay,mpt->nlay,lname);
  if (pl == NULL){
    printf("  name:  %s\n",lname);
    exit_error("POPU_RSP_GEN_GET_LAY_PTR","Population name not found");
  }
  return pl;
}
int ***popu_rsp_gen_cmap_from_w(w,xn,yn,zn,optype,fcrit,rn)
     float ***w;
     int xn,yn,zn;
     char *optype;
     float fcrit;
     int *rn;
{
  int i,j,k;
  int ***cmap,n;
  cmap = get_zero_3d_iarray(xn,yn,zn);
  n = 0;
  for(i=0;i<xn;i++){
    for(j=0;j<yn;j++){
      for(k=0;k<zn;k++){
	if (strcmp(optype,"w_agg_min")==0){
	  if (w[i][j][k] > fcrit){  
	    cmap[i][j][k] = 1;
	    n += 1;
	  }
	}else{
	  printf("  *** optype:  %s\n",optype);
	  exit_error("POPU_RSP_GEN_CMAP_FROM_W","Unknown optype");
	}
      }
    }
  }
  *rn = n;
  return cmap;
}
void popu_rsp_gen_append_by_map(cmap,xn,yn,zn,outfile,name_save,dtype,samp_str,
				name_pre,datid,sline,nnew,w)
     int ***cmap;
     int xn,yn,zn;
     char *outfile,*name_save,*dtype,*samp_str,*name_pre,*datid;
     char *sline;
     int nnew;
     float ***w;  
{
  int i,j,k;
  float wsum;
  char t1[SLEN],t2[SLEN];
  sprintf(t1,"##  %s\n##\n",sline);
  append_string_to_file(outfile,t1);
  sprintf(t1,"##  %d units marked\n##\n",nnew);
  append_string_to_file(outfile,t1);
  if (w != NULL)
    wsum = (float)sum_3d_farray(w,0,xn,0,yn,0,zn);
  for(i=0;i<xn;i++){
    for(j=0;j<yn;j++){
      for(k=0;k<zn;k++){
	if (cmap[i][j][k] >= 1){
	  sprintf(t1,"save_pop_unit_as_%s_%d_%d_%d %s %s",name_save,i,j,k,
		  dtype,samp_str);
	  sprintf(t2,"%s %s %d %d %d %s",t1,name_pre,i,j,k,datid);
	  if (w != NULL)
	    sprintf(t1,"%s   # w %f (%f)\n",t2,w[i][j][k],w[i][j][k]/wsum);
	  else
	    sprintf(t1,"%s\n",t2);
	  append_string_to_file(outfile,t1);
	  cmap[i][j][k] = -1;  
	}
      }
    }
  }
  append_string_to_file(outfile,"\n");
}
int popu_rsp_gen_expand(mpt,m,sline,smap)
     struct pop_top *mpt;
     struct model_struct *m;    
     char *sline;               
     int ****smap;              
{
  int i;
  int xn,yn,zn,ns,nnew,xi,yi,zi,nlay,***c,nsyn;
  float minw,fcrit,***wf,ori_0,ori_1;
  char **slist,*name_save,*dtype,*samp_str,*name_pre,*name_post,*name_pop;
  char *datid,*outfile,*optype;
  char ts[SLEN];
  struct pop_layer *lpre,*lpost,*lay,**llist;
  nnew = 0;
  get_items_from_string(sline,&slist,&ns);
  if (ns > 0){
    outfile = m->marfile;
    if (compare_prefix_string_order("gen_save_pop_lay_ori_as_",
				    slist[0])==1){
      if (ns != 8){
	printf("  ==>  %s\n",sline);
	exit_error("POPU_RSP_GEN_EXPAND","Format error, expecting 7 args");
      }
      name_save = (char *)(slist[0] + 24);
      dtype     =          slist[1];
      samp_str  =          slist[2];
      name_pop  =          slist[3];
      optype    =          slist[4];
      ori_0     =     atof(slist[5]);  
      ori_1     =     atof(slist[6]);  
      datid     =          slist[7];
      printf("    Expanding 'gen_save_pop_lay_ori_as_'\n");
      lay  = popu_rsp_gen_get_lay_ptr(mpt,name_pop);
      c = popu_rsp_gen_get_map(mpt,smap,lay);
      nnew = popc_flag_accum_lay_ori(lay,c,optype,ori_0,ori_1);
      printf("      Including %d units from '%s'  (ori crit:  %.4f %.4f)\n",
	     nnew,name_save,ori_0,ori_1);
      popu_rsp_gen_append_by_map(c,lay->xn,lay->yn,lay->zn,outfile,
				 name_save,dtype,samp_str,name_pre,datid,
				 sline,nnew,NULL);
    }else if (compare_prefix_string_order("gen_save_pop_lay_to_unit_as_",
				    slist[0])==1){
      if (ns != 10){
	printf("  ==>  %s\n",sline);
	exit_error("POPU_RSP_GEN_EXPAND","Format error, expecting 9 args");
      }
      name_save = (char *)(slist[0] + 28);
      dtype     =          slist[1];
      samp_str  =          slist[2];
      name_pre  =          slist[3];
      name_post =          slist[4];
      xi        =     atoi(slist[5]);
      yi        =     atoi(slist[6]);
      zi        =     atoi(slist[7]);
      minw      =     atof(slist[8]);
      datid     =          slist[9];
      printf("    Expanding 'gen_save_pop_lay_to_unit_as_'\n");
      lpre  = popu_rsp_gen_get_lay_ptr(mpt,name_pre);
      lpost = popu_rsp_gen_get_lay_ptr(mpt,name_post);
      c = popu_rsp_gen_get_map(mpt,smap,lpre);
      nnew = popc_flag_accum_lay_to_unit(lpre,lpost,xi,yi,zi,c,minw);
      printf("      Including %d units from '%s'  (minw %.4f)\n",
	     nnew,name_save,minw);
      popu_rsp_gen_append_by_map(c,lpre->xn,lpre->yn,lpre->zn,outfile,
				 name_save,dtype,samp_str,name_pre,datid,
				 sline,nnew,NULL);
    }else if (compare_prefix_string_order("gen_save_pop_lay_to_lay_as_",
					  slist[0])==1){
      if (ns != 7){
	printf("  ==>  %s\n",sline);
	exit_error("POPU_RSP_GEN_EXPAND","Format error, expecting 6 args");
      }
      name_save = (char *)(slist[0] + 27);
      dtype     =          slist[1];
      samp_str  =          slist[2];
      name_pre  =          slist[3];
      name_post =          slist[4];
      minw      =     atof(slist[5]);
      datid     =          slist[6];
      printf("    Expanding 'gen_save_pop_lay_to_lay_as_'\n");
      lpre  = popu_rsp_gen_get_lay_ptr(mpt,name_pre);
      lpost = popu_rsp_gen_get_lay_ptr(mpt,name_post);
      c = popu_rsp_gen_get_map(mpt,smap,lpre);
      nnew = popc_flag_accum_lay_to_lay(lpre,lpost,c,minw);
      printf("      Including %d units from '%s'  (minw %.4f)\n",
	     nnew,name_save,minw);
      popu_rsp_gen_append_by_map(c,lpre->xn,lpre->yn,lpre->zn,outfile,
				 name_save,dtype,samp_str,name_pre,datid,
				 sline,nnew,NULL);
    }else if (compare_prefix_string_order("gen_save_pop_multi_lay_as_",
					  slist[0])==1){
      if (ns < 8){
	printf("  ==>  %s\n",sline);
	exit_error("POPU_RSP_GEN_EXPAND","Format error, expecting 8+ args");
      }
      name_save = (char *)(slist[0] + 26);
      dtype     =          slist[1];
      samp_str  =          slist[2];
      optype    =          slist[ns-3];
      fcrit     =     atof(slist[ns-2]);
      datid     =          slist[ns-1];
      nlay = ns - 6;  
      llist = (struct pop_layer **)myalloc(nlay*sizeof(struct pop_layer *));
      for(i=0;i<nlay;i++){
	llist[i] = popu_rsp_gen_get_lay_ptr(mpt,slist[i+3]);
      }
      lpre = llist[0];
      xn = lpre->xn;
      yn = lpre->yn;
      zn = lpre->zn;
      printf("    Expanding 'gen_save_pop_multi_lay_as_'\n");
      wf = popc_wf_accum_multi_lay(llist,nlay,-1,-1,-1,optype,fcrit);
      c = popu_rsp_gen_cmap_from_w(wf,xn,yn,zn,optype,fcrit,&nnew);
      printf("      Including %d units from '%s'  (fcrit %.4f)\n",
	     nnew,name_save,fcrit);
      popu_rsp_gen_append_by_map(c,lpre->xn,lpre->yn,lpre->zn,outfile,
				 name_save,dtype,samp_str,name_pre,datid,
				 sline,nnew,wf);
    }else if (compare_prefix_string_order("gen_save_pop_mlay_unit_as_",
					  slist[0])==1){
      if (ns < 11){
	printf("  ==>  %s\n",sline);
	exit_error("POPU_RSP_GEN_EXPAND","Format error, expecting 11+ args");
      }
      name_save = (char *)(slist[0] + 26);
      dtype     =          slist[1];
      samp_str  =          slist[2];
      xi        =     atoi(slist[ns-6]);
      yi        =     atoi(slist[ns-5]);
      zi        =     atoi(slist[ns-4]);
      optype    =          slist[ns-3];
      fcrit     =     atof(slist[ns-2]);
      datid     =          slist[ns-1];
      nlay = ns - 9;  
      llist = (struct pop_layer **)myalloc(nlay*sizeof(struct pop_layer *));
      for(i=0;i<nlay;i++){
	llist[i] = popu_rsp_gen_get_lay_ptr(mpt,slist[i+3]);
      }
      lpre = llist[0];
      xn = lpre->xn;
      yn = lpre->yn;
      zn = lpre->zn;
      printf("    Expanding 'gen_save_pop_mlay_unit_as_'\n");
      wf = popc_wf_accum_multi_lay(llist,nlay,xi,yi,zi,optype,fcrit);
      c = popu_rsp_gen_cmap_from_w(wf,xn,yn,zn,optype,fcrit,&nnew);
      printf("      Including %d units from '%s'  (fcrit %.4f)\n",
	     nnew,name_save,fcrit);
      popu_rsp_gen_append_by_map(c,lpre->xn,lpre->yn,lpre->zn,outfile,
				 name_save,dtype,samp_str,name_pre,datid,
				 sline,nnew,wf);
    }
    free_2d_carray(slist,ns);
  }
  return nnew;
}
void popu_rsp_gen(mpt,m,r)
     struct pop_top *mpt;
     struct model_struct *m;        
     struct response_struct *r;     
{
  int i;
  int ns;
  char **sdata,*outfile;
  int ****smap;
  printf("  POPU_RESP_GEN\n");
  smap = (int ****)myalloc(mpt->nlay * sizeof(int ***));
  for(i=0;i<mpt->nlay;i++)
    smap[i] = NULL;
  outfile = m->marfile;
  printf("    Writing response file:  %s\n",outfile);
  remove_file(outfile);  
  read_2d_carray(r->paramfile,1,1,&sdata,&ns);  
  for(i=0;i<ns;i++){
    if (popu_rsp_gen_expand(mpt,m,sdata[i],smap) == 0){
      append_string_to_file(outfile,sdata[i]);  
      append_string_to_file(outfile,"\n");
    }
  }
}
